import sys
from tfpg_viewer.TFPGModel import TFPGModel
from Viewer.viewer import Ui_ViewerMainWindow, ViewerMainWindow
import pkg_resources
import networkx as nx
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from PySide2.QtGui import QIcon
from PySide2.QtWidgets import QAction, QDialog, QPushButton, QMainWindow, QApplication, QWidget, QGridLayout, QVBoxLayout, QLabel, QLineEdit
from PySide2.QtCore import QMetaObject
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT, FigureCanvasQTAgg as FigureCanvas


class Ui_TfpgViewerMainWindow(Ui_ViewerMainWindow):
    def setup_ui(self, MainWindow, tfpg_graph=None):
        super().setup_ui(MainWindow)
        self.EXTENSION = 'txml'
        self.MainWindow = MainWindow
        MainWindow.setWindowTitle("Tfpg Viewer")
        central_widget = QWidget()
        MainWindow.setCentralWidget(central_widget)

        self.standalone = True
        self.tfpg_graph = tfpg_graph

        self.axis_x_left_margin = None
        self.axis_x_right_margin = None
        self.axis_y_bottom_margin = None
        self.axis_y_top_margin = None

        self.m_w11 = QWidget()
        self.m_w12 = QWidget()
        self.lay = QGridLayout(central_widget)

        for w, (r, c) in zip(
            (self.m_w11, self.m_w12),
            ((0, 0), (1, 0)),
        ):
            self.lay.addWidget(w, r, c)
        for c in range(1):
            self.lay.setColumnStretch(c, 1)
        for r in range(1):
            self.lay.setRowStretch(r, 1)

        if self.tfpg_graph:
            self.show_canvas(self.tfpg_graph)


    def show_canvas(self, tfpg_graph):
        canvas = self.pyplot_view_structural(tfpg_graph=tfpg_graph)
        self.canvasLayout = QVBoxLayout(self.m_w11)


        self.canvasLayout.addWidget(canvas)
        lay = QVBoxLayout(self.m_w12)

        self.nodeSettingLayout = QVBoxLayout()
        lay.addLayout(self.nodeSettingLayout)



    def pyplot_view_structural(self, tfpg_graph=None, drawing_area_name='hbox_viewing_area',
                               toolbar_parent_name='vbox_toolbar_viewer'):
        """Shows a structural tfpg using pyplot.

        The figure is plotted to the 'hbox_viewing_area' component
        (if no other component is specified)  of the user interface.

        Two things deserve to be noticed:
        1. The param tfpg_graph must be specified only in case we want to show
           a new TFPG; in case we want to update the existing TFPG, it must be left to None
        2. To the figure is added also a toolbar (containing the buttons for moving, zooming, ...)
           Pay attention to specify properly the name of the widget to which it must be added.

        :param tfpg_graph: the structural tfpg to be shown
        :type  tfpg_graph: data_structure.tfpg.tfpg.Tfpg

        :param drawing_area_name: the name of the widget (an Hbox or a VBox) where the TFPG must
                                  be drawn
        :type  drawing_area_name: gtk.Widget

        :param toolbar_parent_name: the name of the widget which must contain the toolbar
        :type  toolbar_parent_name: gtk.Widget
        """
        self.topToolBar.clear()

        if tfpg_graph is not None:
            self.pos = None
            self.tfpg_graph = None
            self.dpi = None

        canvas = self.create_tfpg_canvas_figure(tfpg_graph=tfpg_graph)

        canvas.mpl_connect('button_press_event', self.onclick)

        toolbar = NavigationToolbar(canvas, None, self)
        self.topToolBar.addWidget(toolbar)
        self.iconInfo = QIcon.fromTheme('dialog-information')
        captionAction = QAction(
            self.iconInfo, "Caption", self.bottomToolBar)
        self.topToolBar.addAction(captionAction)
        captionAction.triggered.connect(self.handleCaptionClick)
        return canvas

    def create_tfpg_canvas_figure(self, tfpg_graph=None):
        """It creates the canvas figure (to be added to a GUI element) corresponding to
           the global TFPG internal representation

           Notice that when we want to update the shown graph (i.e. when zooming, moving it...)
           we call this function with tfpg_graph = None; in this way, the positions of the
           elements are not recomputed and the already populated internatl structure
           is not recreated.

           :param tfpg_graph: the structural tfpg to be represented
           :type  tfpg_graph: data_structure.tfpg.tfpg.Tfpg
        """

        # tfpg_graph is not None only the first time when a TFPG is read from file
        # and it must be initialized
        if tfpg_graph is not None:
            # tfpg_graph is different from None only the first time the TFPG is
            # represented; hence, the internal global variable must be saved
            assert self.tfpg_graph is None
            self.tfpg_graph = tfpg_graph

        # the following dpi number (used to find a proper scale value for the
        # representation) is the result of some (little) empirical evaluation
        if self.dpi is None:
            node_nbr = len(self.tfpg_graph.get_nodes_list())
            self.dpi = ((1/node_nbr)*428)
            if self.dpi <= 4:  # It does not seem possible to have a dpi value lower than 4
                self.dpi = 4

        fig = plt.figure(1, figsize=(15, 12))

        fig.clf()
        fig.set_dpi(self.dpi)
        fig.set_facecolor("gray")
        if self.standalone:
            gridspec.GridSpec(1,1)

            graphical_rep = plt.subplot(autoscalex_on=True, autoscaley_on=True)
        else:
            graphical_rep = plt.subplot(111)

        nx_graph = self.convert_tfpg_struct_to_networkx_graph()
        # once the nx_graph has been created, in case we are drawing for the first time
        # a TFPG, the margins must be initialised
        if tfpg_graph is not None:
            self.initialize_figure_margins()

        graphical_rep.set_xlim(self.axis_x_left_margin,
                               self.axis_x_right_margin)
        graphical_rep.set_ylim(self.axis_y_bottom_margin,
                               self.axis_y_top_margin)
        graphical_rep = self.draw_graph(nx_graph)
        canvas = FigureCanvas(fig)

        return canvas

    def draw_graph(self, nx_graph):
        """Draws the internal graph using networkx library.

        It draws the nx_graph class variable. In this way, when
        this structure is modified, to graphically see the changes
        it is possible to call this function.

        @param nx_graph: the newtorkx representation of the internal
        TFPG struct
        @type  nx_graph: networkx graph
        """

        fig = plt.figure(1)
        nodes_lbl_list = dict()
        edges_lbl_list = dict()

        for nd in nx_graph.nodes(data=True):

            if "lbl" in nd[0]:
                nx.draw_networkx_nodes(nx_graph, self.pos,
                                       nodelist=[nd[0]],
                                       node_shape='s',
                                       node_size=3000,
                                       alpha=0.0)
            else:
                nx.draw_networkx_nodes(nx_graph, self.pos, nodelist=[nd[0]],
                                       node_shape=nd[1]['node_shape'],
                                       node_size=3000,
                                       node_color=nd[1]['node_color'],
                                       linewidths=nd[1]['node_width'])

                nodes_lbl_list[nd[0]] = nd[1]['node_label']


        for edg in nx_graph.edges(data=True, keys=True):

            nx.draw_networkx_edges(nx_graph, self.pos,
                                   edgelist=[(edg[0], edg[1])],
                                   edge_color=edg[3]['edg_color'],
                                   width=edg[3]['edg_width'])

            edges_lbl_list[(edg[0], edg[1])] = edg[3]['edg_label']


        nx.draw_networkx_labels(nx_graph, self.pos, labels=nodes_lbl_list,
                                font_size=10)

        nx.draw_networkx_edge_labels(nx_graph, self.pos,
                                     edge_labels=edges_lbl_list,
                                     bbox=dict(boxstyle='round',
                                               ec='black',
                                               fc='white'),
                                     font_size=10)
        plt.axis('off')
        return fig



    def export_to_file(self, filename, format):
        """Saves the current tfpg in an output file
           Possible formats are: png, pdf, dot
        """
        if format == 'dot':
            # In case of dot format, the graph must be changed a bit
            cleaned_graph = self.clean_graph_for_dot_format()
            networkx_version = pkg_resources.get_distribution(
                'networkx').version
            from distutils.version import StrictVersion
            if StrictVersion(networkx_version) > StrictVersion('1.10'):
                nx.drawing.nx_agraph.to_agraph(cleaned_graph).write(
                    str(filename)+"."+str(format))
            else:
                nx.write_dot(cleaned_graph, str(filename)+"."+str(format))
        else:
            plt.figure(1)  # focus on the tfpg (not on the caption)
            plt.savefig(str(filename)+"."+str(format))

        return

    def clean_graph_for_dot_format(self):
        """Cleans the graph in order to be exported in dot format.

        Essentially, nodes containing node/edge labels (used for better
        show the graph in the GUI) are removed because in dot format
        they behave as different nodes
        """
        cleaned_graph = self.convert_tfpg_struct_to_networkx_graph()

        for nd in list(cleaned_graph.nodes(data=True)):
            if "_lbl" in nd[0]:
                cleaned_graph.remove_node(nd[0])

        for edg in cleaned_graph.edges(data=True, keys=True):
            cleaned_graph[edg[0]][edg[1]][edg[2]
                                          ]['label'] = edg[3]['edg_label']

        return cleaned_graph
    def convert_tfpg_struct_to_networkx_graph(self):
        """It returns the graphical representation of the internal representation
           of the TFPG.

           Notice that for seeing graphical changes, you need first to
           update the internal struct, and then to call this function and use the
           returned value.
        """

        nx_graph = nx.MultiDiGraph(name="graph")

        for nd in self.tfpg_graph.get_nodes_list():
            if nd.get_type() == "FM":
                shape = 's'
                nd_type = "Failure Mode"
            elif nd.get_type() == "AND":
                shape = 's'
                nd_type = "And Discrepancy"
            elif nd.get_type() == "OR":
                shape = 'o'
                nd_type = "Or Discrepancy"

            if nd.is_monitored():
                width = 6.0
            else:
                width = 1.0

            nx_graph.add_node(nd.get_name(), node_type=nd_type,
                              node_shape=shape, node_width=width,
                              node_label=nd.get_name(), node_color='white',
                              node_act_time="''")

        for eg in self.tfpg_graph.get_edges_list():

            tMin = eg.get_tMin()
            tMax = eg.get_tMax()
            if float(tMax) == -1.0:
                tMax = r"$\infty$"  # -1.0 means 'infinity'
            eg_lbl = "[%s, %s] {%s}" % (tMin, tMax, ', '.join(eg.get_modes()))

            nx_graph.add_edge(eg.get_srcNode().get_name(), eg.get_destNode().get_name(),
                              key=eg.get_name(),
                              edg_label=eg_lbl, edg_color='black', edg_act_time="''",
                              edg_width=1.0)

        if self.pos is None:
            self.pos = self._get_graphviz_layout(nx_graph, 'dot')


        # tryng to add a node on the label of each edge
        for eg in self.tfpg_graph.get_edges_list():
            nx_graph.add_node(eg.get_name() + "_edg_lbl")

            eg_src_pos = self.pos[eg.get_srcNode().get_name()]
            eg_dst_pos = self.pos[eg.get_destNode().get_name()]
            eg_lbl_pos = (((abs(eg_src_pos[0]+eg_dst_pos[0]))/2)+10,
                          ((abs(eg_src_pos[1]+eg_dst_pos[1]))/2)-5)
            self.pos[eg.get_name() + "_edg_lbl"] = eg_lbl_pos

        return nx_graph
    def _get_graphviz_layout(self, nx_graph, prog):
        """Return the graphviz layout depending on the version of networkx"""
        networkx_version = pkg_resources.get_distribution('networkx').version
        from distutils.version import StrictVersion
        if StrictVersion(networkx_version) > StrictVersion('1.10'):
            return nx.drawing.nx_agraph.graphviz_layout(nx_graph, prog)
        else:
            return nx.graphviz_layout(nx_graph, prog)
    def handleCaptionClick(self):
        canvas = self.create_caption()
        self.dialog = QDialog()
        self.popup = Ui_Caption(self.dialog)
        self.popup.setup_ui(self, canvas)
        self.dialog.show()

    def create_caption(self):
        """Creates and returns a figure conaining a caption for a given tfpg (and, if enabled, also for a trace).
        Notice that, in order to be shown, the return figure requires to be added to a gtk element."""

        fig = plt.figure(2)
        fig.set_dpi(60)
        fig.set_facecolor("gray")
        label_list = dict()

        graph = nx.MultiDiGraph(name="graph")

        # Adding failure mode/And discrepancy node
        graph.add_node("failure_mode", node_shape='s', node_alpha=1.0,
                       node_width=1.0, node_color='w')

        pos = self._get_graphviz_layout(graph, 'dot')
        fm_pos = pos["failure_mode"]

        # Adding an OR discrepancy node
        pos["or_discrepancy"] = (fm_pos[0], fm_pos[1]-1)
        graph.add_node("or_discrepancy", node_shape='o', node_alpha=1.0,
                       node_width=1.0, node_color='w')

        # Adding a monitored node
        pos["monitored_node"] = (fm_pos[0], fm_pos[1]-2)
        graph.add_node("monitored_node", node_shape='s', node_alpha=1.0,
                       node_width=3.0, node_color='w')

        # Adding failure mode/And discrepancy label node
        pos["failure_mode_lbl"] = (fm_pos[0]+3, fm_pos[1])
        graph.add_node("failure_mode_lbl", node_shape='s', node_alpha=0.0,
                       node_width=0.0, node_color='w')
        label_list["failure_mode_lbl"] = "Failure Mode\nAnd Discrepancy"

        # Adding or discrepancy label node
        pos["or_discrepancy_lbl"] = (fm_pos[0]+3, fm_pos[1]-1)
        graph.add_node("or_discrepancy_lbl", node_shape='s', node_alpha=0.0,
                       node_width=0.0, node_color='w')
        label_list["or_discrepancy_lbl"] = "Or Discrepancy"

        # Adding monitored node label
        pos["monitored_node_lbl"] = (fm_pos[0]+3, fm_pos[1]-2)
        graph.add_node("monitored_node_lbl", node_shape='s', node_alpha=0.0,
                       node_width=0.0, node_color='w')
        label_list["monitored_node_lbl"] = "Monitored node\n(general)"

        # Adding an edge
        pos["tmp_node_A"] = (fm_pos[0]-1, fm_pos[1]-4)
        graph.add_node("tmp_node_A", node_shape='o', node_alpha=1.0,
                       node_width=1.0, node_color='w')
        label_list["tmp_node_A"] = "A"
        pos["tmp_node_B"] = (fm_pos[0]+4, fm_pos[1]-4)
        graph.add_node("tmp_node_B", node_shape='o', node_alpha=1.0,
                       node_width=1.0, node_color='w')
        label_list["tmp_node_B"] = "B"
        graph.add_edge("tmp_node_A", "tmp_node_B")

        pos["edge_lbl"] = (fm_pos[0]+1.5, fm_pos[1]-5)
        graph.add_node("edge_lbl", node_shape='s', node_alpha=0.0,
                       node_width=0.0, node_color='w')
        label_list["edge_lbl"] = "Directed edge (from A to B)"


        for nd in graph.nodes(data=True):

            shape = nd[1]['node_shape']
            color = nd[1]['node_color']
            width = nd[1]['node_width']
            alpha = nd[1]['node_alpha']

            nx.draw_networkx_nodes(graph, pos, nodelist=[nd[0]],
                                   node_shape=shape,
                                   node_size=400,
                                   node_color=color,
                                   linewidths=width, alpha=alpha)

        for edg in graph.edges(data=True):
            nx.draw_networkx_edges(graph, pos,
                                   edgelist=[(edg[0], edg[1])],
                                   edge_color='black')

        nx.draw_networkx_labels(graph, pos, labels=label_list,
                                font_size=10)

        plt.axis("off")

        canvas = FigureCanvas(fig)

        return canvas

    def update_canvas(self):
        canvas = self.pyplot_view_structural()

        for i in reversed(range(self.canvasLayout.count())):
            self.canvasLayout.itemAt(i).widget().setParent(None)

        self.canvasLayout.addWidget(canvas)

    def ui_reset(self):
        """Function wich resets the user interface, i.e. it
        re-initialize the margins and the dpi."""
        # The margins of the figure must be reset
        self.initialize_figure_margins()
        # The DPI must be reset
        self.dpi = None
        # Update the figure
        self.update_canvas()

        return

    def ui_zoom_in(self):
        """Function which implements the increasing zoom feature"""

        self.axis_x_left_margin = self.axis_x_left_margin + 30
        self.axis_x_right_margin = self.axis_x_right_margin - 30
        self.axis_y_top_margin = self.axis_y_top_margin - 30
        self.axix_y_bottom_margin = self.axis_y_bottom_margin + 30
        self.dpi = self.dpi + 10
        # update the figure

        self.update_canvas()
        return

    def ui_zoom_out(self):
        """Function which implements the decreasing zoom feature"""

        self.axis_x_left_margin = self.axis_x_left_margin - 20
        self.axis_x_right_margin = self.axis_x_right_margin + 20
        self.axis_y_top_margin = self.axis_y_top_margin + 20
        self.axis_y_bottom_margin = self.axis_y_bottom_margin - 20
        self.dpi = self.dpi - 10
        if self.dpi < 4:  # it does not seem possible to have a dpi value lower than 4
            self.dpi = 4
        # update the figure
        self.update_canvas()
        return

    def onclick(self, event):
        """Function called when the user presses the mouse button onto
        the graphical representation.

        It is mainly used to check wheter a  user clicks over a TFPG element;
        in this case, the related information is shown.
        """
        component_clicked = self.get_mouse_position(event)
        self.update_component_information(component_clicked)

        return

    def get_mouse_position(self, event):
        """It checks whether the mouse is over a node or an edge.
        In case the mouse is over a node or an edge, it returns its name.
        Otherwise, it returns None.
        """

        for nd in self.pos.keys():

            (nd_x, nd_y) = self.pos[nd]

            if "edg" in nd:
                if event.xdata is not None:
                    if event.xdata >= nd_x - 13 and event.xdata <= nd_x - 8:
                        if event.ydata >= nd_y - 8 and event.ydata <= nd_y + 8:
                            return nd
            elif "nd_lbl" not in nd:
                if event.xdata is not None:
                    if event.xdata >= nd_x - 5 and event.xdata <= nd_x + 5:
                        if event.ydata >= nd_y - 15 and event.ydata <= nd_y + 15:
                            return nd
        return None

    def update_component_information(self, component_clicked, tgt_widget='hbox_info_area'):
        """Displays and adds to the given widget the information related to a particular
        component of the TFPG.

        @param component_clicked: the name of the element for which the info must be shown
        @type  component_clicked: string

        """
        if component_clicked != None:
            self.print_component_info(component_clicked)

        return

    def remove_settings(self):
        for i in reversed(range(self.nodeSettingLayout.count())):
            for j in reversed(range(self.nodeSettingLayout.itemAt(i).layout().count())):
                self.nodeSettingLayout.itemAt(i).layout().itemAt(
                    j).widget().setParent(None)
            self.nodeSettingLayout.itemAt(i).layout().setParent(None)

    def show_settings(self, input, isNode):
        info = ""
        if isNode:
            info = {'name': input.get_name(), 'type': input.get_type(),
                    'isMonitored': input.is_monitored()}
        else:
            info = {'name': input.get_name(), 'src': input.get_srcNode().get_name(),
                    'dst': input.get_destNode().get_name(), 'tMin': input.get_tMin(),
                    'tMax': input.get_tMax(), 'modes': ', '.join(input.get_modes())}

        self.remove_settings()

        self.gridLayout = QGridLayout()

        buttonCancel = QPushButton("&Cancel")
        buttonCancel.clicked.connect(self.handleCancel)
        buttonSave = QPushButton("&Save")
        buttonSave.clicked.connect(lambda: self.handleSave(input, isNode))

        for i in range(len(info)):
            self.label = QLabel()
            self.label.setText(list(info.keys())[i])
            self.line_edit = QLineEdit()
            self.line_edit.setText(str(list(info.values())[i]))
            if i < 3:
                self.gridLayout.addWidget(self.label, i, 0, 1, 1)
                self.gridLayout.addWidget(self.line_edit, i, 1, 1, 1)
            else:
                self.gridLayout.addWidget(self.label, i-3, 2, 1, 1)
                self.gridLayout.addWidget(self.line_edit, i-3, 3, 1, 1)

        self.nodeSettingLayout.addLayout(self.gridLayout)
        self.buttonLayout = QGridLayout()
        self.buttonLayout.addWidget(buttonSave, 0, 0, 1, 1)
        self.buttonLayout.addWidget(buttonCancel, 0, 1, 1, 1)
        self.nodeSettingLayout.addLayout(self.buttonLayout)

    def handleSave(self, input, isNode):
        if isNode:
            old_name = input.get_name()
            new_name = self.gridLayout.itemAtPosition(0, 1).widget().text()
            new_type = self.gridLayout.itemAtPosition(1, 1).widget().text()
            new_is_monitored = self.gridLayout.itemAtPosition(
                2, 1).widget().text()


            input.set_name(new_name)

            input.set_type(new_type)
            if new_is_monitored == "True":
                input.set_isMonitored(True)
            elif new_is_monitored == "False":
                input.set_isMonitored(False)
            else:
                print("Invalid input")
            self.pos[new_name] = self.pos.pop(old_name)



        else:
            old_name = input.get_name()
            new_name = self.gridLayout.itemAtPosition(0, 1).widget().text()
            new_src = self.gridLayout.itemAtPosition(1, 1).widget().text()
            new_dst = self.gridLayout.itemAtPosition(2, 1).widget().text()
            new_tmin = self.gridLayout.itemAtPosition(0, 3).widget().text()
            new_tmax = self.gridLayout.itemAtPosition(1, 3).widget().text()
            new_modes = self.gridLayout.itemAtPosition(2, 3).widget().text()

            input.set_name(new_name)
            input.set_srcNode(self.tfpg_graph.get_node_by_name(new_src))
            input.set_destNode(self.tfpg_graph.get_node_by_name(new_dst))
            input.set_tMin(new_tmin)
            input.set_tMax(new_tmax)
            input.set_modes(new_modes.split(','))
            self.pos[new_name+"_edg_lbl"] = self.pos.pop(old_name+"_edg_lbl")

        self.update_canvas()
        self.remove_settings()

    def handleCancel(self):
        self.remove_settings()

    def print_component_info(self, component, pretty_print=True):
        """Given a component, it returns its related information.

        If the parma pretty_print is set to True (by default), a string
        containing the information in a pretty-print form is returned,
        otherwise a dict is returned; otherwise, a dict is returned.

        For instance, in case component is an edge, the following dict
        is returned:
        {('name': <name>), ('src': <src>), ('dst': <dst>), ('tMin': <tMin>), ('tMax': <tMax>),
         ('modes': <modes>)}

        @param component: the name of the component for which we want to retrieve
                          the information
        @type  component: string
        """


        if "_edg_lbl" in component:
            edg_name = component[:component.find("_edg_lbl")]
            for edg in self.tfpg_graph.get_edges_list():
                if edg.get_name() == edg_name:
                    self.show_settings(edg, False)

        else:
            for nd in self.tfpg_graph.get_nodes_list():
                if nd.get_name() == component:
                    self.show_settings(nd, True)




    def initialize_figure_margins(self):
        """"Sets the bounds of the figure which contains the graphical representation.
        Once the nx_graph internal structure has been populated, this function reads
        the positions of all the nodes, and takes the minumum and maximum values for
        the x and the y ranges.
        This is useful to properly center the figure in the window.
        """

        ymin = 1000
        ymax = 0
        xmin = 1000
        xmax = 0

        for (x, y) in self.pos.values():

            if x < xmin:
                xmin = x
            if x > xmax:
                xmax = x
            if y < ymin:
                ymin = y
            if y > ymax:
                ymax = y

        self.axis_x_left_margin = xmin-20

        self.axis_x_right_margin = xmax+20

        self.axis_y_bottom_margin = ymin-20

        self.axis_y_top_margin = ymax+20

        return

    def open_file(self):

        model = TFPGModel()
        file_path, _ = super().open_file(self.EXTENSION)
        if self.tfpg_graph is None:
            self.tfpg_graph, errors = model.build_graph(file_path)
            self.show_canvas(self.tfpg_graph)
        else:
            self.tfpg_graph, errors = model.build_graph(file_path)
            self.pos = None
            self.update_canvas()


def process_options():
    """Checks whether the executable has been called properly"""

    import os
    import logging
    from optparse import OptionParser


    usage = "usage: tfpg_viewer.py [-t <tfpg_file>]"

    parser = OptionParser(usage)
    parser.add_option("-t", "--tfpg",
                      dest="tfpg_fname",
                      default=None,
                      help="the name of the TFPG file")


    (opts, args) = parser.parse_args()

    if opts.tfpg_fname is not None:
        if not os.path.exists(opts.tfpg_fname):
            parser.error("Invalid TFPG filename: %s" % opts.tfpg_fname)
        else:
            logging.debug("-t option specified with file '%s'" %
                          opts.tfpg_fname)


    return opts






def run(tfpg_fname, out_fname, output_format):
    app = QApplication(sys.argv)
    app.setQuitOnLastWindowClosed(False)

    model = TFPGModel()
    ui = Ui_TfpgViewerMainWindow()
    main_window = ViewerMainWindow()
    tfpg_graph = None
    if tfpg_fname:
        tfpg_graph, errors = model.build_graph(tfpg_fname)

    ui.setup_ui(main_window, tfpg_graph)
    if out_fname:
        if len(errors) > 0:
            print("Error! %s" % str(errors[0]))
        else:

            ui.export_to_file(out_fname, output_format)
            print("Exported to file: %s" % out_fname)
            main_window.quit_terminal_process()
    else:
        main_window.show()
        sys.exit(app.exec_())


class NavigationToolbar(NavigationToolbar2QT):
    """Extends the Matplotlib Navigation Toolbar. This toolbar is added for performing
    several actions on TFPG representations."""

    # Items to be shown on the toolbar; we are simply interested in resetting,
    # moving and zooming
    toolitems = (
        ('Home', 'Reset to original view and size', 'home', 'home'),
        ('Pan',  'Pan axes with left mouse, zoom with right', 'move', 'pan'),
        ('Zoom In', 'Zoom In with a click', 'zoom_to_rect', 'zoom_in'),
        ('Zoom Out', 'Zoom Out with a click', 'zoom_to_rect', 'zoom_out'),
        ('Save', 'Save the figure', 'filesave', 'save_figure'),
    )

    def __init__(self, canvas, win, viewer):
        """Class constructor. Notice that a parameter has been added, i.e. the
        viewer; this is required because we want to override some default functions
        with other functions contained in the TFPGView class."""
        NavigationToolbar2QT.__init__(self, canvas, win)
        self.main_window = viewer
        return

    def home(self):
        """Overrides the default function used to reset the figure; in particular,
        first we need to reset the ui with the custom method defined in the TfpgView.
        """
        self.main_window.ui_reset()
        NavigationToolbar2QT.home(self)

    def zoom_in(self):
        """Ovverides the zooming function. In particular, we do not like too much
        the standard matplotlib way of increasing with a rectangle.
        """
        self.main_window.ui_zoom_in()

    def zoom_out(self):
        """Overrides the standard matplotlib way of zooming out."""
        self.main_window.ui_zoom_out()


pass  # end of NavigationToolbar class
# ----------------------------------------------------------------------


class Ui_Caption(object):

    def __init__(self, Dialog):
        self.Dialog = Dialog

    def setup_ui(self, Dialog, canvas):
        self.Dialog.resize(350, 300)
        self.Dialog.setWindowTitle("Caption")

        self.verticalLayout = QVBoxLayout(self.Dialog)
        self.verticalLayout.addWidget(canvas)

        QMetaObject.connectSlotsByName(self.Dialog)

    def accept(self):
        self.Dialog.close()
