import sys
import argparse

from Viewer.viewer import Ui_ViewerMainWindow, ViewerMainWindow
from trace_viewer.filter_popup import Ui_Dialog
from trace_viewer.filter import Filter
from trace_viewer.nusmv import NuSMVTraceReader
from PySide2.QtWidgets import (
    QApplication,
    QDialog,
    QLineEdit,
    QWidget,
    QTableView,
    QHeaderView,
    QHBoxLayout,
    QVBoxLayout,
    QGridLayout,
    QLabel,
    QComboBox,
    QDialogButtonBox,
    QAction
)
from PySide2.QtGui import (
    QColor,
    QBrush,
    QStandardItem,
    QIcon
)
from PySide2.QtCore import Qt, Signal, QRect, QMetaObject


PROG_VER = (2, 0, 0)


class LineEdit(QLineEdit):
    keyPressed = Signal(Qt.Key)

    def keyPressEvent(self, event):
        key = event.key()
        if key == Qt.Key.Key_Backspace:
            self.keyPressed.emit(Qt.Key.Key_Backspace)
        elif key == Qt.Key.Key_Return:
            self.keyPressed.emit(Qt.Key.Key_Return)
        super().keyPressEvent(event)


class Ui_TraceViewerMainWindow(Ui_ViewerMainWindow):

    def update_ui(self, trace):
        self.comboBox.addItems(trace.get_filter_names())

        self.labelTotalSteps.setText("Total steps: " + str(trace.get_steps()))

        self.comboBox.currentIndexChanged.connect(
            lambda: self.combo_box_changed(trace))

        self.actionCancel.triggered.connect(lambda: self.cancel_pressed(trace))

        self.actionEdit.triggered.connect(lambda: self.show_pop_up(trace))

        self.lineEditNameSearch.keyPressed.connect(
            lambda key: self.on_key_pressed(key, trace))

    def on_key_pressed(self, key, trace):
        if key == Qt.Key.Key_Backspace:
            self.handle_backspace(trace)
        elif key == Qt.Key.Key_Return:
            self.handle_enter(trace)

    def handle_backspace(self, trace):
        old_filter = trace.get_filter(self.comboBox.currentText())
        self.remove_filter(trace)
        self.apply_filter(old_filter, trace)

    def handle_enter(self, trace):
        name = self.lineEditNameSearch.text()
        old_filter = trace.get_active_filter()

        new_filter = Filter()
        new_filter.set_filter_name(old_filter.get_filter_name())
        new_filter.set_ranges(old_filter.get_ranges())
        new_filter.set_types(old_filter.get_types())

        old_names = old_filter.get_names()
        new_names = old_names+[name]
        new_filter.set_names(new_names)

        self.remove_filter(trace)
        self.apply_filter(new_filter, trace)

    def cancel_pressed(self, trace):
        self.comboBox.setCurrentIndex(0)

    def load_table_data(self, trace):

        row_count = trace.get_length()
        col_count = trace.get_steps()

        for row in range(row_count):
            row_name = trace.get_value(row)
            vertical_header = QStandardItem(row_name)
            row_color = QBrush(QColor(trace.get_best_row_color(row)))
            vertical_header.setForeground(row_color)
            trace.setHorizontalHeaderItem(0, QStandardItem("Name"))
            trace.setItem(row, 0, vertical_header)
            vertical_header.setFlags(Qt.ItemFlag.ItemIsEnabled)
            vertical_header.setBackground(trace.get_best_row_background_color(row))
            for col in range(col_count):

                horizontal_header = QStandardItem("step {}".format(col+1))
                horizontal_header.setBackground(
                    trace.get_best_step_color(col+1))
                trace.setHorizontalHeaderItem(col+1, horizontal_header)

                cell = QStandardItem(
                    str(trace.get_value_at_step(row_name, col)))
                cell.setFlags(Qt.ItemFlag.ItemIsEnabled)
                cell.setBackground(trace.get_best_cell_color(row_name, col))
                trace.setItem(row, col+1, cell)

        self.tableWidget.setModel(trace)

    def apply_filter(self, filter, trace):
        # Set the active filter for the trace

        # Reload the trace data
        fname = trace.get_fname()
        trace = NuSMVTraceReader.parse(NuSMVTraceReader, fname)

        # Hide the columns and rows that don't match the filter
        for i in range(trace.get_steps()):
            if not filter.is_valid_step(i):
                self.tableWidget.setColumnHidden(i+1, True)

        for i in range(len(trace.names)):
            if (not filter.is_valid_type(trace.get_type(sorted(list(trace.names))[i]))) or \
                    (not filter.is_valid_name(trace.get_value(i))):
                self.tableWidget.setRowHidden(i, True)

    def combo_box_changed(self, trace):
        self.remove_filter(trace)

    # Check if a filter has been selected
        if self.comboBox.currentText() != 'no filter':
            # Get the selected filter
            filter = trace.get_filter(self.comboBox.currentText())
            trace.set_active_filter(filter)
            self.apply_filter(filter, trace)

    # If no filter is selected, show all data
        else:
            filter = Filter()
            self.apply_filter(filter, trace)

    def remove_filter(self, trace):
        num_rows = trace.get_length()
        num_cols = trace.get_steps()

        for row in range(num_rows):
            self.tableWidget.setRowHidden(row, False)

        for col in range(num_cols):
            self.tableWidget.setColumnHidden(col, False)

    def open_file(self):
        file_path, _ = super().open_file(self.EXTENSION)

        trace = NuSMVTraceReader.parse(NuSMVTraceReader, file_path)

        self.load_table_data(trace)

        self.update_ui(trace)

    def show_pop_up(self, trace):
        self.window = QDialog()
        self.popup = Ui_Dialog(self.window)
        if self.comboBox.currentIndex() == 0:
            filter = Filter()
        else:
            filter = trace.get_filter_at_pos(self.comboBox.currentIndex())

        self.popup.setup_ui(self, trace)
        self.window.show()

    def setup_ui(self, MainWindow):
        super().setup_ui(MainWindow)
        self.EXTENSION = 'xml'
        MainWindow.setWindowTitle("Trace Viewer")
        self.centralwidget = QWidget(MainWindow)

        self.tableWidget = QTableView(self.centralwidget)
        self.tableWidget.setGeometry(QRect(0, 0, 3000, 600))
        self.tableWidget.horizontalHeader().setSectionResizeMode(
            QHeaderView.ResizeMode.ResizeToContents)

        self.horizontalLayout = QHBoxLayout(self.centralwidget)
        self.verticalLayout = QVBoxLayout()
        self.horizontalLayout.addLayout(self.verticalLayout)
        self.verticalLayout.addWidget(self.tableWidget)

        self.setup_toolbars(MainWindow)

        MainWindow.setCentralWidget(self.centralwidget)

    def setup_toolbars(self, MainWindow):
        self.labelNameSearch = QLabel(self.centralwidget)
        self.labelNameSearch.setText("Name: ")
        self.lineEditNameSearch = LineEdit()
        MainWindow.setCentralWidget(self.lineEditNameSearch)

        self.labelFilter = QLabel(self.centralwidget)
        self.labelFilter.setText("Filter: ")
        self.comboBox = QComboBox()
        self.comboBox.setMinimumWidth(100)

        self.iconEdit = QIcon.fromTheme('emblem-system')
        self.actionEdit = QAction(self.iconEdit, "edit", self.centralwidget)

        self.iconCancel = QIcon.fromTheme('edit-clear')
        self.actionCancel = QAction(
            self.iconCancel, "cancel", self.centralwidget)

        self.topToolBar.addWidget(self.labelNameSearch)
        self.topToolBar.addWidget(self.lineEditNameSearch)
        self.topToolBar.addSeparator()
        self.topToolBar.addWidget(self.labelFilter)
        self.topToolBar.addWidget(self.comboBox)
        self.topToolBar.addAction(self.actionEdit)
        self.topToolBar.addAction(self.actionCancel)
        self.topToolBar.addSeparator()

        self.labelTotalSteps = QLabel(self.centralwidget)
        self.bottomToolBar.addWidget(self.labelTotalSteps)

        self.setup_info()

    def setup_info(self):
        self.iconInfo = QIcon.fromTheme('dialog-information')
        self.actionInfo = QAction(self.iconInfo, "info", None)
        self.topToolBar.addAction(self.actionInfo)
        self.actionInfo.triggered.connect(lambda: self.show_info())
    def show_info(self):
        self.dialog = QDialog()
        self.popup = Ui_Info(self.dialog)
        self.popup.setup_ui(self)

        self.dialog.show()


class Ui_Info(object):

    def __init__(self, Dialog):
        self.Dialog = Dialog
        self.captions = {"color: black; background-color: white; border: 1px solid black;": ['engine.rpm','Name is a state variable'],
                      "color: blue; background-color: white; border: 1px solid black;": ['do_delta', 'Name is an input variable'],
                      "color: green; background-color: white; border: 1px solid black;": ['delta','Name is a combinatorial variable'],
                      "color: black; background-color: gray; border: 1px solid black;": ['step 3','Each step label is gray by default'],
                      "color: black; background-color: red; border: 1px solid black;": ['step 1', 'If step is a loop, it\'s displayed in red'],
                      "color: black; background-color: lightpink; border: 1px solid black;": ['delta', 'Name is a timed variable'],
                      "color: black; background-color: 	firebrick; border: 1px solid black;": ['fork[0]', 'Name is a fault variable']}

    def setup_ui(self, Dialog):
        self.Dialog.resize(350, 300)
        self.Dialog.setWindowTitle("Info")

        self.gridLayout = QGridLayout()
        for i in range(len(self.captions)):
            self.cell = QLabel(list(self.captions.values())[i][0])
            self.cell.setStyleSheet(list(self.captions.keys())[i])
            self.description = QLabel(self.Dialog)
            self.description.setText(list(self.captions.values())[i][1])
            self.gridLayout.addWidget(self.cell, i, 0, 1, 1)
            self.gridLayout.addWidget(self.description, i, 1, 1, 1)

        self.buttonBox = QDialogButtonBox(self.Dialog)
        self.buttonBox.setOrientation(Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.accepted.connect(self.accept)

        self.verticalLayout = QVBoxLayout(self.Dialog)
        self.verticalLayout.addLayout(self.gridLayout)
        self.verticalLayout.addWidget(self.buttonBox)

        QMetaObject.connectSlotsByName(self.Dialog)

    def accept(self):
        self.Dialog.close()


def process_options():
    parser = argparse.ArgumentParser(
        description='Trace Viewer'
    )
    parser.add_argument('-t','--trace-file', metavar="TRACE-FILE",
                        help='The trace file name')
    parser.add_argument('-v', '--version', action='version',
                        version='%(prog)s ' + '.'.join(map(str, PROG_VER)))
    parser.add_argument('-ft', '--failure-variables', metavar='FAILURES-FILE', help="The failure variables file name")
    parser.add_argument('-tv', '--timed-variable', metavar='TIMED-VAR', help="The timed variable name")
    args = parser.parse_args()
    return args


def run():
    args = process_options()

    # trace_fname = process_options()
    app = QApplication(sys.argv)
    app.setQuitOnLastWindowClosed(False)

    ui = Ui_TraceViewerMainWindow()
    main_window = ViewerMainWindow()
    ui.setup_ui(main_window)

    if args.trace_file is not None:
        trace = NuSMVTraceReader.parse(NuSMVTraceReader, args.trace_file,
                                       args.failure_variables, args.timed_variable)
        ui.load_table_data(trace)
        ui.update_ui(trace)
    main_window.show()
    sys.exit(app.exec_())


if __name__ == "__main__":
    run()
