
# %HEADER

from builtins import object
import logging
try:
    import xsap.utils.options as options
except ImportError:
    from compass import options # in case we are in COMPASS project
import os

## TODO: consider to add such file?
## import utils.misc
import functools

__all__ = ["xsap", "init", "get_level_from_string"]

xsap = logging.getLogger("Xsap")


# Allow multiple loggers using the same file
# See: http://mail.python.org/pipermail/tutor/2007-December/059158.html
class ManagedFileHandler(logging.FileHandler):
    """Subclass of logging.FileHandler to cope with multiple handlers
    targeting the same file"""

    """Map of open filenames to streams"""
    open_files = {}

    def __init__(self, filename, mode='a', encoding=None, delay=0):
        logging.FileHandler.__init__(self, filename, mode, encoding)

    def _open(self):
        # First check map of open streams
        if self.baseFilename in ManagedFileHandler.open_files:
            return ManagedFileHandler.open_files[self.baseFilename]
        # if not present, call super and store result in map
        stream = logging.FileHandler._open(self)
        ManagedFileHandler.open_files[self.baseFilename] = stream
        return stream

def init():
    """Initialises the logging system according to data in
    options. Note: options must be already set before calling it"""
    global xsap

    xsap = _XsapLogger()
    return

def xsap_logger_add_handler(filename):
    """Adds a file handler with the given filename to the xSAP logger"""
    new_logger = logging.FileHandler(filename, "w")
    new_logger.setLevel(get_level_from_string(options.LOGGING_SYS_LEVEL))
    new_formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    new_logger.setFormatter(new_formatter)
    xsap.logger.addHandler(new_logger)

def xsap_logger_remove_handler(filename):
    """Removes, if existing, the file handler with the given filename
    from the xSAP logger"""
    for h in xsap.logger.handlers:
        if 'baseFilename' in dir(h):
            if h.baseFilename.endswith(filename):
                xsap.logger.removeHandler(h)


def get_level_from_string(level_name):
    """Given a string, returns the corresponding logging
    level. Raise ValueError if bad string is passed"""
    if not level_name: return logging.INFO # default

    LEVELS = {'debug': logging.DEBUG,
              'info': logging.INFO,
              'warning': logging.WARNING,
              'error': logging.ERROR,
              'critical': logging.CRITICAL}

    if level_name not in LEVELS:
        raise ValueError("Logging level must be in '%s'" % ", ".join(list(LEVELS.keys())))

    return LEVELS[level_name]


def log_icd(func):
    """This is a decorator which can be used to mark functions
    as logged for the ICD compliance testing. It will result
	in a message to the testing logger when the function is used."""

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        testing.info(
            "Call to ICD function %(funcname)s." % {
            'funcname': func.__module__ + "." + func.__name__, }
        )
        return func(*args, **kwargs)
    return new_func



# -----------------------------------------------------------
class _Logger (object):
    def __init__(self, logger_name):

        self.logger = logging.getLogger(logger_name)
        self.fch = None

        if options.LOGGING_FNAME:
            logfname = os.path.join(options.TOP_LEVEL_DIR, options.LOGGING_FNAME)
        else:
            ## TODO: consider to add a misc.py file containing such functions?
            # logfname = utils.misc.get_out_fname("xsap.log")
            logfname = os.path.join(options.TOP_LEVEL_DIR, "xsap.log")

        # file handler (if log name)
        if logfname:
            fch = ManagedFileHandler(logfname, "w")
            fch.setLevel(get_level_from_string(options.LOGGING_SYS_LEVEL))
            formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
            fch.setFormatter(formatter)
            self.logger.addHandler(fch)
            self.fch = fch

        return
    pass # end of class
# -----------------------------------------------------------

class _XsapLogger (_Logger):

    user = logging.getLogger("Xsap.user")
    tfpg_behav  = logging.getLogger("Xsap.tfpg.behavioral_validation")
    tfpg_synth  = logging.getLogger("Xsap.tfpg.synthesis")
    tfpg_effect = logging.getLogger("Xsap.tfpg.effectiveness_validation")
    tfpg_tight  = logging.getLogger("Xsap.tfpg.tightening")
    tfpg_other  = logging.getLogger("Xsap.tfpg")

    def __init__(self):
        _Logger.__init__(self, "Xsap")

        # sets up the logger
        self.logger.setLevel(get_level_from_string(options.LOGGING_XSAP_LEVEL))

        #console handler
        ch = logging.StreamHandler()
        ch.setLevel(get_level_from_string(options.LOGGING_XSAP_CONSOLE_LEVEL))
        # Only print simple messages
        formatter = logging.Formatter("%(message)s")
        ch.setFormatter(formatter)
        self.logger.addHandler(ch)

        return

    def debug(self, msg, *args, **kwargs): self.logger.debug(msg, *args, **kwargs)
    def info(self, msg, *args, **kwargs): self.logger.info(msg, *args, **kwargs)
    def warning(self, msg, *args, **kwargs): self.logger.warning(msg, *args, **kwargs)
    def error(self, msg, *args, **kwargs): self.logger.error(msg, *args, **kwargs)
    def critical(self, msg, *args, **kwargs): self.logger.critical(msg, *args, **kwargs)
    pass # end of class
# -----------------------------------------------------------
