
import os
from xsap.utils.processes.process import ShellProcess, TextLineGrepper
from xsap.utils.processes.process import wait_for_condition, get_exec_name
import xsap.utils.misc.log as log
try:
    import xsap.utils.options as options
except ImportError:
    from compass import options # in case we are in COMPASS project


class NuSMVProcess(ShellProcess):
    """This class gives the basic functionalities in order to
    perform model checking"""

    # prefix for generated files
    _fname_prefix = "__nusmv_prefix__"

    def __init__(self, path=None, smv_fname=None, ord_fname=None,
                 args=[], stdout=None, stderr=None, dumpers=[],
                 dynamic_reorder=True, verbose_level=0,
                 boolean_conversion_uses_predicate_normalization=False):

        if ord_fname is None and smv_fname is not None:
            # search if there is one possible ordering available
            basename = os.path.splitext(smv_fname)[0]
            ord_fname = basename + ".ord"
            if not os.path.isfile(ord_fname):
                ord_fname = None

        # check that files exist
        for fname in (smv_fname, ord_fname):
            if fname and not os.path.isfile(fname):
                msg = "File not found: %s" % fname
                log.xsap.user.error(msg)
                raise IOError(msg)


        args = list(args)
        args += ["-pre", "cpp", "-int"]

        self._smv_fname = smv_fname
        self._ord_fname = ord_fname
        self._trace_counter = 0
        self._build_cmd = set()

        # path is different from None in case the function has been called
        # by the parent (i.e. nuXmv);
        self.__degraded = False
        if path is None:
            try:
                path = get_exec_name(options.NUSMV)
            except IOError:
                path = get_exec_name(options.DEGRADED_MODE)
                self.__degraded = True
                pass
        # else path is inherited from the parent

        super(NuSMVProcess, self).__init__(path, args, stdout, stderr, dumpers)

        # a few default commands sent:
        # 1. set verbose level
        # 2. exit immediately on any error. This is important and
        # used in how commands success is checked.
        # 2. enable dynamic BDD var order. By default it is advantageous
        self.send_cmd('set verbose_level %d'
                      % verbose_level)
        self.send_cmd("set on_failure_script_quits")
        if dynamic_reorder:
            self.send_cmd("dynamic_var_ordering -e sift")
        if boolean_conversion_uses_predicate_normalization:
            self.send_cmd("set boolean_conversion_uses_predicate_normalization 1")
        else:
            self.send_cmd("set boolean_conversion_uses_predicate_normalization 0")
        if self._smv_fname:
            self.send_cmd("set input_file '%s'" % self._smv_fname)
        if self._ord_fname:
            self.send_cmd("set input_order_file '%s'" % self._ord_fname)

    def is_degraded_mode(self):
        """Returns True if the process supports full operations"""
        return self.__degraded


    def wait_for_command_finish(self,
                                end_message="END OF NUSMV COMMAND EXECUTION ~",
                                tool="NuSMV"):
        # this messages is used to detect the end of FSAP command
        # execution.  it must be unique (not output by normal Xsap
        # run), without spaces at the ends and simple enough for python's re.
        return super(NuSMVProcess, self).wait_for_command_finish(end_message, tool)

    def _check_handler_errors(self, txt, tool="NuSMV"):
        super(NuSMVProcess, self)._check_handler_errors(txt, tool)

    def _log_returned_text(self, txt, tool="NuSMV"):
        super(NuSMVProcess, self)._log_returned_text(txt, tool, log.xsap.user)

    def reset(self):
        super(NuSMVProcess, self).reset()
        self.build = set()

    def _build(self, cmd):
        if cmd not in self._build_cmd:
            self.send_cmd(cmd)
            self.wait_for_command_finish("DONE_%s" % cmd.upper())
            self._build_cmd.add(cmd)

    def build_bdd_fsm(self):
        self._build("go")

    def build_bmc_fsm(self):
        self._build("go_bmc")

    def build_hierarchy(self):
        """Builds the hierarchy if needed. Returns True iff
        everthing went ok"""

        if (not set(("go", "go_bmc")) &
                self._build_cmd):
            self.read_model()
            self.send_cmd("flatten_hierarchy")
            self.send_cmd("encode_variables")

            return self.wait_for_command_finish()

        return True


    def read_model(self, model=None):
        cmd = "read_model"
        if model:
            cmd += ' -i "%s"' % model
        self.send_cmd(cmd)
        return self.wait_for_command_finish(end_message="DONE READ_MODEL")


    def write_flat_model(self, out_fname):
        """Writes the flattened model to the specified file"""
        self.send_cmd('write_flat_model -o "%s"' % out_fname)
        return

    def model_check_bdd_ctl(self, formula, generate_trace=True):
        """Model checks a formula (string), and returns: (True,
        None) if the formula is true, (False, fname) if
        false. fname is the filename of a xml trace containing the
        countexample. Returns (None, None) if no result could be
        found"""
        self.build_bdd_fsm()
        if not self.is_alive():
            log.xsap.user.debug(" >> error: building bdd-fsm failed")
        cmd = 'check_ctlspec -p "%s"' % formula

        if not generate_trace:
            #disable global trace generation before the check
            env_var_cte = self.get_env_variable_value('counter_examples')
            self.send_cmd("unset counter_examples")

        self.send_cmd(cmd)

        if not generate_trace:
            #undo the disabling of global trace generation after the check
            self.send_cmd("set counter_examples %s" % env_var_cte)

        self.show_property()

        # this grepper checks for TRUE/FALSE, the other for problems in
        # trace generation
        grep = TextLineGrepper(
            ".*\s+is (false|true)")
        m, txt = grep.match(self)  # wait for result!
        if m is None:
            log.xsap.user.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return ('error', None)

        self._check_handler_errors(txt)

        # processes results
        if m.groups()[0] == "true":
            return ('true', None)

        return ('false', self._create_trace())

    def model_check_bdd_ltl(self, formula, prop_idx=None, generate_trace=True):
        """Model checks a formula (string), and returns: (True,
        None) if the formula is true, (False, fname) if
        false. fname is the filename of a xml trace containing the
        countexample. Returns (None, None) if no result could be
        found"""
        assert prop_idx is None or formula is None, "Cannot specify both prop_idx and formula"
        self.build_bdd_fsm()
        if not self.is_alive():
            log.xsap.user.debug(" >> error: building bdd-fsm failed")

        if formula:
            cmd = 'check_ltlspec -p "%s"' % formula
        elif prop_idx is not None:
            cmd = 'check_ltlspec -n %d' % prop_idx
        else:
            cmd = 'check_ltlspec'

        if not generate_trace:
            #disable global trace generation before the check
            env_var_cte = self.get_env_variable_value('counter_examples')
            self.send_cmd("unset counter_examples")

        self.send_cmd(cmd)

        if not generate_trace:
            #undo the disabling of global trace generation after the check
            self.send_cmd("set counter_examples %s" % env_var_cte)

        grep = TextLineGrepper(
            ".*\s+is (false|true)")
        m, txt = grep.match(self)  # wait for result!
        if m is None:
            log.xsap.user.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return ('error', None)
        self._check_handler_errors(txt)

        # processes results
        if m.groups()[0] == "true":
            return ('true', None)

        return ('false', self._create_trace())

    def model_check_bmc_ltl(self, formula, k, use_inc=True, prop_idx=None, generate_trace=True):
        """Model checks a formula (string), and returns: (True,
        None) if the formula is true, (False, fname) if
        false. fname is the filename of a xml trace containing the
        countexample. Returns (None, None) if no result could be
        found"""
        self.build_bmc_fsm()
        if not self.is_alive():
            log.xsap.user.debug(" >> error: building bmc-fsm failed")

        compl_opt = "-c" if options.SBMC_IS_COMPLETE else ""
        if options.BMC_USE_SBMC:
            if formula:
                cmd = ('check_ltlspec_sbmc_inc %s -p "%s" -k %d' %
                       (compl_opt, formula, k))
            elif prop_idx is not None:
                cmd = ('check_ltlspec_sbmc_inc %s -n %d -k %d' %
                       (compl_opt, prop_idx, k))
            else:
                cmd = ('check_ltlspec_sbmc_inc %s -k %d' %
                       (compl_opt, k))
        else:
            if use_inc:
                if formula:
                    cmd = ('check_ltlspec_bmc_inc -p "%s" -k %d' %
                           (formula, k))
                elif prop_idx is not None:
                    cmd = ('check_ltlspec_bmc_inc %s -n %d -k %d' %
                           (compl_opt, prop_idx, k))
                else:
                    cmd = 'check_ltlspec_bmc_inc -k %d' % k
            elif prop_idx is not None:
                cmd = ('check_ltlspec_bmc %s -n %d -k %d' %
                       (compl_opt, prop_idx, k))
            else:
                cmd = ('check_ltlspec_bmc -p "%s" -k %d' %
                       (formula, k))

        if not generate_trace:
            #disable global trace generation before the check
            env_var_cte = self.get_env_variable_value('counter_examples')
            self.send_cmd("unset counter_examples")

        self.send_cmd(cmd)
        # This is for detecting no result found
        self.send_cmd("echo DONE_BMC")

        if not generate_trace:
            #undo the disabling of global trace generation after the check
            self.send_cmd("set counter_examples %s" % env_var_cte)

        result_patterns = [
            "is false",
            "is true",
            "no counterexample found with bound %d" % k
            ]
        grep = TextLineGrepper(r".*\s+(%s)" % '|'.join(result_patterns), "DONE_BMC")
        m, txt = grep.match(self)  # wait for result!
        if m is None or grep.is_match_unwanted(m):
            log.xsap.user.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return ('error', None)

        self._check_handler_errors(txt)

        if m.groups()[0] == "is true":
            return ('true', None)

        if m.groups()[0] == "is false":
            return ('false', self._create_trace())

        return ('unknown', None)

    def model_check_bmc_invar(self, formula, k, generate_trace=True):
        """Model checks a formula (string), and returns: (True,
        None) if the formula is true, (False, fname) if
        false. fname is the filename of a xml trace containing the
        countexample. Returns (None, None) if no result could be
        found"""

        self.build_bmc_fsm()
        if not self.is_alive():
            log.xsap.user.debug(" >> error: building bmc-fsm failed")

        cmd = ('check_invar_bmc_inc -k %d -a falsification' % k)

        if formula:
            cmd += (' -p "%s" ' %  formula)

        # TODO: to be moved to the constructor
        # cmd = self.add_show_properties_cmd(cmd)
        # self._create_cmd_file(cmd)

        if not generate_trace:
            #disable global trace generation before the check
            env_var_cte = self.get_env_variable_value('counter_examples')
            self.send_cmd("unset counter_examples")

        self.send_cmd(cmd)
        # This is for detecting no result found
        self.send_cmd("echo DONE_BMC")

        if not generate_trace:
            #undo the disabling of global trace generation after the check
            self.send_cmd("set counter_examples %s" % env_var_cte)

        result_patterns = [
            "is false",
            "is true",
            "no proof or counterexample found with bound %d" % k
            ]
        grep = TextLineGrepper(r".*\s+(%s)" % '|'.join(result_patterns), "DONE_BMC")
        m, txt = grep.match(self)  # wait for result!
        if m is None or grep.is_match_unwanted(m):
            log.xsap.user.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return ('error', None)

        self._check_handler_errors(txt)

        if m.groups()[0] == "is true":
            return ('true', None)
        if m.groups()[0] == "is false":
            return ('false', self._create_trace())

        return ('unknown', None)

    def _create_trace(self):
        """Invokes the tool to create a xml trace file out of latest
        created trace. Returns the filename containing the trace.
        Raises a utils.misc.exceptions.TraceException if an
        error occurs. Returns None if the trace was not generated
        due to an unknown reason."""
        # searches for a good name for trace
        basename = os.path.splitext(self._smv_fname)[0]
        if self._fname_prefix:
            basename = self._fname_prefix + os.path.splitext(os.path.basename(self._smv_fname))[0]
        while True:
            self._trace_counter += 1
            fname = "%s_trace%d.xml" % (basename, self._trace_counter)
            if not os.path.isfile(fname):
                break

        # creates the trace fname
        cmd = 'show_traces -p 4 -o "%s"' % fname
        self.send_cmd(cmd)

        self.send_cmd('echo "TRACE_DONE"')

        # wait for the file to be generated:
        grep = TextLineGrepper(["There are no traces currently available",
                                "TRACE_DONE"],
                               "Division by zero")
        m, txt = grep.match(self)

        if m is None:
            log.xsap.user.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return None

        if "There are no traces currently available" in txt:
            return None

        if grep.is_match_unwanted(m):
            raise TraceException(m.string)

        res = wait_for_condition(os.path.isfile, [fname], 5)
        assert res, "A trace file was expected to be generated"

        return fname

    def get_traces_num(self):
        """Returns the number (integer) of traces currently loaded
        in the trace database. Returns None if an error occurs"""
        grep = TextLineGrepper(
            ["There (is|are) (\d+|no) traces? currently available."])
        cmd = 'show_traces -t'
        self.send_cmd(cmd)
        m, txt = grep.match(self)
        if m is None:
            log.xsap.user.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return None

        return (0
                if m.group(2) == "no"
                else int(m.group(2)))

    def read_trace(self, fname, build_model=True):
        """Reads given trace from file, and returns the numeric
        index (as integer) which was used to store the trace
        internally. The index can later be used to pick states from
        traces. If an error occurs, returns None"""
        if not self.build_hierarchy():
            return None

        if build_model:
            cmd = "build_flat_model"
            self.send_cmd(cmd)

        cmd = 'read_trace "%s"' % fname
        self.send_cmd(cmd)
        grep = TextLineGrepper("Trace is stored at (\d+) index",
                               "Division by zero")
        m, txt = grep.match(self)
        if m is None:
            log.xsap.user.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return None

        if grep.is_match_unwanted(m):
            raise TraceException(m.string)

        return int(m.group(1))

    def show_dependencies(self, bound, expr):
        """Invokes the command 'show_dependencies'"""

        cmd = 'show_dependencies'
        if bound is not None:
            assert isinstance(bound, int)
            assert bound >= 0
            cmd += ' -k %d' % bound
        cmd += ' -e "%s"' % expr
        self.send_cmd(cmd)
        return

    def show_property(self):
        if self._fname_prefix:
            cmd = 'show_property -F xml -o "%s%s"' % (self._fname_prefix, options.SHOW_PROPERTIES_FILE_NAME)
        else:
            cmd = 'show_property -F xml -o "%s"' % options.SHOW_PROPERTIES_FILE_NAME
        self.send_cmd(cmd)
        return

    def check_fsm_deadlocks(self):
        """Checks the BDD fsm for deadlocks. Returns True if the
        FSM is deadlock-free, False if it contains deadlocks, or
        None if an error occurs."""
        self.build_bdd_fsm()

        cmd = 'check_fsm'
        self.send_cmd(cmd)
        grep = TextLineGrepper(
            ["The transition relation is total: No deadlock state exists",
             "However, all the states without successors are"],
            ["The transition relation is not deadlock-free."],)

        m, txt = grep.match(self)
        if m is None:
            log.system.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return None

        return (False
                if grep.is_match_unwanted(m)
                else True)

    def goto_state_bdd(self, trace_index, step_index):
        """Selects the given trace and step index to be the next
        trace used by the simulation (BDD-based). Creates a new
        trace, and returns the index (integer) of the newly created
        trace."""
        self.build_bdd_fsm()
        self.send_cmd('goto_state "%d.%d"' % (trace_index, step_index))
        self.wait_for_command_finish()
        return self.get_traces_num()

    def goto_state_bmc(self, trace_index, step_index):
        """Selects the given trace and step index to be the next
        trace used by the simulation (BMC-based). Creates a new
        trace, and returns the index (integer) of the newly created
        trace."""
        self.build_bmc_fsm()
        self.send_cmd('bmc_pick_state -s "%d.%d"' % (trace_index, step_index))
        self.wait_for_command_finish()
        return self.get_traces_num()

    def pick_state_random_bmc(self):
        """
        """
        self.build_bmc_fsm()

        grep = TextLineGrepper(
            ["PICK_DONE"],
            ["The set of initial states is EMPTY"])

        cmd = 'bmc_pick_state -r; echo "PICK_DONE"'
        self.send_cmd(cmd)
        m, txt = grep.match(self)
        if m is None:
            log.xsap.user.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return None

        if grep.is_match_unwanted(m):
            return None

        return self.get_traces_num()

    def pick_state_random_bdd(self, constraints):
        """Creates a 1-step trace for starting a simulation (using
        BDDs). The state is randomly selected taking into account
        the list of constraints (list of strings). Returns the index of
        the created trace, or None if no trace was created."""

        self.build_bdd_fsm()

        grep = TextLineGrepper(
            ["PICK_DONE"],
            ["The set of initial states is EMPTY"])

        constr_cmd = ('-c "%s"' % " & ".join(("(%s)" % c for c in constraints))
                      if constraints
                      else "")

        cmd = 'pick_state -r %s; echo "PICK_DONE"' % constr_cmd
        self.send_cmd(cmd)
        m, txt = grep.match(self)
        if m is None:
            log.xsap.user.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return None

        if grep.is_match_unwanted(m):
            return None

        return self.get_traces_num()



    def simulate_random(self, constraints, length):
        """
        """
        idx = self.simulate_random_bdd(constraints, length)
        if idx is None:
            return None
        idx = self.goto_state_bmc(idx, -1)

        return self._create_trace()


    def simulate_random_bdd(self, constraints, length):
        """Constraints is a list of propositional formulae
        (strings). Returns the index of the created trace, or
        None if a problem occurs"""
        self.build_bdd_fsm()

        grep = TextLineGrepper(
            ["SIM_DONE"],
            ["No future state exists: trace not built",
             'No current state set. Use the "pick_state" command.'])

        constr_cmd = ('-t "%s"' % " & ".join(("(%s)" % c for c in constraints))
                      if constraints
                      else "")

        cmd = 'simulate -r %s -k %d; echo SIM_DONE' % (constr_cmd, length)
        self.send_cmd(cmd)
        m, txt = grep.match(self)
        if m is None:
            log.xsap.user.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return None

        if grep.is_match_unwanted(m):
            return None

        return self.get_traces_num()

    def simulate_check_feasability_bmc(self, constraints):
        """Constraints is a list of next formulae
        (strings). Returns a list of boolean values telling
        positionally which constraints are true or false."""

        res = []
        if constraints:
            self.build_bmc_fsm()

            msg = "SIM_FEAS:"
            constr_cmd = " ".join(('-c "%s"' % c for c in constraints))

            grep = TextLineGrepper("%s\s*([01]+)" % msg)
            cmd = (
                'echo -n %s; bmc_simulate_check_feasible_constraints -q %s' %
                (msg, constr_cmd))
            self.send_cmd(cmd)
            m, txt = grep.match(self)
            if m is None:
                log.xsap.user.error("Process failed on command:\n> " + cmd)
                self._log_returned_text(txt)
                return None

            res = [True if c == "1" else False for c in m.group(1)]

        return res

    def get_properties_count(self):
        """Returns the number of properties in the property manager"""
        cmd = 'show_property -s'
        self.send_cmd(cmd)
        grep = TextLineGrepper("Current number of stored properties: (\d+)")
        m, txt = grep.match(self)  # wait for result!
        prop_cnt = int(m.groups()[0])

        return prop_cnt

    def get_env_variable_value(self, env_var_name):
        """Returns the value of the environment variable env_var_name.
           Returns an exception in case the environment variable does not exist"""

        if not self.is_alive():
            # For Windows: it does not kills automatically also the
            # threads created by this process
            self.kill_and_close_streams()
            raise Exception("Process failed")


        msg= "ENV_VAR_VALUE:"
        cmd= 'echo %s$%s' % (msg, env_var_name)
        self.send_cmd(cmd)
        grep = TextLineGrepper('%s\s*(.*)'% msg)
        m, txt = grep.match(self)

        if m is None:
            raise Exception("Process failed on command:\n> " + cmd)

        if m.groups()[0] == '$'+ env_var_name:
            raise Exception("Environment variable " + '$'+ env_var_name + " does not exist\n")

        env_var_value = m.groups()[0]

        return env_var_value


    # TODO: _create_cmd_file(self, cmd)
    # TODO: def remove_properties_file()
    # TODO: def get_properties_file()

    # ===
    # TODO: the following commands are offered by NuSMV but there is
    #       no related interface
    # ===
    # add_property
    # alias
    # bmc_inc_simulate
    # bmc_setup
    # bmc_simulate
    # build_boolean_model
    # build_flat_model
    # build_module
    # check_compute
    # check_invar
    # check_invar_bmc
    # check_ltlspec_bmc_onepb
    # check_pslspec
    # check_pslspec_bmc
    # check_pslspec_bmc_inc
    # check_pslspec_sbmc
    # check_pslspec_sbmc_inc
    # check_sexp2bdd_cache
    # compass_gen_sigref
    # compute
    # compute_reachable
    # convert_property_to_invar
    # dump_expr
    # dump_fsm
    # execute_partial_traces
    # execute_traces
    # gen_invar_bmc
    # gen_ltlspec_bmc
    # gen_ltlspec_bmc_onepb
    # gen_ltlspec_sbmc
    # get_internal_status
    # hrc_dump_model
    # hrc_write_model
    # process_model
    # set_bdd_parameters
    # show_plugins
    # show_vars
    # write_boolean_model
    # write_coi_model
    # write_flat_model_udg
    # write_order
