
from builtins import zip
from builtins import range
import os
import re
from xsap.utils.processes.nusmv_process import NuSMVProcess
from xsap.utils.processes.process import get_exec_name, TextLineGrepper
from xsap.utils.processes.process import get_exec_name
from xsap.utils.misc.exceptions import ProcessInternalException
import xsap.utils.misc.log as log
try:
    import xsap.utils.options as options
except ImportError:
    from compass import options # in case we are in COMPASS project


class nuXmvProcess(NuSMVProcess):
    """This class gives the basic functionalities in order to
    perform model checking"""

    # these are literals used in is_state_zeno result
    (ZENO, NON_ZENO, UNREACHABLE, UNKNOWN_NOT_FOUND,
     UNKNOWN_NOT_PROVED, ERROR) = list(range(6))

    # these are literals used in check_clock_divergence
    (BOUNDED, UNBOUNDED, UNKNOWN) = list(range(3))

     # prefix for generated files
    _fname_prefix = "__nuxmv_prefix__"

    def __init__(self, path=None, smv_fname=None, ord_fname=None, need_msat=False,
                 args=[], stdout=None, stderr=None, dumpers=[],
                 dynamic_reorder=True, verbose_level=0,
                 boolean_conversion_uses_predicate_normalization=False):

        self._need_msat = need_msat

        # path is different from None in case the function has been called
        # by the parent (i.e. xSAP)
        self.__degraded = False
        if path is None:
            try:
                path = get_exec_name(options.NUXMV)
            except IOError:
                path = get_exec_name(options.DEGRADED_MODE)
                self.__degraded = True
                pass
        # else path is inherited from the parent

        super(nuXmvProcess, self).__init__(
            path, smv_fname, ord_fname, args, stdout, stderr,
            dumpers, dynamic_reorder, verbose_level,
            boolean_conversion_uses_predicate_normalization)
        return


    def wait_for_command_finish(self,
                                end_message="END OF NUXMV COMMAND EXECUTION ~",
                                tool="nuXmv"):
        # this messages is used to detect the end of FSAP command
        # execution.  it must be unique (not output by normal Xsap
        # run), without spaces at the ends and simple enough for python's re.
        return super(nuXmvProcess, self).wait_for_command_finish(end_message, tool)

    def _check_handler_errors(self, txt, tool="nuXmv"):
        super(nuXmvProcess, self)._check_handler_errors(txt, tool)

    def _log_returned_text(self, txt, tool="nuXmv"):
        super(nuXmvProcess, self)._log_returned_text(txt, tool)


    def needs_msat(self):
        """Returns True if current xsap process needs mathsat
        (deals with infinite domains)"""
        return self._need_msat

    def build_bdd_fsm(self):
        assert not self._need_msat, "Only MATHSAT can be used"
        super(nuXmvProcess, self).build_bdd_fsm()
        return

    def build_bmc_fsm(self):
        assert not self._need_msat, "Only MATHSAT can be used"
        super(nuXmvProcess, self).build_bmc_fsm()
        return

    def build_msat_fsm(self):
        self._build("go_msat")


    @classmethod
    def zeno_result_to_string(cls, res):
        _map = dict(list(zip(
            [cls.ZENO, cls.NON_ZENO, cls.UNREACHABLE, cls.UNKNOWN_NOT_FOUND,
             cls.UNKNOWN_NOT_PROVED, cls.ERROR],
            ["ZENO", "NON-ZENO", "UNREACHABLE", "NOT_FOUND",
             "NOT_PROVED", "ERROR"])))
        return _map[res]

    @classmethod
    def clock_diverge_result_to_string(cls, res):
        _map = dict(list(zip(
            [cls.BOUNDED, cls.UNBOUNDED, cls.UNKNOWN, cls.ERROR],
            ["BOUNDED", "UNBOUNDED", "UNKNOWN", "ERROR"])))
        return _map[res]


    def model_check_bmc_ltl(self, formula, k, use_inc=True, prop_idx=None, generate_trace=True):
        """Model checks a formula (string), and returns: (True,
        None) if the formula is true, (False, fname) if
        false. fname is the filename of a xml trace containing the
        countexample. Returns (None, None) if no result could be
        found"""
        assert prop_idx is None or formula is None, "Cannot specify both prop_idx and formula"
        # Use MATHSAT if needed
        if self._need_msat:
            return self.model_check_msat_bmc_ltl(formula, k, prop_idx=prop_idx, generate_trace=generate_trace)

        # else use BMC
        return super(nuXmvProcess, self).model_check_bmc_ltl(formula, k,
                                                             use_inc,
                                                             prop_idx=prop_idx,
                                                             generate_trace=generate_trace)


    def model_check_bmc_invar(self, formula, k, generate_trace=True):
        """Model checks a formula (string), and returns: (True,
        None) if the formula is true, (False, fname) if
        false. fname is the filename of a xml trace containing the
        countexample. Returns (None, None) if no result could be
        found"""

        # Use MATHSAT if needed
        if self._need_msat:
            return self.model_check_msat_bmc_invar(formula, k, generate_trace)

        # else use BMC
        return super(nuXmvProcess, self).model_check_bmc_invar(formula, k, generate_trace)

    def model_check_ic3_invar(self, formula, k, generate_trace=True):
        """Model checks a formula (string), and returns: (True,
        None) if the formula is true, (False, fname) if
        false. fname is the filename of a xml trace containing the
        countexample. Returns (None, None) if no result could be
        found"""

        if self._need_msat:
            self.build_msat_fsm()
            if not self.is_alive():
                log.xsap.user.debug(" >> error: building msat-fsm failed")

            cmd = ('check_invar_ic3 -i -k %d' % k)

            if formula:
                cmd += (' -p "%s" ' %  formula)
        else:
            self.build_bmc_fsm()
            if not self.is_alive():
                log.xsap.user.debug(" >> error: building bmc-fsm failed")

            cmd = ('check_invar_ic3 -k %d' % k)

            if formula:
                cmd += (' -p "%s" ' %  formula)

        # TODO: to be moved to the constructor
        # cmd = self.add_show_properties_cmd(cmd)
        # self._create_cmd_file(cmd)

        if not generate_trace:
            #disable global trace generation before the check
            env_var_cte = self.get_env_variable_value('counter_examples')
            self.send_cmd("unset counter_examples")

        self.send_cmd(cmd)
        # This is for detecting no result found
        self.send_cmd("echo DONE_BMC")

        if not generate_trace:
            #undo the disabling of global trace generation after the check
            self.send_cmd("set counter_examples %s" % env_var_cte)

        result_patterns = [
            "is false",
            "is true",
            ]
        grep = TextLineGrepper(r".*\s+(%s)" % '|'.join(result_patterns), "DONE_BMC")
        m, txt = grep.match(self)  # wait for result!
        if m is None or grep.is_match_unwanted(m):
            log.xsap.user.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return ('error', None)

        self._check_handler_errors(txt)

        if m.groups()[0] == "is true":
            return ('true', None)

        if m.groups()[0] == "is false":
            return ('false', self._create_trace())

        return ('unknown', None)


    def model_check_klive_ltl(self, formula, ic3_bound,
                              klive_bound=None,
                              force_smt_use=True,
                              prop_idx=None,
                              generate_trace=True):
        """Model checks a formula (string), and returns: (True,
        None) if the formula is true, (False, fname) if
        false. fname is the filename of a xml trace containing the
        countexample. Returns (None, None) if no result could be
        found
        """
        assert prop_idx is None or formula is None, "Cannot specify both prop_idx and formula"
        if force_smt_use:
            self.build_msat_fsm()
            if not self.is_alive():
                log.xsap.user.debug(" >> error: building msat-fsm failed")
        else:
            self.build_bmc_fsm()
            if not self.is_alive():
                log.xsap.user.debug(" >> error: building bmc-fsm failed")

        cmd = 'check_ltlspec_ic3'

        if force_smt_use:
            cmd += ' -i'
        if formula:
            cmd += ' -p "%s"' % formula
        if prop_idx is not None:
            cmd += ' -n "%d"' % prop_idx
        if ic3_bound:
            cmd += ' -k %d' % ic3_bound
        if klive_bound:
            cmd += ' -l %d' % klive_bound
        # TODO: to be moved to the constructor
        # cmd = self.add_show_properties_cmd(cmd)

        # self._create_cmd_file(cmd)

        if not generate_trace:
            #disable global trace generation before the check
            env_var_cte = self.get_env_variable_value('counter_examples')
            self.send_cmd("unset counter_examples")

        self.send_cmd(cmd)
        self.send_cmd("echo DONE KLIVE")  # This is for detecting no result found

        if not generate_trace:
            #undo the disabling of global trace generation after the check
            self.send_cmd("set counter_examples %s" % env_var_cte)

        result_patterns = [
            "is false",
            "is true",
            "is unknown"
            ]
        grep = TextLineGrepper(r".*\s+(%s)" % '|'.join(result_patterns), "DONE_KLIVE")
        m, txt = grep.match(self)  # wait for result!

        # processes results
        if m is None or grep.is_match_unwanted(m):
            log.xsap.user.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return ('error', None)

        self._check_handler_errors(txt)

        if m.groups()[0] == "is true":
            return ('true', None)

        if m.groups()[0] == "is false":
            return ('false', self._create_trace())

        return ('unknown', None)


    def model_check_msat_bmc_ltl(self, formula, k, prop_idx=None, generate_trace=True):
        """Model checks a formula (string), and returns: (True,
        None) if the formula is true, (False, fname) if
        false. fname is the filename of a xml trace containing the
        countexample. Returns (None, None) if no result could be
        found"""
        assert prop_idx is None or formula is None, "Cannot specify both prop_idx and formula"
        self.build_msat_fsm()
        if not self.is_alive():
            log.xsap.user.debug(" >> error: building msat-fsm failed")

        if options.BMC_USE_SBMC:
            if formula:
                cmd = 'msat_check_ltlspec_sbmc_inc -p "%s" -k %d' % (formula, k)
            elif prop_idx is not None:
                cmd = 'msat_check_ltlspec_sbmc_inc -n %d -k %d' % (prop_idx, k)
            else:
                cmd = 'msat_check_ltlspec_sbmc_inc -k %d' % k
        else:
            if formula:
                cmd = 'msat_check_ltlspec_bmc -p "%s" -k %d' % (formula, k)
            elif prop_idx is not None:
                cmd = 'msat_check_ltlspec_bmc -n %d -k %d' % (prop_idx, k)
            else:
                cmd = 'msat_check_ltlspec_bmc -k %d' % k

        if not generate_trace:
            #disable global trace generation before the check
            env_var_cte = self.get_env_variable_value('counter_examples')
            self.send_cmd("unset counter_examples")

        self.send_cmd(cmd)
        self.send_cmd("echo DONE BMC")  # This is for detecting no result found

        if not generate_trace:
            self.send_cmd("set counter_examples %s" % env_var_cte)

        result_patterns = [
            "is false",
            "is true",
            "terminating with bound %d" % k,
            "no counterexample found with bound %d" %k
            ]
        grep = TextLineGrepper(r".*\s+(%s)" % '|'.join(result_patterns),
                               "DONE_BMC")
        m, txt = grep.match(self)  # wait for result!

        # processes results
        if m is None or grep.is_match_unwanted(m):
            log.xsap.user.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return ('error', None)

        self._check_handler_errors(txt)

        if m.groups()[0] == "is true":
            return ('true', None)

        if m.groups()[0] == "is false":
            return ('false', self._create_trace())

        return ('unknown', None)


    def model_check_msat_bmc_invar(self, formula, k,
                                   use_inc=False,
                                   generate_trace=True):
        """Model checks a formula (string), and returns: (True,
        None) if the formula is true, (False, fname) if
        false. fname is the filename of a xml trace containing the
        countexample. Returns (None, None) if no result could be
        found"""
        self.build_msat_fsm()
        if not self.is_alive():
            log.xsap.user.debug(" >> error: building msat-fsm failed")

        cmd = 'msat_check_invar_bmc -k %d -a falsification' % (k)

        if use_inc:
            cmd += ' -i'
        if formula:
            cmd += ' -p "%s"' % (formula)

        # TODO: to be moved to the constructor
        # cmd = self.add_show_properties_cmd(cmd)
        # self._create_cmd_file(cmd)

        if not generate_trace:
            #disable global trace generation before the check
            env_var_cte = self.get_env_variable_value('counter_examples')
            self.send_cmd("unset counter_examples")

        self.send_cmd(cmd)
        self.send_cmd("echo DONE BMC")  # This is for detecting no result found

        if not generate_trace:
            #undo the disabling of global trace generation after the check
            self.send_cmd("set counter_examples %s" % env_var_cte)

        result_patterns = [
            "is false",
            "is true",
            "no counterexample found for invariant .* up to %d" % k,
            ]
        grep = TextLineGrepper(r".*\s+(%s)" % '|'.join(result_patterns), "DONE_BMC")
        m, txt = grep.match(self)  # wait for result!

        # processes results
        if m is None or grep.is_match_unwanted(m):
            log.xsap.user.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return ('error', None)

        self._check_handler_errors(txt)

        if m.groups()[0] == "is true":
            return ('true', None)

        if m.groups()[0] == "is false":
            return ('false', self._create_trace())

        return ('unknown', None)


    def goto_state(self, trace_index, step_index):
        """Picks state from existing trace using BDD/BMC or MSAT, and
        creates a new trace. Returns the filename of the
        newly created trace."""
        # Use MATHSAT if needed

        # TODO: temporary hack: handle the case in which trace_index is invalid
        if trace_index is None or trace_index <= 0:
            raise ProcessInternalException(
                self.name,
                "Command goto-state received an invalid trace. "
                "This may be due to a known issue in nusmv. "
                "However, please report this bug.")

        if self._need_msat:
            idx = self.goto_state_msat(trace_index, step_index)
        else:
            idx = self.goto_state_bdd(trace_index, step_index)
            if idx is None:
                return None
            idx = self.goto_state_bmc(idx, -1)

        assert self.get_traces_num() == idx, \
            "The generated trace is not the last one as expected"
        res = self._create_trace()
        if res is None:
            raise ProcessInternalException(
                self.name,
                "Command goto-state seems to be not working correctly."
                "This may be due to a known issue in nusmv. "
                "However, please report this bug.")
        return res


    def pick_state_random(self, constraints):
        """Picks state randomly using BDD or MSAT.
        @param constraints is a list of smv constraints (list of
        strings or list of tuples <logic, str>)
        @return the file name of the created trace, or
        None if no trace was created."""

        # if it is a list of tuples, get smv (2nd el of tuple)
        if len(constraints) > 0:
            if isinstance(constraints[0], tuple):
                assert isinstance(constraints[0][1], str)
                constraints = [c[1] for c in constraints]

        # Use MATHSAT if needed
        if self._need_msat:
            idx = self.pick_state_random_msat(constraints)
        else:
            idx = self.pick_state_random_bdd(constraints)

        if idx is None:
            return None
        # picks also in BMC if possible
        if not self._need_msat:
            idx = self.goto_state_bmc(idx, 1)

        return self._create_trace()


    def simulate_random(self, constraints, length):
        """Simulate using BDD or MSAT. The state is randomly
        selected taking into account the list of SMV constraints
        (list of strings). Returns the name of the created xml trace, or
        None if a problem occurs"""

        # Use MATHSAT if needed
        if self._need_msat:
            idx = self.simulate_random_msat(constraints, length)
            return self._create_trace()
        else:
            return super(nuXmvProcess, self).simulate_random(constraints, length)

    def simulate_check_feasability(self, constraints):
        """Constraints is a list of next formulae
        (strings). Returns a list of boolean values telling
        positionally which constraints are true or false."""
        # Use MATHSAT if needed
        if self._need_msat:
            return self._simulate_check_feasability_msat(constraints)
        else:
            return super(nuXmvProcess, self).simulate_check_feasability_bmc(constraints)


    def _check_pattern_aux(self, pattern_name, specs,
                           time_bound, time_var,
                           max_search_depth, max_klive_bound):
        """Service for all time property checking functions, as all of them
        share a common base code. specs is a list of one or two
        formula strings
        """

        self.build_msat_fsm()

        cmd_map = {"absence": "hasdel_check_pattern_absence",
                   "existence": "hasdel_check_pattern_existence",
                   "universality": "hasdel_check_pattern_universality",
                   "precedence": "hasdel_check_pattern_precedence",
                   "responseGlobal": "hasdel_check_pattern_response_global"}

        assert pattern_name in cmd_map, "Unsupported pattern: " + pattern_name

        formulas = ['%s "%s"' % p
                    for p in zip(("-p", "-q"), specs)]

        cmd = ('{cmd_name} {formulas} -T "{time_bound}" -t "{time_var}" ' +
               '-k {max_search_depth} -l {max_klive_bound}').format(
            cmd_name=cmd_map[pattern_name],
            formulas=" ".join(formulas),
            time_bound=time_bound,
            time_var=time_var,
            max_search_depth=options.BMC_LENGTH,
            max_klive_bound=options.KLIVE_BOUND)

        self.send_cmd(cmd)
        ack_msg = "DONE HYCOMP " + pattern_name
        # This is for detecting no result found
        self.send_cmd("echo %s" % ack_msg)

        grep = TextLineGrepper(
            [".*\s*The property is (false|true|unknown)",
             ack_msg])
        m, txt = grep.match(self)  # wait for result!

        # processes results
        if m is None or grep.is_match_unwanted(m):
            log.xsap.user.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return (None, None)

        self._check_handler_errors(txt)

        # processes results
        if m.groups()[0] == "true":
            return (True, None)
        elif m.groups()[0] == "unknown":
            return (None, None)
        assert m.groups()[0] == "false"
        return (False, self._create_trace())


    def check_pattern_absence(self, formula, time_bound, time_var,
                              max_search_depth, max_klive_bound):
        """Invokes hycomp_check_pattern_absence.
        Returns:
           (True, None) if property is true
           (False, trace_fname) with trace_fname xml file of generated trace
           (None, None) if checking result is unknown
        """
        return self._check_pattern_aux("absence",
                                       [formula],
                                       time_bound, time_var,
                                       max_search_depth, max_klive_bound)

    def check_pattern_existence(self, formula, time_bound, time_var,
                                max_search_depth, max_klive_bound):
        """Invokes hycomp_check_pattern_existence.
        Returns:
           (True, None) if property is true
           (False, trace_fname) with trace_fname xml file of generated trace
           (None, None) if checking result is unknown
        """
        return self._check_pattern_aux("existence",
                                       [formula],
                                       time_bound, time_var,
                                       max_search_depth, max_klive_bound)

    def check_pattern_universality(self, formula, time_bound, time_var,
                                   max_search_depth, max_klive_bound):
        """Invokes hycomp_check_pattern_universality.
        Returns:
           (True, None) if property is true
           (False, trace_fname) with trace_fname xml file of generated trace
           (None, None) if checking result is unknown
        """
        return self._check_pattern_aux("universality",
                                       [formula],
                                       time_bound, time_var,
                                       max_search_depth, max_klive_bound)

    def check_pattern_precedence(self, formula_p, formula_q,
                                 time_bound, time_var,
                                 max_search_depth, max_klive_bound):
        """Invokes hycomp_check_pattern_precedence.
        Returns:
           (True, None) if property is true
           (False, trace_fname) with trace_fname xml file of generated trace
           (None, None) if checking result is unknown
        """
        return self._check_pattern_aux("precedence",
                                       [formula_p, formula_q],
                                       time_bound, time_var,
                                       max_search_depth, max_klive_bound)

    def check_pattern_response_global(self, formula_p, formula_q,
                                      time_bound, time_var,
                                      max_search_depth, max_klive_bound):
        """Invokes hycomp_check_pattern_response_global.
        Returns:
           (True, None) if property is true
           (False, trace_fname) with trace_fname xml file of generated trace
           (None, None) if checking result is unknown
        """
        return self._check_pattern_aux("responseGlobal",
                                       [formula_p, formula_q],
                                       time_bound, time_var,
                                       max_search_depth, max_klive_bound)

    def is_state_zeno(self, formula, time_var,
                      max_search_depth=options.BMC_LENGTH,
                      max_klive_bound=options.KLIVE_BOUND):
        """Performs state-zeno check.
        Takes:
           formula: for state to check (string)
           time_var: name of time variable
           max_seatch_depth: depth of search bound (integer)
           max_klive_bound: bound for k-liveness check (integer)

        Returns one of these pairs:
           (NuSMVProcess.UNREACHABLE, None)
           (NuSMVProcess.NON_ZENO, trace_fname): trace_name is a
                                                 filename (string)
           (NuSMVProcess.ZENO, trace_fname): trace_name is a filename (string)
           (NuSMVProcess.UNKNOWN_NOT_FOUND, steps): steps is an integer value
           (NuSMVProcess.UNKNOWN_NOT_PROVED, time): time is a rational
                 value (string) in default time unit used by SmvConverter
        """

        self.build_msat_fsm()

        cmd = ('{cmd_name} -s "{formula}" -t "{time_var}" ' +
               '-k {max_search_depth} -l {max_klive_bound}').format(
            cmd_name="hasdel_is_state_zeno",
            formula=formula,
            time_var=time_var,
            max_search_depth=max_search_depth,
            max_klive_bound=max_klive_bound)

        self.send_cmd(cmd)
        ack_msg = "DONE HYCOMP zeno"
        # This is for detecting no result found
        self.send_cmd("echo %s" % ack_msg)

        grep = TextLineGrepper(
            ["The state is (unreachable|ZENO|NON-ZENO)",
             "Cannot (find) a zeno-path in (\d+) steps",
             "Cannot (prove) the absence of zeno paths up-to ([f'\d./]+)",
             ack_msg])
        m, txt = grep.match(self)  # wait for result!

        # processes results
        if m is None or grep.is_match_unwanted(m):
            log.xsap.user.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return (self.ERROR, None)

        self._check_handler_errors(txt)

        # processes results
        if m.groups()[0] == "unreachable":
            return (self.UNREACHABLE, None)
        elif m.groups()[0] == "NON-ZENO":
            return (self.NON_ZENO, self._create_trace())
        elif m.groups()[0] == "ZENO":
            return (self.ZENO, self._create_trace())
        elif m.groups()[0] == "find":
            return (self.UNKNOWN_NOT_FOUND, int(m.group(2)))
        elif m.groups()[0] == "prove":
            return (self.UNKNOWN_NOT_PROVED, m.group(2))
        assert False, "Bad state" + txt

    def check_clock_divergence(self, clock_name, clock_bound=None,
                               max_search_depth=options.BMC_LENGTH):

        """Takes clock_name (string), clock_bound (string for rational
        number, or None for no-bound-specification),
        max_search_depth (integer, -1 for no bound).

        Searches for clock diverge and returns one pair of:
          * (NuSMVProcess.BOUNDED, bound)
            Where bound is a string with rational expression (smv syntax)

          * (NuSMVProcess.UNBOUNDED, trace)
            Where trace is a file name containing produced trace (iff
            clock_bound was specified), or None if no clock_bound was
            specified.

          * (NuSMVProcess.UNKNOWN, None)
        """

        self.build_msat_fsm()

        clock_bound = ("-s"
                       if clock_bound is None
                       else '-b "%s"' % clock_bound)

        cmd = ('{cmd_name} -c "{clock_name}" {clock_bound} ' +
               '-k {max_search_depth}').format(
            cmd_name="hasdel_check_clock_divergence",
            clock_name=clock_name,
            clock_bound=clock_bound,
            max_search_depth=max_search_depth)

        self.send_cmd(cmd)
        ack_msg = "DONE HYCOMP divergence"
        # This is for detecting no result found
        self.send_cmd("echo %s" % ack_msg)

        grep = TextLineGrepper(
            ["The clock %s is (BOUNDED)" % clock_name,
             "The clock %s is (UNBOUNDED)" % clock_name,
             "The check on the clock divergence for %s is (UNKNOWN)"
             % clock_name,
             ack_msg])

        m, txt = grep.match(self)  # wait for result!

        # processes results
        if m is None or grep.is_match_unwanted(m):
            log.xsap.user.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return (self.ERROR, None)

        self._check_handler_errors(txt)
        # processes results
        if m.groups()[0] == "BOUNDED":
            # searches the bound:
            m2 = re.search("The bound is ([\d.f' /*]+)", txt)
            assert m2 is not None, "Error matching bound, text is:\n" + txt
            return (self.BOUNDED, m2.group(1))

        elif m.groups()[0] == "UNBOUNDED":
            # if bound is specified, trace is also created
            trace = clock_bound and self._create_trace()
            return (self.UNBOUNDED,  trace)
        elif m.groups()[0] == "UNKNOWN":
            return (self.UNKNOWN, None)
        assert False, "Bad state" + txt

    def goto_state_msat(self, trace_index, step_index):
        """Selects the given trace and step index to be the next
        trace used by the simulation (MSAT-based). Creates a new
        trace, and returns the index (integer) of the newly created
        trace."""
        self.build_msat_fsm()
        self.send_cmd('msat_pick_state -s "%d.%d"' % (trace_index, step_index))
        self.wait_for_command_finish()
        return self.get_traces_num()

    def pick_state_random_msat(self, constraints):
        """Creates a 1-step trace for starting a simulation (using
        mathsat). The state is randomly selected taking into
        account the list of constraints (strings). Returns the
        index of the the created trace, or None if no trace was
        created."""
        self.build_msat_fsm()

        grep = TextLineGrepper(
            ["PICK_DONE"],
            ["No trace: initial state is inconsistent",
             "No trace: constraint and initial state are inconsistent"])

        constr = ('"%s"' % " & ".join(("(%s)" % c for c in constraints))
                  if constraints
                  else "")

        cmd = ('msat_pick_state %s; echo "PICK_DONE"' %
               (("-c " + constr)
                if constr
                else ""))

        self.send_cmd(cmd)
        m, txt = grep.match(self)
        if m is None:
            log.xsap.user.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return None

        if grep.is_match_unwanted(m):
            return None

        return self.get_traces_num()

    def simulate_random_msat(self, constraints, length):
        """Constraints is a list of propositional formulae
        (strings). Returns the index of the created trace, or None
        if a problem occurs"""
        self.build_msat_fsm()

        grep = TextLineGrepper(
            ["SIM_DONE"],
            ["No trace: constraint and initial state are inconsistent",
             "Simulation is UNSAT"])

        constr = ('"%s"' % " & ".join(("(%s)" % c for c in constraints))
                  if constraints
                  else "")

        cmd = ('msat_simulate %s -k %d; echo SIM_DONE' %
               (("-t " + constr) if constr else "", length))
        self.send_cmd(cmd)
        m, txt = grep.match(self)
        if m is None:
            log.xsap.user.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return None

        if grep.is_match_unwanted(m):
            return None

        return self.get_traces_num()

    def _simulate_check_feasability_msat(self, constraints):
        """Constraints is a list of next formulae
        (strings). Returns a list of boolean values telling
        positionally which constraints are true or false."""

        res = []
        if constraints:
            self.build_msat_fsm()

            msg = "SIM_FEAS:"
            constr_cmd = " ".join(('-c "%s"' % c for c in constraints))

            grep = TextLineGrepper("([^0-9]*)([01]+)")
            cmd = (
                'echo -n %s; msat_simulate_check_feasible_constraints -q %s' %
                (msg, constr_cmd))
            self.send_cmd(cmd)
            m, txt = grep.match(self)
            if m is None:
                log.xsap.user.error("Process failed on command:\n> " + cmd)
                self._log_returned_text(txt)
                return None

            res = [True if c == "1" else False for c in m.group(0)]

        return res

    # ===
    # TODO: the following commands are offered by nuXmv but there is
    #       no related interface
    # ===
    # abw_check_bmc
    # abw_mh_makegraph
    # abw_print_smv
    # abw_print_smv_new
    # abw_psl_to_aut
    # abw_psl_to_smv
    # abw_psl_to_waba
    # abw_vis_get_nbw
    # abw_zigzag_inc_bmc
    # add_abstraction_preds
    # automata_generate_test
    # automata_performs_test
    # automata_print_automaton
    # build_abstract_model
    # build_simplified_property
    # check_invar_bmc_itp
    # check_invar_cegar_predabs
    # check_invar_guided
    # check_invar_inc_coi
    # check_invar_inc_coi_bdd
    # check_invar_inc_coi_bmc
    # check_invar_local
    # check_ltlspec_compositional
    # check_ltlspec_inc_coi
    # check_ltlspec_inc_coi_bdd
    # check_ltlspec_inc_coi_bmc
    # check_ltlspec_kzeno
    # check_ltlspec_on_trace
    # check_ltlspec_simplify
    # check_property_as_invar_ic3
    # check_pslspec_ba
    # check_traces_properties
    # compile_patterns
    # compute_reachable_guided
    # config_abstraction
    # extract_core_ltlspec
    # invgen_go
    # ltlspec_zigzag_inc_bmc
    # msat_check_fair
    # msat_check_fp_bmc
    # msat_check_invar_bmc_cegar_implabs
    # msat_check_invar_bmc_implabs
    # msat_check_invar_inc_coi
    # msat_check_ltlspec_inc_coi
    # msat_check_sere_bmc
    # msat_encode_variables
    # psl_printSONFSMV
    # psl_printSONFSMV_date07
    # qe_go
    # read_abw
    # read_aiger_model
    # reqan_check_assertion
    # reqan_check_consistency
    # reqan_check_possibility
    # show_param_region
    # show_param_synth_problems
    # sugar_check_pslspec_bmc
    # sugar_mh_makegraph
    # sugar_pslspec_zigzag_inc_bmc
    # synth_param
    # write_abstract_model
    # write_aiger_model
    # write_countacc_model
    # write_hier_coi_model
    # write_range_reduced_model
    # write_simplified_model_func
    # write_simplified_model_rel
    # write_trace_exec_c_model
    # write_vmt_model
    # write_xmi_model
