# Handy launcher for the Fault Tree Viewer

from pathlib import Path
import sys
import logging
import _paths


def get_suitable_ft_files(events_fname, gates_fname, outdir):
    """Automatically retrieve files when partially specified.
    Returns pair (events_fname, gates_fname)
    """

    events_suffix = "events.txt"
    gates_suffix = "gates.txt"

    # searches base (if any exists)
    base = None
    if gates_fname is None and events_fname is not None:
        _base = Path(events_fname).name
        if _base.endswith(events_suffix):
            base = _base[:-len(events_suffix)]
            logging.info("Found base name " + base)

    elif gates_fname is not None and events_fname is None:
        _base = Path(gates_fname).name
        if _base.endswith(gates_suffix):
            base = _base[:-len(gates_suffix)]
            logging.info("Found base name " + base)

    if gates_fname is None and Path(outdir).is_dir():
        logging.info("Searching for suitable gates file")

        pat = (base if base else "*") + gates_suffix
        gates_files = list(Path(outdir).glob(pat))
        if len(gates_files) == 0:
            logging.error("No gates file found in output dir '%s'. "
                          "Specify explicitly" % outdir)
            sys.exit(1)
        if len(gates_files) > 1:
            logging.error("Multiple gates files found in output dir '%s'. "
                          "Specify explicitly" % outdir)
            sys.exit(1)
        gates_fname = str(gates_files[0])
        logging.info("Found gates file '%s'" % gates_fname)

    if events_fname is None and Path(outdir).is_dir():
        logging.info("Searching for suitable events file")

        pat = (base if base else "*") + events_suffix
        events_files = list(Path(outdir).glob(pat))
        if len(events_files) == 0:
            logging.error("No events file found in output dir '%s'. "
                          "Specify explicitly" % outdir)
            sys.exit(1)
        if len(events_files) > 1:
            logging.error("Multiple events files found in output dir '%s'. "
                          "Specify explicitly" % outdir)
            sys.exit(1)
        events_fname = str(events_files[0])
        logging.info("Found events file '%s'" % events_fname)

    return (events_fname, gates_fname)


def handle_options():
    import argparse

    OUTPUT_FORMATS = "png svg pdf".split()
    PNG_FORMAT, SVG_FORMAT, PDF_FORMAT = OUTPUT_FORMATS

    parser = argparse.ArgumentParser(description="Opens the Fault Tree Viewer"
                                     "with specified or found files")

    parser.add_argument("--events-file", "-e",
                        metavar="EVENTS-FILE",
                        help="The input events file name")

    parser.add_argument("--gates-file", "-g",
                        metavar="GATES-FILE",
                        help="The input gates file name")

    parser.add_argument("--xml-file", "-x",
                        metavar="XML-FILE",
                        help="The input fault tree xml file name")

    parser.add_argument(
        "-d", "--outdir",
        default="out",
        metavar="PATH",
        help="Input directory, where gates and events files are searched "
        "when not specified (default:%(default)s)")

    parser.add_argument(
        "-o", "--output-file",
        default=None,
        metavar="OUTPUT-FILE",
        help="Output file where fault tree is dumped")

    parser.add_argument(
        "-f", "--output-format",
        default=PNG_FORMAT,
        metavar="OUTPUT-FORMAT",
        help="Output format for the fault tree (default: %(default)s)")

    parser.add_argument('--verbose', '-v',
                        action='store_true',
                        default=False,
                        help='Sets verbosity to high level')

    args = parser.parse_args()

    for fn in (args.gates_file, args.events_file, args.xml_file):
        if fn and not Path(fn).is_file():
            parser.error("File not found: " + str(fn))

    if ((args.gates_file or args.events_file) and args.xml_file):
        parser.error("Cannot specify both the xml file and events and gates files. "
                     "Please specify only the xml file or both events and gates files.")

    if ((args.gates_file and args.events_file is None) or
        (args.gates_file is None and args.events_file)):
        parser.error("Please provide both the events and gates file or the xml file.")

    return args


def main(events_fname, gates_fname, xml_fname, output_fname, output_format):
    _paths.setup_path()
    from fault_tree_viewer import fault_tree_viewer as ftv
    ftv.run(events_fname, gates_fname, xml_fname, output_fname, output_format)


if "__main__" == __name__:
    args = handle_options()
    if args.verbose:
        logging.getLogger().setLevel(logging.INFO)

    if args.xml_file is None:

        events_fname, gates_fname = get_suitable_ft_files(
            args.events_file,
            args.gates_file,
            args.outdir)
        xml_fname = None

    else:
        events_fname = None
        gates_fname = None
        xml_fname = args.xml_file

    main(events_fname, gates_fname, xml_fname, args.output_file, args.output_format)
