'''
@author: Roverto Cavada
'''

from lxml import etree
from feic.fe.data_structure.model.xmlBound import XmlBound, XmlBoundLineno


class CommonCauses(XmlBoundLineno):
    CAUSE = "cause"

    def __init__(self, **kwargs):
        XmlBound.__init__(self, {},
                          "common_causes",
                          **kwargs)
        setattr(self, "causes", kwargs.get('causes', []))
        return

    def toXML(self):
        """
        Creates a lxml element based on the values contained in the current
        instance of the element. Assigns parameters and processes subelements.
        """
        root = etree.Element(self.tag)
        self._toXMLAttr(root, iter(self.m_map.values()))

        self._toXMLSubElems(root, self.causes)
        return root

    def fromXML(self, root):
        """
        Creates the root element and lists the needed parameters.
        If exists calls the creation of all the sub elements.

        @param root: lxml Element from which we start picking elements.
        """
        self._fromXMLAttrs(root, iter(self.m_map.values()))
        self.causes = (
            self._fromXMLSubElems(
                self.get_all_sub_elems(root, ".")))
        return

    def generateSubelement(self, elem):
        """
        Instantiates an element depending on the tag name.

        @param elem: The element to be instantiated.
        """
        _map = {self.CAUSE: CommonCause}

        # instantiates the element in the correct class
        obj = _map[elem.tag]() if elem.tag in _map else None
        assert obj
        obj.fromXML(elem)
        return(obj)

    pass  # end of class


class CommonCause(XmlBoundLineno):
    CC_MODULE = "cc_module"
    DESC = "desc"
    PROB = "probability"

    def __init__(self, **kwargs):
        XmlBoundLineno.__init__(self, {"name": "name",
                                       self.PROB: "probability",
                                       self.DESC: "desc"},
                                "cause",
                                **kwargs)
        setattr(self, "cc_modules", kwargs.get('cc_modules', []))

        self.probability = kwargs.get(self.PROB, "0")
        return

    def toXML(self):
        """
        Creates a lxml element based on the values contained in the current
        instance of the element. Assigns parameters and processes subelements.
        """
        root = etree.Element(self.tag)
        self._toXMLAttr(root, iter(self.m_map.values()))
        self._toXMLSubElems(root, self.cc_modules)
        if XmlBound.dumping_fe:
            self.deleteAttrs(root, ["desc"])
        return root

    def fromXML(self, root):
        """
        Creates the root element and lists the needed parameters.
        If exists calls the creation of all the sub elements.

        @param root: lxml Element from which we start picking elements.
        """
        self._fromXMLAttrs(root, iter(self.m_map.values()))
        self.cc_modules = (
            self._fromXMLSubElems(
                self.get_all_sub_elems(root, ".")))
        return

    def generateSubelement(self, elem):
        """
        Instantiates an element depending on the tag name.

        @param elem: The element to be instantiated.
        """
        _map = {self.CC_MODULE: CCModule}
        # instantiates the element in the correct class
        obj = _map[elem.tag]() if elem.tag in _map else None
        assert obj
        obj.fromXML(elem)
        return(obj)

    pass  # end of class


class CCModule(XmlBoundLineno):
    CC_MODE = "cc_mode"
    DESC = "desc"
    INSTANCES = "instances"
    INSTANCE_PAT = "pat"

    def __init__(self, **kwargs):
        XmlBoundLineno.__init__(self, {"name": "name", self.DESC: "desc"},
                                "cc_module",
                                **kwargs)
        setattr(self, "cc_modes", kwargs.get('cc_modes', []))
        self.instances = kwargs.get('instances', None)
        return

    def toXML(self):
        """
        Creates a lxml element based on the values contained in the current
        instance of the element. Assigns parameters and processes subelements.
        """
        root = etree.Element(self.tag)

        if self.instances is not None:
            instances = etree.SubElement(root, self.INSTANCES)
            for pat in self.instances:
                el_pat = etree.SubElement(instances, self.INSTANCE_PAT)
                el_pat.text = pat

        self._toXMLAttr(root, iter(self.m_map.values()))
        self._toXMLSubElems(root, self.cc_modes)
        if XmlBound.dumping_fe:
            self.deleteAttrs(root, ["desc"])
        return root

    def fromXML(self, root):
        """
        Creates the root element and lists the needed parameters.
        If exists calls the creation of all the sub elements.

        @param root: lxml Element from which we start picking elements.
        """
        self._fromXMLAttrs(root, iter(self.m_map.values()))

        self.instances = (root.xpath("./%s/%s/text()" % (self.INSTANCES,
                                                         self.INSTANCE_PAT))
                          if root.xpath("count(%s) > 0" % self.INSTANCES)
                          else None)

        self.cc_modes = (
            self._fromXMLSubElems(
                self.get_all_sub_elems(root, ".")))

        return

    def generateSubelement(self, elem):
        """
        Instantiates an element depending on the tag name.

        @param elem: The element to be instantiated.
        """
        _map = {self.CC_MODE: CCMode}
        # instantiates the element in the correct class
        obj = _map[elem.tag]() if elem.tag in _map else None
        if obj is None and elem.tag == self.INSTANCES:
            # instances already handled in fromXML
            return None

        assert obj
        obj.fromXML(elem)
        return obj

    pass  # end of class


class CCMode(XmlBoundLineno):
    CC_RANGE = "cc_range"

    def __init__(self, **kwargs):
        XmlBoundLineno.__init__(self, {"slice": "slice", "mode": "mode"},
                                "cc_mode",
                                **kwargs)
        self.low = 0
        self.high = 0
        return

    def toXML(self):
        """
        Creates a lxml element based on the values contained in the current
        instance of the element. Assigns parameters and processes subelements.
        """
        root = etree.Element(self.tag)
        _range = etree.SubElement(root, self.CC_RANGE)
        _range.set("low", str(self.low))
        _range.set("high", str(self.high))

        self._toXMLAttr(root, iter(self.m_map.values()))
        if XmlBound.dumping_fe:
            self.deleteAttrs(root, ["desc"])
        return root

    def fromXML(self, root):
        """
        Creates the root element and lists the needed parameters.
        If exists calls the creation of all the sub elements.

        @param root: lxml Element from which we start picking elements.
        """
        self._fromXMLAttrs(root, iter(self.m_map.values()))

        self.low = int(self.get_mandatory_singleton_elem_param(
            root, "./%s" % self.CC_RANGE, "low"))
        self.high = int(self.get_mandatory_singleton_elem_param(
            root, "./%s" % self.CC_RANGE, "high"))
        return

    pass  # end of class
