'''
Created on Oct 6, 2011

@author: rayan
'''

from lxml import etree

from feic.fe.data_structure.model.xmlBound import XmlBoundLineno
from feic.fe.data_structure.model.input import Input
from feic.fe.data_structure.model.output import Output
from feic.fe.data_structure.model.raw_smv import RawSmv
from feic.fe.data_structure.model.effect import Effect
from feic.fe.data_structure.model.template import Template


class EffectModel(XmlBoundLineno):
    """Manages the creation of the structure effect model. Can generate a
    new structure based on its value or read from an existing
    one. Also handles the dumping to FE.
    """
    NAME = "name"
    DESC = "desc"
    EM = "effects_model"
    INP = "input"
    OUT = "output"
    VALS = "values"
    RAW = "raw"
    EFCT = "effect"
    TMPL = "template"
    TMPLS = "templates"

    def __init__(self, **kwargs):
        XmlBoundLineno.__init__(self, {self.NAME: "name",
                                       self.DESC: "desc"},
                                self.EM, **kwargs)

        setattr(self, "values", kwargs.get('values', []))
        setattr(self, "effect", kwargs.get('effect', []))
        setattr(self, "templates", kwargs.get('templates', []))
        setattr(self, "raw", kwargs.get('raw', []))
        #self.lib_manager = kwargs.get('lib_manager', None)

    def reloadEM(self, root):
        """
        Reloads missing parts of the EffectModel Structure
        from libraries.

        @param root: Effects Model top level element.
        """
        libname = root.xpath("./@name")[0]
        lib_manager = self.lib_manager
        libs = lib_manager.getEmlObjectify()
        emlib = libs[libname] if libname in libs else None
        if emlib is not None:
            self.effect = (
                self._fromXMLSubElems(
                    [self.get_mandatory_singleton_elem(
                        emlib, "./%s" % self.EFCT)]))
            self.raw = (
                self._fromXMLSubElems(
                    [self.get_optional_singleton_elem(
                        emlib, "./%s" % self.RAW)]))
        else:
            raise Exception("Invalid name of library: " + libname)

    def toXML(self):
        """Creates a lxml element based on the values contained in the
        current instance of the element. Assigns parameters and
        processes subelements.

        """
        root = etree.Element(self.tag)  # Creates the top lavel node
        values = etree.SubElement(root, self.VALS)
        if not XmlBoundLineno.dumping_fe:
            templates = (etree.SubElement(root, self.TMPLS)
                         if self.templates
                         else None)
            self._toXMLSubElems(templates, self.templates)
            # assigning attributes
            self._toXMLAttr(root, iter(self.m_map.values()))
            self._toXMLSubElems(values, self.values)

        elif XmlBoundLineno.dumping_fe:
            map = self.getTemplateMap()
            # assigning attributes
            self._toXMLAttr(root, iter(self.m_map.values()))
            self._toXMLSubElems(values, self.values)
            self._toXMLSubElems(root, self.effect, map)
            self._toXMLSubElems(root, self.raw, map)
            self.deleteAttrs(root, ["desc"])

        return root

    def fromXML(self, root):
        """
        Creates the root element and lists the needed parameters.
        If exists calls the creation of all the sub elements.

        @param root: lxml Element from which we start picking elements.
        """
        # reading all attributes
        self._fromXMLAttrs(root, iter(self.m_map.values()))
        self.templates = (
            self._fromXMLSubElems(
                self.get_all_sub_elems(root, "./%s" % self.TMPLS)))
        self.values = self._fromXMLSubElems(root.xpath("./%s/*" % self.VALS))
        self.effect = self._fromXMLSubElems(root.xpath("./%s" % self.EFCT))
        self.raw = self._fromXMLSubElems(root.xpath("./%s" % self.RAW))
        if not (self.effect and self.raw):
            self.reloadEM(root)

    def generateSubelement(self, elem):
        """
        Instantiates an element depending on the tag name.

        @param elem: The element to be instantiated.
        """
        map = {self.INP: Input,
               self.OUT: Output,
               self.RAW: RawSmv,
               self.EFCT: Effect,
               self.TMPL: Template}

        # instantiates the element in the correct class
        obj = (map[elem.tag]()
               if elem.tag in map
               else None)
        assert obj
        obj.fromXML(elem)

        return obj

    def getTemplateMap(self):
        """Returns a dictionary containing a map of name value of all
        templates in the local dynamics model.
        """
        return dict((template.name, template.value)
                    for template in self.templates
                    if template.value != "")

    pass  # end of class
