'''
Created on Oct 10, 2011

@author: rayan
'''

from lxml import etree
from feic.fe.data_structure.model.nominal_comp_ext import NCExtension
from feic.fe.data_structure.model.common_cause import CommonCauses
from feic.fe.data_structure.model.xmlBound import XmlBound, XmlBoundLineno


class FaultEstensionInfo(XmlBoundLineno):
    """
    Top level structure, manages the creation of the structure fault extension
    info. Can generate a new structure based on its value or read from an
    existing one. Also handles the dumping to FE.
    """
    NAME = "name"
    DESC = "desc"
    SMVF = "smv_file"
    FEI = "fault_extension_info"
    EXTS = "extensions"
    CAUSES = "common_causes"
    NCX = "nominal_component_extensions"
    FEITXT = "fei_txt"

    def __init__(self, lib_manager=None, **kwargs):
        """
        lib_manager is the LibraryManager instance which is set for all the
        XmlBound nodes.
        """
        XmlBoundLineno.__init__(self,
                                {self.NAME: "name",
                                 self.DESC: "desc",
                                 self.SMVF: "smv_file",
                                 self.FEITXT: "fei_txt"},
                                "fault_extension_info",
                                **kwargs)

        if lib_manager:
            XmlBound.lib_manager = lib_manager

        setattr(self, "extensions", kwargs.get('extensions', []))
        setattr(self, "causes", kwargs.get('causes', None))
        return

    def toXML(self):
        """
        Creates a lxml element based on the values contained in the current
        instance of the element. Assigns parameters and processes subelements.
        """
        root = etree.Element(self.tag)
        extensions = etree.SubElement(root, self.EXTS)
        self._toXMLAttr(root, iter(self.m_map.values()))
        self._toXMLSubElems(extensions, self.extensions)

        if self.causes is not None:
            self._toXMLSubElems(root, [self.causes])

        if XmlBound.dumping_fe:
            self.deleteAttrs(root, ["desc"])
            root.tag = "fault_extension"

        return root

    def fromXML(self, root):
        """
        Assigns values to the parameters and instantiates the subelements.

        @param root: lxml Element from which values and subelements are
        obtained.
        """
        self._fromXMLAttrs(root, iter(self.m_map.values()))
        self.extensions = self._fromXMLSubElems(
            self.get_all_sub_elems(root, "./%s" % self.EXTS))

        causes = self.get_optional_singleton_elem(root, "./%s" % self.CAUSES)
        self.causes = (self.generateSubelement(causes)
                       if causes is not None
                       else None)
        return

    def generateSubelement(self, elem):
        """
        Creates an instance of an element based on its tag name.

        @param elem: lxml Element from which element type is obtained.
        """
        _map = {self.NCX: NCExtension,
                self.CAUSES: CommonCauses}
        # instantiates the element in the correct class
        obj = _map[elem.tag]() if elem.tag in _map else None

        assert obj, "elem: " + elem.tag  # should always pass
        obj.fromXML(elem)
        return(obj)

    def parse_from_feinfo(self, fname):
        tree = etree.parse(fname)
        self.fromXML(tree.getroot())
        return

    def save_to_feinfo(self, fname):
        with open(fname, "wb") as out_file:
            out_file.write(etree.tostring(self.toXML(), pretty_print=True))
            pass
        return

    def save_to_fe(self, fname):
        with open(fname, "wb") as out_file:
            out_file.write(etree.tostring(self.dumpFE(), pretty_print=True))
            pass
        return

    pass  # end of class
