from lxml import etree

from feic.fe.data_structure.model.local_dynamics import LocalDynamics
from feic.fe.data_structure.model.effect_model import EffectModel
from feic.fe.data_structure.model.user_info import UserInfo
from feic.fe.data_structure.model.xmlBound import XmlBoundLineno


class FaultMode(XmlBoundLineno):
    """
    Manages the creation of the structure fault mode. Can generate a new
    structure based on its value or read from an existing one. Also handles the
    dumping to FE.
    """
    LDM = "local_dynamics_model"
    UINF = "user_info"
    FMOD = "fault_mode"
    EM = "effects_model"
    SMBLS = "symbols"
    ENABLED = "enabled"
    PROB = "probability"
    LATENT = "latent"

    def __init__(self, **kwargs):
        XmlBoundLineno.__init__(self,
                                {self.ENABLED: "enabled",
                                 self.PROB: "probability"},
                                self.FMOD, **kwargs)

        self.ldm = kwargs.get('ldm', None)
        self.em = kwargs.get('em', None)
        self.user_info = kwargs.get('user_info', None)

        self.enabled = kwargs.get('enabled', "yes")
        self.probability = kwargs.get(self.PROB, "0")
        self.latent = kwargs.get(self.LATENT, None)

    def toXML(self):
        """
        Creates a lxml element based on the values contained in the current
        instance of the element. Assigns parameters and processes subelements.
        """
        root = None

        if self.enabled == "yes" or not XmlBoundLineno.dumping_fe:
            root = etree.Element(self.tag)
            root.append(self.user_info.toXML())
            root.append(self.em.toXML())
            root.append(self.ldm.toXML())
            if self.latent is not None:
                root.append(self.latent.toXML())

            if XmlBoundLineno.dumping_fe:
                name = (
                    self.get_mandatory_singleton_elem_text(
                        root, "./user_info/name"))
                root.set("name", name)
                root.remove(self.get_mandatory_singleton_elem(root,
                                                              "./user_info"))
                self.get_symbols_from_fm(root)

            self._toXMLAttr(root, iter(self.m_map.values()))

            if XmlBoundLineno.dumping_fe:
                self.deleteAttrs(root, [self.ENABLED])

        return root

    def fromXML(self, root):
        """
        Creates the root element and lists the needed parameters.
        If exists calls the creation of all the sub elements.

        @param root: lxml Element from which we start picking elements.
        """
        self._fromXMLAttrs(root, iter(self.m_map.values()))

        self.user_info = (
            self.generateSubelement(
                self.get_mandatory_singleton_elem(root, "./%s" % self.UINF)))
        self.em = (
            self.generateSubelement(
                self.get_mandatory_singleton_elem(root, "./%s" % self.EM)))
        self.ldm = (
            self.generateSubelement(
                self.get_mandatory_singleton_elem(root, "./%s" % self.LDM)))
        elem = self.get_optional_singleton_elem(root, "./%s" % self.LATENT)
        self.latent = (self.generateSubelement(elem)
                       if elem is not None
                       else None)

    def generateSubelement(self, elem):
        """
        Instantiates an element depending on the tag name.

        @param elem: The element to be instantiated.
        """
        _map = {self.UINF: UserInfo,
                self.EM: EffectModel,
                self.LDM: LocalDynamics,
                self.LATENT: LatentFaultInfo, }

        # instantiates the element in the correct class
        obj = (_map[elem.tag]()
               if elem.tag in _map
               else None)

        assert obj
        obj.fromXML(elem)
        return obj


class LatentFaultInfo(XmlBoundLineno):
    """
    For latent fault information, which can be associated to fault modes.
    """
    LATENT = "latent"
    ENABLED = "is_latent"
    PROB = "probability"

    def __init__(self, **kwargs):
        XmlBoundLineno.__init__(self,
                                {self.ENABLED: "is_latent",
                                 self.PROB: "probability"},
                                self.LATENT, **kwargs)

        self.is_latent = kwargs.get(self.ENABLED, "no")
        self.probability = kwargs.get(self.PROB, "0")

    def toXML(self):
        """
        Creates a lxml element based on the values contained in the current
        instance of the element. Assigns parameters and processes subelements.
        """
        root = etree.Element(self.tag)
        self._toXMLAttr(root, iter(self.m_map.values()))
        return root

    def fromXML(self, root):
        """
        Creates the root element and lists the needed parameters.
        If exists calls the creation of all the sub elements.

        @param root: lxml Element from which we start picking elements.
        """
        self._fromXMLAttrs(root, iter(self.m_map.values()))
