'''
Created on Oct 8, 2011

@author: rayan
'''

from lxml import etree

from feic.fe.data_structure.model.fault_mode import FaultMode
from feic.fe.data_structure.model.xmlBound import XmlBoundLineno
from feic.fe.data_structure.model.global_dynamics import GlobalDynamics


class FaultModel(XmlBoundLineno):
    """
    Manages the creation of the structure fault model. Can generate a new
    structure based on its value or read from an existing one. Also handles the
    dumping to FE.
    """
    NAME = "name"
    SMBLS = "symbols"
    FMDL = "fault_model"
    FMOD = "fault_mode"
    GDM = "global_dynamics_model"
    FMODS = "fault_modes"
    DESC = "desc"
    INSTANCES = "instances"
    INSTANCE_PAT = "pat"
    ENABLED = "enabled"

    def __init__(self, **kwargs):
        XmlBoundLineno.__init__(self,
                                {self.NAME: "name",
                                 self.DESC: "desc",
                                 self.ENABLED: "enabled"},
                                self.FMDL, **kwargs)

        # None means 'all instances' (semantics taken from the XML schema)
        self.instances = kwargs.get('instances', None)

        self.fault_modes = kwargs.get('fault_modes', [])
        self.gdm = kwargs.get("gdm", None)
        self.enabled = kwargs.get('enabled', "yes")

        # this is the pointer to the containing NCExtension (it will be set
        # later)
        self.extension = None
        return

    def toXML(self):
        """
        Creates a lxml element based on the values contained in the current
        instance of the element. Assigns parameters and processes subelements.
        """
        #assert self.gdm
        root = None

        if self.enabled == "yes" or not XmlBoundLineno.dumping_fe:
            root = etree.Element(self.tag)
            if self.instances is not None:
                instances = etree.SubElement(root, self.INSTANCES)
                for pat in self.instances:
                    el_pat = etree.SubElement(instances, self.INSTANCE_PAT)
                    el_pat.text = pat
                    pass
                pass

            fms = etree.SubElement(root, self.FMODS)
            self._toXMLSubElems(fms, self.fault_modes)

            if self.gdm:
                self._toXMLSubElems(root, [self.gdm])

            self._toXMLAttr(root, iter(self.m_map.values()))

            if XmlBoundLineno.dumping_fe:
                self.deleteAttrs(root, [self.DESC, self.ENABLED])
                pass
            pass

        return root

    def fromXML(self, root):
        """
        Creates the root element and lists the needed parameters.
        If exists calls the creation of all the sub elements.

        @param root: lxml Element from which we start picking elements.
        """
        self._fromXMLAttrs(root, iter(self.m_map.values()))

        if root.xpath("count(%s) > 0" % self.INSTANCES):
            self.instances = root.xpath("%s/%s/text()" % (self.INSTANCES,
                                                          self.INSTANCE_PAT))
        else:
            self.instances = None
            pass

        self.get_symbols_from_fms(root)

        self.fault_modes = (
            self._fromXMLSubElems(
                self.get_all_sub_elems(root, "./%s" % self.FMODS)))
        self.gdm = GlobalDynamics()

        elem = self.get_optional_singleton_elem(root, "./%s" % self.GDM)
        if elem is not None:
            self.gdm.fromXML(elem)

    def generateSubelement(self, elem):
        """
        Instantiates an element depending on the tag name.

        @param elem: The element to be instantiated.
        """
        _map = {self.FMOD: FaultMode,
                }
        # instantiates the element in the correct class
        obj = _map[elem.tag]() if elem.tag in _map else None
        assert obj, "offending tag: " + elem.tag + " " + str(self) + " " + self.name
        obj.fromXML(elem)
        return(obj)

    pass  # end of class
