'''
Created on Oct 7, 2011

@author: rayan
'''

from lxml import etree
from feic.fe.data_structure.model.template import Template
from feic.fe.data_structure.model.event import Event
from feic.fe.data_structure.model.xmlBound import XmlBoundLineno
from feic.fe.data_structure.model.transition import Transition
from feic.fe.data_structure.model.raw_smv import RawSmv
from feic.fe.data_structure.model.input import Input
from feic.fe.data_structure.model.output import Output


class LocalDynamics (XmlBoundLineno):
    """Manages the creation of the structure local dynamics. Can generate
    a new structure based on its value or read from an existing
    one. Also handles the dumping to FE.
    """
    NAME = "name"
    DESC = "desc"
    VAL = "value"
    VALS = "values"
    LDM = "local_dynamics_model"
    TMPL = "template"
    TMPLS = "templates"
    EVNT = "event"
    EVNS = "events"
    RAW = "raw"
    TRNS = "transitions"
    TRN = "transition"
    INP = "input"
    OUT = "output"

    def __init__(self, **kwargs):
        XmlBoundLineno.__init__(self, {self.NAME: "name", self.DESC: "desc"},
                                "local_dynamics_model", **kwargs)

        setattr(self, "values", kwargs.get('values', []))
        setattr(self, "events", kwargs.get('events', []))
        setattr(self, "templates", kwargs.get('templates', []))
        setattr(self, "raw", kwargs.get('raw', []))
        setattr(self, "transitions", kwargs.get('transitions', []))

    def reloadLDM(self, root):
        """
        Reloads missing parts of the LocalDynamics Structure
        from libraries.

        @param root: Local Dynamics top level element.
        """
        libname = root.xpath("./@name")[0]
        libs = self.lib_manager.getLdlObjectify()
        ldmlib = libs[libname] if libname in libs else None
        if ldmlib is not None:
            self.transitions = (
                self._fromXMLSubElems(ldmlib.xpath("./%s/*" % self.TRNS)))
            self.raw = self._fromXMLSubElems(ldmlib.xpath("./%s" % self.RAW))
            templates = self._fromXMLSubElems(ldmlib.xpath("./%s/*" %
                                                           self.TMPLS))

            req_temp = set(t.name for t in templates)
            in_temp = set(t.name for t in self.templates)
            # check that all required templates are provided
            mis_temp = req_temp - in_temp
            if mis_temp:
                raise Exception("Missing templates in LDM '%s': '%s'" %
                                (libname, ", ".join(mis_temp)))

            # checks that there are no unrequested inputs
            unreq_temp = in_temp - req_temp
            if unreq_temp:
                raise Exception("Unrequested templates in LDM '%s': '%s'" %
                                (libname, ", ".join(unreq_temp)))

        else:
            raise Exception("Invalid name of library")

    def toXML(self):
        """Creates a lxml element based on the values contained in the
        current instance of the element. Assigns parameters and
        processes subelements.

        """
        root = etree.Element(self.tag)
        values = etree.SubElement(root, self.VALS) if self.values else None
        events = etree.SubElement(root, self.EVNS)
        self._toXMLAttr(root, iter(self.m_map.values()))
        if not XmlBoundLineno.dumping_fe:
            templates = (etree.SubElement(root, self.TMPLS)
                         if self.templates
                         else None)
            self._toXMLSubElems(templates, self.templates)
            self._toXMLSubElems(events, self.events)
            self._toXMLSubElems(values, self.values)
        else:
            trns = etree.SubElement(root, self.TRNS)
            _map = self.getTemplateMap()
            self._toXMLSubElems(events, self.events, _map)
            self._toXMLSubElems(values, self.values, _map)
            self._toXMLSubElems(root, self.raw, _map)
            self._toXMLSubElems(trns, self.transitions, _map)
            self.deleteAttrs(root, ["desc"])

        return root

    def fromXML(self, root):
        """
        Creates the root element and lists the needed parameters.
        If exists calls the creation of all the sub elements.

        @param root: lxml Element from which we start picking elements.
        """
        self._fromXMLAttrs(root, iter(self.m_map.values()))
        self.templates = (
            self._fromXMLSubElems(
                self.get_all_sub_elems(root, "./%s" % self.TMPLS)))
        self.events = (
            self._fromXMLSubElems(
                self.get_all_sub_elems(root, "./%s" % self.EVNS)))
        self.transitions = (
            self._fromXMLSubElems(
                self.get_all_sub_elems(root, "./%s" % self.TRNS)))
        self.values = (
            self._fromXMLSubElems(
                self.get_all_sub_elems(root, "./%s" % self.VALS)))
        self.raw = self._fromXMLSubElems(root.xpath("./%s" % "raw"))

        if not (self.transitions and self.raw):
            self.reloadLDM(root)

    def generateSubelement(self, elem):
        """
        Instantiates an element depending on the tag name.

        @param elem: The element to be instantiated.
        """
        _map = {
            self.TMPL: Template,
            self.EVNT: Event,
            self.TRN: Transition,
            "raw": RawSmv,
            self.INP: Input,
            self.OUT: Output,
            }  # mapping the tagname to corrisponding class

        # instantiates the element in the correct class
        obj = _map[elem.tag]() if elem.tag in _map else None
        assert obj
        obj.fromXML(elem)

        return obj

    def getTemplateMap(self):
        """Returns a dictionary containing a map of name value of all
        templates in the local dynamics model.
        """
        return dict((template.name, template.value)
                    for template in self.templates
                    if template.value != "")

    pass  # end of class
