from __future__ import print_function
from builtins import object
import copy
from lxml import etree
from string import Template


class XmlBound(object):
    """
    Class containing some faundamental services used to build the structure
    of the Fault Extension.

    @ivar dumping_fe: Class instance, influences the toXML method. If true will
    perform some extra operation depending on the element calling the toXML.

    @ivar lib_manager: Istatnza di library manager che viene utilizzata per l'
    interazione con le Effect Model e Local Dynamics Model.
    """
    dumping_fe = False
    lib_manager = None

    def __init__(self, xml2member_map, tag, **kwargs):
        """
        Initializes the XmlBound function, declares two vital datastructures
        m_map and sub_elements. Assigns parameter if specified.

        @param xml2member: Contains the attributes to be declared.

        @param tag: The tag name of the current node to be declared.

        @param **kwargs: Optional parameters used to get named variables.
        """
        self.m_map = xml2member_map
        setattr(self, "tag", tag)
        for m in xml2member_map.values():
            setattr(self, m, kwargs.get(m, ""))
            pass
        return

    def toXML(self):
        """
        Creates a lxml element based on the values contained in the current
        instance of the element. Assigns parameters and processes subelements.
        """
        root = etree.Element(self.tag)
        self._toXMLAttr(root, iter(self.m_map.values()))
        if XmlBound.dumping_fe:
            self.deleteAttrs(root, ["desc"])
        return root

    def _toXMLAttr(self, element, attrs):
        """
        Fills the structure with the attributes given in the list. The
        corresponding members must be existing already and set to the desired
        value.

        @param element: The top element of the current subtree

        @param attrs: List of attributes to be added/modified in the given
        element
        """
        for attr in attrs:
            v = getattr(self, attr)
            element.set(attr, v if v is not None else "")
        return

    def _toXMLSubElems(self, root, elems, _map=None):
        """
        Fills the structure by generating the subelements in the list.
        Creates the children nodes and adds them to the parent.

        @param root: The parent node.

        @param elems: List of elements to be created.

        @param _map: Map containing the template map, if not
        None templates will be solved.
        """
        for x in elems:
            if _map:
                x.solveTemplates(_map)
            _x = x.toXML()

            if _x is not None:
                root.append(_x)
                pass
        return

    def fromXML(self, root):
        """
        Gets the values from an existing data structure. Fills the attributes
        and recursively manages the subelements.

        @param root: The beganning node, will start reading from here.
        """
        self._fromXMLAttrs(root, iter(self.m_map.values()))
        return

    def _fromXMLAttrs(self, element, attrs):
        """
        Fills the structure with the attributes given in the list.
        The corresponding members must be existing already.

        If the XML attribute is not found, the member keeps the current value
        if not None.

        @param element: Subject element in which the attributes will be applied
        @param attrs: list of attributes to be set.
        """
        for attr in attrs:
            assert hasattr(self, attr)
            v = element.xpath("./@%s" % attr)
            if len(v) > 0 or getattr(self, attr) is None:
                setattr(self, self.m_map[attr], v[0] if v else "")
                pass
            pass
        return

    def _fromXMLSubElems(self, elems):
        """
        Processes the subelements by instantiating the elements present
        in the given list.

        @param elems: list of element to be processed.
        """
        return [x
                for x in
                (self.generateSubelement(elem)
                 for elem in elems)
                if x is not None]

    def generateSubelement(self, elem):
        """
        Instantiates an element depending on the tag name.

        @param elem: The element to be instantiated.
        """
        raise NotImplementedError
        return

    def makeCopy(self):
        """
        Returns a deep copy of the self instance and all the underlaying
        structure. Each subnode is copied with the current values.
        """
        obj = copy.deepcopy(self)
        return obj

    def dumpFE(self):
        """
        Returns the root element of an lxml tree containing a
        FE tree structure obtained dumping a copy of the FEINFO.
        """
        try:
            XmlBound.dumping_fe = True
            root = self.makeCopy().toXML()  # put exception
        finally:
            XmlBound.dumping_fe = False
        return root

    def deleteAttrs(self, elem, attrs):
        """
        Deletes the specified attributes from the element.

        @param elem: element from which the attributes will be deleted.
        @param attrs: List of attributes from which elem will be deleted if
        exists.
        """
        for attr in attrs:
            if attr in elem.attrib:
                del elem.attrib[attr]

    def solveTemplates(self, _map):
        """
        For each attribute in the current element finds and replaces templates.

        @param _map: The template _map containing the replace strings
        """
        for member in self.m_map.values():
            setattr(self, member,
                    Template(getattr(self, member)).substitute(_map))

    def get_mandatory_singleton_elem(self, root, xpath):
        """
        Given an element path, returns it. The path
        MUST be found, as the element is mandatory.

        @param root: the root element to search from.

        @param xpath: string to identify the xpath of the element.

        @raise ValueError: ValueError if the given path does not lead to an
        element, or if it is not a singleton.
        """
        opt = root.xpath(xpath)
        if len(opt) == 0:
            print(etree.tostring(root, pretty_print=True))
            print(root, xpath)
            raise ValueError("Mandatory element not found in in path %s" %
                             xpath)
        if len(opt) > 1:
            raise ValueError("Mandatory element in path %s is "
                             "not a singleton (%d elements found)" %
                             (xpath, len(opt)))
        return opt[0]

    def get_optional_singleton_elem(self, root, xpath):
        """
        Given an element path, returns it. None
        is returned if the path is not found (i.e. the element is
        not mandatory)
        @param root: the root element to search from.
        @param xpath: string to identify the xpath of the element.

        @raise ValueError: if the given path does not lead neither
        to a singleton or to an empty result"""
        opt = root.xpath(xpath)
        if len(opt) > 1:
            raise ValueError("Optional element in path %s is not a singleton"
                             "(%d elements found)" % (xpath, len(opt)))
        return opt[0] if len(opt) else None

    def get_mandatory_singleton_elem_text(self, root, xpath):
        """
        Given an element path, returns the text in it. The path
        MUST be found, as the element is mandatory.

        @param root: the root element to search from.
        @param xpath: string to identify the xpath of the element.

        Raises ValueError if the given path does not lead to an
        element, or if it is not a singleton."""
        elem = self.get_mandatory_singleton_elem(root, xpath)
        return elem.text

    def get_optional_singleton_elem_text(self, root, xpath):
        """
        Given an element path, returns the text in it. Empty string
        is returned if the path is not found (i.e. the element is
        not mandatory)
        @param root: the root element to search from.
        @param xpath: string to identify the xpath of the element.

        Raises ValueError if the given path does not lead neither
        to a singleton or to an empty result"""
        elem = self.get_optional_singleton_elem(root, xpath)
        return elem.text if elem is not None else ""

    def get_mandatory_singleton_elem_param(self, root, xpath, param):
        """
        Given an element path, returns the text in it. The path
        MUST be found, as the element is mandatory.

        @param root: the root element to search from.
        @param xpath: string to identify the xpath of the element.

        Raises ValueError if the given path does not lead to an
        element, or if it is not a singleton.
        """
        elem = root.xpath("%s/@%s" % (xpath, param))[0]
        return elem

    def get_optional_singleton_elem_param(self, root, xpath, param):
        """
        Given an element path, returns the text in it. The path
        MUST be found, as the element is mandatory.

        @param root: the root element to search from.
        @param xpath: string to identify the xpath of the element.

        Raises ValueError if the given path does not lead to an
        element, or if it is not a singleton.
        """
        elem = root.xpath("%s/@%s" % (xpath, param))
        return elem[0] if elem != [] else ""

    def get_all_sub_elems(self, root, xpath):
        """
        Given an element path, returns all its subelements.

        @param root: the root element to search from.
        @param xpath: string to identify the xpath of the element.
        """
        opt = root.xpath("%s/*" % xpath)
        return opt

    def get_symbols_from_fms(self, root):
        """
        Loads the symbols contained in the structure of the referenced fault
        model, so will go through the fault_modes extracting the effect_models
        then outputs from it.

        @param root: lxml object pointing a Fault Model from wich symbols will
        be extracted.
        """
        setattr(self, "symbols", "")
        symbol_list = []
        self.m_map["symbols"] = "symbols"

        for fm in self.get_all_sub_elems(root, "./fault_modes"):
            em_symbol = (
                self.get_optional_singleton_elem_param(
                    fm,
                    "./effects_model/values/output",
                    "binds"))
            symbol_list.append(em_symbol)

        symbol = " ".join(list(set(symbol_list)))
        setattr(self, "symbols", symbol)
        return

    def get_symbols_from_fm(self, root):
        """
        Loads the symbols contained in the structure of the referenced fault
        mode. So will extract all the outputs and get symbols from it.

        @param root: lxml object pointing the root of a Fault Mode from which
        symbols will be extracted.
        """
        setattr(self, "symbols", "")
        self.m_map["symbols"] = "symbols"
        for output in root.xpath("./effects_model/values/output"):
            symbol = self.get_optional_singleton_elem_param(output,
                                                            ".", "binds")
            setattr(self, "symbols",
                    str.lstrip("%s %s" %
                               (getattr(self, "symbols"), symbol)))
        return

    pass  # end of class
#----------------------------------------------------------------------


class XmlBoundLineno(XmlBound):
    """XmlBound with lineno attribute"""
    line_attr_name = "lineno"

    def __init__(self, xml2member_map, tag, **kwargs):
        """
        Initializes the XmlBound function, declares two vital datastructures
        m_map and sub_elements. Assigns parameter if specified.

        @param xml2member: Contains the attributes to be declared.

        @param tag: The tag name of the current node to be declared.

        @param **kwargs: Optional parameters used to get named variables.
        """

        if self.line_attr_name not in xml2member_map:
            xml2member_map = xml2member_map.copy()
            xml2member_map[self.line_attr_name] = self.line_attr_name

        XmlBound.__init__(self, xml2member_map, tag, **kwargs)
        self.set_lineno(kwargs.get(self.line_attr_name, None))

    def set_lineno(self, line):
        setattr(self, self.line_attr_name,
                str(line) if line is not None else None)

    pass  # end of class
