'''
Created on Nov 8, 20    11

@author: rayan
'''
from gtk import ListStore

from gtkmvc import Model

from feic.fe.data_structure.model.fault_ext_info \
    import FaultEstensionInfo

from feic.fe.data_structure.model.nominal_comp_ext import NCExtension
from feic.fe.fault_model import FaultModelModel
from feic.fe.data_structure.model.fault_model import FaultModel
from feic.fe.data_structure.model.global_dynamics import GlobalDynamics


class FEModel(Model):
    """
    Creates the gui logics and the data structures to allow the visualization
    of the Extensions and the Fault Models contained in the current
    Fault Extension Info.
    """
    models = {}  # Will store models connected to the fault_models

    fei = None
    curr_ext = None

    __observables__ = ("curr_ext", )

    def __init__(self, smv_model):
        Model.__init__(self)
        self.smv_model = smv_model

        # list of NCExtension
        self.nce_list = ListStore(object)

        # list of FaultModel
        self.fm_list = ListStore(object)
        return

    def get_fei(self):
        """Returns the internal FaultExtensionInfo structure"""
        return self.fei

    def create_structure(self, fei):
        """
        Fills the ListStore models containing the Extensions and Fault Models.

        @param fei: Fault Extension Info, used to obtain information on
        Extensions and Fault Models.
        """
        assert (fei and isinstance(fei, FaultEstensionInfo))
        self.fei = fei
        self.nce_list.clear()
        self.fm_list.clear()
        for ext in fei.extensions:
            self.nce_list.append([ext])

            for fm in ext.fault_models:
                fm.extension = ext
                self.fm_list.append([fm])
                fm.get_symbols_from_fms(fm.toXML())
                self.models[fm] = FaultModelModel(self.smv_model)
                pass
            pass
        return

    def set_current_extension(self, extension):
        """
        Sets the current nominal_component_extension to the selected one.

        @param extension: NCExtension instance
        """
        self.curr_ext = extension
        return

    def get_current_extension(self):
        """Returns the currently set NCExtension"""
        return self.curr_ext

    def create_fault_model(self):
        """Creates a new FaultModel for the currently set NCExtension model,
        and returns it. If None is currently selected, None is returned."""

        extension = self.get_current_extension()
        if extension is None:
            return None

        fault_model = FaultModel(gdm=GlobalDynamics())
        fault_model.enabled = "yes"
        fault_model.extension = extension
        fault_model.get_symbols_from_fms(fault_model.toXML())
        return fault_model

    def filter_func(self, model, _iter):
        """
        Filters the specified model, containing only fault models. will return
        true if the fault model is contained in the current extension
        false otherwise.

        @param model:  GtkListStore containing the fault models.

        @param _iter: GtkTreeIterator indicating the element to analyze.
        """

        fault_model = model.get_value(_iter, 0)
        if self.curr_ext is not None:
            return True if fault_model in self.curr_ext.fault_models else False
        return True

    def add_new_extension(self, name):
        """
        Creates a new NominalComponentExtension with empty parameters, adds it
        to the datastructure and to the ncs_list to allow its visualization.

        @param name: name of the module to be added.
        """
        extension = NCExtension()
        extension.module = name
        if self.fei is None:
            self.fei = FaultEstensionInfo()
        self.fei.extensions.append(extension)
        self.create_structure(self.fei)
        return

    def remove_extension(self, extension):
        """Removes the given extension from the list of extensions,
        and then rebuilds all the structure"""
        self.fei.extensions.remove(extension)
        self.create_structure(self.fei)
        return

    def find_and_replace_liststore(self, tree_model, column, old_elem,
                                   new_elem):
        """
        Finds and substitutes the old element with a new given. Returns true
        is element is found and replaced false otherwise.
        """
        _iter = tree_model.get_iter_first()
        while iter is not None:
            if tree_model.get_value(_iter, column) == old_elem:
                tree_model.set_value(_iter, column, new_elem)
                self.models[new_elem] = self.models[old_elem]
                self.models.pop(old_elem)
                return True
            else:
                _iter = tree_model.iter_next(_iter)
        return False

    def save_to_fe(self, fname):
        self.fei.save_to_fe(fname)
        return

    def save_to_feinfo(self, fname):
        self.fei.save_to_feinfo(fname)
        return

    def edit_fault_model(self, fault_model, old_fault_model):
        """
        If corteoller confirmed state is true replaces the current fault model
        with the edited one, adds data_structure and to the liststore to make
        it visible.

        @param windget: GtkTopLevelWindow calling the function.
        @param old_fault_model: The old fault_model to be removed if necessary.
        @param controller: Controller containing the confirmation state.
        """
        extension = old_fault_model.extension
        extension.fault_models.remove(old_fault_model)
        # adding to the current extension element
        extension.fault_models.append(fault_model)
        fault_model.get_symbols_from_fms(fault_model.toXML())
        fault_model.extension = extension
        self.find_and_replace_liststore(self.fm_list, 0,
                                        old_fault_model, fault_model)
        return

    def add_fault_model(self, fault_model, gui_fault_model):
        """
        @param fault_model: the new created fault model, to be added.
        @param gui_fault_model: the corresponding GUI FaultModelModel
        """
        extension = fault_model.extension
        # adding to the current extension element
        extension.fault_models.append(fault_model)
        fault_model.get_symbols_from_fms(fault_model.toXML())
        # adding to the tree to make it visible
        self.fm_list.append([fault_model])
        self.models[fault_model] = gui_fault_model
        return

    def remove_selected_element(self, tree):
        """
        Removes the selected element from the specified GtkTreeView widget.

        @param tree: The GtkTreeView in which element will be retrieved and
        deleted.
        """
        try:
            selection = tree.get_selection()
            tree_model, tree_iter = selection.get_selected()
            fault_model = tree_model.get_value(tree_iter, 0)
            extension = fault_model.extension
            # removing from fault model list
            extension.fault_models.remove(fault_model)
            tree_model.refilter()
        except TypeError:
            print "No fault model was selected"
        return

    pass  # class delimiter
