'''
Created on Nov 18, 2011

@author: rayan
'''
from gtk import ListStore
from gtkmvc.model import Model

from feic.fe.data_structure.model.output import Output


class FaultModeModel(Model):
    """
    Creates the gui logics and the data structures to allow the visualization
    of the Effect Model and the Local Dynamics Model contained in the current
    Fault Mode.
    """
    fault_mode = None
    fault_mode_name = None
    __observables__ = ("fault_mode_name", )

    def __init__(self, fault_mode, smv_model):
        """Parent fault mode is the NCExtension instance, smv_model is a
        valid SMVModel instance"""
        Model.__init__(self)

        self.fault_mode = fault_mode
        self.smv_model = smv_model

        self.em_list = ListStore(str, str)
        self.ldm_list = ListStore(str, str)
        self.templates_ldm = ListStore(object)
        self.templates_em = ListStore(object)
        self.events = ListStore(object)
        self.outputs = ListStore(object)
        self.inputs = ListStore(object)
        return

    def create_structure(self, fm):
        """
        Fills the ListStore models containing the Effect Models and
        Local Dynamics.

        @param fm: Fault Mode, used to obtain information on the Effect Models
        and Local Dynamics.
        """
        self.fault_mode = fm
        self.fault_mode_name = self.fault_mode.user_info.name
        self.get_libraries(fm.lib_manager)
        self.get_ldm_templates_model(fm, self.templates_ldm)
        self.get_em_templates_model(fm, self.templates_em)
        self.get_events_model(fm, self.events)
        self.get_outputs_model(fm, self.outputs)
        self.get_inputs_model(fm, self.inputs)
        pass

    def get_libraries(self, lib_mgr):
        """
        Retrives libraries using the given library manager and manages to
        build the two library dipendent elements effect_model list and local
        dynamics model list.

        @param lib_mgr: Library manager with the correct path to libraries.
        """
        em_libs = lib_mgr.getEmlObjectify()["effects_model_library"]
        ldm_libs = lib_mgr.getLdlObjectify()["local_dynamic_library"]
        em_lib_list = self.get_libs_info(em_libs, "effect_model")
        ldm_lib_list = self.get_libs_info(ldm_libs, "local_dynamics_model")
        for title, desc in em_lib_list.iteritems():
            self.em_list.append([title, desc])
            pass
        for title, desc in ldm_lib_list.iteritems():
            self.ldm_list.append([title, desc])
            pass
        return

    def get_ldm_templates_model(self, fm, tree):
        """Finds the templates_ldm in the given fault_model and manages to
        store them in the given ListStore.

        @param fm: fault_model from which templates_ldm will be retrived.
        @param tree: Tree where found element will be stored.

        """
        tree.clear()
        for template in fm.ldm.templates:
            tree.append([template])

    def get_em_templates_model(self, fm, tree):
        """Finds the templates_ldm in the given fault_model and manages to
        store them in the given ListStore.

        @param fm: fault_model from which templates_ldm will be retrived.
        @param tree: Tree where found element will be stored.

        """
        tree.clear()
        for template in fm.em.templates:
            tree.append([template])

    def get_events_model(self, fm, tree):
        """
        Finds the events in the given fault_model and manages to store
        them in the given ListStore.

        @param fm: fault_model from which events will be retrived.
        @param tree: Tree where found element will be stored.
        """
        tree.clear()
        for event in fm.ldm.events:
            tree.append([event])

    def get_outputs_model(self, fm, tree):
        """
        Finds the outputs in the given fault_model and manages to store them
        in the given ListStore.

        @param fm: fault_model from which outputs will be retrived.
        @param tree: Tree where found element will be stored.
        """
        tree.clear()
        for value in fm.em.values:
            if value.tag == "output":
                tree.append([value])

    def lookup_output_type(self, output_name, exclude_fm=None):
        """Looks up the type for the given output_name in all Fault Modes in
        the parent Fault Model. Returns a list of possible types (having
        different types is an error, but we need to support this)"""

        _types = []

        mod_name = self.fault_mode.fault_model.extension.module

        # Lookup in the SMV model
        nominal_type = self.smv_model.get_variable_type(mod_name, output_name)
        if nominal_type is not None:
            _types.append(nominal_type)
            pass

        # Lookup in other fault modes
        for fm in self.fault_mode.fault_model.fault_modes:
            if exclude_fm != fm:
                for v in fm.em.values:
                    if isinstance(v, Output) and \
                            v.binds == output_name and \
                            str(v.type) not in _types:
                        _types.append(str(v.type))
                        pass
                    pass
                pass
            pass

        return _types

    def get_inputs_model(self, fm, tree):
        """
        Finds the outputs in the given fault_model and manages to store them
        in the given ListStore.

        @param fm: fault_model from which outputs will be retrived.
        @param tree: Tree where found element will be stored.
        """
        tree.clear()
        for value in fm.em.values:
            if value.tag == "input":
                tree.append([value])

    def get_libs_info(self, lib, type):
        """
        Creates a map<title, description> given the library manager.

        @param lib: Element lxml from which libraries will be retrived
        @param type: Type of the library to get
        """
        lib_list = lib.xpath("./%s" % type)
        lib_map = {}
        for elem in lib_list:
            desc = elem.xpath("./description")[0]
            lib_map[elem.xpath("./@name")[0]] = desc.text

        return lib_map

    def has_templates(self):
        """Utility which tells the caller if the model has templates_ldm"""
        ldm_templates = self.templates_ldm.get_iter_first() is not None
        em_templates = self.templates_em.get_iter_first() is not None
        return ldm_templates or em_templates

    def set_fm_name(self, text):
        self.fault_mode_name = self.fault_mode.user_info.name = text
        return

    def set_fm_desc(self, text):
        self.fault_mode.user_info.desc = text
        return

    def get_iter_from_model(self, model, text):
        """
        Gets the iterator corrisponding to the given text picked the
        list store model.

        @param model: A ListStore that contains the given text
        @param text: The string to look for
        """
        for row in model:
            if row[0] == text:
                return row.iter

    pass  # end of class
