'''
Created on Nov 8, 2011

@author: rayan
'''
from gtk import ListStore

import gobject
from string import Template
from completion import CompletionModel

from gtkmvc.model import Model

from feic.fe.data_structure.model.fault_model import FaultModel
from feic.fe.data_structure.model.transition import Transition
from feic.fe.data_structure.model.event import Event
from feic.fe.data_structure.model.input import Input
from feic.fe.data_structure.model.output import Output

from feic.fe.fault_mode import FaultModeModel


class FaultModelModel(Model):
    """
    Creates the gui logics and the data structures to allow the visualization
    of the Fault Modes and the Transitions in the Global Dynamics Model
    contained in the current Fault Model.
    """

    dest_fault_mode = None
    fault_model = None
    fault_model_name = None
    models = {}

    __observables__ = ("dest_fault_mode", "fault_model", "fault_model_name",
                       "affect_all_instances")

    def __init__(self, smv_model):
        #TODO: where is the associated FaultModel?!!

        Model.__init__(self)

        self.smv_model = smv_model
        self.fault_model = None  # FaultModel assigned in create_structure

        #TODO: comment the content of these stores!
        self.fm_list = ListStore(object)  # fault mode list
        # Transition, from, to, trigger names
        self.trns_list = ListStore(object, str, str, str)
        # FaultMode (None for GDM), name of mode (e.g. "nominal)
        self.faults = ListStore(object, str)

        self.fault_events = ListStore(object, str)
        self.input = ListStore(str)
        self.user_events = ListStore(object, str, str)  # Event, name, type

        self.instances = InstancesListModel()
        self.completion = CompletionModel()

        # support for logical OP affect_all_instances:
        self.__affect_all_instances = True
        return

    def create_structure(self, fault_model):
        """
        Fills the ListStore models containing the Fault Modes and the
        Transitions.

        @param fault_model: FaultModel, used to obtain information on the
        Fault Models and Global Dynamics Transitions.
        """
        assert (fault_model and isinstance(fault_model, FaultModel))
        module_name = fault_model.extension.module

        self.fault_model = fault_model
        self.generate_fault_modes()
        self.generate_transitions()
        self.generate_user_events()
        self.create_fault_events_list()
        self.generate_input_list()

        self.instances.fill(self.smv_model.navigator.
                            module_instances_list.
                            get_mod_instances(module_name))
        self.instances.select(fault_model.instances)

        self.fault_model_name = fault_model.name
        self.__affect_all_instances = self.fault_model.instances is None
        return

    def generate_fault_modes(self):
        """
        Generates a fault modes retrieved from the current fault_model.
        """
        self.fm_list.clear()
        for fm in self.fault_model.fault_modes:
            self.fm_list.append([fm])
            #TODO: that's definitely weird! Maggie what have you done here?!!
            fm.get_symbols_from_fm(fm.toXML())
            fm.fault_model = self.fault_model
            self.models[fm] = FaultModeModel(fm, self.smv_model)
            self._add_fm_completion_info(fm)
            pass
        return

    def generate_transitions(self):
        """
        Generates a transition list retrieved from the current fault_model.
        """
        self.trns_list.clear()
        for trns in self.fault_model.gdm.transitions:
            self.trns_list.append([trns,
                                   getattr(trns, "from"),  # from is a keyword!
                                   trns.to, trns.trigger])
            pass
        self.create_fault_list()
        return

    def create_fault_list(self):
        """
        Populates the fault list retrieving data from the current fault_model.
        """
        self.faults.clear()  # clear the fault_events_list
        self.faults.append([None, "nominal"])  # adding nominal component
        for row in self.fm_list:
            fault_mode = self.fm_list.get_value(row.iter, 0)
            fault_name = "%s.fault" % fault_mode.user_info.name
            self.faults.append([fault_mode, fault_name])  # getting fm name
            self._add_fm_completion_info(fault_mode)
            pass  # while iter is not None
        return

    def create_fault_events_list(self):
        """
        Populates the fault event liststore. events are retrieved from the
        current fault mode.
        """
        fault_mode = self.dest_fault_mode
        self.fault_events.clear()

        # empty event to unselect
        self.fault_events.append([None, ""])

        if fault_mode is not None:
            template_map = fault_mode.ldm.getTemplateMap()
            for event in fault_mode.ldm.events:
                event_name = Template(event.name).substitute(template_map)
                self.fault_events.append([event,
                                          "%s.%s" % (fault_mode.user_info.name,
                                                     event_name)])
                pass
            pass
        for event in self.fault_model.gdm.events:
            self.fault_events.append([event, event.name])
            pass
        return

    def generate_user_events(self):
        """
        Populates the user_events liststore. events are retrieved from the
        current fault mode.
        """
        self.user_events.clear()

        for event in self.fault_model.gdm.events:
            self.user_events.append([event, event.name, event.type])
            pass
        return

    def add_fault_mode(self, fault_mode):
        """
        Adds the created fault mode to
        the data_structure and to the liststore to make it visible.

        @param fault_mode: the new created fault mode, to be added.
        """
        fault_model = fault_mode.fault_model
        fault_mode.enabled = "yes"
        # adding to fault model list
        fault_model.fault_modes.append(fault_mode)
        # adding to tree to make it visible
        self.fm_list.append([fault_mode])
        fault_mode.get_symbols_from_fm(fault_mode.toXML())
        self.create_fault_list()
        self.create_fault_events_list
        self._add_fm_completion_info(fault_mode)
        return

    def edit_fault_mode(self, fault_mode, old_fault_mode):
        """
        Replaces the current fault model
        with the edited one, adds data_structure and to the liststore to make
        it visible.

        @param fault_mode: The new fault_mode to be used
        @param old_fault_mode: The old fault_mode to be removed.
        """
        fault_model = old_fault_mode.fault_model
        fault_model.fault_modes.remove(old_fault_mode)
        # adding to the current extension element
        fault_model.fault_modes.append(fault_mode)
        fault_mode.get_symbols_from_fm(fault_mode.toXML())
        fault_mode.fault_model = fault_model
        self.find_and_replace_liststore(self.fm_list, 0,
                                        old_fault_mode, fault_mode)
        self.generate_transitions()
        self.create_fault_list()
        self.create_fault_events_list()
        return

    def find_and_replace_liststore(self, tree_model, column, old_elem,
                                   new_elem):
        """
        Finds and substitutes the old element with a new given. Returns true
        is element is found and replaced false otherwise.
        """
        _iter = tree_model.get_iter_first()
        while _iter is not None:
            if tree_model.get_value(_iter, column) == old_elem:
                tree_model.set_value(_iter, column, new_elem)
                if self.dest_fault_mode == old_elem:
                    # Change dest. fault_mode if needed
                    self.dest_fault_mode = new_elem
                return True
            else:
                _iter = tree_model.iter_next(_iter)
                pass
            pass

        return False

    def set_fm_name(self, text):
        """
        Aligns the text of the name
        """
        self.fault_model.name = text
        self.fault_model_name = self.fault_model.name
        return

    def set_fm_desc(self, text):
        """
        Manages to update the fault model description each time there is a
        change in the GtkTextBuffer contained by the GtkTextView widget.

        @param text: the new description
        """
        self.fault_model.desc = text
        return

    def set_instances(self):
        """Align selected instances with the instances list found in the
        FaultModel"""
        self.fault_model.instances = \
            None if self.__affect_all_instances \
            else \
            [row[InstancesListModel.COL_INSTANCE_NAME]
             for row in self.instances
             if row[InstancesListModel.COL_INSTANCE_SELECTED]]
        return

    def remove_selected_element(self, tree, _list):
        """
        Removes the selected element from the specified GtkTreeView widget.

        @param tree: The GtkTreeView in which element will be retrieved and
        deleted.
        @param _list: _list from which the element will be removed.
        """
        try:
            selection = tree.get_selection()
            tree_model, tree_iter = selection.get_selected()
            elem = tree_model.get_value(tree_iter, 0)
            _list.remove(elem)  # removing selected from fault model _list
            tree_model.remove(tree_iter)
        except TypeError:
            print "No fault mode was selected"
        return

    def add_new_transition(self):
        """Adds a new transition to the list in GDM"""
        transition = Transition()

        setattr(transition, "from", self.faults[0][1])
        setattr(transition, "to", self.faults[0][1])

        # adding to the data structure
        self.fault_model.gdm.transitions.append(transition)
        # adding to the tree_model
        self.trns_list.append([transition,
                               getattr(transition, "from"),
                               transition.to,
                               ""])
        return

    def add_new_user_event(self):
        """
        Adds as empty event and adds it to the global_dynamics events.
        """
        event = Event()
        event.type = "output"
        self.fault_model.gdm.events.append(event)

        self.generate_user_events()
        self.create_fault_events_list()
        return

    def remove_user_event(self, _iter):
        """@param _iter point to the event in the list store to be removed"""
        self.fault_model.gdm.events.remove(self.user_events[_iter][0])
        self.user_events.remove(_iter)
        return

    def generate_input_list(self):
        self.input.clear()
        self.input.append(["output"])
        self.input.append(["input"])
        return

    def _add_fm_completion_info(self, fault_mode):
        """Given a Fault Mode, adds to the completion model all needed words
        from the fault mode"""
        _name = str(fault_mode.user_info.name)

        self.completion.add_word(_name, ("Fault Mode", ))

        for io in fault_mode.em.values:
            if isinstance(io, Input):
                self.completion.add_word(_name + "." + str(io.reads),
                                         ("Input reads", ))
                pass
            elif isinstance(io, Output):
                self.completion.add_word(_name + "." + str(io.reads),
                                         ("Output reads", ))
                self.completion.add_word(_name + "." + str(io.writes),
                                         ("Output writes", ))
                pass
            else:
                print "Could not add completion data for ", io
                pass
            pass
        return

    # -----------------------------------------------------------------------
    # Logical properties
    @Model.getter("affect_all_instances")
    def _get_affect_all_instances(self, _name):
        return self.__affect_all_instances

    @Model.setter("affect_all_instances")
    def _set_affect_all_instances(self, _name, value):
        if value == self.__affect_all_instances:
            return
        self.__affect_all_instances = value
        self.set_instances()
        return

    pass  # end of class
# ---------------------------------------------------------------------------


class InstancesListModel(ListStore):
    """
    This class is a TreeModel which is a list of instances
    """

    _col_types = (gobject.TYPE_PYOBJECT, str, gobject.TYPE_BOOLEAN)
    (COL_INSTANCE, COL_INSTANCE_NAME, COL_INSTANCE_SELECTED) = \
        range(len(_col_types))

    def __init__(self):
        ListStore.__init__(self, *self._col_types)
        return

    def fill(self, instances):
        """
        Fills the ListStore taking the instances from the given list.
        That list is taken from the ModuleInstancesListModel instance
        contained in the SmvModel"""
        self.clear()
        for inst, path in instances:
            self.append((inst, path, True))
            pass
        return

    def select(self, instances):
        """Selects only the instances found in the given list.
        If instances is None, selects all instances.
        @param instances is a list of names, or None
        """
        if instances is None:
            for row in self:
                row[self.COL_INSTANCE_SELECTED] = True

        for row in self:
            row[self.COL_INSTANCE_SELECTED] = \
                row[self.COL_INSTANCE_NAME] in instances

    pass  # end of class
