'''
Created on Sep 21, 2011

@author: Cristian Mattarei
'''
from lxml import etree

### Tokens definition ###

EFFECT_MODEL = "effect_model"
EFFECTS_MODEL_LIBRARY = "effects_model_library"
DESCRIPTION = "description"
DEFINITION = "definition"
VALUES = "values"
EFFECT = "effect"
INPUT = "input"
OUTPUT = "output"
ENTERING = "entering"
DURING = "during"
NAME = "name"
TYPE = "type"
READS = "reads"
WRITES = "writes"
DESC = "desc"
LOCAL = "local"
FALSE = "false"
TRUE = "true"
LOC = "loc"
RAW = "raw"
TEMPLATES = "templates"

###


VKINDS = [INPUT, OUTPUT]
EKINDS = [ENTERING, DURING]


class Value(object):

    _kind = None
    _name = None
    _type = None
    _reads = None
    _writes = None
    _description = None

    def __init__(self, kind, name, type_, reads, writes, description):

        if not isinstance(kind, str):
            raise Exception("String instance expected (got %s instead)"
                            % type(kind))

        if (kind not in VKINDS):
            raise Exception("\"input\" or \"output\" expected (got %s instead)"
                            % kind)

        if (type_ is not None) and not isinstance(type_, str):
            raise Exception("String instance expected (got %s instead)"
                            % type(type_))

        if reads is not None and not isinstance(reads, str):
            raise Exception("String instance expected (got %s instead)"
                            % type(reads))

        if writes is not None and not isinstance(writes, str):
            raise Exception("String instance expected (got %s instead)"
                            % type(writes))

        if name is not None and not isinstance(name, str):
            raise Exception("String instance expected (got %s instead)"
                            % type(name))

        if not isinstance(description, str):
            raise Exception("String instance expected (got %s instead)"
                            % type(description))

        self._kind = kind
        self._name = name
        self._type = type_
        self._reads = reads
        self._writes = writes
        self._description = description

    def __str__(self):
        return etree.tostring(self.getObjectify(), pretty_print=True)

    def getObjectify(self):
        value = etree.Element(str(self._kind))
        if self._type is not None:
            value.attrib[TYPE] = self._type
        if self._reads is not None:
            value.attrib[READS] = self._reads
        if self._writes is not None:
            value.attrib[WRITES] = self._writes
        if self._name is not None:
            value.attrib[NAME] = self._name

        value.attrib[DESC] = self._description

        return value


class Effect(object):

    _kind = None
    _type = None
    _description = None

    def __init__(self, kind, type_, description):
        if not isinstance(kind, str):
            raise Exception("String instance expected (got %s instead)"
                            % type(kind))

        if (kind not in EKINDS):
            raise Exception("\"entering\" or \"during\" expected "
                            "(got %s instead)" % kind)

        if not isinstance(type_, str):
            raise Exception("String instance expected (got %s instead)"
                            % type(type_))

        if not isinstance(description, str):
            raise Exception("String instance expected (got %s instead)"
                            % type(description))

        self._kind = kind
        self._type = type_
        self._description = description

    def __str__(self):
        return etree.tostring(self.getObjectify(), pretty_print=True)

    def getObjectify(self, local=True, location=None):
        effect = etree.Element(str(self._kind))
        if self._type is not None:
            effect.attrib[TYPE] = self._type
            effect.attrib[LOCAL] = TRUE
        effect.text = self._description

        if local and location is not None:
            effect.attrib[LOCAL] = FALSE
            effect.text = location

        return effect


class Raw(object):

    _type = None
    _value = None

    def __init__(self, type_, value):
        if type_ is not None and not isinstance(type_, str):
            raise Exception("String instance expected (got %s instead)"
                            % type(type_))

        if not isinstance(value, str):
            raise Exception("String instance expected (got %s instead)"
                            % type(value))

        self._type = type_
        self._value = value

    def getObjectify(self, local=True, location=None):
        raw = etree.Element(RAW)
        if self._type is not None:
            raw.attrib[LOCAL] = TRUE
            raw.attrib[TYPE] = self._type

        if self._value is not None:
            raw.text = self._value

        if local and location is not None:
            raw.attrib[LOCAL] = FALSE
            raw.text = location

        return raw

    def __str__(self):
        return etree.tostring(self.getObjectify(), pretty_print=True)


class EDefinition(object):
    _name = None
    _templates = None
    _values = None
    _effects = None
    _raw = None

    def __init__(self, name, templates, values, effects, raw):
        if not isinstance(name, str):
            raise Exception("String instance expected (got %s instead)"
                            % type(name))

        if not isinstance(templates, list):
            raise Exception("List of Effect instance expected (got %s instead)"
                            % type(templates))

        if not isinstance(values, list):
            raise Exception("List of Value instance expected (got %s instead)"
                            % type(values))

        if not isinstance(effects, list):
            raise Exception("List of Effect instance expected (got %s instead)"
                            % type(effects))

        if not isinstance(raw, Raw):
            raise Exception("Raw instance expected (got %s instead)"
                            % type(raw))

        self._templates = templates
        self._name = name
        self._values = values
        self._effects = effects
        self._raw = raw

    def getValues(self):
        return self._values

    def getTemplates(self):
        return self._templates

    def getEffects(self):
        return self._effects

    def getRaw(self):
        return self._raw

    def getName(self):
        return self._name

    def getObjectify(self):
        eDefinition = etree.Element(EFFECT_MODEL)
        eDefinition.attrib[NAME] = self._name

        templates = etree.SubElement(eDefinition, TEMPLATES)
        for el in self._templates:
            templates.append(el.getObjectify())

        values = etree.SubElement(eDefinition, VALUES)
        for el in self._values:
            values.append(el.getObjectify())

        effects = etree.SubElement(eDefinition, EFFECT)
        for el in self._effects:
            effects.append(el.getObjectify())

        eDefinition.append(self._raw.getObjectify())

        return eDefinition

    def __str__(self):
        return etree.tostring(self.getObjectify(), pretty_print=True)


class EffectModel(object):
    '''
    classdocs
    '''

    _name = None
    _templates = None
    _description = None
    _values = None
    _effects = None
    _raw = None

    def __init__(self, name, templates, description, values, effects, raw):
        '''
        Constructor
        '''

        self._name = None
        self._description = None
        self._values = None
        self._effects = None
        self._raw = None

        if not isinstance(name, str):
            raise Exception("String instance expected (got %s instead)"
                            % type(name))

        if not isinstance(description, str):
            raise Exception("String instance expected (got %s instead)"
                            % type(description))

        if not isinstance(values, list):
            raise Exception("List of Value instance expected (got %s instead)"
                            % type(values))

        if not isinstance(effects, list):
            raise Exception("List of Effect instance expected (got %s instead)"
                            % type(effects))

        if not isinstance(raw, Raw):
            raise Exception("Raw instance expected (got %s instead)"
                            % type(raw))

        if not isinstance(templates, list):
            raise Exception("Raw instance expected (got %s instead)"
                            % type(templates))

        self._name = name
        self._templates = templates
        self._description = description
        self._values = values
        self._effects = effects
        self._raw = raw

    def getObjectify(self, definition=None):
        eModel = etree.Element(EFFECT_MODEL)
        eModel.attrib[NAME] = self._name
        des = etree.SubElement(eModel, DESCRIPTION)
        des.text = self._description

        if definition is None:
            templates = etree.SubElement(eModel, TEMPLATES)
            for el in self._templates:
                templates.append(el.getObjectify())

        if definition is None:
            values = etree.SubElement(eModel, VALUES)
            for el in self._values:
                values.append(el.getObjectify())

            effects = etree.SubElement(eModel, EFFECT)
            for el in self._effects:
                effects.append(el.getObjectify())

            eModel.append(self._raw.getObjectify())
        else:
            def_ = etree.SubElement(eModel, DEFINITION)
            def_.attrib[LOC] = definition

        return eModel

    def getName(self):
        return self._name

    def __str__(self):
        return etree.tostring(self.getObjectify(), pretty_print=True)
