from __future__ import print_function
import sys
import os.path


def main(in_fname, out_fname, lib_path, text_format=False):
    lib_manager = get_library_manager(lib_path)

    if text_format:
        from feic.translate import fei_translator
        xml_fname = in_fname + ".xml"
        fei_translator.convert_fei_text_to_xml_format(
            in_fname, xml_fname, lib_path)
        in_fname = xml_fname

    generate_structure(lib_manager, in_fname, out_fname)


def get_default_fault_lib_path():
    """Returns the default fault library path"""
    prog_dir = os.path.dirname(__file__)
    def_lib_path = os.path.abspath(
        os.path.join(prog_dir, "../../data/fm_library"))
    return def_lib_path


def get_library_manager(prefix):
    """
    Generates an instance of the library manager that is
    used to retrive the EffectModel and LocalDynamicsModel
    libraries.

    @param prefix: Path where the libraries are stored.
    """

    from feic.fe.lib_mgr.libraries.manager import LibraryManager

    eml = os.path.join(prefix, "eml.xml")
    emlS = os.path.join(prefix, "eml.xsd")
    ldl = os.path.join(prefix, "ldl.xml")
    ldlS = os.path.join(prefix, "ldl.xsd")
    eDefinS = os.path.join(prefix, "eml_behavior/definition.xsd")
    lDefinS = os.path.join(prefix, "ldl_behavior/definition.xsd")
    libManager = LibraryManager(prefix, eml, emlS, ldl, ldlS, eDefinS, lDefinS)

    return libManager


def generate_structure(lib_manager, _file, out_fname):
    """
    Reads the FEI from the file and generates the FE.

    @param _file: The input file containing the FEI.
    """
    from feic.fe.data_structure.model.fault_ext_info import \
        FaultEstensionInfo

    FEI = FaultEstensionInfo(lib_manager)
    FEI.parse_from_feinfo(_file)
    FEI.save_to_fe(out_fname)


__OPT_NAME_LIB_PATH = "lib_path"
__OPT_NAME_OUT_FNAME = "out_fname"


def handle_options():
    """returns (options, input_file):
    values (as a map) and the name of the input file"""

    from optparse import OptionParser

    parser = OptionParser(usage="Usage: %prog [options] <fe-info>.xml",
                          description="Converts a FE Info file into a FE file",
                          version="%prog 0.1")
    parser.add_option("-o", "--output", dest=__OPT_NAME_OUT_FNAME,
                      metavar="OUTPUT",
                      help="write output to OUTPUT")
    parser.add_option("-p", "--path", dest=__OPT_NAME_LIB_PATH,
                      default=get_default_fault_lib_path(),
                      metavar="PATH",
                      help="Path to the extension library (default:%default")
    parser.add_option("--text-fei",
                      action="store_true",
                      default=False,
                      help="Instead of taking xml format as input, "
                      "takes fei in text format")

    options, args = parser.parse_args()

    if len(args) != 1:
        print("Input file argument required")
        parser.print_help()
        sys.exit(1)

    in_fname = args[0]
    if not os.path.isfile(in_fname):
        print("File not found:", in_fname)
        sys.exit(1)

    if not getattr(options, __OPT_NAME_OUT_FNAME):
        root, ext = os.path.splitext(os.path.basename(in_fname))
        setattr(options, __OPT_NAME_OUT_FNAME, root + "_out" + ext)

    if not os.path.isdir(getattr(options, __OPT_NAME_LIB_PATH)):
        print("Lib path not found:", getattr(options, __OPT_NAME_LIB_PATH))
        sys.exit(1)

    return (options, in_fname)


if "__main__" == __name__:
    scripts_dir = os.path.abspath(os.path.join(
        os.path.dirname(__file__), "../../scripts"))
    sys.path.append(scripts_dir)
    import _paths
    _paths.setup_path()

    options, in_fname = handle_options()
    main(in_fname,
         getattr(options, __OPT_NAME_OUT_FNAME),
         getattr(options, __OPT_NAME_LIB_PATH),
         getattr(options, "text-fei"))
