
import os
import os.path
import itertools
from posixpath import join
import re


# ----------------------------------------------------------------------
if __name__ == "__main__":
    # setting up path
    import sys
    code_dir = os.path.abspath(os.path.join(
        os.path.dirname(__file__), "../.."))
# ----------------------------------------------------------------------


from feic.fe.data_structure.model import (
    fault_model, fault_ext_info,
    nominal_comp_ext, fault_mode,
    effect_model, user_info, global_dynamics,
    event, transition, local_dynamics, template,
    input as _input, output, common_cause)



import antlr4
try:
    from feic.translate.generated4.feiParser import feiParser
    from feic.translate.generated4.feiLexer import feiLexer
    from feic.translate.generated4.feiVisitor import feiVisitor
except ImportError as e:
    raise ImportError(
        "Error importing generated module: did you build xSAP?\n" + str(e))

class Visitor(feiVisitor):

    def __init__(self, lib_manager, smv=None, fei_txt=None):
            super(Visitor, self).__init__()
            self.lib_manager = lib_manager

            self.fe = fault_ext_info.FaultEstensionInfo(self.lib_manager, fei_txt=fei_txt)

            if smv:
                self.fe.smv_file = smv

    def visitStart(self, ctx):
        return self.visitFault_extension(ctx.fault_extension())


    def visitFault_extension(self, ctx):
        self.fe.set_lineno(ctx.FAULT().getSymbol().line)

        fe_name = self.visitFull_id(ctx.full_id())
        #print(ctx.FAULT().getText() + " " + ctx.EXTENSION().getText() + " " + fe_name)
        self.fe.name = fe_name

        mod_extensions = []
        for el in ctx.extension():
            mod_extensions.append(self.visitExtension(el))

        cc_causes = None
        ctx_common_causes = ctx.common_causes()
        if ctx_common_causes:
            cc_causes = self.visitCommon_causes(ctx_common_causes)

        self.fe.extensions = mod_extensions
        self.fe.causes = cc_causes
        #return


    # Visit a parse tree produced by feiParser#common_causes.
    def visitCommon_causes(self, ctx):
        ext = common_cause.CommonCauses()
        ext.causes = [self.visitCause(ctx_cause) for ctx_cause in ctx.cause()]
        return ext


    # Visit a parse tree produced by feiParser#cause.
    def visitCause(self, ctx):
        line_number = ctx.CAUSE().getSymbol().line

        ext = common_cause.CommonCause()
        ext.name = self.visitCause_id(ctx.cause_id())
        #print(ctx.CAUSE().getText() + " " + ext.name)
        ext.set_lineno(line_number)

        ctx_opt = ctx.cause_opt()
        if ctx_opt:
            pv = self.visitCause_opt(ctx_opt)
            #print("prob: "+ str(pv))
            ext.probability = pv

        ext.cc_modules = [self.visitCc_module(ctx_mod) for ctx_mod in ctx.cc_module()]
        return ext


    # Visit a parse tree produced by feiParser#cc_module.
    def visitCc_module(self, ctx):
        ext = common_cause.CCModule()
        ext.name = self.visitFull_id(ctx.full_id())
        #print(ctx.MODULE().getText() + " " + ext.name)
        ext.set_lineno = ctx.MODULE().getSymbol().line

        ctx_instances = ctx.instances()
        if ctx_instances:
            ext.instances = self.visitInstances(ctx_instances)

        ext.cc_modes = [self.visitCc_mode(ctx_mode) for ctx_mode in ctx.cc_mode()]
        return ext


    # Visit a parse tree produced by feiParser#cc_mode.
    def visitCc_mode(self, ctx):
        ext = common_cause.CCMode()
        ext.slice = self.visitSlice_id(ctx.slice_id())
        #print(ext.slice)
        ext.mode = self.visitMode_id(ctx.mode_id())
        #print(ext.mode)
        ext.set_lineno = ctx.MODE().getSymbol().line

        ext.low, ext.high = self.visitCc_range(ctx.cc_range())
        #print("[" + str(ext.low) + "," + str(ext.high) +  "]")
        return ext


    # Visit a parse tree produced by feiParser#cc_range.
    def visitCc_range(self, ctx):
        l = ctx.NUMBER()[0].getText() #TODO : check
        r = ctx.NUMBER()[1].getText()
        return int(l), int(r)


    # Visit a parse tree produced by feiParser#extension.
    def visitExtension(self, ctx):
        #print(ctx.EXTENSION().getText() + " " + ctx.OF().getText())
        return self.visitMod_extension(ctx.mod_extension())


    def visitMod_extension(self, ctx):
        ext = nominal_comp_ext.NCExtension()
        name_module = self.visitFull_id(ctx.full_id())
        #print(ctx.MODULE().getText() + " " + name_module)
        ext.module = name_module
        ext.set_lineno(ctx.MODULE().getSymbol().line)

        instances = []
        if ctx.instances():
            instances = self.visitInstances(ctx.instances())
        else:
            for s in ctx.r_slice():
                fm_res = self.visitR_slice(s)
                fm_res.instances += instances
                ext.fault_models.append(fm_res)

        return ext


    def visitR_slice(self, ctx):
        fm = fault_model.FaultModel()

        name = ctx.BASIC_ID().getText()
        #print(ctx.SLICE().getText() + " " + name)
        fm.name = name
        fm.set_lineno(ctx.SLICE().getSymbol().line)

        fm.instances = []

        if ctx.instances():
            fm.instances = self.visitInstances(ctx.instances())

        _vars = self.visitR_vars(ctx.r_vars())
        #print(ctx.AFFECTS().getText() + " " + str(_vars))


        fm.fault_modes = self.visitFault_modes(ctx.fault_modes())
        if ctx.global_dynamics():
            fm.gdm = self.visitGlobal_dynamics(ctx.global_dynamics())

        symbs = set()
        for _fm in fm.fault_modes:
            for v in itertools.chain(_fm.em.values, _fm.ldm.values):
                if isinstance(v, _input.Input):
                    symbs.add(v.value)
                else:
                    symbs.add(v.binds)

            for e in _fm.ldm.events:
                symbs.add(e.binds)

            for t in itertools.chain(_fm.em.templates, _fm.ldm.templates):
                symbs.add(t.name)
                pass

        #print("SYMBS " + str(symbs))

        res_set = _vars - symbs
        if len(res_set) > 0:
            f = res_set.pop()
            self.raise_error_context(ctx, "Affected variable '" + f + "' not found in "
                        "any fault mode of fault model '" + fm.name + "'")

        return fm


    # Visit a parse tree produced by feiParser#instances.
    def visitInstances(self, ctx):
        return self.visitFull_id_list(ctx.full_id_list())


    # Visit a parse tree produced by feiParser#fault_modes.
    def visitFault_modes(self, ctx):
        res = []
        for ctx_fm in ctx.fault_mode():
            fm = self.visitFault_mode(ctx_fm)
            res.append(fm)
        #print("RES FAULT MODES " + str(res))
        return res


    # Visit a parse tree produced by feiParser#fault_mode.
    def visitFault_mode(self, ctx):
        fm = fault_mode.FaultMode()
        fm.set_lineno(ctx.MODE().getSymbol().line)

        name_mode = self.visitMode_id(ctx.mode_id())
        #print(ctx.MODE().getText() + " " + name_mode)

        fm.user_info = user_info.UserInfo(name=name_mode)

        if ctx.mode_opts():
            _opts, line = self.visitMode_opts(ctx.mode_opts())

            pt = "0"
            for k in _opts.keys():
                if k == feiLexer.OPT_PROB:
                    pt = _opts[k]

            fm.probability = pt
            #print("prob: " + str(pt))

            if (feiLexer.OPT_LATENT not in _opts and feiLexer.OPT_LATENT_PROB not in _opts):
                fm.latent = None
            else:
                is_latent = _opts.get(feiLexer.OPT_LATENT, "yes")
                prob = _opts.get(feiLexer.OPT_LATENT_PROB, "0")
                fm.latent = fault_mode.LatentFaultInfo(is_latent=is_latent, probability=prob)
                fm.latent.set_lineno(line)

        fm.ldm = self.visitLocal_dynamics_model_id(ctx.local_dynamics_model_id())
        fm.em, values, events, templates = self.visitEffect(ctx.effect())

        #print("FM EM VALUES " + str(fm.em.values))
        self.process_fault_mode_library_context(ctx, fm, values, events, templates)
        return fm

    def get_ctx_data(self, ctx, print_col=True):
        if ctx is None:
            return ""
        res = "Line " + str(ctx.start.line)
        if print_col:
            res += " Col:" + str(ctx.start.column)
        return res

    def raise_error_context(self, ctx, msg):
        raise RuntimeError(self.get_ctx_data(ctx) + ": " + msg)

    def process_fault_mode_library_context(self, ctx, fm, values, events, templates):

        libs = self.lib_manager.getEmlObjectify()
        lib = libs.get(fm.em.name)

        ldm_libs = self.lib_manager.getLdlObjectify()
        ldm_lib = ldm_libs.get(fm.ldm.name)

        if lib is None:
            self.raise_error_context(ctx, "EM Library '" + fm.em.name + "' not found")

        if ldm_lib is None:
            self.raise_error_context(ctx, "LDM Library '" + fm.ldm.name + "' not found")

        em_inputs = set(lib.xpath("./values/input/@reads"))
        em_outputs = set(lib.xpath("./values/output/@writes"))

        ldm_inputs = set(ldm_lib.xpath("./values/input/@reads"))
        ldm_outputs = set(ldm_lib.xpath("./values/output/@writes"))

        for v, ns, ctx, _ in values:
            if isinstance(v, _input.Input):
                if ns:
                    if ns == fm.em.name:
                        if v.reads in em_inputs:
                            fm.em.values.append(v)
                        else:
                            self.raise_error_context(ctx, "Input '" + v.reads + "' "
                                        "not found in the specified '" +
                                        ns + "' EM library")
                    elif ns == fm.ldm.name:
                        if v.reads in ldm_inputs:
                            fm.ldm.values.append(v)
                        else:
                            self.raise_error_context(ctx, "Input '" + v.reads + "' "
                                            "not found in the specified '" +
                                            ns + "' LDM library")
                    else:
                        self.raise_error_context(ctx, "Specified namespace '" +
                                    ns + "' does not exist")

                else:
                    if v.reads in em_outputs and v.reads in ldm_outputs:
                        fm.em.values.append(v)
                        fm.ldm.values.append(v)
                    elif v.reads in em_inputs:
                        fm.em.values.append(v)
                    elif v.reads in ldm_inputs:
                        fm.ldm.values.append(v)
                    else:
                        self.raise_error_context(ctx,
                                    "Input '" + v.reads + "' "
                                    "not found in the library (neither in EM and LD)")

            else:
                if ns:
                    if ns == fm.em.name:
                        if v.writes in em_outputs:
                            fm.em.values.append(v)
                        else:
                            self.raise_error_context(ctx, "Output '" + v.writes + "' "
                                        "not found in the specified '" +
                                        ns + "' EM library")
                    elif ns == fm.ldm.name:
                        if v.writes in ldm_outputs:
                            fm.ldm.values.append(v)
                        else:
                            self.raise_error_context(ctx, "Output '" + v.writes + "' "
                                        "not found in the specified '" +
                                        ns + "' LDM library")
                    else:
                        self.raise_error_context(ctx, "Specified namespace '" +
                                    ns + "' does not exist")
                else:
                    assert isinstance(v, output.Output)
                    if v.writes in em_outputs and v.writes in ldm_outputs:
                        fm.em.values.append(v)
                        fm.ldm.values.append(v)
                    elif v.writes in em_outputs:
                        fm.em.values.append(v)
                    elif v.writes in ldm_outputs:
                        fm.ldm.values.append(v)
                    else:
                        self.raise_error_context(ctx, "Output '" + v.writes + "' "
                                    "not found in the library (neither in EM and LD)")

        # here templates are filled and checked
        em_templates = set(lib.xpath("templates/template/@name"))
        ld_templates = set(ldm_lib.xpath("templates/template/@name"))

        for t, ns, t_ctx in templates:
            if ns:
                if ns == fm.em.name:
                    if t.name in em_templates:
                        fm.em.templates.append(t)
                    else:
                        self.raise_error_context(t_ctx, "Template '" + t.name + "' "
                                    "not found in the specified '" +
                                    ns + "' EM library")
                elif ns == fm.ldm.name:
                    if t.name in em_templates:
                        fm.em.templates.append(t)
                    else:
                        self.raise_error_context(t_ctx,
                                    "Template '" + t.name + "' "
                                    "not found in the specified '" +
                                    ns + "' EM library")
            else:
                if t.name in em_templates and t.name in ld_templates:
                    fm.em.templates.append(t)
                    fm.ldm.templates.append(t)
                if t.name in em_templates:
                    fm.em.templates.append(t)
                elif t.name in ld_templates:
                    fm.ldm.templates.append(t)
                else:
                    self.raise_error_context(t_ctx, "The template '%s' could not be found "
                                "in the library (neither EM and LD)" % t.name)

        # checks that wrong names are not provided
        req_temp = set(em_templates) | set(ld_templates)
        prov_temp = set(t.name for t, _, _ in templates)
        mis_temp = req_temp - prov_temp
        if mis_temp:
            self.raise_error_context(ctx,
                        "In Fault Mode '%s', one or more required templates "
                        "were not provided: '%s'" % (fm.user_info.name,
                                                    ", ".join(mis_temp)))

        # checks that templates' values are also defined
        symbols = set()
        for v, ns, ctx, _ in values:
            if isinstance(v, _input.Input):
                symbols.add(v.reads)
                #symbols.add(v.value)
            else:
                assert isinstance(v, output.Output)
                symbols.add(v.reads)
                symbols.add(v.writes)
        for e, ns, ev_ast in events:
            symbols.add(e.name)

        for t, ns, t_ctx in templates:
            if t.value not in symbols:
                self.raise_error_context(t_ctx,
                            "Template '%s' has value '%s' which has no "
                            "corresponency in the Fault Mode '%s' interface.\n"
                            "Recognized values are: %s\n"
                            "For events template, you may want to add "
                            "an event '%s' to the FM's interface." %
                            (t.name, t.value, fm.user_info.name,
                            ", ".join(sorted(symbols)), t.value))

        # events il ldm
        ld_events = set(ldm_lib.xpath("events/event/@name"))
        for e, ns, ev_ctx in events:
            if ns:
                if ns == fm.em.name:
                    fm.em.events.append(e)
                elif ns == fm.ldm.name:
                    fm.ldm.events.append(e)
                else:
                    self.raise_error_context(ev_ctx, "Specified namespace '" +
                                ns + "' is not valid")
            else:
                if e.name in ld_events:
                    fm.ldm.events.append(e)
                else:
                    ref_temp = None
                    for t, ns, _ in templates:
                        if t.value == e.name:
                            ref_temp = t
                            break

                    if ref_temp is None:
                        self.raise_error_context(ev_ctx, "Event '" + e.name +
                                    "' not found in ldm " + fm.ldm.name)

                    if t in fm.ldm.templates:
                        fm.ldm.events.append(e)
                    else:
                        assert t in fm.em.templates
                        fm.em.events.append(e)

        return




    # Visit a parse tree produced by feiParser#global_dynamics.
    def visitGlobal_dynamics(self, ctx):
        gm = global_dynamics.GlobalDynamics()
        gm.set_lineno(ctx.GLOBAL().getSymbol().line)

        gm.transitions = []

        for ctx_event in ctx.new_event():
            gm.events.append(self.visitNew_event(ctx_event)[0])

        for ctx_trans in ctx.trans():
            gm.transitions.append(self.visitTrans(ctx_trans))

        return gm


    # Visit a parse tree produced by feiParser#new_event.
    def visitNew_event(self, ctx):
        return self.visitEvent_par(ctx.event_par())


    # Visit a parse tree produced by feiParser#mode_id.
    def visitMode_id(self, ctx):
        return ctx.BASIC_ID().getText()


    # Visit a parse tree produced by feiParser#slice_id.
    def visitSlice_id(self, ctx):
        return ctx.BASIC_ID().getText()


    # Visit a parse tree produced by feiParser#cause_id.
    def visitCause_id(self, ctx):
        return ctx.BASIC_ID().getText()


    # Visit a parse tree produced by feiParser#local_dynamics_model_id.
    def visitLocal_dynamics_model_id(self, ctx):
        ldm = local_dynamics.LocalDynamics()
        ldm.name = ctx.BASIC_ID().getText()
        #print("LDM NAME " + ldm.name)
        ldm.set_lineno(ctx.BASIC_ID().getSymbol().line)
        return ldm

    # Visit a parse tree produced by feiParser#effect.
    def visitEffect(self, ctx):
        em = effect_model.EffectModel()

        em.name = ctx.BASIC_ID().getText()
        #print(em.name)
        em.set_lineno(ctx.LPAR().getSymbol().line)

        values, events, templates = self.visitParams(ctx.params())
        return em, values, events, templates

    def visitTrans(self, ctx):
        trans = transition.Transition()
        trans.set_lineno(trans.TRANS().getSymbol().line)

        setattr(trans, "from", self.visitTrans_mode_id(ctx.trans_mode_id()[0]))
        trans.to = self.visitTrans_mode_id(ctx.trans_mode_id()[1])

        trans.trigger = None
        trans.guard = None

        id, exp = self.visitTrans_guard(ctx.trans_guard())

        if id:
            trans.trigger = id
        if exp:
            trans.guard = exp

        return trans


    def visitTrans_guard(self, ctx):
        id = exp = None
        if ctx.full_id():
            id = self.visitFull_id(ctx.full_id())
        if ctx.trans_guard_when():
            exp = self.visitTrans_guard_when(ctx.trans_guard_when())
        return id, exp

    def visitTrans_guard_when(self, ctx):
        exp, _ = self.visitSimple_expression(ctx.simple_expression())
        return exp

    def visitTrans_mode_id(self, ctx):
        if ctx.mode_id():
            id = self.visitMode_id(ctx.mode_id())
            if ctx.NOMINAL():
                return id + "." + ctx.NOMINAL().getText()
            else:
                return id + "." + ctx.LOWER_FAULT().getText()


    def visitR_vars(self, ctx):
        _vars = set()
        for id in ctx.BASIC_ID():
            _vars.add(id.getText())
        return _vars


    def visitParams(self, ctx):
        maps_to = {}
        values = []
        events = []
        templates = []

        for ctx_par in ctx.par():
            i, t = self.visitPar(ctx_par)

            if i == feiLexer.DATA:
                io, isvar, value, d, ns = t
                #print("IO: " +str(io) + " VALUE " + str(value) + " ISVAR " + str(isvar))

                #if isinstance(d, _input.Input):
                #    print("READS " + d.reads + " VALUE " + d.value)
                #else:
                #    print("WRITES " + d.writes + " BINDS " + d.binds)

                if isvar:
                    if value in maps_to:
                        last, _, ic, oc = maps_to[value]
                        assert oc is None

                        if isinstance(last, _input.Input):
                            assert isinstance(d, output.Output), \
                            "Object {0} not an output instance, last is: {1}, is var is: {2}".format(d, last, isvar)
                            d.reads = last.reads
                            maps_to[value] = (d, ns, ic, None)
                        else:
                            assert isinstance(last, output.Output)
                            assert isinstance(d, _input.Input)
                            last.reads = d.reads
                    else:
                        maps_to[value] = (d, ns, ctx_par.data_par(), None)
                else:
                    # bare input
                    assert io == "input"
                    values.append((d, ns, ctx_par.data_par(), None))
            elif i == feiLexer.EVENT:
                ev, ns = t
                events.append((ev, ns, ctx_par.event_par()))
            elif i == feiLexer.TEMPLATE:
                t0, ns = t
                templates.append((t0, ns, ctx_par.template_par()))

        for k in maps_to:
            # Add the namespace to the values
            v, ns, ic, oc = maps_to[k]
            values.append((v, ns, ic, oc))

        return values, events, templates

    # Visit a parse tree produced by feiParser#par.
    def visitPar(self, ctx):
        if ctx.data_par():
            return feiLexer.DATA, self.visitData_par(ctx.data_par())
        elif ctx.event_par():
            return feiLexer.EVENT, self.visitEvent_par(ctx.event_par())
        elif ctx.template_par():
            return feiLexer.TEMPLATE, self.visitTemplate_par(ctx.template_par())


    # Visit a parse tree produced by feiParser#data_par.
    def visitData_par(self, ctx):
        rw_var, ns = self.visitNamespace_id(ctx.namespace_id())
        #print(rw_var)
        io, value, isvar = self.visitRw_op(ctx.rw_op())
        d = None
        if io == "input":
            d = _input.Input()
            d.reads = rw_var
            d.value = value
        else:
            assert io == "output"
            assert isvar
            d = output.Output()
            d.writes = rw_var
            d.binds = value

        return io, isvar, value, d, ns


    def visitEvent_par(self, ctx):
        ev = event.Event()
        ev.name, ns = self.visitNamespace_id(ctx.namespace_id())
        #print(ev.name)
        ev.set_lineno(ctx.EVENT().getSymbol().line)

        if ctx.rw_op():
            io, ev.binds, isvar = self.visitRw_op(ctx.rw_op())
            assert isvar
            connected = "yes"
        else:
            io, ev.binds = "output", ""
            connected = "no"

        ev.type = io
        ev.connected = connected

        #print("EV.NAME " + ev.name + " EV.BINDS " + ev.binds + " EV.TYPE " + ev.type + " EV.CONNECTED " + ev.connected)
        return ev, ns


    def visitRw_op(self, ctx):
        if ctx.RW_OP_LT():
            io = "input"
            value, isvar = self.visitSimple_expression(ctx.simple_expression())
            #print(ctx.RW_OP_LT().getText() + " " + value)
            return io, value, isvar
        else:
            io = "output"
            #print(ctx.RW_OP_GT().getText() + " " + ctx.BASIC_ID().getText())
            return io, ctx.BASIC_ID().getText(), True


    def visitTemplate_par(self, ctx):
        t = template.Template()
        t.name, ns = self.visitNamespace_id(ctx.namespace_id())
        #print(t.name)
        t.value = ctx.BASIC_ID().getText()
        #print(t.value)
        return t, ns


    def visitNamespace_id(self, ctx):
        num_basic_id = len(ctx.BASIC_ID())
        if num_basic_id == 1:
            first_name = ctx.BASIC_ID()[0].getText()
            return first_name, None
        else:
            return (ctx.BASIC_ID()[1].getText(), ctx.BASIC_ID()[0].getText())


    def visitMode_opts(self, ctx):
        line = ctx.LBRA().getSymbol().line
        res = {}
        for ctx_mode_opt in ctx.mode_opt():
            opt_type, opt_value = self.visitMode_opt(ctx_mode_opt)
            assert opt_type not in res, \
            "At line %d, duplicated option '%s'" % (line, opt_type)
            res[opt_type] = opt_value
        return res, line


    def visitMode_opt(self, ctx):
        if ctx.prob:
            return feiLexer.OPT_PROB, ctx.REAL_NUMBER().getText()
        elif ctx.OPT_PROB():
            return ctx.OPT_PROB, ctx.REAL_NUMBER().getText()
        elif ctx.OPT_LATENT():
            return ctx.OPT_LATENT, self.visitYesno(ctx.yesno())
        else:
            return ctx.OPT_LATENT_PROB, ctx.REAL_NUMBER().getText()


    def visitCause_opt(self, ctx):
        p = ctx.REAL_NUMBER().getText()
        return p


    def visitYesno(self, ctx):
        if ctx.YES():
            return ctx.YES().getText()
        else:
            return ctx.NO().getText()


    def visitFull_id_list(self, ctx):
        full_id_list = []
        for w in ctx.wildcard_full_id():
            wildcard_id = self.visitWildcard_full_id(w)
            full_id_list.append(wildcard_id)
        #print(full_id_list)
        return full_id_list


    def visitWildcard_full_id(self, ctx):
        res = []
        for base in ctx.base():
            base_id = self.visitBase(base)
            res.append(base_id)
        return ".".join(res)


    def visitBase(self, ctx):
        left_square_bracket = ctx.LSQBRA()
        if left_square_bracket:
            list_id = []
            if ctx.external:
                list_id.append(ctx.external.text)
            list_id.append(left_square_bracket.getText())
            if ctx.internal:
                list_id.append(ctx.internal.text)
            else:
                list_id.append(ctx.NUMBER().getText())
            list_id.append(ctx.RSQBRA().getText())
            return "".join(list_id)
        elif ctx.ASTERISK():
            if ctx.BASIC_ID():
                return ctx.BASIC_ID()[0].getText() + ctx.ASTERISK().getText()
            else:
                return ctx.ASTERISK().getText()
        else:
            return ctx.BASIC_ID()[0].getText()



    # Visit a parse tree produced by feiParser#full_id.
    def visitFull_id(self, ctx):
        list_basic_id = []
        for ctx_basic_id in ctx.BASIC_ID():
            basic_id_text = ctx_basic_id.getText()
            list_basic_id.append(basic_id_text)

        return '.'.join(list_basic_id)


    def visitSimple_expression(self, ctx):
        if ctx.tr:
            return ctx.tr.text, False
        elif ctx.fl:
            return ctx.fl.text, False
        elif ctx.WORD_NUMBER():
            return ctx.WORD_NUMBER().getText(), False
        elif ctx.NUMBER():
            return ctx.NUMBER().getText(), False
        elif ctx.REAL_NUMBER():
            return ctx.REAL_NUMBER().getText(), False
        elif ctx.BASIC_ID():
            return ctx.BASIC_ID().getText(), True
        elif ctx.bang:
            internal_exp, _ = self.visitSimple_expression(ctx.simple_expression()[0])
            return (ctx.bang.text + internal_exp).strip(), False
        elif ctx.opt:
            internal_exp, _ = self.visitSimple_expression(ctx.simple_expression()[0])
            return (internal_exp + ctx.opt.getText()).strip(), False
        elif ctx.lbra:
            internal_exp, _ = self.visitSimple_expression(ctx.simple_expression()[0])
            return (ctx.lbra.text + internal_exp + ctx.rbra.text).strip(), False
        else:
            left_exp, _ = self.visitSimple_expression(ctx.left)
            op = ctx.op.text
            right_exp, _ = self.visitSimple_expression(ctx.right)
            return (left_exp + op + right_exp).strip(), False

def parse_file_antlr4(fname, lib_manager=None, smv_file=None):
    from feic import feicm
    import os
    assert(os.path.exists(fname))

    stream = antlr4.FileStream(fname)
    lexer = feiLexer(stream)
    tokens = antlr4.CommonTokenStream(lexer)
    parser = feiParser(tokens)

    r = parser.start()

    visitor = Visitor(lib_manager, fei_txt=fname, smv=smv_file)
    visitor.visit(r)

    return visitor.fe



def convert_fei_text_to_xml_format(fei_txt_in_fname,
                                   fei_xml_out_fname,
                                   lib_path=None,
                                   smv_file=None):
    convert_fei4_text_to_xml_format(fei_txt_in_fname, fei_xml_out_fname, lib_path, smv_file)


def convert_fei4_text_to_xml_format(fei_txt_in_fname,
                                   fei_xml_out_fname,
                                   lib_path=None,
                                   smv_file=None):
    """Converts given FEI in textual format, to the corresponding xml format
    raises RuntimeError if an error is encountered while parsing
    """
    from lxml import etree
    from feic import feicm

    lib_manager = get_library_manager(
        lib_path or feicm.get_default_fault_lib_path())

    fe = parse_file_antlr4(fei_txt_in_fname, lib_manager, smv_file)

    fe_xml = fe.toXML()
    with open(fei_xml_out_fname, "wb") as ofile:
        ofile.write(etree.tostring(fe_xml, pretty_print=True))


def raise_error(ast, msg):
    raise RuntimeError(get_ast_data(ast) + ": " + msg)


# TODO: Tags

def process_id(ast):
    if ast.type == feiLexer.BASIC_ID:
        return str(ast)

    assert ast.type == feiLexer.FULL_ID
    return ".".join(str(i) for i in ast.getChildren())


def process_namespace_id(ast):
    """
    return tuple: (var_name, namespace)
    namespace is None if no namespace is provided
    """

    assert ast.type == feiLexer.NAMESPACE_ID

    children = ast.getChildren()
    if len(children) == 2:
        return (process_id(children[1]), process_id(children[0]))

    assert len(children) == 1

    return (process_id(children[0]), None)


def process_effects(ast, lib_manager):
    assert ast.type == feiLexer.EFFECT

    em = effect_model.EffectModel()

    # Children:
    # 0: em name
    # 1: probability (optional)
    # [12]: parameters
    children = ast.getChildren()
    em.name = process_id(children[0])
    em.set_lineno(ast.line)

    maps_to = {}
    events = []
    values = []
    templates = []

    for c in children[1].getChildren():
        if c.type == feiLexer.DATA:
            io, value, isvar = process_rw(c.getChildren()[1])
            rw_var, ns = process_namespace_id(c.getChildren()[0])

            d = None

            if io == "input":
                d = _input.Input()
                d.reads = rw_var
                d.value = value
            else:
                assert io == "output"
                assert isvar
                d = output.Output()
                d.writes = rw_var
                d.binds = value

            # TODO: d.type = ??

            # Here we check if we already had a binding for 'var'. If
            # it is the case, merge them in one unique Output data.
            if isvar:
                if value in maps_to:
                    last, last_ns, ic, oc = maps_to[value]
                    assert oc is None

                    if isinstance(last, _input.Input):
                        assert isinstance(d, output.Output), \
                            "Object {0} not an output instance, last is: {1}, is var is: {2}".format(d, last, isvar)
                        d.reads = last.reads
                        maps_to[value] = (d, ns, ic, None)
                    else:
                        assert isinstance(last, output.Output)
                        assert isinstance(d, _input.Input)
                        last.reads = d.reads
                else:
                    maps_to[value] = (d, ns, c, None)
            else:
                # bare input
                assert io == "input"
                values.append((d, ns, c, None))

        elif c.type == feiLexer.EVENT:
            ev, ns = process_event(c)
            events.append((ev, ns, c))

        else:
            assert c.type == feiLexer.TEMPLATE
            t = template.Template()
            t.name, ns = process_namespace_id(c.getChild(0))
            t.value = process_id(c.getChild(1))

            templates.append((t, ns, c))

    for v, ns, ic, oc in maps_to.itervalues():
        # Add the namespace to the values
        values.append((v, ns, ic, oc))

    return em, values, events, templates


def process_local_dynamics(ast):
    ldm = local_dynamics.LocalDynamics()
    ldm.name = process_id(ast)
    ldm.set_lineno(ast.line)

    return ldm


def process_fault_mode(ast, lib_manager):
    assert ast.type == feiLexer.MODE

    fm = fault_mode.FaultMode()
    fm.set_lineno(ast.line)

    # Children:
    # 0: fm name
    # 1: (optional) probability
    # [12]: Local dynamics id
    # [23]: effects
    children = ast.getChildren()
    fm.user_info = user_info.UserInfo(name=process_id(children[0]))

    idx = 1
    if children[1].type == feiLexer.MODE_OPTS:
        idx += 1
        _opts = process_opt_list(children[1])
        pt, pv = process_opt_probability(_opts)
        fm.probability = pt
        fm.latent = process_opt_latent(_opts)
        if fm.latent is not None:
            fm.latent.set_lineno(children[1].line)

    fm.ldm = process_local_dynamics(children[idx])
    fm.em, values, events, templates = process_effects(
        children[idx+1], lib_manager)

    process_fault_mode_library(ast, lib_manager, fm, values, events, templates)
    return fm


def get_ast_data(ast, print_col=False):
    if ast is None:
        return ""
    res = "Line " + str(ast.line)
    if print_col:
        res += " Col:" + str(ast.charPositionInLine)
    return res


def process_fault_mode_library(ast, lib_manager, fm, values,
                               events, templates):

    libs = lib_manager.getEmlObjectify()
    lib = libs.get(fm.em.name)

    ldm_libs = lib_manager.getLdlObjectify()
    ldm_lib = ldm_libs.get(fm.ldm.name)

    if lib is None:
        raise_error(ast, "EM Library '" + fm.em.name + "' not found")

    if ldm_lib is None:
        raise_error(ast, "LDM Library '" + fm.ldm.name + "' not found")

    em_inputs = set(lib.xpath("./values/input/@reads"))
    em_outputs = set(lib.xpath("./values/output/@writes"))

    ldm_inputs = set(ldm_lib.xpath("./values/input/@reads"))
    ldm_outputs = set(ldm_lib.xpath("./values/output/@writes"))

    #print "FOR EM:", em_inputs, em_outputs
    #print "FOR LDM:", ldm_inputs, ldm_outputs

    for v, ns, ast1, ast2 in values:
        if isinstance(v, _input.Input):
            if ns:
                if ns == fm.em.name:
                    if v.reads in em_inputs:
                        fm.em.values.append(v)
                    else:
                        raise_error(ast1, "Input '" + v.reads + "' "
                                    "not found in the specified '" +
                                    ns + "' EM library")
                elif ns == fm.ldm.name:
                    if v.reads in ldm_inputs:
                        fm.ldm.values.append(v)
                    else:
                        raise_error(ast1, "Input '" + v.reads + "' "
                                           "not found in the specified '" +
                                           ns + "' LDM library")
                else:
                    raise_error(ast1, "Specified namespace '" +
                                ns + "' does not exist")

            else:
                if v.reads in em_outputs and v.reads in ldm_outputs:
                    fm.em.values.append(v)
                    fm.ldm.values.append(v)
                elif v.reads in em_inputs:
                    fm.em.values.append(v)
                elif v.reads in ldm_inputs:
                    fm.ldm.values.append(v)
                else:
                    raise_error(ast1,
                                "Input '" + v.reads + "' "
                                "not found in the library (neither in EM and LD)")

        else:
            if ns:
                if ns == fm.em.name:
                    if v.writes in em_outputs:
                        fm.em.values.append(v)
                    else:
                        raise_error(ast1, "Output '" + v.writes + "' "
                                    "not found in the specified '" +
                                    ns + "' EM library")
                elif ns == fm.ldm.name:
                    if v.writes in ldm_outputs:
                        fm.ldm.values.append(v)
                    else:
                        raise_error(ast1, "Output '" + v.writes + "' "
                                    "not found in the specified '" +
                                    ns + "' LDM library")
                else:
                    raise_error(ast1, "Specified namespace '" +
                                ns + "' does not exist")
            else:
                assert isinstance(v, output.Output)
                if v.writes in em_outputs and v.writes in ldm_outputs:
                    fm.em.values.append(v)
                    fm.ldm.values.append(v)
                elif v.writes in em_outputs:
                    fm.em.values.append(v)
                elif v.writes in ldm_outputs:
                    fm.ldm.values.append(v)
                else:
                    raise_error(ast1, "Output '" + v.writes + "' "
                                "not found in the library (neither in EM and LD)")

    # here templates are filled and checked
    em_templates = set(lib.xpath("templates/template/@name"))
    ld_templates = set(ldm_lib.xpath("templates/template/@name"))

    for t, ns, t_ast in templates:
        if ns:
            if ns == fm.em.name:
                if t.name in em_templates:
                    fm.em.templates.append(t)
                else:
                    raise_error(t_ast, "Template '" + t.name + "' "
                                "not found in the specified '" +
                                ns + "' EM library")
            elif ns == fm.ldm.name:
                if t.name in em_templates:
                    fm.em.templates.append(t)
                else:
                    raise_error(t_ast,
                                "Template '" + t.name + "' "
                                "not found in the specified '" +
                                ns + "' EM library")
        else:
            if t.name in em_templates and t.name in ld_templates:
                fm.em.templates.append(t)
                fm.ldm.templates.append(t)
            if t.name in em_templates:
                fm.em.templates.append(t)
            elif t.name in ld_templates:
                fm.ldm.templates.append(t)
            else:
                raise_error(t_ast, "The template '%s' could not be found "
                            "in the library (neither EM and LD)" % t.name)

    # checks that wrong names are not provided
    req_temp = set(em_templates) | set(ld_templates)
    prov_temp = set(t.name for t, _, _ in templates)
    mis_temp = req_temp - prov_temp
    if mis_temp:
        raise_error(ast,
                    "In Fault Mode '%s', one or more required templates "
                    "were not provided: '%s'" % (fm.user_info.name,
                                                 ", ".join(mis_temp)))

    # checks that templates' values are also defined
    symbols = set()
    for v, ns, ast1, ast2 in values:
        if isinstance(v, _input.Input):
            symbols.add(v.reads)
            #symbols.add(v.value)
        else:
            assert isinstance(v, output.Output)
            symbols.add(v.reads)
            symbols.add(v.writes)
    for e, ns, ev_ast in events:
        symbols.add(e.name)

    for t, ns, t_ast in templates:
        if t.value not in symbols:
            raise_error(t_ast,
                        "Template '%s' has value '%s' which has no "
                        "corresponency in the Fault Mode '%s' interface.\n"
                        "Recognized values are: %s\n"
                        "For events template, you may want to add "
                        "an event '%s' to the FM's interface." %
                        (t.name, t.value, fm.user_info.name,
                         ", ".join(sorted(symbols)), t.value))

    # events il ldm
    ld_events = set(ldm_lib.xpath("events/event/@name"))
    for e, ns, ev_ast in events:
        if ns:
            if ns == fm.em.name:
                fm.em.events.append(e)
            elif ns == fm.ldm.name:
                fm.ldm.events.append(e)
            else:
                raise_error(ev_ast, "Specified namespace '" +
                            ns + "' is not valid")
        else:
            if e.name in ld_events:
                fm.ldm.events.append(e)
            else:
                ref_temp = None
                for t, ns, _ in templates:
                    if t.value == e.name:
                        ref_temp = t
                        break

                if ref_temp is None:
                    raise_error(ev_ast, "Event '" + e.name +
                                "' not found in ldm " + fm.ldm.name)

                if t in fm.ldm.templates:
                    fm.ldm.events.append(e)
                else:
                    assert t in fm.em.templates
                    fm.em.events.append(e)

    return


def process_rw(ast):
    assert ast.type in (feiLexer.RW_OP_GT, feiLexer.RW_OP_LT)

    io = "input" if ast.type == feiLexer.RW_OP_LT else "output"

    if ast.type == feiLexer.RW_OP_GT:
        return io, process_id(ast.getChildren()[0]), True

    # instead, RW_OP_LT operator has a simple_expression as right
    # child.
    value = " ".join(str(t) for t in ast.getChildren()).strip()
    #print(value)
    isvar = re.match("[a-zA-Z_]\w*", value) is not None
    #print(isvar)

    return io, value, isvar


def process_event(ast):
    assert ast.type == feiLexer.EVENT

    # Children:
    # 0: event id
    # 1: RW_operator
    children = ast.getChildren()

    ev = event.Event()
    ev.name, ns = process_namespace_id(children[0])
    ev.set_lineno(ast.line)

    if len(children) >= 2:
        io, ev.binds, isvar = process_rw(children[1])
        assert isvar
        connected = "yes"
    else:
        io, ev.binds = "output", ""
        connected = "no"

    ev.type = io
    ev.connected = connected
    return ev, ns


def process_trans_mode(ast):
    assert ast.type == feiLexer.TRANS_MODE

    c = ast.getChildren()
    assert c[0].type in (feiLexer.NOMINAL, feiLexer.LOWER_FAULT)

    _mode = str(c[0])

    if len(c) > 1:
        return process_id(c[1]) + "." + _mode

    return _mode


def process_trans(ast):
    assert ast.type == feiLexer.TRANS

    trans = transition.Transition()
    trans.set_lineno(ast.line)

    # Children:
    # 0: trans_mode (from)
    # 1: trans_guard
    # 2: trans_mode (to)
    children = ast.getChildren()

    setattr(trans, "from", process_trans_mode(children[0]))
    trans.to = process_trans_mode(children[2])

    trans_ast = children[1].getChildren()

    trans.trigger = None
    trans.guard = None

    for c in trans_ast:
        if c.type == feiLexer.WHEN:
            trans.guard = " ".join(str(s) for s in c.getChildren())
        else:
            trans.trigger = process_id(c)

    return trans


def process_global_dynamics(ast):
    assert ast.type == feiLexer.GLOBAL_DYNAMICS

    gm = global_dynamics.GlobalDynamics()
    gm.set_lineno(ast.line)

    # Children:
    # 0: events
    # 1+: trans
    children = ast.getChildren()
    #assert children[0].type == feiLexer.EVENTS

    gm.transitions = []
    for c in children:
        if c.type == feiLexer.EVENT:
            # TODO: process namespace
            gm.events.append(process_event(c)[0])
        else:
            assert c.type == feiLexer.TRANS
            gm.transitions.append(process_trans(c))

    return gm


def process_instance(ast):
    assert ast.type == feiLexer.WILDCARD_ID

    return "".join(str(c) for c in ast.getChildren())


def process_fault_model(ast, lib_manager, instances):
    assert ast.type == feiLexer.SLICE

    fm = fault_model.FaultModel()

    # Children:
    # 0: FM name
    # 1: Variables
    # 2: fault modes
    # 3?: global dynamics
    children = ast.getChildren()

    fm.name = process_id(children[0])
    fm.set_lineno(ast.line)

    fm.instances = []

    _vars = {}
    for c in children[1:]:
        if c.type == feiLexer.VARS:

            # TODO: ONLY Check that vars are actually used (right side
            # of the >> or <<)
            for v, s in ((process_id(s), s) for s in c.getChildren()):
                _vars[v] = s
            pass
        elif c.type == feiLexer.FAULT_MODES:
            fm.fault_modes = [process_fault_mode(n, lib_manager)
                              for n in c.getChildren()]
        elif c.type == feiLexer.GLOBAL_DYNAMICS:
            fm.gdm = process_global_dynamics(c)
        elif c.type == feiLexer.INSTANCES:
            fm.instances = [process_instance(n)
                            for n in c.getChildren()]
        else:
            assert False

    # Collect all right-sided symbols in data, template, events of the
    # fault modes of this fault model. All vars declared in the
    # "AFFECTED" field have to be listed in the fault modes (and the
    # opposite?)

    symbs = set()
    for _fm in fm.fault_modes:
        for v in itertools.chain(_fm.em.values, _fm.ldm.values):
            if isinstance(v, _input.Input):
                symbs.add(v.value)
            else:
                symbs.add(v.binds)

        for e in _fm.ldm.events:
            symbs.add(e.binds)

        for t in itertools.chain(_fm.em.templates, _fm.ldm.templates):
            symbs.add(t.name)
            pass

    res_set = set(_vars.iterkeys()) - symbs
    if len(res_set) > 0:
        f = res_set.pop()
        raise_error(_vars[f], "Affected variable '" + f + "' not found in "
                    "any fault mode of fault model '" + fm.name + "'")

    fm.instances += instances
    #print("FM INSTANCES " + fm.instances)

    #if fm.gdm is None:
    #    fm.gdm = global_dynamics.GlobalDynamics()

    return fm


def process_extension(ast, lib_manager):
    ext = None

    # Children:
    # 0: extension module/instance
    # 1..N: Slices
    children = ast.getChildren()

    assert ast.type == feiLexer.MOD_EXTENSION
    ext = nominal_comp_ext.NCExtension()
    ext.module = process_id(children[0])
    ext.set_lineno(ast.line)

    instances = []
    for c in children[1:]:
        if c.type == feiLexer.SLICE:
            ext.fault_models.append(process_fault_model(c,
                                                        lib_manager,
                                                        instances))
        else:
            assert c.type == feiLexer.INSTANCES
            instances = [process_instance(n)
                         for n in c.getChildren()]

    return ext


def process_causes(ast):
    assert ast.type == feiLexer.COMMON_CAUSES
    children = ast.getChildren()
    ext = common_cause.CommonCauses()
    ext.causes = [process_cause(ext_ast)
                  for ext_ast in children]
    return ext


def process_cause(ast):
    assert ast.type == feiLexer.CC_CAUSE
    children = ast.getChildren()
    ext = common_cause.CommonCause()
    ext.name = process_id(children[0])
    ext.set_lineno(ast.line)
    idx = 1
    if children[1].type == feiLexer.CAUSE_OPTS:
        idx += 1
        _opts = process_opt_list(children[1])
        pt, pv = process_opt_probability(_opts)
        ext.probability = pt

    ext.cc_modules = [process_cc_module(ext_ast)
                      for ext_ast in children[idx:]]
    return ext


def process_opt_list(ast):
    """Process mode and cause options list. Returns a map with name and
    value"""

    res = {}
    assert ast.type in (feiLexer.CAUSE_OPTS, feiLexer.MODE_OPTS)
    for opt in ast.getChildren():
        assert feiLexer.OPT_PARAM == opt.type
        name, val = opt.getChildren()
        tok = name.token.type
        assert tok not in res, \
            "At line %d, duplicated option '%s'" % (opt.line, name.text)
        res[tok] = val.text

    return res


def process_opt_probability(_map):
    """Takes an option map as returned by process_opt_list.
    Returns a pair (string, float)"""
    val = (_map[feiLexer.OPT_PROB]
           if feiLexer.OPT_PROB in _map
           else "0")

    return (val, float(val))


def process_opt_latent(_opts):
    """Takes an option map as returned by process_opt_list. Returns a
       LatentFaultInfo instance if information about latency is found,
       or else returns None
    """

    if (feiLexer.OPT_LATENT not in _opts and
            feiLexer.OPT_LATENT_PROB not in _opts):
        return None

    is_latent = _opts.get(feiLexer.OPT_LATENT, "yes")
    prob = _opts.get(feiLexer.OPT_LATENT_PROB, "0")

    return fault_mode.LatentFaultInfo(is_latent=is_latent,
                                      probability=prob)


def process_cc_module(ast):
    assert ast.type == feiLexer.CC_MODULE
    children = ast.getChildren()

    ext = common_cause.CCModule()
    ext.name = process_id(children[0])
    ext.set_lineno(ast.line)
    if children[1].type == feiLexer.INSTANCES:
        ext.instances = [process_instance(n)
                         for n in children[1].getChildren()]
    ext.cc_modes = [process_cc_mode(n)
                    for n in children
                    if n.type == feiLexer.CC_MODE]
    return ext


def process_cc_mode(ast):
    assert ast.type == feiLexer.CC_MODE
    children = ast.getChildren()
    ext = common_cause.CCMode()
    ext.slice = process_id(children[0])
    ext.mode = process_id(children[1])
    ext.set_lineno(ast.line)
    ext.low, ext.high = process_cc_range(children[2])
    return ext


def process_cc_range(ast):
    assert ast.type == feiLexer.CC_RANGE
    children = ast.getChildren()
    assert len(children) == 2
    return [int(c.text) for c in children]


def traverse_ast(ast, lib_manager, smv=None, fei_txt=None):
    assert ast.type == feiLexer.FAULT_EXTENSION

    fe = fault_ext_info.FaultEstensionInfo(lib_manager,
                                           fei_txt=fei_txt)
    if smv:
        fe.smv_file = smv
    fe.set_lineno(ast.line)

    # Children:
    # 0: FE name
    # 1..N: Extensions
    children = ast.getChildren()
    fe.name = process_id(children[0])
    mod_extensions = (ext_ast
                      for ext_ast in children[1:]
                      if ext_ast.type == feiLexer.MOD_EXTENSION)

    cc_causes = (process_causes(children[-1])
                 if children[-1].type == feiLexer.COMMON_CAUSES
                 else None)

    fe.extensions = [process_extension(ext_ast, lib_manager)
                     for ext_ast in mod_extensions]

    fe.causes = cc_causes
    return fe


def get_library_manager(prefix):
    """
    Generates an instance of the library manager that is
    used to retrive the EffectModel and LocalDynamicsModel
    libraries.

    @param prefix: Path where the libraries are stored.
    """
    from feic.fe.lib_mgr.libraries.manager import LibraryManager

    eml = os.path.join(prefix, "eml.xml")
    emlS = os.path.join(prefix, "eml.xsd")
    ldl = os.path.join(prefix, "ldl.xml")
    ldlS = os.path.join(prefix, "ldl.xsd")
    eDefinS = os.path.join(prefix, "eml_behavior/definition.xsd")
    lDefinS = os.path.join(prefix, "ldl_behavior/definition.xsd")
    libManager = LibraryManager(prefix, eml, emlS, ldl, ldlS, eDefinS, lDefinS)
    return libManager


__OPT_NAME_LIB_PATH = "lib_path"


def handle_options():
    import argparse
    from feic import feicm

    parser = argparse.ArgumentParser(
        description="Translates FEI documents from TXT to XML (yet)")

    parser.add_argument('-i', '--input',
                        action="store",
                        required=True,
                        help="The input file")

    parser.add_argument('-o', '--output',
                        action="store",
                        required=True,
                        help="The output file")

    parser.add_argument('-s', '--smv',
                        action="store",
                        help="The SMV file")

    parser.add_argument("-p", "--path", dest=__OPT_NAME_LIB_PATH,
                        default=feicm.get_default_fault_lib_path(),
                        metavar="PATH",
                        help="Path to the extension library (default:%default")

    args = parser.parse_args()

    # args check
    if not os.path.isdir(getattr(args, __OPT_NAME_LIB_PATH)):
        parser.error("Lib path not found:", getattr(args, __OPT_NAME_LIB_PATH))

    return args





'''
if __name__ == "__main__":

    args = handle_options()
    try:
        convert_fei_text_to_xml_format(args.input, args.output,
                                       getattr(args, __OPT_NAME_LIB_PATH),
                                       args.smv)
    except RuntimeError, e:
        print "ERROR:", e
'''
