
# %HEADER

import os


class TfpgAssociations(dict):
    """TfpgAssociations class extends the dict class.
       The idea is that TfpgAssociations are represented as a dict in which:
       + a key represents the name of a variable
       + a value is a 3-tuple where:
         - the 1st element is a value in ("OK", "ERROR") saying whether
           the associated expression is correct
         - the 2nd element is the expression
         - the 3rd element is a value in ("MonitoredDiscr",
           "UnmonitoredDiscr", "FailureMode", "TfpgMode") specifying
           the type of expression
    """

    def __init__(self, *args, **kw):
        super(TfpgAssociations, self).__init__(*args, **kw)
        self.itemlist = list(super(TfpgAssociations, self).keys())

    @staticmethod
    def parse_assoc(fn_assoc):
        """ This is a parsing function for TFPG associations accepting both XML
        and TXT formats. Error checking is performed.

        @param   fn_assoc: name of file containing TFPG associations
        @type    fn_assoc: string

        @return: The parsed TfpgAssociations object, or None in case of errors.
        @rtype:  TfpgAssociations / None
        """

        assoc_file_ext = os.path.splitext(fn_assoc)[1]
        assoc_struct = None

        if assoc_file_ext in [".axml", ".xml"]:
            from xsap.data_structure.tfpg.xml.tfpg_associations import (
                AssociationsDocument,
                parse_assoc_xml_file_to_internal_struct)

            # TODO: the following check on input file errors is a temporary fix
            #       and should be handled natively in the parsing function
            assoc_doc, validation_errors = AssociationsDocument().parse(fn_assoc)
            if len(validation_errors) > 0:
                return None
            if len(AssociationsDocument.get_syntax_errors_in_assoc_document(assoc_doc)) > 0:
                return None

            assoc_struct = parse_assoc_xml_file_to_internal_struct(fn_assoc)

        elif assoc_file_ext == ".tfpga":
            from xsap.data_structure.tfpg.txt.tfpg_associations import (
                parse_assoc_txt_file_to_internal_struct)
            assoc_struct = parse_assoc_txt_file_to_internal_struct(fn_assoc)

        else:
            # unknown file format
            return None

        assert assoc_struct is not None
        return assoc_struct

    def __setitem__(self, key, value):
        if key in list(self.keys()):
            raise Exception("Two ore more associations with the same "
                            "key have been specified.")
        if key != '':
            self.itemlist.append(key)
            super(TfpgAssociations, self).__setitem__(key, value)

    def __iter__(self):
        return iter(self.itemlist)

    def keys(self):
        return self.itemlist

    def values(self):
        return [self[k] for k in self.itemlist]

    def itervalues(self):
        return (self[k] for k in self.itemlist)

    def get_expression_from_var(self, var):
        """Given the name of the variable, if stored, it returns
           the linked expression.

           @param var: the name of the variable for which we want to get
                       the associated expression
           @type  var: string

           @return: the associated expression
           @rtype:  string
        """
        if var not in list(self.keys()):
            return None
        return self[var][1]

    def get_var_names_of_nodes(self):
        """Returns the names of all TFPG nodes specified in this association
        structure."""
        return self.get_var_names_of_monitored_discr() + \
               self.get_var_names_of_unmonitored_discr() + \
               self.get_var_names_of_failure_modes()

    def get_var_names_of_monitored_discr(self):
        """Returns the names of the variables related to associations about
        monitored discrepancies"""

        mon_discr = []
        for k in self.itemlist:
            if self[k][2] == "MonitoredDiscr":
                mon_discr.append(k)
        return mon_discr

    def get_var_names_of_unmonitored_discr(self):
        """Returns the names of the variables related to associations about
        unmonitored discrepancies"""

        unmon_discr = []
        for k in self.itemlist:
            if self[k][2] == "UnmonitoredDiscr":
                unmon_discr.append(k)
        return unmon_discr

    def get_var_names_of_failure_modes(self):
        """Returns the names of the variables related to associations about
        failure modes"""

        failure_modes = []
        for k in self.itemlist:
            if self[k][2] == "FailureMode":
                failure_modes.append(k)
        return failure_modes

    def get_var_names_of_tfpg_modes(self):
        """Returns the names of the variables related to associations about
        tfpg modes"""

        tfpg_modes = []
        for k in self.itemlist:
            if self[k][2] == "TfpgMode":
                tfpg_modes.append(k)
        return tfpg_modes
