from xsap.data_structure.tfpg.tfpg_errors import UnreachableDiscrepancy
from xsap.data_structure.tfpg.tfpg_errors import NodeNotInEdge


def antlr4_parse_tfpg_txt_file_to_internal_struct(filename):

    from xsap.data_structure.tfpg.txt.grammars.generated4.TfpgVisitor import TfpgVisitor
    from xsap.data_structure.tfpg.tfpg import Tfpg, OPEN_INFINITY
    from xsap.data_structure.tfpg.tfpg import CLOSED_INFINITY, INFINITY

    class Visitor(TfpgVisitor):
        def __init__(self):
            super(Visitor, self).__init__()
            self.tfpg_graph = Tfpg()
            self.errors = []


        def visitTfpg(self, ctx):
            name = self.visitTfpg_Name(ctx.tfpg_name())
            infinity_semantics = self.visitInfty_semantics(ctx.infty_semantics())

            #print("NAME " + name)
            #print("INFINITY SEMANTICS " + infinity_semantics)

            nodes = ctx.node()
            for node in nodes:
                n = self.visitNode(node)
                #print("NODE " + n)

            modes = ctx.modes_list()
            for mode in modes:
                mode_string = self.visitModes_list(mode)
                #print("MODES " + mode_string)

            edges = ctx.edge()
            for edge in edges:
                ed = self.visitEdge(edge)
                #print("EDGE " + ed)


        def visitTfpg_Name(self, ctx):
            name_ident = ctx.IDENT()
            id = name_ident.getText()
            self.tfpg_graph.set_name(id)
            return id


        def visitInfty_semantics(self, ctx):
            inf_sem_open = ctx.INFINITY_SEMANTICS_OPEN()
            if inf_sem_open:
                self.tfpg_graph.set_infinity_semantics(OPEN_INFINITY)
                return inf_sem_open.getText()
            else:
                inf_sem_close_token = ctx.INFINITY_SEMANTICS_CLOSED()
                assert inf_sem_close_token is not None, "Invalid value for infinity semantics"
                self.tfpg_graph.set_infinity_semantics(CLOSED_INFINITY)
                return inf_sem_close_token.getText()


        def visitNode(self, ctx):

            fm_token = ctx.FM()
            and_token = ctx.AND()
            or_token = ctx.OR()
            node_id = ctx.IDENT()

            if fm_token:
                self.tfpg_graph.add_fm_node(node_id.getText())
            elif and_token:
                monitored = True if ctx.MONITORED() else False
                self.tfpg_graph.add_and_node(node_id.getText(), monitored)
            elif or_token:
                monitored = True if ctx.MONITORED() else False
                self.tfpg_graph.add_or_node(node_id.getText(), monitored)

            return node_id.getText()


        def visitModes_list(self, ctx):
            modes = []
            for id in ctx.IDENT():
                modes.append(id.getText())
            self.tfpg_graph.set_modes_list(modes)
            return ' '.join(modes)


        def visitEdge(self, ctx):
            edg_name = self.visitEdge_name(ctx.edge_name())
            edg_src = self.visitEdge_src(ctx.edge_src())
            edg_dst = self.visitEdge_dst(ctx.edge_dst())

            edg_tMin = self.visitMin_time(ctx.min_time())
            edg_tMax = self.visitMax_time(ctx.max_time())

            edg_tMin_value = float(edg_tMin)
            edg_tMax_value = INFINITY if edg_tMax == 'INFINITY' else float(edg_tMax)
            edg_modes = []
            modes_string = self.visitEdge_modes_list(ctx.edge_modes_list())
            if modes_string == "ALL":
                for m in self.tfpg_graph.get_modes_list():
                    edg_modes.append(m)
            else:
                for m in modes_string.split():
                    edg_modes.append(m)

            if self.tfpg_graph.get_node_by_name(edg_src) is None:
                self.errors.append(NodeNotInEdge(edg_src, edg_name))

            if self.tfpg_graph.get_node_by_name(edg_dst) is None:
                self.errors.append(NodeNotInEdge(edg_dst, edg_name))

            if self.tfpg_graph.get_node_by_name(edg_src) and \
            self.tfpg_graph.get_node_by_name(edg_dst):
                self.tfpg_graph.add_edge(self.tfpg_graph.get_node_by_name(edg_src),
                                    edg_tMin_value, edg_tMax_value, edg_modes,
                                    self.tfpg_graph.get_node_by_name(edg_dst),
                                    edg_name)
            return edg_src + " " + \
                str(edg_tMin_value) + " " + \
                    str(edg_tMax_value) + \
                        " " + " ".join(edg_modes) + \
                            " " + edg_dst + " " + edg_name


        def visitEdge_name(self, ctx):
            ident = ctx.IDENT()
            return ident.getText()


        def visitEdge_src(self, ctx):
            ident = ctx.IDENT()
            return ident.getText()


        def visitEdge_dst(self, ctx):
            ident = ctx.IDENT()
            return ident.getText()


        def visitEdge_modes_list(self, ctx):
            all_token = ctx.ALL()
            if all_token:
                return all_token.getText()
            else:
                identifiers = ctx.IDENT()
                ids = identifiers[0].getText()
                for id in identifiers[1:]:
                    ids = ids + " " + id.getText()
                return ids


        def visitMin_time(self, ctx):
            min_time = ctx.FLOAT()
            return min_time.getText()


        def visitMax_time(self, ctx):
            infinity_token = ctx.INFINITY()
            if infinity_token:
                return infinity_token.getText()
            else:
                max_time = ctx.FLOAT()
                return max_time.getText()


        def get_tfpg_graph(self):
            return self.tfpg_graph

        def get_errors(self):
            return self.errors

    import antlr4
    from xsap.data_structure.tfpg.txt.grammars.generated4.TfpgLexer import TfpgLexer
    from xsap.data_structure.tfpg.txt.grammars.generated4.TfpgParser import TfpgParser

    char_stream = antlr4.FileStream(filename)
    
    lexer = TfpgLexer(char_stream)
    tokens = antlr4.CommonTokenStream(lexer)
   
    parser = TfpgParser(tokens)
    
    tree = parser.tfpg()
    visitor = Visitor()
    visitor.visit(tree)
    
    tfpg_graph = visitor.get_tfpg_graph()
    
    errors = visitor.get_errors()
    
    return tfpg_graph, errors

def parse_tfpg_txt_file_to_internal_struct(filename):
    """Parses a file written in human readable format and returns the
    corresponding tfpg internal structure.
    @param   filename:  The path of the tfpg file.
    @type    filename:  string

    @return   tfpg_graph:  The Tfpg structural.
    @rtype    tfpg_graph:  Tfpg
    """
    errors = None
    tfpg_graph = None

    
    tfpg_graph, errors = antlr4_parse_tfpg_txt_file_to_internal_struct(filename)

    for tfpg_err in tfpg_graph.get_errors_list():
        errors.append(tfpg_err)
    if errors != []:
        for err in errors:
            if not isinstance(err, UnreachableDiscrepancy):
                raise err

    return tfpg_graph, errors


def dump_tfpg_to_txt_file(tfpg_struct, out_filename):
    """Saves the given tfpg in human readable format.

    @param tfpg_struct: the internal representation of the tfpg
    @type  tfpg_struct: data_structure.tfpg.tfpg.Tfpg

    @param out_filename: The name of the file in which the tfpg will be dumped.
    @type  out_filename: string
    """
    from xsap.data_structure.tfpg.tfpg import INFINITY

    with open(out_filename, "w") as f:
        name = tfpg_struct.get_name()
        if name == "":
            name = "tfpg"
        f.write("NAME %s\n" % name)

        if tfpg_struct.has_open_infinity_semantics():
            f.write("INFINITY_SEMANTICS_OPEN\n")
        else:
            assert tfpg_struct.has_closed_infinity_semantics()
            f.write("INFINITY_SEMANTICS_CLOSED\n")

        for n in tfpg_struct.get_nodes_list():
            _type = n.get_type()
            name = n.get_name()
            if n.is_monitored():
                mon = "MONITORED"
            else:
                mon = ""
            f.write("%s %s %s\n" % (_type, name, mon))

        if tfpg_struct.get_modes_list() != []:
            f.write("MODES %s\n" % (",".join(tfpg_struct.get_modes_list())))

        for e in tfpg_struct.get_edges_list():
            if float(e.get_tMax()) == INFINITY:
                max_val = "INFINITY"
            else:
                max_val = e.get_tMax()
            f.write("EDGE %s %s %s %s %s (%s)\n" % (
                    e.get_name(),
                    e.get_srcNode().get_name(),
                    e.get_destNode().get_name(),
                    e.get_tMin(),
                    max_val,
                    (",".join(e.get_modes())))
                    )
        return
