
# %HEADER

from xsap.data_structure.tfpg.tfpg import NodeType
from xsap.features.common.tfpg_symb_converter import (
    compute_smv_em_params_name,
    compute_smv_etmin_params_name,
    compute_smv_etmax_real_params_name,
    compute_smv_etmax_bool_params_name,
    compute_smv_edge_clock_name,
    compute_smv_edge_state_name,
    compute_smv_mode_name,
    compute_smv_node_name)
from xsap.features.common.ProofObligationMTL import ProofObligationMTL


class ProofObligationINVAR(ProofObligationMTL):

    def __init__(self, po_mtl):
        """
        @param po_mtl: the MTL proof obligation from which to derive the information
        @type  po_mtl: ProofObligationMTL
        """

        super(ProofObligationINVAR, self).__init__(po_mtl.tfpg, po_mtl.assoc,
                                                   po_mtl.parametric, po_mtl.tags)

    # INTERFACES

    def __str__(self):
        """
        @return: Returns an invariant property corresponding to the completeness proof obligation
                 in MTL, assuming that the model has been extended with the symbols that
                 get_extension_data() returns (see below).
        @rtype:  string
        """

        # nb: due to the overridden methods, the formula does not contain the auxiliary
        # placeholders used in the MTL case; thus, we don't need to call additionally
        # the function replace_placeholders().
        return self.get_completeness_po()

    def get_extension_data(self):
        """
        @return: Returns the list of symbols (history variables, defines, edge timers, and
                 parameters) that the SMV model needs to be extended with such that the invariant
                 property that can be obtained with get_invar_reduction_proof_obligation() can be
                 checked.
        @rtype:  dict
        """

        ls_hvar = []
        ls_defines = []
        ls_edge_timers = []
        ls_parameters = []

        # nodes history variables (defines for virtual nodes)
        for n in self.tfpg.get_nodes_list():

            assoc_expr = self.assoc.get_expression_from_var(n.get_name())

            if assoc_expr is None:
                # n is a virtual discrepancy
                preds = self.tfpg.get_predecessors(n)
                assert len(preds) > 0

                if n.get_type() == NodeType.OR:
                    assoc_expr = " | ".join([compute_smv_node_name(v) for v in preds])
                else:
                    assoc_expr = " & ".join([compute_smv_node_name(v) for v in preds])

                ls_defines.append({"name": compute_smv_node_name(n),
                                   "expression": assoc_expr})

            else:
                ls_hvar.append({"name": compute_smv_node_name(n),
                                "expression": assoc_expr})

        # mode defines
        for m in self.tfpg.get_modes_list():
            ls_defines.append({"name": compute_smv_mode_name(m),
                               "expression": self.assoc.get_expression_from_var(m)})

        # edge timers and edge states
        for e in self.tfpg.get_edges_list():

            # edge modes
            ls_mode_expr = []
            for m in e.get_modes():

                # edge mode parameter (if necessary)
                if self.parametric and len(e.get_modes()) > 1:
                    em_param = compute_smv_em_params_name(e, m)
                    ls_mode_expr.append("(%s & %s)" % (em_param, compute_smv_mode_name(m)))
                    ls_parameters.append({"name": em_param, "type": "boolean"})
                else:
                    ls_mode_expr.append(compute_smv_mode_name(m))

            # edge state
            edge_state_expr = "%s & (%s)" % (compute_smv_node_name(e.get_srcNode()),
                                             " | ".join(ls_mode_expr))
            ls_defines.append({"name": compute_smv_edge_state_name(e),
                               "expression": edge_state_expr})

            # edge timer
            ls_edge_timers.append({"name": compute_smv_edge_clock_name(e),
                                   "state": compute_smv_edge_state_name(e),
                                   "target": compute_smv_node_name(e.get_destNode())})

            # tmin/tmax parameters
            if self.parametric:
                ls_parameters.append({"name": compute_smv_etmin_params_name(e), "type": "real"})
                ls_parameters.append({"name": compute_smv_etmax_real_params_name(e), "type": "real"})
                ls_parameters.append({"name": compute_smv_etmax_bool_params_name(e), "type": "boolean"})

        return {'history_variables': ls_hvar,
                'defines': ls_defines,
                'edge_timers': ls_edge_timers,
                'parameters': ls_parameters}

    # INTERNAL FUNCTIONS

    def get_edge_tmin_expr(self, tfpg_edge):

        if self.parametric and "tmin" in self.tags:
            tmin_symb = compute_smv_etmin_params_name(tfpg_edge)
        else:
            tmin_symb = "%f" % tfpg_edge.get_tMin()

        etimer = compute_smv_edge_clock_name(tfpg_edge)
        return "%s >= %s" % (etimer, tmin_symb)

    def get_tmin_po_for_discrepancy(self, discrepancy):

        # G (symb_node_D -> O (symb_node_D & (
        #       (symb_edge_FM1_D S[1,inf) symb_edge_FM1_D) |
        #       (symb_edge_FM2_D S[0,inf) symb_edge_FM2_D)
        # )))

        d_active = compute_smv_node_name(discrepancy)

        ls_edges_to_d = self.tfpg.get_edges_from_dest_node(discrepancy.get_name())
        ls_edge_tmin_sat = [self.get_edge_tmin_expr(e) for e in ls_edges_to_d]

        join_op = " | " if discrepancy.get_type() == NodeType.OR else " & "
        edges_tmin_sat = join_op.join(ls_edge_tmin_sat)

        return "%s -> (%s)" % (d_active, edges_tmin_sat)

    def get_edge_tmax_expr(self, tfpg_edge):

        if tfpg_edge.get_tMax() == -1 and (not self.parametric or "tmax" not in self.tags):
            # the constraint is dropped for tmax=inf, if we don't produce a parametric version
            # of the proof obligation or if tmax is not to be parameterized; ie, we substitute
            # the expression with the FALSE constant.
            return "FALSE"

        if self.parametric and "tmax" in self.tags:
            tmax_symb = compute_smv_etmax_real_params_name(tfpg_edge)
        else:
            tmax_symb = "%f" % tfpg_edge.get_tMax()

        etimer = compute_smv_edge_clock_name(tfpg_edge)
        full_edge_expr = "%s > %s" % (etimer, tmax_symb)

        if self.parametric and "tmax" in self.tags:
            tmax_is_inf = compute_smv_etmax_bool_params_name(tfpg_edge)
            # expression says: "not tmax=inf and the finite tmax is exceeded"
            return "(!%s & (%s))" % (tmax_is_inf, full_edge_expr)
        else:
            return "(%s)" % full_edge_expr

    def get_tmax_po_for_discrepancy(self, discrepancy):

        ls_edge_expr = []

        for edge in self.tfpg.get_edges_from_dest_node(discrepancy.get_name()):
            tmax_expr = self.get_edge_tmax_expr(edge)

            if tmax_expr == "FALSE":
                if discrepancy.get_type() == NodeType.AND:
                    # In this case we will have a conjunction where one of the conjuncts is FALSE,
                    # which causes the whole conjunction to be FALSE. The initial negation inverts
                    # this, such that the tmax proof obligation is always TRUE.
                    return "TRUE"
                if len(self.tfpg.get_predecessors(discrepancy)) == 1:
                    # The same applies to OR nodes which have only one predecessor.
                    return "TRUE"

            ls_edge_expr.append(tmax_expr)

        join_op = " | " if discrepancy.get_type() == NodeType.OR else " & "
        formula = "! (%s)" % join_op.join(ls_edge_expr)

        return formula
