
# %HEADER

from xsap.data_structure.tfpg.tfpg import NodeType
from xsap.features.common.tfpg_symb_converter import (
    compute_smv_em_params_name,
    compute_smv_etmin_params_name,
    compute_smv_etmax_real_params_name,
    compute_smv_etmax_bool_params_name)

from collections import OrderedDict


class ProofObligationMTL(object):

    def __init__(self, tfpg_struct, assoc_struct, parametric, tags=None):
        """
        @param tfpg_struct: the TFPG for which proof obligations need to be generated
        @type  tfpg_struct: data_structure.tfpg.tfpg.Tfpg

        @param assoc_struct: the association map of reference for the proof obligations
        @type  assoc_struct: data_structure.tfpg.tfpg_associations.TfpgAssociations

        @param parametric: flag indicating whether the proof obligation should be parametric
        @type  parametric: bool

        @param tags: List of parameter tags; subset of ("tmin", "tmax", "modes"); if parametric
                     is true, values are only parameterized if their class is in the tag list.
        @type  tags: tuple of strings, or None
        """

        if parametric:
            assert tags is not None
            for t in tags:
                assert t in ("tmin", "tmax", "modes")

        self.parametric = parametric
        self.tags = tags

        assert tfpg_struct is not None
        self.tfpg = tfpg_struct

        assert assoc_struct is not None
        self.assoc = assoc_struct

        self.mtl_placeholder_map = self.generate_mtl_placeholder_map()

    # INTERFACES

    def __str__(self):
        """
        @return: Get MTL proof obligation string expressing completeness, according to
                 the parameterization settings (see class constructor).
        @rtype:  string
        """

        return self.replace_placeholders(self.get_completeness_po())

    # INTERNAL FUNCTIONS

    # example TFPG used in code comments
    #
    # [FM1] --[1,8]{mA}--> (D)
    #                       ^
    #                       |
    # [FM2] --[0,inf]{mB}---'

    def get_po_conjunct_list(self):
        """ Generates a list with individual MTL proof obligation strings for each discrepancy.
        The TFPG is complete iff all of them hold. """

        discrepancies = self.tfpg.get_or_nodes() + self.tfpg.get_and_nodes()
        return [self.replace_placeholders(self.get_po_for_discrepancy(d)) for d in discrepancies]

    def get_placeholder_map(self):
        return self.mtl_placeholder_map

    def get_completeness_po(self):
        """
        @return: MTL proof obligation expressing completeness. For various subformulae placeholder
                 strings are used, which can be resolved using the placeholder map (see function
                 get_placeholder_map()).
        @rtype:  string
        """

        discrepancies = self.tfpg.get_or_nodes() + self.tfpg.get_and_nodes()
        proof_obligations = [self.get_po_for_discrepancy(d) for d in discrepancies]

        return " & ".join(proof_obligations)

    def replace_placeholders(self, mtl_formula):

        # first pass: the symbols of all node and edge expressions are replaced
        for symb in self.mtl_placeholder_map:
            mtl_formula = mtl_formula.replace(symb, self.mtl_placeholder_map[symb])

        # second pass: the symbols inside edge expressions and virtual nodes are replaced
        for symb in self.mtl_placeholder_map:
            mtl_formula = mtl_formula.replace(symb, self.mtl_placeholder_map[symb])

        # at this point, no placeholder should be present in the formula
        for symb in self.mtl_placeholder_map:
            assert symb not in mtl_formula

        return mtl_formula

    @staticmethod
    def get_mtl_placeholder_node_active(n_name):
        return "symb_node_%s" % n_name

    @staticmethod
    def get_mtl_placeholder_mode_active(m_name):
        return "symb_mode_%s" % m_name

    @staticmethod
    def get_mtl_placeholder_edge_active(src_name, dest_name):
        return "symb_edge_%s_to_%s" % (src_name, dest_name)

    def get_edge_tmin_expr(self, tfpg_edge):

        if self.parametric and "tmin" in self.tags:
            tmin_symb = compute_smv_etmin_params_name(tfpg_edge)
        else:
            tmin_symb = "%f" % tfpg_edge.get_tMin()

        src_name = tfpg_edge.get_srcNode().get_name()
        dest_name = tfpg_edge.get_destNode().get_name()
        symb_edge_active = self.get_mtl_placeholder_edge_active(src_name, dest_name)
        # nb: "S^{>= tmin}" is MTL syntactic sugar for "S^{[tmin,inf)}"
        return "((%s) S[%s,inf) (%s))" % (symb_edge_active, tmin_symb, symb_edge_active)

    def get_tmin_po_for_discrepancy(self, discrepancy):

        # G (symb_node_D -> O (symb_node_D & (
        #       (symb_edge_FM1_D S[1,inf) symb_edge_FM1_D) |
        #       (symb_edge_FM2_D S[0,inf) symb_edge_FM2_D)
        # )))

        d_active = self.get_mtl_placeholder_node_active(discrepancy.get_name())

        ls_edges_to_d = self.tfpg.get_edges_from_dest_node(discrepancy.get_name())
        ls_edge_tmin_sat = [self.get_edge_tmin_expr(e) for e in ls_edges_to_d]

        join_op = " | " if discrepancy.get_type() == NodeType.OR else " & "
        edges_tmin_sat = join_op.join(ls_edge_tmin_sat)

        return "G ((%s) -> O ((%s) & (%s)))" % (d_active, d_active, edges_tmin_sat)

    def get_edge_tmax_expr(self, tfpg_edge):

        if tfpg_edge.get_tMax() == -1 and (not self.parametric or "tmax" not in self.tags):
            # the constraint is dropped for tmax=inf, if we don't produce a parametric version
            # of the proof obligation or if tmax is not to be parameterized; ie, we substitute
            # the expression with the FALSE constant.
            return "FALSE"

        if self.parametric and "tmax" in self.tags:
            tmax_symb = compute_smv_etmax_real_params_name(tfpg_edge)
        else:
            tmax_symb = "%f" % tfpg_edge.get_tMax()

        src_name = tfpg_edge.get_srcNode().get_name()
        dest_name = tfpg_edge.get_destNode().get_name()

        symb_edge_active = self.get_mtl_placeholder_edge_active(src_name, dest_name)
        dest_active = self.get_mtl_placeholder_node_active(dest_name)

        e_active_no_propagation = "(%s) & !(%s)" % (symb_edge_active, dest_active)

        # nb: "S^{> tmax}" is MTL syntactic sugar for "S^{(tmax,inf)}"
        full_edge_expr = "(%s) S(%s,inf) (%s)" % (e_active_no_propagation, tmax_symb,
                                                  e_active_no_propagation)

        if self.parametric and "tmax" in self.tags:
            tmax_is_inf = compute_smv_etmax_bool_params_name(tfpg_edge)
            # expression says: "not tmax=inf and the finite tmax is exceeded"
            return "(!%s & (%s))" % (tmax_is_inf, full_edge_expr)
        else:
            return "(%s)" % full_edge_expr

    def get_tmax_po_for_discrepancy(self, discrepancy):

        # G ! ((symb_edge_FM1_D & !symb_node_D S(8,inf) symb_edge_FM1_D & !symb_node_D) |
        #      (FALSE))

        ls_edge_expr = []

        for edge in self.tfpg.get_edges_from_dest_node(discrepancy.get_name()):
            tmax_expr = self.get_edge_tmax_expr(edge)

            if tmax_expr == "FALSE":
                if discrepancy.get_type() == NodeType.AND:
                    # In this case we will have a conjunction where one of the conjuncts is FALSE,
                    # which causes the whole conjunction to be FALSE. The initial negation inverts
                    # this, such that the tmax proof obligation is always TRUE.
                    return "TRUE"
                if len(self.tfpg.get_predecessors(discrepancy)) == 1:
                    # The same applies to OR nodes which have only one predecessor.
                    return "TRUE"

            ls_edge_expr.append(tmax_expr)

        join_op = " | " if discrepancy.get_type() == NodeType.OR else " & "
        formula = "G ! (%s)" % join_op.join(ls_edge_expr)

        return formula

    def get_po_for_discrepancy(self, discrepancy):
        tmin_po = self.get_tmin_po_for_discrepancy(discrepancy)
        tmax_po = self.get_tmax_po_for_discrepancy(discrepancy)
        return "(%s) & (%s)" % (tmin_po, tmax_po)

    def generate_mtl_placeholder_map(self):

        # placeholders used in generation of proof obligation
        #
        # symb_node_D := O (expr_D)
        # symb_node_FM1 := O (expr_FM1)
        # symb_node_FM2 := O (expr_FM2)
        # symb_mode_mA := expr_mA
        # symb_mode_mB := expr_mB
        #
        # symb_edge_FM1_to_D := symb_FM1_active & (symb_mode_mA)
        # symb_edge_FM2_to_D := symb_FM2_active & (symb_mode_mB)

        mtl_placeholder_map = {}

        for node in self.tfpg.get_nodes_list():
            n_key = self.get_mtl_placeholder_node_active(node.get_name())

            assoc_expr = self.assoc.get_expression_from_var(node.get_name())

            # virtual node case
            if assoc_expr is None:
                assert node.get_type() != NodeType.FM

                predecessor_placeholders = []

                for n_pred in self.tfpg.get_predecessors(node):
                    n_pred_placeholder = self.get_mtl_placeholder_node_active(n_pred.get_name())
                    predecessor_placeholders.append(n_pred_placeholder)

                # sort nodes for improved predictability of resulting n_expr
                predecessor_placeholders.sort()

                if node.get_type() == NodeType.OR:
                    n_expr = " | ".join(predecessor_placeholders)
                else:
                    n_expr = " & ".join(predecessor_placeholders)

            # standard case
            else:
                n_expr = "O (%s)" % assoc_expr

            mtl_placeholder_map[n_key] = n_expr

        for mode in self.tfpg.get_modes_list():
            m_key = self.get_mtl_placeholder_mode_active(mode)
            m_expr = "(%s)" % self.assoc.get_expression_from_var(mode)
            mtl_placeholder_map[m_key] = m_expr

        for edge in self.tfpg.get_edges_list():

            # placeholder name
            src_name = edge.get_srcNode().get_name()
            dest_name = edge.get_destNode().get_name()
            symb_e = self.get_mtl_placeholder_edge_active(src_name, dest_name)

            # source node active placeholder
            symb_src_active = self.get_mtl_placeholder_node_active(edge.get_srcNode().get_name())

            # modes active placeholders
            ls_symb_modes_active = []

            for mode in edge.get_modes():
                m_expr = self.get_mtl_placeholder_mode_active(mode)
                if self.parametric and len(edge.get_modes()) > 1 and "modes" in self.tags:
                    em_param_name = compute_smv_em_params_name(edge, mode)
                    m_expr = "(%s & %s)" % (em_param_name, m_expr)

                ls_symb_modes_active.append(m_expr)

            # full edge expression
            e_expr = "%s & (%s)" % (symb_src_active, " | ".join(ls_symb_modes_active))

            mtl_placeholder_map[symb_e] = e_expr

        return OrderedDict(sorted(mtl_placeholder_map.items()))
