
# %HEADER

""" This file contains high-level functions to analyze the reachability of
individual or sets of TFPG nodes. """


import os
import subprocess
import tempfile

from xsap.data_structure.tfpg.tfpg_associations import TfpgAssociations
from xsap.utils.misc import log
from xsap.utils.processes.xsap_process import XsapProcess


def ensure_discrepancy_is_unreachable(smv_model, discr_smv,
                                      bmc_length=10, outdir='.'):
    """This function is used to perform an additional check performed during
    synthesis. If the synthesis warns the user that a particular discrepancy
    is not reachable, this function can be called to check that this is indeed
    true.

    @param smv_model: the name of the smv model on which run the check
    @type  smv_model: string

    @param discr_smv: the smv discrepancy expression to be analyzed
    @type  discr_smv: string

    @param bmc_length: bound for bmc analysis
    @type  bmc_length: int
    """

    model_fname = os.path.splitext(os.path.basename(smv_model))[0]
    log_out, logfname = tempfile.mkstemp(suffix=".tfpgout",
                                         prefix=model_fname,
                                         dir=outdir, text=True)
    log.xsap.tfpg_other.debug("Invoking xSAP, with redirect to " + logfname)

    need_msat = XsapProcess.has_infinite_domain_vars(smv_model)
    xsap_process = XsapProcess(
        smv_model,
        fail_obs_alm_fname=None,
        ord_fname=None,
        need_msat=need_msat,
        args=[],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        dumpers=[os.fdopen(log_out, "w")])

    res = xsap_process.assert_discrepancy_unreachability(discr_smv, bmc_length)

    xsap_process.kill_and_close_streams()

    return res


def identify_unreachable_nodes(smv_model, need_msat, assoc_struct, analysis_bound, outdir='.'):
    """ This function iterates over all nodes (both failure modes and
    discrepancies) of the given association structure and checks whether the
    respective property holds in some reachable state. It returns a list of
    nodes for which the associated property has been proven to be unreachable.
    In particular this means that when the analysis is inconclusive and
    unreachability cannot be proven, the respective node is not added to the
    returned list.

    @param smv_model: SMV model on which the reachability needs to be studied
    @type  smv_model: string

    @param assoc_struct: TFPG association structure
    @type  assoc_struct: TfpgAssociations

    @param analysis_bound: analysis bound (None for unbounded)
    @type  analysis_bound: int / None

    @param outdir: directory for file output
    @type  outdir: string

    @return: List of TFPG nodes (their names) which have been proven not to be
             reachable in the given model, or None in case of errors.
    @rtype:  list / None
    """

    assert analysis_bound is None or (isinstance(analysis_bound, int) and analysis_bound >= 0)

    unreach_nodes = None

    # prepare list of (node,expr) tuples

    ls_nodes = []
    for name in assoc_struct.get_var_names_of_nodes():
        expression = assoc_struct.get_expression_from_var(name)
        assert expression is not None
        ls_nodes.append({'name': name, 'expression': expression})

    # create xsap process

    model_fname = os.path.splitext(os.path.basename(smv_model))[0]
    log_out, logfname = tempfile.mkstemp(suffix="_node_reachability.tfpgout",
                                         prefix=model_fname,
                                         dir=outdir, text=True)

    log.xsap.tfpg_other.debug("Invoking xSAP, with redirect to " + logfname)

    xsap_process = XsapProcess(
        smv_model,
        fail_obs_alm_fname=None,
        ord_fname=None,
        need_msat=need_msat,
        args=["-dcx"],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        dumpers=[os.fdopen(log_out, "w")])

    # call reachability function in xsap process

    unreach_nodes = xsap_process.identify_unreachable_nodes(ls_nodes, analysis_bound)

    xsap_process.kill_and_close_streams()

    return unreach_nodes


def generate_pruned_assoc_struct(smv_model, need_msat, tfpg_assoc_struct, analysis_bound, outdir='.'):
    """ This function takes an SMV model and a TFPG association structure and
    generates a new association structure that contains only nodes whose
    non-reachability couldn't be proven (are or might be reachable).

    @param smv_model: SMV model on which the reachability needs to be studied
    @type  smv_model: string

    @param tfpg_assoc_struct: TFPG association structure
    @type  tfpg_assoc_struct: TfpgAssociations

    @return: TFPG association structure without unreachable nodes.
    @rtype:  TfpgAssociations / None
    """

    if tfpg_assoc_struct is None:
        return None

    # identify names of unreachable nodes

    unreach_nodes = identify_unreachable_nodes(smv_model, need_msat, tfpg_assoc_struct, analysis_bound, outdir)
    unreach_nodes = [node['name'] for node in unreach_nodes]

    # instantiate new association structure

    pruned_assoc_struct = TfpgAssociations()
    for name in tfpg_assoc_struct.keys():
        if name not in unreach_nodes:
            pruned_assoc_struct[name] = tfpg_assoc_struct[name]

    # sanity checks

    num_failure_modes = len(tfpg_assoc_struct.get_var_names_of_failure_modes())
    num_discrepancies = \
        len(tfpg_assoc_struct.get_var_names_of_unmonitored_discr()) + \
        len(tfpg_assoc_struct.get_var_names_of_monitored_discr())
    num_modes = len(tfpg_assoc_struct.get_var_names_of_tfpg_modes())

    num_failure_modes_pruned = len(pruned_assoc_struct.get_var_names_of_failure_modes())
    num_discrepancies_pruned = \
        len(pruned_assoc_struct.get_var_names_of_unmonitored_discr()) + \
        len(pruned_assoc_struct.get_var_names_of_monitored_discr())
    num_modes_pruned = len(pruned_assoc_struct.get_var_names_of_tfpg_modes())

    assert num_failure_modes_pruned <= num_failure_modes
    assert num_discrepancies_pruned <= num_discrepancies
    assert num_modes_pruned == num_modes

    # done

    return pruned_assoc_struct
