
# %HEADER

""" This file contains the common functionalities used for symbol's
    conversion (tfpg_symb2smv_symb and smv_symb2tfpg_symb).
    In particular, the following interfaces are provided:
    + convert_tfpg_symbol_to_smv_symbol -> encodes a tfpg symbol in smv
    + convert_smv_symbol_to_tfpg_symbol -> encodes a smv symbol in tfpg
    + create_tfpg2smv_map               -> returns a map which associates
                                           to each tfpg symbol its smv encoding
"""
from __future__ import print_function

import re
import sys
import os

EDGE_VAR = "EDGE_VAR_%s"
EDGE_STATE = "EDGE_STATE_%s"
EDGE_CLOCK = "EDGE_CLOCK_%s"


# INTERFACES

def convert_tfpg_symbol_to_smv_symbol(tfpg_symb, symb_type):
    """Encodes a tfpg symbol into its corrispective smv symbol.

       Notice that <symb_type> is a parameter used to express the type of
       the desired conversion. Possible values are:
       ["node", "edge", "edge_state", "edge_clock", "mode"]

       @param tfpg_symb: the tfpg symbol to be converted
       @type  tfpg_symb: TfpgNode, TfpgEdge or str (for modes)

       @param symb_type: an enumerative to specify which kind of symbol
                         we want to convert
       @type  symb_type: string

       @return: the smv encoding of the given tfpg symbol
       @rtype:  string
    """

    from xsap.data_structure.tfpg.tfpg import Edge as TfpgEdge
    from xsap.data_structure.tfpg.tfpg import Node as TfpgNode

    assert symb_type in ["node", "edge", "edge_state", "edge_clock", "mode"]
    if symb_type == "node":
        assert isinstance(tfpg_symb, TfpgNode)
        return compute_smv_node_name(tfpg_symb)
    elif symb_type == "edge":
        assert isinstance(tfpg_symb, TfpgEdge)
        return compute_smv_edge_name(tfpg_symb)
    elif symb_type == "edge_state":
        assert isinstance(tfpg_symb, TfpgEdge)
        return compute_smv_edge_state_name(tfpg_symb)
    elif symb_type == "edge_clock":
        assert isinstance(tfpg_symb, TfpgEdge)
        return compute_smv_edge_clock_name(tfpg_symb)
    else:
        assert symb_type == "mode"
        assert isinstance(tfpg_symb, str)
        return compute_smv_mode_name(tfpg_symb)


def get_tfpg_symbol_patterns():

    return {'fm_pattern': r'failuremode_(\w+)',
            'or_discr_pattern': r'or_node_(\w+)',
            'and_discr_pattern': r'and_node_(\w+)',
            'edge_name_pattern': r'edge_\w+_to_\w+_(\w+)',
            'estate_pattern': r'EState_(\w+)',
            'etime_pattern': r'ETime_(\w+)',
            'mode_pattern': r'Mode_(\w+)'}


def convert_smv_symbol_to_tfpg_symbol(smv_symb):
    """Encodes a smv symbol into its respective tfpg symbol

       @param smv_symb: the smv symbol to be converted
       @type  smv_symb: string

       @return: the tfpg encoding of the given smv symbol
       @rtype:  string
    """

    symb_patterns = get_tfpg_symbol_patterns()

    for pat_name in symb_patterns:
        m = re.match(symb_patterns[pat_name], smv_symb)
        if m is not None:
            return m.group(1)

    print("Unknown symbol: %s" % smv_symb)
    return None


def create_tfpg2smv_mapping(tfpg_struct):
    """Given a TFPG, it creates a 1:1 mapping for each TFPG symbol to
       its SMV encoding

       Note that, for each edge, three entries are created:
       + EDGE_VAR_<edge_name>: the smv encoding of the edge name
       + EDGE_STATE_<edge_name>: the smv encoding of the edge state
       + EDGE_CLOCK_<edge_name>: the smv encoding of the edge time

       @param tfpg_struct: the tfpg for which the map must be created
       @type  tfpg_struct: data_structure.tfpg.tfpg.Tfpg

       @return: a map linking tfpg symbols to smv symbols
       @rtype:  dict()
    """

    tfpg2smv_map = dict()  # Dict: keys = TFPG symbols, values = SMV encoding

    for nd in tfpg_struct.get_nodes_list():
        tfpg2smv_map[nd.get_name()] = compute_smv_node_name(nd)

    for eg in tfpg_struct.get_edges_list():
        tfpg2smv_map[EDGE_VAR % eg.get_name()] = compute_smv_edge_name(eg)
        tfpg2smv_map[
            EDGE_STATE %
            eg.get_name()] = compute_smv_edge_state_name(eg)
        tfpg2smv_map[
            EDGE_CLOCK %
            eg.get_name()] = compute_smv_edge_clock_name(eg)

    for md in tfpg_struct.get_modes_list():
        tfpg2smv_map[md] = compute_smv_mode_name(md)

    return tfpg2smv_map


# HELPERS

def compute_smv_node_name(n):
    t = n.get_type().lower()
    if "fm" == t:
        return "failuremode_%s" % n.get_name()
    else:
        return "%s_node_%s" % (t, n.get_name())


def compute_smv_edge_name(e):
    return "edge_%s_to_%s_%s" % (
        e.get_srcNode().get_name(), e.get_destNode().get_name(), e.get_name())


def compute_smv_edge_state_name(e):
    return "EState_%s" % e.get_name()


def compute_smv_edge_clock_name(e):
    return "ETime_%s" % e.get_name()


def compute_smv_mode_name(m):
    return "Mode_%s" % m


def compute_smv_em_params_name(e, m):
    return "param_edgemode_%s_%s" % (e.get_name(), m)


def compute_smv_etmin_params_name(e):
    return "param_tmin_%s" % e.get_name()


def compute_smv_etmax_real_params_name(e):
    return "param_tmax_real_%s" % e.get_name()


def compute_smv_etmax_bool_params_name(e):
    return "param_tmax_bool_%s" % e.get_name()


# TESTING

if __name__ == "__main__":

    sys.path.append(
        os.path.join(
            os.path.dirname(
                os.path.realpath(__file__)),
            "../../../../scripts"))
    import _paths
    _paths.setup_path()

    from xsap.data_structure.tfpg.tfpg import Tfpg

    # build random tfpg
    tfpg = Tfpg()
    # nodes
    sys_dead = tfpg.add_and_node("SYS_DEAD", True)
    s1_wo = tfpg.add_or_node("S1_WO", False)
    b1_dead = tfpg.add_or_node("B1_DEAD", False)
    b1_low = tfpg.add_or_node("B1_LOW", True)
    g1_dead = tfpg.add_or_node("G1_DEAD", False)
    g1 = tfpg.add_fm_node("G1_OFF")
    s1 = tfpg.add_fm_node("S1_OFF")
    s2_wo = tfpg.add_or_node("S2_WO", False)
    b2_dead = tfpg.add_or_node("B2_DEAD", False)
    b2_low = tfpg.add_or_node("B2_LOW", True)
    g2_dead = tfpg.add_or_node("G2_DEAD", False)
    g2 = tfpg.add_fm_node("G2_OFF")
    s2 = tfpg.add_fm_node("S2_OFF")
    # modes
    primary = "primary"
    secondary1 = "secondary_1"
    secondary2 = "secondary_2"
    default = [primary, secondary1, secondary2]
    tfpg.set_modes_list(default)
    # edges
    tfpg.add_edge(g1, 0, 0, default, g1_dead, "edge1")
    tfpg.add_edge(g1_dead, 0, 100, [primary, secondary1], b1_low, "edge2")
    tfpg.add_edge(b1_low, 5, 10, [primary, secondary1], b1_dead, "edge3")
    tfpg.add_edge(b1_dead, 0, 1, [primary, secondary1], s1_wo, "edge4")
    tfpg.add_edge(s1, 0, 0, default, s1_wo, "edge5")
    tfpg.add_edge(s1_wo, 0, 1, default, sys_dead, "edge6")
    tfpg.add_edge(g2, 0, 0, default, g2_dead, "edge7")
    tfpg.add_edge(g2_dead, 0, 100, [primary, secondary2], b2_low, "edge8")
    tfpg.add_edge(b2_low, 5, 10, [primary, secondary2], b2_dead, "edge9")
    tfpg.add_edge(b2_dead, 0, 1, [primary, secondary2], s2_wo, "edge10")
    tfpg.add_edge(s2, 0, 0, default, s2_wo, "edge11")
    tfpg.add_edge(s2_wo, 0, 1, default, sys_dead, "edge12")
    tfpg.add_edge(b1_dead, 0, 1, [secondary1], s2_wo, "edge13")
    tfpg.add_edge(b2_dead, 0, 1, [secondary2], s1_wo, "edge14")

    errors = tfpg.get_errors_list()
    assert not errors

    # check results
    tfpg2smv_map = create_tfpg2smv_mapping(tfpg)
    for nd in tfpg.get_nodes_list():
        assert convert_tfpg_symbol_to_smv_symbol(
            nd, "node") == tfpg2smv_map[nd.get_name()]
        assert convert_smv_symbol_to_tfpg_symbol(
            tfpg2smv_map[nd.get_name()]) == nd.get_name()
    for eg in tfpg.get_edges_list():
        assert convert_tfpg_symbol_to_smv_symbol(eg, "edge") == tfpg2smv_map[
            EDGE_VAR % eg.get_name()]
        assert convert_tfpg_symbol_to_smv_symbol(
            eg, "edge_state") == tfpg2smv_map[
            EDGE_STATE %
            eg.get_name()]
        assert convert_tfpg_symbol_to_smv_symbol(
            eg, "edge_clock") == tfpg2smv_map[
            EDGE_CLOCK %
            eg.get_name()]
        assert convert_smv_symbol_to_tfpg_symbol(
            tfpg2smv_map[
                EDGE_VAR %
                eg.get_name()]) == eg.get_name()
        assert convert_smv_symbol_to_tfpg_symbol(
            tfpg2smv_map[
                EDGE_STATE %
                eg.get_name()]) == eg.get_name()
        assert convert_smv_symbol_to_tfpg_symbol(
            tfpg2smv_map[
                EDGE_CLOCK %
                eg.get_name()]) == eg.get_name()
    for md in tfpg.get_modes_list():
        assert convert_tfpg_symbol_to_smv_symbol(
            md, "mode") == tfpg2smv_map[md]
        assert convert_smv_symbol_to_tfpg_symbol(tfpg2smv_map[md]) == md

    print("TEST SUCCESSFULLY PASSED!!")
