
# %HEADER

""" The functionalities contained in this file allow to extend a
    smv model w.r.t some associations.
"""
from __future__ import print_function

import subprocess
import sys
import re

from xsap.data_structure.tfpg.tfpg import NodeType
from xsap.utils.processes.xsap_process import XsapProcess

from xsap.features.common.tfpg_symb_converter import (
    compute_smv_node_name,
    compute_smv_mode_name,
    compute_smv_edge_clock_name,
    compute_smv_edge_state_name,
    compute_smv_em_params_name,
    compute_smv_etmin_params_name,
    compute_smv_etmax_real_params_name,
    compute_smv_etmax_bool_params_name,
    get_tfpg_symbol_patterns
)


# INTERFACES

def append_smv_extension(fn_smv, output_fname, extensions, t_delta_var):
    """ Extends the given SMV file by generating SMV code for the required
    extensions and appending it to the given SMV file. The result is stored
    in a new file.

    @param fn_smv: the SMV file to be flattened and extended
    @type  fn_smv: str

    @param output_fname: the target path for the extended SMV file
    @type  output_fname: str

    @param extensions: the dictionary with the extension data, generated for instance by the
                       function get_extension_data() of the class ProofObligationINVAR.
    @type  extensions: dict

    @param t_delta_var: time delta variable used for edge timer generation
    @type  t_delta_var: str or None

    @return "ok" or an error message in case of errors
    @rtype  str
    """

    assert set(extensions.keys()) == {'history_variables', 'defines', 'edge_timers', 'parameters'}

    # parse and flatten model

    xsap_process = XsapProcess(fn_smv, fail_obs_alm_fname=None, ord_fname=None, need_msat=False,
                               args=[], stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                               dumpers=[])

    xsap_process.flatten_and_dump(output_fname)
    xsap_process.wait_for_command_finish()

    if not xsap_process.is_alive():
        xsap_process.kill_and_close_streams()
        return "an error occurred while parsing the SMV file"

    xsap_process.kill_and_close_streams()

    # compute SMV extension

    smv_extension = ""

    # dump dummy t_delta_var if needed

    if len(extensions["edge_timers"]) > 0 and t_delta_var is None:

        # compute unique id for t_delta variable
        t_delta_var = "t_#delta"
        i = 0
        while is_var_already_in_model(t_delta_var, fn_smv):
            t_delta_var = "t_#delta_%d" % i
            i += 1

        # create dummy definition for t_delta
        smv_extension += get_define_smv_code(t_delta_var, "0") + "\n"

    # generate SMV code for all required extensions

    if len(extensions["history_variables"]) > 0:
        smv_extension += "\n-- history variables\n\n"

    for hvar in extensions["history_variables"]:
        smv_extension += get_hvar_smv_code(hvar["name"], hvar["expression"]) + "\n"

    if len(extensions["defines"]) > 0:
        smv_extension += "\n-- defines\n\n"

    for define in extensions["defines"]:
        smv_extension += get_define_smv_code(define["name"], define["expression"]) + "\n"

    if len(extensions["edge_timers"]) > 0:
        smv_extension += "\n-- edge timers\n\n"

    for etimer in extensions["edge_timers"]:
        smv_extension += get_etimer_smv_code(etimer["name"], etimer["state"], etimer["target"],
                                             t_delta_var) + "\n"

    if len(extensions["parameters"]) > 0:
        smv_extension += "\n-- parameters\n\n"

    for param in extensions["parameters"]:
        smv_extension += get_param_smv_code(param["name"], param["type"]) + "\n"

    # append extension to flattened SMV file

    with open(output_fname, 'a') as f_out:
        f_out.write(smv_extension)

    return "ok"


def dump_extended_sys_as_smv(
        tfpg_struct,
        associations,
        fn_smv,
        output_fname,
        delta_var_name=None,
        nodes_as_defines=False,
        dump_edge_timers=False,
        edge_modes_parameterized=False,
        dump_edge_timebound_parameters=False,
        filtered_nodes=()):
    """ Dumps a SLIM model, extended with variables for mapping system
    traces to TFPG traces, to SMV. Optionally also edge timers can be dumped
    and edge modes parameterized for certain parts of behavioral validation
    and synthesis.

    @param tfpg_struct: internal tfpg struct
    @type  tfpg_struct: data_structure.tfpg.tfpg.Tfpg

    @param associations: the TFPG SMV associations
    @type  associations: data_structure.tfpg.tfpg_associations.TfpgAssociations

    @param fn_smv: the name of the smv model
    @type  fn_smv: string

    @param output_fname: the name of the file where to dump smv code
    @type  output_fname: string

    @param delta_var_name: the name of the delta variable
    @type  delta_var_name: string or None

    @param nodes_as_defines: whether to declare a simple DEFINE instead of a
                           history variable VAR for each TFPG node and the
                           associated expression
    @type  nodes_as_defines: boolean

    @param dump_edge_timers: flag to enable dumping of edge timers
    @type  dump_edge_timers: boolean

    @param edge_modes_parameterized: edge modes are parameterized
                                     (else: grounded as in TFPG)
    @type  edge_modes_parameterized: boolean

    @param dump_edge_timebound_parameters: flag to enable dumping of edge time parameter variables
    @type  dump_edge_timebound_parameters: boolean

    @param filtered_nodes: discrepancies to be dumped (others are ignored)
    @type  filtered_nodes: list
    """

    assert not (nodes_as_defines and dump_edge_timers)
    assert not dump_edge_timebound_parameters or dump_edge_timers

    # identify modes and edges

    edges = tfpg_struct.get_edges_list() if tfpg_struct is not None else None
    modes = tfpg_struct.get_modes_list() if tfpg_struct is not None else None

    # get list of TFPG nodes ..

    nodes_list = []

    # .. from TFPG structure
    if tfpg_struct is not None:
        # list contains Node objects
        nodes_list = tfpg_struct.get_nodes_list()

    # .. from association map
    else:
        # list contains node names (strings)
        for k in associations.keys():
            (error, _, _type) = associations[k]
            assert error == "OK"

            # modes are not nodes
            if _type == "TfpgMode":
                continue

            # skip nodes not in the filter
            if k not in filtered_nodes:
                continue

            nodes_list.append(k)

    # parse and flatten model
    xsap_process = XsapProcess(fn_smv, fail_obs_alm_fname=None,
                               ord_fname=None, need_msat=False,
                               args=[], stdout=subprocess.PIPE,
                               stderr=subprocess.STDOUT, dumpers=[])

    xsap_process.flatten_and_dump(output_fname)
    xsap_process.wait_for_command_finish()

    xsap_process.kill_and_close_streams()

    # variable for output smv code

    main_extension = "\n"

    # compute history variables or defines for discrepancies

    main_extension += \
        "    -- discrepancy node definitions\n\n"

    for n in nodes_list:

        # synthesis symbols
        if tfpg_struct is None:
            name = "tfpgsynth_%s" % n
            assoc = associations.get_expression_from_var(n)

        # behavioral validation symbols
        else:
            name = compute_smv_node_name(n)
            assoc = associations.get_expression_from_var(n.get_name())
            if assoc is None:
                assoc = get_assoc_for_virtual_discrepancy(n, edges)

        assert assoc is not None

        assert isinstance(assoc, str)

        # add symbol prefix if we are in synthesis mode

        if nodes_as_defines:
            main_extension += get_define_smv_code(name, assoc)
        else:
            main_extension += get_hvar_smv_code(name, assoc)

    if dump_edge_timers:

        # identify t_delta variable

        if delta_var_name is None:

            # compute unique id for t_delta variable
            delta_var = "t_#delta"
            i = 0
            while is_var_already_in_model(delta_var, fn_smv):
                delta_var = "t_#delta_%d" % i
                i += 1

            # create dummy definition for t_delta
            main_extension += get_define_smv_code(delta_var, "0") + "\n"

        else:
            delta_var = delta_var_name

        # compute mode definitions

        main_extension += "    -- Mode definitions\n\n"

        for m in modes:
            mode_assoc = associations.get_expression_from_var(m)
            mode_varname = compute_smv_mode_name(m)
            main_extension += get_define_smv_code(mode_varname, mode_assoc)

        main_extension += "\n"

        # compute edge status and edge time definitions

        main_extension += \
            "    -- EState/ETime definitions\n"

        for e in edges:

            main_extension += "\n"

            source = compute_smv_node_name(e.get_srcNode())
            target = compute_smv_node_name(e.get_destNode())

            # EState

            estate = compute_smv_edge_state_name(e)

            if edge_modes_parameterized and len(e.get_modes()) > 1:

                # compute mode expression
                em_pairs = [
                    ("(%s & %s)" %
                     (compute_smv_mode_name(m),
                      compute_smv_em_params_name(e, m))) for m in e.get_modes()]
                mode_expr = ' | '.join(em_pairs)

                # set up edge mode parameters
                for m in e.get_modes():
                    main_extension += \
                        "    FROZENVAR %s : boolean;\n" % \
                        compute_smv_em_params_name(e, m)

            else:
                mode_expr = ' | '.join([compute_smv_mode_name(m)
                                        for m in e.get_modes()])

            main_extension += get_define_smv_code(estate, "%s & (%s)" % (source, mode_expr))

            # ETime

            etime_name = compute_smv_edge_clock_name(e)
            main_extension += get_etimer_smv_code(etime_name, estate, target, delta_var)

            if dump_edge_timebound_parameters:
                main_extension += \
                    get_param_smv_code(compute_smv_etmin_params_name(e), "real")
                main_extension += \
                    get_param_smv_code(compute_smv_etmax_real_params_name(e), "real")
                main_extension += \
                    get_param_smv_code(compute_smv_etmax_bool_params_name(e), "boolean")

        main_extension += "\n"

    # append extension to smv file

    out = open(output_fname, 'a')
    out.write(main_extension)
    out.close()

    return


# HELPERS

def get_param_smv_code(pname, ptype):
    assert ptype in ("boolean", "real")
    return "    FROZENVAR %s : %s;\n" % (pname, ptype)


def get_etimer_smv_code(etime_name, estate, target, delta_var):

    smv = "    VAR %s : real;\n" % etime_name

    # ASSIGN init(etime) := case
    #  estate : 0;    -- edge is active
    #  TRUE   : -1;   -- edge is inactive
    #  esac;

    smv += \
        "    ASSIGN init(%s) := case\n" % etime_name + \
        "      %s : 0;\n" % estate + \
        "      TRUE : -1;\n" + \
        "      esac;\n"

    # ASSIGN next(etime) := case
    #   target   : etime;       -- target is already active
    #   !estate' : -1;          -- edge will be inactive
    #   !estate & estate' : 0;  -- edge will become active
    #   TRUE : etime + delta    -- edge stays active
    #   esac;

    smv += \
        "    ASSIGN next(%s) := case\n" % etime_name + \
        "      %s : %s;\n" % (target, etime_name) + \
        "      !next(%s) : -1;\n" % estate + \
        "      !%s & next(%s) : 0;\n" % (estate, estate) + \
        "      TRUE : %s + %s;\n" % (etime_name, delta_var) + \
        "      esac;\n"

    return smv


def get_hvar_smv_code(hvar_name, hvar_expr):

    if "next(" in hvar_expr:
        smv = "    VAR %s : boolean;\n" % hvar_name
        smv += "    DEFINE def_%s := %s;" % (hvar_name, hvar_expr)
        smv += "    INIT !%s;" % hvar_name
        smv += "    TRANS next(%s) <-> (%s | def_%s);" % (hvar_name, hvar_name, hvar_name)

    else:
        smv = "    VAR %s : boolean;\n" % hvar_name
        smv += "    INIT %s <-> (%s);\n" % (hvar_name, hvar_expr)
        smv += "    TRANS next(%s) <-> (%s | next(%s));\n\n" % (hvar_name, hvar_name, hvar_expr)

    return smv


def get_define_smv_code(def_name, def_expr):
    return "    DEFINE %s := %s;\n" % (def_name, def_expr)


def get_assoc_for_virtual_discrepancy(d, edges):

    # this can be only true for discrepancies
    assert d.get_type() != NodeType.FM

    # d is a virtual discrepancy; create an association based on its
    # type and the nodes that have edges towards it.

    # 1. collect all nodes d' that have an edge towards d

    previous_nodes = []

    for e in edges:
        if e.get_destNode() == d:
            # compute smv name of source node and add it to list
            previous_nodes.append(compute_smv_node_name(e.get_srcNode()))

    assert len(previous_nodes) > 0

    # 2. create association based on d's type and all history monitors
    #    of source nodes of incoming edges

    if d.get_type() == NodeType.AND:
        return " & ".join(previous_nodes)
    else:
        return " | ".join(previous_nodes)


def all_fm_nodes_have_association(tfpg_struct, tfpg_assoc_struct):
    """ Checks whether all failure mode nodes in the TFPG have a
    corresponding expression in the associations that defines them.

    Only discrepancies can be left undefined, and will then be treated
    as virtual discrepancies.

    @param  tfpg_struct: internal tfpg struct
    @type   tfpg_struct: data_structure.tfpg.tfpg.Tfpg

    @param  tfpg_assoc_struct: the TFPG SMV associations
    @type   tfpg_assoc_struct: TfpgAssociations

    @return result of the check
    @rtype  Boolean
    """

    discrepancy_nodes = tfpg_struct.get_nodes_list()

    for n in discrepancy_nodes:
        if n.get_type() == NodeType.FM:
            assoc = tfpg_assoc_struct.get_expression_from_var(n.get_name())
            if assoc is None:
                return False

    return True


def is_var_already_in_model(var_name, smv_model):
    """Checks whether the model passed as parameter already contains a
    variable named as the first parameter

    @param var_name: the name of the variable to look for
    @type  var_name: string

    @param smv_model: the name of the SMV file to be inspected
    @type  svm_model: string
    """

    with open(smv_model, "r") as f:
        for line in f.readlines():

            # strip comments
            if '--' in line:
                line = line[:line.find('--')]

            if var_name in line:
                return True

    return False


def model_contains_reserved_symbol_pattern(smv_model):
    """Checks whether the model passed as parameter contains a variable
    with a reserved tfpg symbol pattern.

    @param smv_model: the name of the SMV file to be inspected
    @type  svm_model: string
    """

    symb_patterns = get_tfpg_symbol_patterns()

    # add whitespace in front of pattern to match tokens that start with
    # a reserved pattern
    symb_patterns = [(' ' + symb_patterns[symb]) for symb in symb_patterns]

    line_num = 0
    f = open(smv_model, "r")

    for line in f.readlines():

        line_num += 1

        # strip comments
        if '--' in line:
            line = line[:line.find('--')]

        for pat in symb_patterns:
            for name_symb_type in ["VAR", "IVAR", "DEFINE"]:
                if re.search(r"%s\s+%s" %
                             (name_symb_type, pat), line) is not None:
                    f.close()
                    return {'result': True, 'line_num': line_num}

    f.close()
    return {'result': False, 'line_num': None}


# TESTING

if __name__ == "__main__":

    if len(sys.argv) != 5:
        print("Usage: %s tfpg_file assoc_file smv_model out_file" % sys.argv[0])
        sys.exit(1)

    tfpg_fname = sys.argv[1]
    assoc_fname = sys.argv[2]
    smv_fname = sys.argv[3]
    out_fname = sys.argv[4]

    from xsap.data_structure.tfpg.xml.tfpg import (
        parse_tfpg_xml_file_to_internal_struct)
    from xsap.data_structure.tfpg.xml.tfpg_associations import (
        parse_assoc_xml_file_to_internal_struct)

    tfpg, errors = parse_tfpg_xml_file_to_internal_struct(tfpg_fname)
    if len(tfpg.get_errors_list()) > 0:
        for err in tfpg.get_errors_list():
            raise err
    if len(errors) > 0:
        for err in errors:
            raise err
    test_assoc = parse_assoc_xml_file_to_internal_struct(assoc_fname)

    dump_extended_sys_as_smv(tfpg, test_assoc, smv_fname, out_fname)
    print("FILE '%s' SUCCESSFULLY CREATED" % out_fname)
