from __future__ import print_function

# %HEADER

import sys
import os
import subprocess
import tempfile

try:
    from xsap.utils.options import MC_ENGINE_MSAT_BMC, MC_ENGINE_IC3
except ImportError:
    # in case we are in COMPASS project
    from compass.options import MC_ENGINE_MSAT_BMC, MC_ENGINE_IC3
from xsap.features.common.tfpg_sys_extender import (
    all_fm_nodes_have_association,
    dump_extended_sys_as_smv,
    model_contains_reserved_symbol_pattern)
from xsap.features.tfpg_behavioral_validation.proof_obligations_generator import (
    generate_proof_obligations_completeness,
    generate_proof_obligations_edge_tightness_tmintmax,
    generate_proof_obligations_edge_tightness_tmaxinfty,
    generate_proof_obligations_edge_tightness_modes)
from xsap.features.tfpg_behavioral_validation.trace_splitter import (
    split_behval_completeness_tracefile)

from xsap.utils.misc import log
from xsap.utils.processes.xsap_process import XsapProcess


# INTERFACES

def perform_behavioral_validation(
        property_to_validate,
        tfpg_struct,
        tfpg_assoc_struct,
        smv_model,
        bmc_length=10,
        delta_t_var_name=None,
        engine=MC_ENGINE_MSAT_BMC,
        boolean_conversion=False,
        dynamic_reorder=False,
        monolithic_check=False,
        generate_only=False,
        parametric_dump=False,
        outdir='.',
        xsap_process=None):
    """Implements the functionality of behavioral validation.

    @param property_to_validate: the property to validate (completeness or
                                 edge tightness)
    @type  property_to_validate: string

    @param tfpg_struct: the internal representation of the tfpg
    @type  tfpg_struct: data_structure.tfpg.tfpg.Tfpg

    @param tfpg_assoc_struct: the internal representation of the
                              tfpg associations
    @type  tfpg_assoc_struct: TfpgAssociations

    @param smv_model: the name of the smv model
    @type  smv_model: string

    @param bmc_length: bound for bmc algorithm
    @type  bmc_length: int

    @param outdir: directory for all generated files
    @type  outdir: string

    @param xsap_process: the xsap process we may have created and
                         we want to use
    @type  xsap_process: utils.processes.xsap_process.xSAPProcess

    @return res: the result of the analysis and a possible counterexample
    @type   res: dict
    """

    assert property_to_validate in ('completeness', 'edge_tightness')

    # check presence of reserved symbols
    symb_check_res = model_contains_reserved_symbol_pattern(smv_model)
    if symb_check_res['result']:
        print ("--> Error in SMV model: Line %d contains reserved "
               "symbol name." % symb_check_res['line_num'])
        return {
            'result': "error",
            'counterexample': None,
            'counterexample_description': None}

    # check if all failure modes are defined in the associations file.
    if not all_fm_nodes_have_association(tfpg_struct, tfpg_assoc_struct):
        print ("--> Error in associations file: Some failure modes "
               "are not defined.")
        return {
            'result': "error",
            'counterexample': None,
            'counterexample_description': None}

    if property_to_validate == 'completeness':
        return validate_completeness(
            tfpg_struct,
            tfpg_assoc_struct,
            smv_model,
            bmc_length,
            delta_t_var_name,
            engine,
            boolean_conversion,
            dynamic_reorder,
            monolithic_check,
            generate_only,
            parametric_dump,
            outdir,
            xsap_process)
    else:
        return validate_edge_tightness(
            tfpg_struct,
            tfpg_assoc_struct,
            smv_model,
            bmc_length,
            delta_t_var_name,
            engine,
            boolean_conversion,
            dynamic_reorder,
            generate_only,
            outdir,
            xsap_process)


# INTERNAL

def validate_completeness(
        tfpg_struct,
        tfpg_assoc_struct,
        smv_model,
        bmc_length,
        delta_t_var_name,
        engine,
        boolean_conversion,
        dynamic_reorder,
        monolithic_check,
        generate_only,
        parametric_dump,
        outdir,
        xsap_process):

    if parametric_dump and not generate_only:
        return {'result':'error (parametric_dump only available in generate_only mode)'}

    untimed = delta_t_var_name is None

    model_fname = get_model_fname(smv_model)

    log_out, logfname = tempfile.mkstemp(suffix=".tfpgout",
                                         prefix=model_fname,
                                         dir=outdir, text=True)
    log.xsap.tfpg_behav.debug("Invoking xSAP, with redirect to " + logfname)

    print()
    print("TFPG Behavioral Validation (completeness)")
    print()

    # if we specify a xsap_process, we want to handle manually the process
    user_spec_proc = xsap_process is not None

    # extend file names with output directory
    extended_model_fname = os.path.abspath(
        os.path.join(outdir, model_fname + "_bv_extended_completeness.smv"))
    proof_obligations_fname = os.path.abspath(
        os.path.join(outdir,
                     model_fname + "_proof_obligations_completeness.txt"))

    # extend the model
    dump_extended_sys_as_smv(
        tfpg_struct,
        tfpg_assoc_struct,
        smv_model,
        extended_model_fname,
        delta_t_var_name,
        nodes_as_defines=False,
        dump_edge_timers=True,
        edge_modes_parameterized=parametric_dump,
        dump_edge_timebound_parameters=parametric_dump)

    print("  > extended smv model created (%s)" % extended_model_fname)

    # generate proof obligations
    proof_obl = generate_proof_obligations_completeness(
        tfpg_struct, untimed, proof_obligations_fname, monolithic_check,
        parametric_dump)
    print ("  > %d proof obligation%s generated (%s)" %
           (len(proof_obl),
            ('' if len(proof_obl) == 1 else 's'), proof_obligations_fname))
    print()

    if generate_only:
        os.close(log_out)
        return {'result': 'generated all problem files'}

    # instantiate the process
    if not user_spec_proc:
        need_msat = XsapProcess.has_infinite_domain_vars(extended_model_fname)
        xsap_process = XsapProcess(
            extended_model_fname,
            fail_obs_alm_fname=None,
            ord_fname=None,
            need_msat=need_msat,
            args=[],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            dumpers=[os.fdopen(log_out, "w")],
            dynamic_reorder=dynamic_reorder,
            verbose_level=0,
            boolean_conversion_uses_predicate_normalization=boolean_conversion)
        xsap_process.set_tmp_filename_prefix(outdir)
    else:
        # if we already own a process, we need to read the
        # extended model
        xsap_process.reset()
        xsap_process.read_model(model=extended_model_fname)
        # the log file can be closed as it is not needed
        os.close(log_out)

    # perform check
    res_completeness = xsap_process.check_tfpg_completeness(
        proof_obl, bmc_length, engine, boolean_conversion)

    # terminate process
    # in case we already own a xsap process, we want it to
    # be kept alive
    if not user_spec_proc:
        xsap_process.kill_and_close_streams()

    # xsap terminated with error
    if res_completeness is None:
        return {'result': 'error', 'message': 'xSAP terminated with error'}
    if res_completeness['result'] == 'error':
        return res_completeness

    # process trace if necessary
    if res_completeness['counterexample'] is not None:
        res_completeness['counterexample'] = \
            split_behval_completeness_tracefile(
                res_completeness['counterexample'])

    return res_completeness


def validate_edge_tightness(
        tfpg_struct,
        tfpg_assoc_struct,
        smv_model,
        bmc_length,
        delta_t_var_name,
        engine,
        boolean_conversion,
        dynamic_reorder,
        generate_only,
        outdir,
        xsap_process,
        tags=("tmin", "tmax", "tmax-inf", "modes"),
        tightness_precision=None):

    if tags is None or len(tags) == 0:
        return {'result': 'error', 'description': 'no tag for tightening given'}

    for t in tags:
        if t not in ("tmin", "tmax", "tmax-inf", "modes"):
            return {'result': 'error', 'description': 'unknown tag (%s)' % t}

    assert tightness_precision is None or tightness_precision <= 1

    untimed = delta_t_var_name is None

    res_tightness_tmintmax = None
    res_tightness_tmax_inf = None
    res_tightness_modes = None

    model_fname = get_model_fname(smv_model)

    # if we specify a xsap_process, we want to handle manually the process
    user_spec_proc = xsap_process is not None

    # check tightness of tmin/tmax (no infty) for all edges

    print()
    print ("tfpg behavioral validation (edge tightness 1/3, tightness of "
           "finite propagation bounds)")
    print()

    if untimed:
        print ("  > skipping, edge times are ignored (time delta variable "
               "was not specified).")

    elif "tmin" not in tags and "tmax" not in tags:
        print ("  > skipping tightness check of tmin/tmax values (per user request)")

    else:
        # compute file names
        extended_model_fname = os.path.abspath(
            os.path.join(outdir,
                         model_fname + "_bv_extended_edge_tightness_1_3.smv"))
        proof_obligations_fname = os.path.abspath(
            os.path.join(outdir,
                         model_fname + "_proof_obligations_edge_tightness_1_3.txt"))

        # dump extended model
        dump_extended_sys_as_smv(
            tfpg_struct,
            tfpg_assoc_struct,
            smv_model,
            extended_model_fname,
            delta_t_var_name,
            nodes_as_defines=False,
            dump_edge_timers=True,
            edge_modes_parameterized=False)
        print("  > extended smv model created (%s)" % extended_model_fname)

        # generate and dump proof obligations
        proof_obl = generate_proof_obligations_edge_tightness_tmintmax(
            tfpg_struct, proof_obligations_fname, tightness_precision)
        print("  > %d proof obligation%s generated (%s)" % \
            (len(proof_obl),
             ('' if len(proof_obl) == 1 else 's'), proof_obligations_fname))
        print()

        if not generate_only:

            # instantiate the process
            log_out, logfname = tempfile.mkstemp(
                suffix=".tfpgout",
                prefix=model_fname,
                dir=outdir,
                text=True)
            log.xsap.tfpg_behav.debug(
                "Invoking xSAP, with redirect to " + logfname)
            if not user_spec_proc:
                need_msat = engine in [MC_ENGINE_MSAT_BMC, MC_ENGINE_IC3]
                xsap_process = XsapProcess(
                    extended_model_fname,
                    fail_obs_alm_fname=None,
                    ord_fname=None,
                    need_msat=need_msat,
                    args=[],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    dumpers=[os.fdopen(log_out, "w")],
                    dynamic_reorder=dynamic_reorder,
                    verbose_level=0,
                    boolean_conversion_uses_predicate_normalization=boolean_conversion)
                xsap_process.set_tmp_filename_prefix(outdir)
            else:
                # if we already own a process, we need to read the
                # extended model
                xsap_process.reset()
                xsap_process.read_model(model=extended_model_fname)
                # the log file can be closed as it is not needed
                os.close(log_out)

            # perform check
            res_tightness_tmintmax = xsap_process.check_tfpg_edge_tightness(
                proof_obl, bmc_length, engine, boolean_conversion, tags)

            # in case we already own a xsap process, we want it to
            # be kept alive
            if not user_spec_proc:
                xsap_process.kill_and_close_streams()

            if res_tightness_tmintmax['result'] == "error":
                return res_tightness_tmintmax
            if res_tightness_tmintmax['result'] == "not_tight":
                return res_tightness_tmintmax
            if res_tightness_tmintmax['result'] == "not_tight_bounded":
                return res_tightness_tmintmax

    # check tightness of tmax=infty (if any)

    print()
    print ("tfpg behavioral validation (edge tightness 2/3, tightness "
           "of infinite propagation bounds)")
    print()

    # flag to indicate status of tmax=inf tightness checks (if any)
    tmaxinf_tight = None

    if untimed:
        print ("  > skipping, edge times are ignored (time "
               "delta variable was not specified).")

    elif "tmax-inf" not in tags:
        print ("  > skipping tightness check of infinite tmax values (per user request)")

    else:
        # compute file names
        proof_obligations_fname = os.path.abspath(
            os.path.join(outdir,
                         model_fname + "_proof_obligations_edge_tightness_2_3.txt"))

        # generate and dump proof obligations
        proof_obl = generate_proof_obligations_edge_tightness_tmaxinfty(
            tfpg_struct,
            tfpg_assoc_struct,
            delta_t_var_name,
            proof_obligations_fname)

        if len(proof_obl) == 0:
            print("  > no edges with tmax=infinity, nothing to check;")

        else:
            # compute file names
            extended_model_fname = os.path.abspath(
                os.path.join(
                    outdir, model_fname + "_bv_extended_edge_tightness_2_3.smv"))

            # dump extended model
            dump_extended_sys_as_smv(
                tfpg_struct,
                tfpg_assoc_struct,
                smv_model,
                extended_model_fname,
                delta_t_var_name,
                nodes_as_defines=False,
                dump_edge_timers=False,
                edge_modes_parameterized=False)
            print("  > extended smv model created (%s)" % extended_model_fname)

            # info on proof obligations file
            print("  > %d proof obligation%s generated (%s)" % \
                (len(proof_obl),
                 ('' if len(proof_obl) == 1 else 's'), proof_obligations_fname))

            if not generate_only:

                # fall-back to msat_bmc if not already chosen
                tmax_infty_engine = engine
                if tmax_infty_engine == MC_ENGINE_IC3:
                    print ("  > cannot use ic3 engine for LTL properties, "
                           "fall-back to msat_bmc")
                    tmax_infty_engine = MC_ENGINE_MSAT_BMC

                print()

                # instantiate the process
                log_out, logfname = tempfile.mkstemp(
                    suffix=".tfpgout",
                    prefix=model_fname,
                    dir=outdir,
                    text=True)
                log.xsap.tfpg_behav.debug(
                    "Invoking xSAP, with redirect to " + logfname)
                if not user_spec_proc:
                    need_msat = XsapProcess.has_infinite_domain_vars(extended_model_fname)
                    xsap_process = XsapProcess(
                        extended_model_fname,
                        fail_obs_alm_fname=None,
                        ord_fname=None,
                        need_msat=need_msat,
                        args=[],
                        stdout=subprocess.PIPE,
                        stderr=subprocess.STDOUT,
                        dumpers=[os.fdopen(log_out, "w")],
                        boolean_conversion_uses_predicate_normalization=boolean_conversion)
                    xsap_process.set_tmp_filename_prefix(outdir)
                else:
                    # if we already own a process, we need to read the
                    # extended model
                    xsap_process.reset()
                    xsap_process.read_model(model=extended_model_fname)
                    # the log file can be closed as it is not needed
                    os.close(log_out)

                # check proof obligations
                res_tightness_tmax_inf = xsap_process.check_tfpg_edge_tightness(
                    proof_obl, bmc_length,
                    tmax_infty_engine, boolean_conversion, tags)

                # in case we already own a xsap process, we want it to
                # be kept alive
                if not user_spec_proc:
                    xsap_process.kill_and_close_streams()

                if res_tightness_tmax_inf['result'] == "error":
                    return res_tightness_tmax_inf
                if res_tightness_tmax_inf['result'] == "not_tight":
                    return res_tightness_tmax_inf
                if res_tightness_tmax_inf['result'] == "not_tight_bounded":
                    return res_tightness_tmax_inf

                assert res_tightness_tmax_inf['result'] == "tight"
                assert res_tightness_tmax_inf['tmaxinf_tight'] in ("tight", "unknown")
                tmaxinf_tight = res_tightness_tmax_inf['tmaxinf_tight']

    # check tightness of modes for all edges

    print()
    print ("tfpg behavioral validation (edge tightness 3/3, "
           "tightness of edge modes)")
    print()

    if "modes" not in tags:
        print ("  > skipping tightness check of modes (per user request)")

    else:
        # compute file names
        extended_model_fname = os.path.abspath(
            os.path.join(outdir,
                         model_fname + "_bv_extended_edge_tightness_3_3.smv"))
        proof_obligations_fname = os.path.abspath(
            os.path.join(outdir,
                         model_fname + "_proof_obligations_edge_tightness_3_3.txt"))

        # dump extended model
        dump_extended_sys_as_smv(
            tfpg_struct,
            tfpg_assoc_struct,
            smv_model,
            extended_model_fname,
            delta_t_var_name,
            nodes_as_defines=False,
            dump_edge_timers=True,
            edge_modes_parameterized=True)
        print("  > extended smv model created (%s)" % extended_model_fname)

        # generate and dump proof obligations
        proof_obl = generate_proof_obligations_edge_tightness_modes(
            tfpg_struct, untimed, proof_obligations_fname)
        print("  > %d proof obligation%s generated (%s)" % \
            (len(proof_obl),
             ('' if len(proof_obl) == 1 else 's'), proof_obligations_fname))
        print()

        if generate_only:
            return {'result': 'generated all problem files'}

        # instantiate the process
        log_out, logfname = tempfile.mkstemp(suffix=".tfpgout",
                                             prefix=model_fname,
                                             dir=outdir, text=True)
        log.xsap.tfpg_behav.debug("Invoking xSAP, with redirect to " + logfname)
        if not user_spec_proc:
            need_msat = XsapProcess.has_infinite_domain_vars(extended_model_fname)
            xsap_process = XsapProcess(
                extended_model_fname,
                fail_obs_alm_fname=None,
                ord_fname=None,
                need_msat=need_msat,
                args=[],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                dumpers=[os.fdopen(log_out, "w")],
                dynamic_reorder=dynamic_reorder,
                verbose_level=0,
                boolean_conversion_uses_predicate_normalization=boolean_conversion)
            xsap_process.set_tmp_filename_prefix(outdir)
        else:
            # if we already own a process, we need to read the
            # extended model
            xsap_process.reset()
            xsap_process.read_model(model=extended_model_fname)

        # perform check
        res_tightness_modes = xsap_process.check_tfpg_edge_tightness(
            proof_obl, bmc_length, engine, boolean_conversion, tags)

        # in case we already own a xsap process, we want it to
        # be kept alive
        if not user_spec_proc:
            xsap_process.kill_and_close_streams()

    res_tightness = {'tmaxinf_tight': tmaxinf_tight, 'result': 'tight', 'description': None}

    # possible results: not_tight (tighter value exists for some parameter),
    #                   not_tight_bounded (tighter value might exist for some parameter),
    #                   tight (tighter values don't exist for any parameter).

    def select_result(res_a, res_b):
        if res_a == "not_tight" or res_b == "not_tight":
            return "not_tight"
        if res_a == "not_tight_bounded" or res_b == "not_tight_bounded":
            return "not_tight_bounded"
        assert res_a == "tight" and res_b == "tight"
        return "tight"

    if res_tightness_tmintmax is not None:
        assert res_tightness_tmintmax['result'] in ('tight', 'not_tight_bounded', 'not_tight')
        res_tightness['result'] = select_result(res_tightness_tmintmax['result'], res_tightness['result'])

    if res_tightness_tmax_inf is not None:
        assert res_tightness_tmax_inf['result'] in ('tight', 'not_tight_bounded', 'not_tight')
        res_tightness['result'] = select_result(res_tightness_tmax_inf['result'], res_tightness['result'])

    if res_tightness_modes is not None:
        assert res_tightness_modes['result'] in ('tight', 'not_tight_bounded', 'not_tight')
        res_tightness['result'] = select_result(res_tightness_modes['result'], res_tightness['result'])

    return res_tightness

def get_model_fname(smv_model):
    return os.path.splitext(os.path.basename(smv_model))[0]

# TESTING

if __name__ == "__main__":

    if len(sys.argv) != 4:
        print("Usage: %s tfpg_file assoc_file smv_model" % sys.argv[0])
        sys.exit(1)

    tfpg_fname = sys.argv[1]
    assoc_fname = sys.argv[2]
    smv_fname = sys.argv[3]

    from xsap.data_structure.tfpg.xml.tfpg import (
        parse_tfpg_xml_file_to_internal_struct)
    from xsap.data_structure.tfpg.xml.tfpg_associations import (
        parse_assoc_xml_file_to_internal_struct)

    tfpg, errors = parse_tfpg_xml_file_to_internal_struct(tfpg_fname)
    assoc = parse_assoc_xml_file_to_internal_struct(assoc_fname)

    res = perform_behavioral_validation('completeness', tfpg, assoc, smv_fname)

    print("Result:", res['result'])
    print("Trace:", res['counterexample'])
    print("TEST SUCCESSFULLY PASSED!")
