from __future__ import print_function

# %HEADER

from builtins import range
from xsap.features.common.tfpg_symb_converter import (
    compute_smv_em_params_name,
    compute_smv_node_name,
    compute_smv_edge_clock_name,
    compute_smv_etmin_params_name,
    compute_smv_etmax_real_params_name,
    compute_smv_etmax_bool_params_name
)


# INTERFACES (COMPLETENESS)

def generate_proof_obligations_completeness(
        tfpg_struct,
        untimed,
        out_fname=None,
        monolithic_check=False,
        parametric=False):
    """Given a TFPG, it returns a list containing the proof obligation
    built on top of its discrepancies and edges.

    @param tfpg_struct: the tfpg to be considered
    @type  tfpg_struct: data_structure.tfpg.tfpg.Tfpg

    @param out_fname: the name of the output file where to dump the
                      proof obligations
    @type  out_fname: string

    @param untimed: specifies whether the analysis should ignore times
    @type  untimed: boolean

    @return proof_obligations: the generated proof obligations
    @rtype  proof_obligations: list
    """

    discrepancies = tfpg_struct.get_or_nodes() + tfpg_struct.get_and_nodes()
    edges = tfpg_struct.get_edges_list()

    proof_obligations = []

    # for each discrepancy node
    for d in discrepancies:
        proof_obligations += \
            gen_completeness_props_for_node(d, edges, untimed, parametric)

    if monolithic_check:
        # completeness props are all invarspecs; monolithic check means
        # glueing them all together and verifying them in one go, which
        # usually speeds up verification;
        all_props = [("(%s)" % p['prop']) for p in proof_obligations]
        proof_obligations = [
            {'object': 'monolithic_check',
             'prop': " & ".join(all_props),
             'proptype': 'invarspec',
             'description': 'monolithic check',
             'tmax-infinity': False}]

    if out_fname is not None:
        dump_proof_obligations_to_file(
            proof_obligations, out_fname, 'invarspec')

    return proof_obligations


# INTERFACES (TIGHTNESS)

def generate_proof_obligations_edge_tightness_modes(
        tfpg_struct, untimed, out_fname=None):

    proof_obligations = []

    # discrepancies = tfpg_struct.get_or_nodes() + tfpg_struct.get_and_nodes()
    edges = tfpg_struct.get_edges_list()

    # for every edge
    for e in edges:

        # skip if edge has only one mode (dropping it would result in an
        # illegal TFPG)
        if len(e.get_modes()) == 1:
            continue

        # for every mode of the edge
        for m in e.get_modes():

            # goal: check if completeness is broken by dropping the mode.

            # compute the completeness proof obligations of the edge's target
            # node
            compl_props = gen_completeness_props_for_node(
                e.get_destNode(), edges, untimed, False)

            # compute property disabling exactly the current edge mode
            mode_selec = compute_mode_disabling_expression(edges, e, m)

            # create proof obligation by and-ing the two completeness proof
            # obligations, implied by the mode selection expression

            prop = "(%s)" % compl_props[0]['prop']

            if len(compl_props) > 1:
                assert len(compl_props) == 2
                prop += " & (%s)" % compl_props[1]['prop']

            prop = "(%s) -> (%s)" % (mode_selec, prop)

            proof_obligations.append(
                {'object': compute_smv_em_params_name(e, m),
                 'prop': prop,
                 'proptype': 'invarspec',
                 'description': 'can drop mode %s from edge %s' % (m, e.get_name()),
                 'tmax-infinity': False})

    if out_fname is not None:
        dump_proof_obligations_to_file(
            proof_obligations, out_fname, 'invarspec')

    return proof_obligations


def generate_proof_obligations_edge_tightness_tmaxinfty(
        tfpg_struct, tfpg_assoc_struct, t_delta_var, out_fname=None):
    """ Generate proof obligations for each edge with tmax=infinity to
    check whether no finite bound for the propagation delay exists.
    It means there must be an infinite trace on which the edge is,
    at some point, always active, but the propagation never occurs.

    @param tfpg_struct: the TFPG to be considered
    @type  tfpg_struct: data_structure.tfpg.tfpg.Tfpg

    @param tfpg_assoc_struct: the internal representation of the
                              tfpg associations
    @type  tfpg_assoc_struct: TfpgAssociations

    @param out_fname: the name of the output file where to dump the
                      proof obligations
    @type  out_fname: string
    """

    proof_obligations = []
    edges = tfpg_struct.get_edges_list()

    for e in edges:

        if float(e.get_tMax()) != -1:
            continue

        e_source = compute_smv_node_name(e.get_srcNode())
        e_target = compute_smv_node_name(e.get_destNode())

        expr_modes = ' | '.join(
            [tfpg_assoc_struct.get_expression_from_var(m)
             for m in e.get_modes()])

        # it is not possible that the edge is active forever but the target
        # never activates

        ltlspec = "! F (G (%s & (%s) & !%s) & G (F %s > 0))" % (
            e_source, expr_modes, e_target, t_delta_var)
        descr = "can tighten tmax of edge %s" % e.get_name()

        proof_obligations.append({'object': e.get_name(),
                                  'prop': ltlspec,
                                  'proptype': 'ltlspec',
                                  'description': descr,
                                  'tmax-infinity': True})

    if out_fname is not None:
        dump_proof_obligations_to_file(proof_obligations, out_fname, 'ltlspec')

    return proof_obligations


def generate_proof_obligations_edge_tightness_tmintmax(
        tfpg_struct, proof_obligations_fname, search_precision=None):
    """ Generate proof obligations for each edge to check edge's tmin/tmax
    tightness. This means to generate, for each edge, the completeness proof
    obligations wrt the target node of that edge, but with tightened tmin and
    tmax; if a property holds, it means the tightened tmin or tmax can
    effectively be tightened in the tfpg without breaking completeness.

    @param tfpg_struct: the tfpg to be considered
    @type  tfpg_struct: data_structure.tfpg.tfpg.Tfpg

    @param proof_obligations_fname: the name of the output file where to dump
                                    the proof obligations
    @type  proof_obligations_fname: string

    @param search_precision: precision of tightness checking
    @type  search_precision: int or None

    @return proof_obligations: the generated proof obligations
    @rtype  proof_obligations: list
    """

    proof_obligations = []

    if search_precision is not None:
        smallest_num = search_precision
    else:
        smallest_num = get_smallest_decimal(tfpg_struct)
    assert smallest_num == 1 or smallest_num <= 0.1

    edges = tfpg_struct.get_edges_list()

    for e in edges:

        e_name = e.get_name()
        e_target = e.get_destNode()
        e_tmin = float(e.get_tMin())
        e_tmax = float(e.get_tMax())
        tmax_infty = e_tmax == -1.0

        # check if tightening is possible at all: if tmax is infinity,
        # we can always (try to) tighten tmin; else, we need to be sure
        # that when tightening tmin we don't exceed tmax (or, vice-versa,
        # that when tightening tmax we don't exceed tmin).
        if not tmax_infty and e_tmin + smallest_num > e_tmax:
            continue

        # tighten the tmin/tmax parameters of the current edge
        e_tmin += smallest_num
        if not tmax_infty:
            e_tmax -= smallest_num

        ls_below_tmin = []
        ls_exceeding_tmax = []

        # compute partial property for current edge
        etime = compute_smv_edge_clock_name(e)
        ls_below_tmin.append("%s < %s" % (etime, e_tmin))
        if not tmax_infty:
            ls_exceeding_tmax.append("%s > %s" % (etime, e_tmax))

        # compute partial properties for other edges pointing to same target
        # node
        incoming_edges = [
            e2 for e2 in edges
            if e2.get_destNode() is e_target and e2 is not e]
        for e2 in incoming_edges:

            e2_etime = compute_smv_edge_clock_name(e2)
            e2_tmin = e2.get_tMin()
            e2_tmax = e2.get_tMax()

            ls_below_tmin.append("%s < %s" % (e2_etime, e2_tmin))

            # skip edges with tmax=infty
            if not tmax_infty and float(e2_tmax) >= 0:
                ls_exceeding_tmax.append("%s > %s" % (e2_etime, e2_tmax))

        new_proof_obligations = gen_completeness_aux(
            e_target, ls_exceeding_tmax, ls_below_tmin, for_tightness=True)

        for new_po in new_proof_obligations:
            new_po['object'] = e_name
            new_po['description'] = "can tighten " + \
                new_po['description'] + " of edge %s" % e_name

        proof_obligations += new_proof_obligations

    if proof_obligations_fname is not None:
        dump_proof_obligations_to_file(
            proof_obligations,
            proof_obligations_fname,
            'invarspec')

    return proof_obligations


# INTERNAL

def compute_mode_disabling_expression(edges, chosen_edge, chosen_mode):

    chosen_mode_param = None
    other_mode_params = []

    for e in edges:
        if len(e.get_modes()) == 1:
            continue

        for m in e.get_modes():
            if e is chosen_edge and m is chosen_mode:
                chosen_mode_param = compute_smv_em_params_name(e, m)
            else:
                other_mode_params.append(compute_smv_em_params_name(e, m))

    return " & ".join(other_mode_params) + " & !" + chosen_mode_param


def dump_proof_obligations_to_file(proof_obligations, out_fname, spec_type):
    """ Dumps a list of proof obligations to file.

    @param proof_obligations: the dict representing proof obligations
    @type  proof_obligations: list of dict

    @param out_fname: the name of the output file
    @type  out_fname: string
    """

    assert spec_type in ('ltlspec', 'invarspec')

    out_file = open(out_fname, "w")

    for proof_obl_idx in range(len(proof_obligations)):

        proofobl = proof_obligations[proof_obl_idx]

        out_file.write(
            "-- proof obligation %d/%d\n" %
            (proof_obl_idx + 1, len(proof_obligations)))
        out_file.write("-- %s?\n" % proofobl['description'])
        if spec_type == 'ltlspec':
            out_file.write("LTLSPEC (%s);\n\n" % proofobl['prop'])
        else:
            out_file.write("INVARSPEC (%s);\n\n" % proofobl['prop'])

    out_file.close()
    return


def gen_completeness_props_for_node(tfpg_node, edges, untimed, parametric):

    incoming_edges = [e for e in edges if e.get_destNode() is tfpg_node]
    assert len(incoming_edges) > 0

    ls_below_tmin = []
    ls_exceeding_tmax = []

    # for each incoming edge e
    for e in incoming_edges:

        etime = compute_smv_edge_clock_name(e)

        tmin = 0 if untimed else e.get_tMin()
        tmax_real = -1 if untimed else e.get_tMax()
        tmax_bool = None

        if parametric:
            tmin = compute_smv_etmin_params_name(e)
            tmax_real = compute_smv_etmax_real_params_name(e)
            tmax_bool = compute_smv_etmax_bool_params_name(e)

        ls_below_tmin.append("%s < %s" % (etime, tmin))

        if parametric:
            # (!guard & etime > param_tmax_real)
            ls_exceeding_tmax.append("(!%s & %s > %s)" % (tmax_bool, etime, tmax_real))

        elif float(tmax_real) >= 0:
            # (etime > const_tmax)
            ls_exceeding_tmax.append("%s > %s" % (etime, tmax_real))

        else:
            # the expression "etime(e) > infinity" simplifies to FALSE,
            # as it can never be true.
            assert float(tmax_real) == -1

            if not untimed:
                ls_exceeding_tmax.append("FALSE")

            else:
                # in the untimed case, the tmax proof obligations trivially
                # evaluate to true and can thus be ignored. here it means that
                # the tmax expressions are not added to the list of relevant
                # expressions (ls_exceeding_tmax), and the corresponding proof
                # obligations are therefore not built at all.
                pass

    return gen_completeness_aux(
        tfpg_node,
        ls_exceeding_tmax,
        ls_below_tmin,
        for_tightness=False)


def gen_completeness_aux(
        tfpg_node,
        ls_exceeding_tmax,
        ls_below_tmin,
        for_tightness):
    """ Generates completeness proof obligations for a given TFPG node from
    lists of partial constraints.

    @param tfpg_node: the TFPG node whose completeness po's should be generated
    @type  tfpg_node: data_structure.tfpg.tfpg.Node

    @param ls_exceeding_tmax: list of expressions for propagation delays
                              exceedingtmax of edges pointing to tfpg_node
    @type  ls_exceeding_tmax: list

    @param ls_below_tmin: list of expressions for propagation delays being
                          lower than tmin of edges pointing to tfpg_node
    @type  ls_below_tmin: list
    """

    tmin_descr = 'tmin' if for_tightness \
        else 'unexpected activation of node %s is possible' % \
        tfpg_node.get_name()
    tmax_descr = 'tmax' if for_tightness \
        else 'missed activation of node %s is possible' % tfpg_node.get_name()

    props = []

    node_name = compute_smv_node_name(tfpg_node)

    node_type = tfpg_node.get_type().lower()
    assert node_type in ["and", "or"]

    if node_type == "or":

        # missed activation
        if len(ls_exceeding_tmax) > 0:
            prop = "!(%s)" % ' | '.join(ls_exceeding_tmax)
            props.append({'object': tfpg_node.get_name(),
                          'prop': prop,
                          'proptype': 'invarspec',
                          'description': tmax_descr,
                          'tmax-infinity': False})

        # unexpected activation
        prop = "!(%s & (%s))" % (node_name, ' & '.join(ls_below_tmin))
        props.append({'object': tfpg_node.get_name(),
                      'prop': prop,
                      'proptype': 'invarspec',
                      'description': tmin_descr,
                      'tmax-infinity': False})

    if node_type == "and":

        # missed activation
        if len(ls_exceeding_tmax) > 0:
            prop = "!(%s)" % ' & '.join(ls_exceeding_tmax)
            props.append({'object': tfpg_node.get_name(),
                          'prop': prop,
                          'proptype': 'invarspec',
                          'description': tmax_descr,
                          'tmax-infinity': False})

        # unexpected activation
        prop = "!(%s & (%s))" % (node_name, ' | '.join(ls_below_tmin))
        props.append({'object': tfpg_node.get_name(),
                      'prop': prop,
                      'proptype': 'invarspec',
                      'description': tmin_descr,
                      'tmax-infinity': False})

    return props


def get_smallest_decimal(tfpg_struct):
    """ Computes precision needed to express any tmin/tmax constant and returns
    the smallest number corresponding to that precision. E.g.: for 0.004 and
    0.2 it will return 0.001.

    @param tfpg_struct: the tfpg to be considered
    @type  tfpg_struct: data_structure.tfpg.tfpg.Tfpg

    @return [anonymous]: the smallest decimal corresponding to precision of
                         tmin/tmax constraints
    @rtype  [anonymous]: float
    """

    def get_precision(tbound):
        f_tbound = float(tbound)
        assert f_tbound == -1 or f_tbound >= 0
        decpart = str(f_tbound).split('.')[1]
        if len(decpart) == 1 and decpart == '0':
            return 0
        assert len(decpart) > 1
        return len(decpart)

    max_precision = 0

    for e in tfpg_struct.get_edges_list():

        prec = get_precision(e.get_tMin())
        if prec > max_precision:
            max_precision = prec

        prec = get_precision(e.get_tMax())
        if prec > max_precision:
            max_precision = prec

    if max_precision == 0:
        return 1
    else:
        assert max_precision > 0
        return float('0.' + (max_precision - 1) * '0' + '1')


# TESTING

if __name__ == "__main__":

    from xsap.data_structure.tfpg.tfpg import Tfpg

    # build random tfpg
    tfpg = Tfpg()
    # nodes
    sys_dead = tfpg.add_and_node("SYS_DEAD", True)
    s1_wo = tfpg.add_or_node("S1_WO", False)
    b1_dead = tfpg.add_or_node("B1_DEAD", False)
    b1_low = tfpg.add_or_node("B1_LOW", True)
    g1_dead = tfpg.add_or_node("G1_DEAD", False)
    g1 = tfpg.add_fm_node("G1_OFF")
    s1 = tfpg.add_fm_node("S1_OFF")
    s2_wo = tfpg.add_or_node("S2_WO", False)
    b2_dead = tfpg.add_or_node("B2_DEAD", False)
    b2_low = tfpg.add_or_node("B2_LOW", True)
    g2_dead = tfpg.add_or_node("G2_DEAD", False)
    g2 = tfpg.add_fm_node("G2_OFF")
    s2 = tfpg.add_fm_node("S2_OFF")
    # modes
    primary = "primary"
    secondary1 = "secondary_1"
    secondary2 = "secondary_2"
    default = [primary, secondary1, secondary2]
    tfpg.set_modes_list(default)
    # edges
    tfpg.add_edge(g1, 0, 0, default, g1_dead, "edge1")
    tfpg.add_edge(g1_dead, 0, 100, [primary, secondary1], b1_low, "edge2")
    tfpg.add_edge(b1_low, 5, 10, [primary, secondary1], b1_dead, "edge3")
    tfpg.add_edge(b1_dead, 0, 1, [primary, secondary1], s1_wo, "edge4")
    tfpg.add_edge(s1, 0, 0, default, s1_wo, "edge5")
    tfpg.add_edge(s1_wo, 0, 1, default, sys_dead, "edge6")
    tfpg.add_edge(g2, 0, 0, default, g2_dead, "edge7")
    tfpg.add_edge(g2_dead, 0, 100, [primary, secondary2], b2_low, "edge8")
    tfpg.add_edge(b2_low, 5, 10, [primary, secondary2], b2_dead, "edge9")
    tfpg.add_edge(b2_dead, 0, 1, [primary, secondary2], s2_wo, "edge10")
    tfpg.add_edge(s2, 0, 0, default, s2_wo, "edge11")
    tfpg.add_edge(s2_wo, 0, 1, default, sys_dead, "edge12")
    tfpg.add_edge(b1_dead, 0, 1, [secondary1], s2_wo, "edge13")
    tfpg.add_edge(b2_dead, 0, 1, [secondary2], s1_wo, "edge14")

    errors = tfpg.get_errors_list()
    assert not errors

    proof_obl = generate_proof_obligations_completeness(tfpg, True)
    assert len(proof_obl) == 2 * \
        len(tfpg.get_and_nodes() + tfpg.get_or_nodes())
    print("TEST SUCCESSFULLY PASSED!!")
