from __future__ import absolute_import

import tempfile
import os
import subprocess


# INTERFACES

def perform_effectiveness_validation(
        tfpg_struct,
        sampling_rate,
        target_fm_set,
        target_system_mode,
        use_bmc,
        bmc_length,
        outdir,
        xsap_process=None):
    """Implements the functionality of effectiveness validation.

    @param tfpg_struct: the tfpg to be analyzed
    @type  tfpg_struct: data_structure.tfpg.tfpg.Tfpg

    @param sampling_rate: the interval at which the system is monitored
    @type  sampling_rate: float

    @param target_fm_set: the set of FMs that need to be diagnosed
                          (colon-separated)
    @type  target_fm_set: string

    @param target_system_mode: the system mode for which effectiveness
                               should be analyzed
    @type  target_system_mode: string

    @param use_bmc: tells if BMC must be used
    @type  use_bmc: boolean

    @param bmc_length: maximum path length for BMC
    @type  bmc_length: int

    @param outdir: output directory for generated files
    @type  outdir: string

    @param xsap_process: the xsap process we may have created and
                         we want to use
    @type  xsap_process: utils.processes.xsap_process.xSAPProcess
    """
    from xsap.utils.misc import log

    # dump TFPG as SMV -> TODO: check something? i.e. that the variables are
    # not contained in the model?
    from xsap.features.common.tfpg2smv import dump_tfpg_as_smv
    system_mode_var_name = "system_mode"
    time_tick = "time_tick"
    frozen_sys_mode = True
    dump_validation_props = False

    tfpg_name = tfpg_struct.get_name().lower()
    tfpg_smv_fname = os.path.join(outdir, tfpg_name + ".smv")

    log_out, logfname = tempfile.mkstemp(
        suffix=".tfpgout", prefix=tfpg_name,
        dir=outdir, text=True)
    log.xsap.tfpg_effect.debug("Invoking xSAP, with redirect to " + logfname)

    # if we specify a xsap_process, we want to handle manually the process
    user_spec_proc = xsap_process is not None

    dump_tfpg_as_smv(
        tfpg_struct,
        tfpg_smv_fname,
        time_tick,
        sampling_rate,
        frozen_sys_mode,
        dump_validation_props)

    # get observables
    from xsap.features import tfpg_effectiveness_validation
    from .observables_extractor import extract_and_dump_observables

    obs_fname = os.path.join(outdir, tfpg_name + "_diag_obs.tmp")
    log.xsap.tfpg_effect.debug(
        'Creating helper file for observables (%s).' % obs_fname)
    extract_and_dump_observables(
        tfpg_struct,
        obs_fname,
        time_tick,
        system_mode_var_name)

    # create target FMs list
    from xsap.features import tfpg_effectiveness_validation
    from .diagnosability_cond_generator import (
        generate_diag_conditions_and_get_smv_prop)

    diagcond_fname = os.path.join(outdir, tfpg_name + "_diag_cond.tmp")
    log.xsap.tfpg_effect.debug(
        'Creating helper file for diagnosability condition (%s).' %
        diagcond_fname)
    smv_prop = generate_diag_conditions_and_get_smv_prop(
        target_fm_set,
        tfpg_struct,
        diagcond_fname,
        system_mode_var_name,
        target_system_mode)

    # instantiate the process
    from xsap.utils.processes.xsap_process import XsapProcess
    if not user_spec_proc:
        xsap_process = XsapProcess(tfpg_smv_fname, fail_obs_alm_fname=None,
                                   ord_fname=None, need_msat=False,
                                   args=[], stdout=subprocess.PIPE,
                                   stderr=subprocess.STDOUT,
                                   dumpers=[os.fdopen(log_out, "w")])
        xsap_process.set_tmp_filename_prefix(outdir)
    else:
        # if we already own a process, we need to read the
        # extended model
        xsap_process.reset()
        xsap_process._need_msat = False
        xsap_process.read_model(model=tfpg_smv_fname)
        # the log file can be closed as it is not needed
        os.close(log_out)

    res = xsap_process.check_tfpg_diagnosability(
        obs_fname, diagcond_fname, smv_prop, use_bmc, bmc_length)

    # split the trace and update the results
    if xsap_process.is_alive(): # killed from outside
        orig_trace = res['critical_pair']
        if orig_trace is not None:
            from xsap.features.tfpg_effectiveness_validation import trace_splitter
            from .trace_splitter import split_effectiveness_tracefile
            res['critical_pair'] = split_effectiveness_tracefile(orig_trace)

    if not user_spec_proc:
        xsap_process.kill_and_close_streams()

    return res
