from pysmt.shortcuts import (Symbol, And, Or, LE, Minus, Real, TRUE,
                             Implies, ExactlyOne, Equals, Iff, GE)
from pysmt.typing import REAL, BOOL

from xsap.data_structure.tfpg.tfpg import NodeType

#
# Helper functions for the SMT encoding
#


def edge_tmin_symbol(edge):
    """Returns the symbol associated to tMin of the given edge.

    @param edge: The Edge
    @type  edge: xsap.data_structure.tfpg.tfpg.Edge

    @return    : A Symbol for tMin
    @rtype     : pysmt.fnode.FNode
    """
    src, dst = edge.get_srcNode(), edge.get_destNode()
    name = "$tmin_%s_%s" % (src.get_name(), dst.get_name())
    return Symbol(name, REAL)


def edge_tmax_symbol(edge):
    """Returns the symbol associated to tMax of the given edge.

    @param edge: The Edge
    @type  edge: xsap.data_structure.tfpg.tfpg.Edge

    @return    : A Symbol for tMax
    @rtype     : pysmt.fnode.FNode
    """
    src, dst = edge.get_srcNode(), edge.get_destNode()
    name = "$tmax_%s_%s" % (src.get_name(), dst.get_name())
    return Symbol(name, REAL)


def node_state_symbol(node):
    """Returns the symbol associated to the activation state of the Node.

    @param node: The node name
    @type  node: string

    @param node: The node
    @type  edge: xsap.data_structure.tfpg.tfpg.Node

    @return    : A Symbol for the activation state of the Node
    @rtype     : pysmt.fnode.FNode
    """
    if isinstance(node, str):
        s = node
    else:
        s = node.get_name()
    return Symbol("%s__state" % s, BOOL)


def monitor_state_symbol(node):
    """Returns the symbol of the activation state of the monitor of a Node.

    @param node: The node name
    @type  node: string

    @param node: The node
    @type  edge: xsap.data_structure.tfpg.tfpg.Node

    @return    : A Symbol for the activation state of the monitor of the Node
    @rtype     : pysmt.fnode.FNode
    """
    if isinstance(node, str):
        s = node
    else:
        s = node.get_name()
    return Symbol("%s__monitor_state" % s, BOOL)


def get_primed_symbol(symbol):
    """Returns a primed version of a symbol.

    @param symbol: A symbol
    @type  symbol: pysmt.fnode.FNode

    @return      : An associated new symbol with the same type
    @type        : pysmt.fnode.FNode
    """
    return Symbol("%s_primed" % str(symbol), typename=symbol.symbol_type())


def node_time_point_symbol(node):
    """Returns the symbol associated to the activation time of the Node.

    @param node: The node name
    @type  node: string

    @param node: The node
    @type  edge: xsap.data_structure.tfpg.tfpg.Node

    @return    : A Symbol for the activation time of the Node
    @rtype     : pysmt.fnode.FNode
    """
    if isinstance(node, str):
        s = node
    else:
        s = node.get_name()
    return Symbol("%s__time" % s, REAL)


def monitor_time_point_symbol(node):
    """Returns the symbol of the activation time of the monitor of the Node.

    @param node: The node name
    @type  node: string

    @param node: The node
    @type  edge: xsap.data_structure.tfpg.tfpg.Node

    @return    : A Symbol for the activation time of the monitor of the Node
    @rtype     : pysmt.fnode.FNode
    """
    if isinstance(node, str):
        s = node
    else:
        s = node.get_name()
    return Symbol("%s__monitor_time" % s, REAL)


def monitor_health_symbol(node):
    """Returns the symbol of the health status of the monitor of the Node.

    @param node: The node name
    @type  node: string

    @param node: The node
    @type  edge: xsap.data_structure.tfpg.tfpg.Node

    @return    : A Symbol for the health status of the monitor of the Node
    @rtype     : pysmt.fnode.FNode
    """
    if isinstance(node, str):
        s = node
    else:
        s = node.get_name()
    return Symbol("%s__monitor_health" % s, REAL)


def mode_symbol(mode):
    """Returns the symbol associated to the given Mode.

    @param node: The Mode name
    @type  node: string

    @return    : A Symbol for the Mode
    @rtype     : pysmt.fnode.FNode
    """
    return Symbol("%s__mode" % mode, BOOL)


def tfpg_to_formula(tfpg, positive=False):
    """Converts the TFPG into an SMT formula.

    This returns the formula not considering the observable part of
    the TFPG (i.e., without monitors).

    @param tfpg: The TFPG to convert
    @type  tfpg: xsap.data_structure.tfpg.tfpg.TFPG

    @param positive: Whether to require all activation times to be >= 0
    @type  positive: Boolean

    @return    : The formula representing the TFPG
    @rtype     : pysmt.fnode.FNode

    """
    constraints = []
    for node in tfpg.get_nodes_list():
        if node.type == NodeType.OR:
            constraints.append(
                get_or_constraint(
                    node, tfpg, positive=positive))
        elif node.type == NodeType.AND:
            constraints.append(
                get_and_constraint(
                    node, tfpg, positive=positive))
        else:
            # Do nothing for FM
            assert node.type == NodeType.FM
    ftfpg = And(constraints)

    # Only one mode can be true
    cardinality_constraint = ExactlyOne([mode_symbol(mode)
                                         for mode in tfpg.get_modes_list()])
    ftfpg = And(ftfpg, cardinality_constraint).simplify()
    return ftfpg


def mode_compatible_with_edge(edge):
    """Given an edge returns the disjunction of modes.

    This formula is used to indicate that only one of the modes
    defined in the edge can be active.

    @param edge: The edge from which to take the modes
    @type  edge: xsap.data_structure.tfpg.tfpg.Edge

    @return    : A disjunction of the modes of the edge.
    @rtype     : pysmt.fnode.Fnode
    """
    disj = [mode_symbol(mode) for mode in edge.get_modes()]
    return Or(disj)


def get_or_constraint(dst, tfpg, positive=False):
    """Returns a formula encoding the constraints on a OR node.

    Wrt the paper, here we perform the computation of the Boolean
    constraints and Temporal constraints at the same time. Moreover,
    we split the temporal constraints in two. The second part is used
    to define the upper-bounds (minimum_ub) and is not defined if tmax
    is INFINITY.

    @param dst: The node to consider
    @type  dst: xsap.data_structure.tfpg.tfpg.Node

    @param tfpg: The Tfpg to which the node belongs to
    @type  tfpg: xsap.data_structure.tfpg.tfpg.Tfpg

    @param positive: Whether to require all activation times to be >= 0
    @type  positive: Boolean

    @return    : A disjunction of the modes of the edge.
    @rtype     : pysmt.fnode.Fnode
    """
    # Get all predecessors of node
    srcs = tfpg.get_predecessors(dst)
    assert len(srcs) > 0, "Dst has no predecessor: %s" % dst.name

    constraints = []
    time_constraints = []
    minimum_ub_constraints = []
    for src in srcs:
        edge = tfpg.get_edge(src, dst)
        mode = mode_compatible_with_edge(edge)
        diff = Minus(node_time_point_symbol(dst), node_time_point_symbol(src))
        tmin = Real(edge.get_tMin())
        tmax = Real(edge.get_tMax())

        timing = LE(tmin, diff)
        if edge.get_tMax() >= 0:
            timing = And(LE(tmin, diff), LE(diff, tmax))
            minimum_ub_constraints.append(Implies(And(node_state_symbol(src),
                                                      node_state_symbol(dst)),
                                                  LE(diff, tmax)))
        if positive:
            timing = And(timing, GE(node_time_point_symbol(dst), Real(0)))

        constraints.append(And(node_state_symbol(src), mode))
        time_constraints.append(And(node_state_symbol(src), timing))

    tim = Implies(node_state_symbol(dst), Or(time_constraints))
    boo = Iff(node_state_symbol(dst), Or(constraints))
    f_or = And(tim,  # T_or
               boo,  # B_or
               And(minimum_ub_constraints))
    return f_or


def get_and_constraint(dst, tfpg, positive=False):
    """Returns a formula encoding the constraints on an AND node.

    Wrt the paper, here we perform the computation of the Boolean
    constraints and Temporal constraints at the same time.

    @param dst: The node to consider
    @type  dst: xsap.data_structure.tfpg.tfpg.Node

    @param tfpg: The Tfpg to which the node belongs to
    @type  tfpg: xsap.data_structure.tfpg.tfpg.Tfpg

    @param positive: Whether to require all activation times to be >= 0
    @type  positive: Boolean

    @return    : A disjunction of the modes of the edge.
    @rtype     : pysmt.fnode.Fnode

    """
    srcs = tfpg.get_predecessors(dst)
    assert len(srcs) > 0
    disj_constraints = []
    constraints = []
    lb_time_constraints = []
    ub_time_constraints = []

    # Add At-Least-One tmax constraint
    for src in srcs:
        edge = tfpg.get_edge(src, dst)
        mode = mode_compatible_with_edge(edge)
        diff = Minus(node_time_point_symbol(dst), node_time_point_symbol(src))
        tmax = Real(edge.get_tMax())

        disj_constraints.append(And(node_state_symbol(src), mode))

        timing = TRUE()
        if edge.get_tMax() >= 0:
            timing = LE(diff, tmax)
        ub_time_constraints.append(And(node_state_symbol(src), timing))
    constraints.append(Or(disj_constraints))

    # Add All-tmin constraint
    for src in srcs:
        edge = tfpg.get_edge(src, dst)
        mode = mode_compatible_with_edge(edge)
        diff = Minus(node_time_point_symbol(dst), node_time_point_symbol(src))
        tmin = Real(edge.get_tMin())

        constraints.append(And(node_state_symbol(src), mode))
        lb_time_constraints.append(And(node_state_symbol(src), GE(diff, tmin)))

    if positive:
        constraints.append(GE(node_time_point_symbol(dst), Real(0)))

    tim = Implies(node_state_symbol(dst),
                  And(Or(ub_time_constraints),
                      And(lb_time_constraints)))
    boo = Iff(node_state_symbol(dst), And(constraints))
    f_and = And(tim, boo)
    return f_and


def get_monitors_formula(tfpg, include_health):
    """Returns the part of the encoding dealing with monitored discrepancies.

    Note that health variables are encoded as Reals and not
    boolean. This makes it extremly simple to express cardinality
    constraints on those variables (i.e., at most 2 monitor can fail).

    @param tfpg: The TFPG @type tfpg:
    xsap.data_structure.tfpg.tfpg.Tfpg

    @param include_health: Whether to include health variables in the encoding.
    @type  include_health: Boolean

    @return  : The formula encoding the monitored discrepancies
    @rtype   : pysmt.fnode.FNode

    """
    monitored_nodes = [node for node in tfpg.get_nodes_list()
                       if node.is_monitored()]
    constraints = []
    for monitor in monitored_nodes:
        if include_health:
            health = Equals(monitor_health_symbol(monitor), Real(1))
        else:
            health = TRUE()
        odt = monitor_time_point_symbol(monitor)
        udt = node_time_point_symbol(monitor)
        od = monitor_state_symbol(monitor)
        ud = node_state_symbol(monitor)

        same_behaviour = And(Iff(od, ud), Equals(odt, udt))
        constraints.append(Implies(health, same_behaviour))

    if include_health:
        healths = [monitor_health_symbol(m) for m in monitored_nodes]
        constraints.append(And([Or(Equals(h, Real(1)),
                                   Equals(h, Real(0))) for h in healths]))
    return And(constraints)


def obs_tfpg_formula(tfpg, include_health=True):
    """Returns the observable version of the encoding of the TFPG.

    The return formula is defined as:
      otfpg(odt, udt, m, h)

    and is the orginal encoding with the addition of the observable part.

    @param tfpg: The TFPG
    @type  tfpg: xsap.data_structure.tfpg.tfpg.Tfpg

    @param include_health: Whether to include health information in the
                           monitors
    @type  include_health: Boolean

    @return  : The observable encoding of the TFPG
    @rtype   : pysmt.fnode.FNode
    """
    ftfpg = tfpg_to_formula(tfpg)
    monitors = get_monitors_formula(tfpg, include_health=include_health)
    otfpg = And(ftfpg, monitors)
    return otfpg


def get_mode_in_model(tfpg, smt_model):
    """Utility function to extract the Mode from an SMT Model.

    @param tfpg: The TFPG
    @type  tfpg: xsap.data_structure.tfpg.tfpg.Tfpg

    @param smt_model: A model or solver instance
    @type  smt_model: pysmt Model

    @return  : The mode in the model
    @rtype   : string
    """
    modes = [mode for mode in tfpg.get_modes_list()
             if smt_model.get_value(mode_symbol(mode)) == TRUE()]
    assert len(modes) == 1,\
        "Expected a single mode but the following where found" + \
        str(modes)
    return modes[0]
