# This implementation is based on:
#   M. Bozzano, A. Cimatti, M. Gario, A. Micheli,
#    "SMT-based Validation of Timed Failure Propagation Graphs",
#   AAAI 2015
#
# where we explain the intuition and details behind the encoding.
#
from pysmt.logics import QF_LRA
from pysmt.shortcuts import Bool, Real
from pysmt.shortcuts import Not, And, Iff, Equals, ForAll
from pysmt.shortcuts import Solver, qelim, get_model

from xsap.features.tfpg_smt.encoding import (tfpg_to_formula,
                                             obs_tfpg_formula,
                                             node_time_point_symbol,
                                             node_state_symbol,
                                             get_mode_in_model,
                                             mode_symbol,
                                             monitor_time_point_symbol,
                                             monitor_state_symbol,
                                             monitor_health_symbol,
                                             get_primed_symbol)

from xsap.data_structure.tfpg.tfpg_errors import UnsupportedInfinitySemantics


class TFPGReasoner(object):
    """ Class representing a TFPG Reasoner.

    The reasoner takes a TFPG and builds the SMT representation.
    Moreover, it provides facilities to perform different types
    of analysis on the TFPG.
    """

    def __init__(self, tfpg, include_health=True):
        """Build a TFPGReasoner for the given TFPG structure.

        @param tfpg: The TFPG on which to perform reasoning
        @type  tfpg: data_structure.tfpg.tfpg.Tfpg

        @param include_health: Enables the use of health variables for
                               monitored discrepancies.
        @type  include_health: Boolean.
        """
        if not tfpg.has_open_infinity_semantics():
            raise UnsupportedInfinitySemantics(
                "Expecting Open Infinity semantics.")

        self.tfpg = tfpg
        self.ftfpg = None
        self.otfpg = None
        self.include_health = include_health

    def get_tfpg_formula(self, positive=False):
        """Returns the SMT encoding of the TFPG: ftfpg(udt, m).

        The encoding is expressed on all non-observable variables.

        @param positive: Enforce activation times to be positive.
        @type  positive: Boolean

        @return   self.ftfpg: The SMT formula encoding the TFPG.
        @rtype    self.ftfpg: pysmt.fnode.FNode
        """

        if self.ftfpg is None:
            self.ftfpg = tfpg_to_formula(self.tfpg, positive=positive)
        return self.ftfpg

    def get_otfpg_formula(self):
        """Returns the SMT encoding of the observable TFPG: otfpg(odt, udt, m, h)

        @return   self.otfpg: The SMT formula encoding the observable TFPG.
        @rtype    self.otfpg: pysmt.fnode.FNode
        """

        if self.otfpg is None:
            self.otfpg = obs_tfpg_formula(self.tfpg,
                                          include_health=self.include_health)
        return self.otfpg

    # MG: This is here because we check that all values of the
    # scenario match the TFPG. This could be moved to the utility
    # functions, if we check the correctness afterwards.
    # Also, this might need to be integrated with existing functionalities for
    # partial traces of XSAP.
    #
    def scenario_to_formula(self, scenario):
        """Converts a scenario into a formula.

        Syntax:
          #mode        Defines the mode
          #__mon_D1    Monitor of D1 state
          __mon_D1     Monitor of D1 time-point
          __H_mon_D1   Health of Monitor of D1
          #D1          State of D1
          D1           Time-point of D1

        If the boolean encoding is enabled, specifying a time-point
        automatically forces the state to be true.

        @param scenario: The scenario to be converted
        @type  scenario: dict

        @return  The SMT formula encoding the scenario.
        @rtype   pysmt.fnode.FNode
        """

        sc = []
        for (element, value) in scenario.items():
            if element == "#mode":
                assert self.tfpg.has_mode(value), value
                sc.append(mode_symbol(value))
            elif element[0:7] == "#__mon_":  # Monitor state
                node_name = element[7:]
                assert self.tfpg.has_node_by_name(node_name), node_name
                sc.append(Iff(monitor_state_symbol(node_name),
                              Bool(value)))
            elif element[0:6] == "__mon_":  # Monitor Time
                node_name = element[6:]
                assert self.tfpg.has_node_by_name(node_name), node_name
                sc.append(Equals(monitor_time_point_symbol(node_name),
                                 Real(value)))
                sc.append(monitor_state_symbol(node_name))
            elif element[0:8] == "__H_mon_":  # Monitor Health
                node_name = element[8:]
                assert self.tfpg.has_node_by_name(node_name), node_name
                sc.append(Equals(monitor_health_symbol(node_name),
                                 Real(value)))
            elif element[0] == "#":  # element is a state info
                node_name = element[1:]
                assert self.tfpg.has_node_by_name(node_name), node_name
                node = node_state_symbol(node_name)
                value = Bool(value)
                sc.append(Iff(node, value))
            else:  # Time-point information
                node_name = element
                assert self.tfpg.has_node_by_name(node_name), node_name
                sc.append(Equals(node_time_point_symbol(node_name),
                                 Real(value)))
                sc.append(node_state_symbol(node_name))
        res = And(sc)

        return res

    def certain_diagnosis(self, scenario, beta):
        """Checks if a condition beta is implied by all diagnosis of the scenario.

        @param scenario: A scenario
        @type  scenario: dict

        @param beta: A diagnosis condition
        @type  beta: pysmt.fnode.FNode

        @return A tuple (res, cex). res represents whether beta
                appears in all diagnosis. If res is False, cex contains a
                counter-example, None otherwise.
        @rtype (res, cex): (Boolean, Model)
        """

        fscenario = self.scenario_to_formula(scenario)
        query = And(Not(beta),
                    fscenario,
                    self.get_otfpg_formula())

        model = get_model(query)
        if model is not None:
            model = self.get_scenario_from_model(model, include_obs=True)
            return (False, model)
        else:
            return (True, None)

    def check_scenario(self, scenario, obs=False):
        r"""Checks the consistency of a TFPG wrt a Scenario.

        This runs a SAT check on TFPG /\ Scenario, and returns
        the result. If no scenario is provided, SAT of TFPG is
        tested.

        The scenario is provided as a dictionary:
          { "NodeName": time, "#mode": mode }

        @param scenario: A scenario
        @type  scenario: dict

        @param obs: Use observable tfpg model
        @param obs: Boolean

        @return (res, cex): res represents whether the scenario is consistent
                            wrt the TFPG. If res is False, cex contains a
                            counter-example, None otherwise.
        @rtype  (res, cex): (Boolean, Scenario)
        """

        sc = self.scenario_to_formula(scenario)
        return self.check_consistency(sc, obs=obs)

    def check_scenario_necessity(self, scenario, obs=False):
        """Checks the necessity of a scenario in the TFPG.

        @param scenario: A scenario
        @type  scenario: dict

        @param obs: Use observable tfpg model
        @param obs: Boolean

        @return (res, cex): res represents whether the scenario is
                            necessary wrt the TFPG. If res is False,
                            cex contains a counter-example, None
                            otherwise.
        @rtype (res, cex): (Boolean, Scenario)
        """
        sc = self.scenario_to_formula(scenario)
        n_res, model = self.check_consistency(Not(sc), obs=obs)
        return (not n_res, model)

    def check_consistency(self, condition=None, obs=False):
        r"""Checks the consistency of a TFPG wrt a condition.

        This runs a SAT check on TFPG /\ Condition, and returns the
        result. If no condition is provided, SAT of TFPG is tested.

        @param condition: The condition to test
        @type  condition: pysmt.fnode.FNode

        @param obs: Use observable tfpg model
        @param obs: Boolean

        @return (res, cex): res represents whether the condition is satisfiable
                            wrt the TFPG. If res is True, cex contains a
                            witness, None otherwise.
        @rtype (res, cex): (Boolean, Scenario)
        """

        if condition is None:
            condition = Bool(True)
        if obs:
            tfpg_formula = self.get_otfpg_formula()
        else:
            tfpg_formula = self.get_tfpg_formula()

        query = And(tfpg_formula, condition)

        model = get_model(query)
        if model is not None:
            model = self.get_scenario_from_model(model, include_obs=obs)
            return (True, model)
        return (False, None)

    def get_scenario_from_model(self, smt_model, include_obs=False):
        """Extracts a scenario over the TFPG elements from the SMT Model.

        @param smt_model: An SMT model
        @type  smt_model: pysmt Model

        @param include_obs: Include observable discrepancies in the scenario.
        @type  include_obs: Boolean

        @return sc: The scenario represented in the SMT Model
        @rtype  sc: dict
        """

        sc = {}

        for n in self.tfpg.get_nodes_list():
            sc[n.get_name()] = smt_model.get_value(node_time_point_symbol(n))
            sc["#" + n.get_name()] = smt_model.get_value(node_state_symbol(n))

        # Exit if we are not considering Observable information
        if not include_obs:
            return sc

        # Monitors
        for n in self.tfpg.get_monitored_discrepancies():
            mon = "__mon_" + n.get_name()
            sc[mon] = smt_model.get_value(monitor_time_point_symbol(n))
            sc["#" + mon] = smt_model.get_value(monitor_state_symbol(n))
            if self.include_health:
                sc["__H_mon_" +
                   n.get_name()] = smt_model.get_value(
                    monitor_health_symbol(n))

        # Convert from PySMT types to Python Types
        for k in sc:
            assert sc[k].is_constant()
            sc[k] = sc[k].constant_value()

        # Add mode information -- This is done here because the function
        # get_mode_in_model returns a string, and not a pySMT object
        sc["#mode"] = get_mode_in_model(self.tfpg, smt_model)
        return sc

    def get_diagnosis_from_model(self, smt_model):
        """Extracts the FMs and modes from the SMT Model.

        @param smt_model: An SMT model
        @type  smt_model: pysmt Model

        @return sc: The scenario represented in the SMT Model
        @rtype  sc: dict
        """

        sc = {}

        # Mode
        sc["#mode"] = get_mode_in_model(self.tfpg, smt_model)
        # Failure Modes
        for n in self.tfpg.get_fm_nodes():
            smt_value = smt_model.get_value(node_state_symbol(n))
            assert smt_value.is_constant()
            sc["#" + n.get_name()] = smt_value.constant_value()
        return sc

    def get_all_diagnosis(self, observation):
        r"""Returns all diagnosis for a given observation.

        Performs an All-SAT on TFPG /\ observation, enumerating FM and modes.

        @param observation: An observation to diagnose
        @type  observation: pysmt.fnode.Fnode

        @return diagnoses: A list of scenarios over FM and modes (diagnoses).
        @rtype  diagnoses: List of scenarios
        """

        diagnoses = []
        with Solver(logic=QF_LRA) as solver:
            solver.add_assertion(self.get_otfpg_formula())
            solver.add_assertion(observation)
            while solver.solve():
                diag = self.get_diagnosis_from_model(solver)
                diagnoses.append(diag)
                solver.add_assertion(Not(self.scenario_to_formula(diag)))
        return diagnoses

    def get_all_modes(self, condition=None):
        r""" Returns the list of modes that satisfy the given condition.

        This runs an All-SAT on TFPG /\ condition, enumerating on modes.

        @param condition: The condition to check modes against
        @type  condition: pysmt.fnode.Fnode

        @return diagnoses: A list of modes compatible with the condition
        @rtype  diagnoses: List of modes
        """

        modes = []
        with Solver(logic=QF_LRA) as solver:
            solver.add_assertion(self.get_tfpg_formula())
            if condition:
                solver.add_assertion(condition)
            while solver.solve():
                mode_name = get_mode_in_model(self.tfpg, solver)
                solver.add_assertion(Not(mode_symbol(mode_name)))
                modes.append(mode_name)
        return modes

    def check_refinement(self, tfpg1, mapping):
        r"""Checks that tfpg1 refines self, using the given mapping.

        \forall u1, m1 . TFPG1 -> (\exists u2,m2 . (mapping /\ TFPG2))
        \forall u1, m1 . \exists u2,m2 . TFPG1 -> (mapping /\ TFPG2)
        \neg \exists u1, m1 . TFPG1 /\ (\forall u2,m2 .
        (mapping -> \neg TFPG2))

        @param tfpg1: The TFPG that refines self
        @type  tfpg1: xsap.data_structure.tfpg.tfpg.Tfpg

        @param mapping: Mapping of nodes between the TFPGs
        @type  mapping: xsap.features.tfpg_smt.utils.Mapping

        @return (res, cex): res is True if tfpg1 refines self wrt the
                            mapping. If res is False, a
                            counter-example expressed over tfpg1.
        @rtype (res, cex): (Boolean, Model)
        """

        ftfpg_1 = tfpg_to_formula(tfpg1)
        ftfpg_2 = self.get_tfpg_formula()

        # Substitute symbols in ftfpg_2
        vars2 = ftfpg_2.get_free_variables()
        vars2_renaming = {var: get_primed_symbol(var) for var in vars2}
        ftfpg_2 = ftfpg_2.substitute(vars2_renaming)
        ftfpg_2 = ftfpg_2.substitute(mapping.get_substitution())
        vars2 = [get_primed_symbol(var) for var in vars2]

        # Create refinement problem
        qproblem = And(ftfpg_1, ForAll(vars2, Not(ftfpg_2)))
        qf_problem = qelim(qproblem)
        model = get_model(qf_problem)

        if model is None:
            return True, None
        else:
            reduced_model = dict([(v, model.get_value(v).constant_value())
                                  for v in ftfpg_1.get_free_variables()])
            return False, reduced_model

    def check_activability(self):
        """ Check that each discrepancies is activable.

        @return A list of non activable discrepancies.
        @rtype List of nodes
        """

        non_activable = []
        with Solver() as solver:
            solver.add_assertion(self.get_tfpg_formula())
            for d in (self.tfpg.get_or_nodes() + self.tfpg.get_and_nodes()):
                node = node_state_symbol(d)
                res = solver.solve(assumptions=[node])
                if not res:
                    non_activable.append(d)
        return non_activable

    def check_diagnosability(self):
        """Returns the list of failure modes that are not diagnosable.

        Note: This code does not support the health variables.

        @return (m-fms, cexs): A list of modes-failure modes pairs. For each
                               element in the list mode in the list, a
                               corresponding element in the dictionary cexs
                               provides a counter-example to diagnosability.
        @rtype (fms, cexs): (List of Mode-Nodes pairs, Dict From nodes to
                             Models)
        """
        assert self.include_health is False, \
            ("Diagnosability testing does not currently support health "
             "information")

        non_diagnosable = {}
        non_obs_nodes = self.tfpg.get_unmonitored_discrepancies() +\
            self.tfpg.get_fm_nodes()
        udt = []
        ud = []
        for node in non_obs_nodes:
            udt.append(node_time_point_symbol(node))
            ud.append(node_state_symbol(node))

        # Create primed version of unobservable vars for the Right plant
        uvars_L = udt + ud
        uvars_R = [get_primed_symbol(v) for v in uvars_L]

        # Create a renaming map
        renaming = dict(zip(uvars_L, uvars_R))

        # Perform substitution to obtain the Right Plant
        otfpg_L = self.get_otfpg_formula()
        otfpg_R = otfpg_L.substitute(renaming)

        # Solve for each failure mode and mode combination
        with Solver() as solver:
            solver.add_assertion(otfpg_L)
            solver.add_assertion(otfpg_R)
            for fm in self.tfpg.get_fm_nodes():
                node_state = node_state_symbol(fm)
                cond_L = node_state
                cond_R = Not(node_state).substitute(renaming)
                solver.push()
                solver.add_assertion(And(cond_L, cond_R))
                for mode in self.tfpg.get_modes_list():
                    solver.push()
                    solver.add_assertion(mode_symbol(mode))
                    res = solver.solve()
                    if res:
                        cex_L = [(v, solver.get_py_value(v)) for v in uvars_L]
                        cex_R = [(v, solver.get_py_value(v)) for v in uvars_R]
                        key = (str(mode), str(fm))
                        non_diagnosable[key] = (cex_L, cex_R)
                    solver.pop()
                solver.pop()

        return (list(non_diagnosable.keys()), non_diagnosable)
