from __future__ import print_function
from builtins import object
from pysmt.shortcuts import Implies, Iff, Equals, And

from xsap.features.tfpg_smt.encoding import (node_time_point_symbol,
                                             node_state_symbol,
                                             get_primed_symbol)


class Mapping(object):
    """Describes a mapping between nodes of two TFPGs.

    This is used, e.g., in the definition of the Refinement problem.
    """

    def __init__(self, mapping=None):
        if mapping is not None:
            self.mapping = mapping
        else:
            self.mapping = []

    # MG: This should check that there are no multiple mappings of
    #     the same node. It might make sense to convert this into a dict.
    def add(self, n1, n2):
        """Adds the mapping from node n1 to n2."""
        self.mapping.append((n1, n2))

    def as_formula(self):
        """Returns the formula describing the mapping.

        The formula forces equality of both the boolean and temporal part.
        """
        states = [Iff(node_state_symbol(n1),
                      get_primed_symbol(node_state_symbol(n2)))
                  for (n1, n2) in self.mapping]
        tps = [Implies(node_state_symbol(n1),
                       Equals(node_time_point_symbol(n1),
                              get_primed_symbol(node_time_point_symbol(n2))))
               for (n1, n2) in self.mapping]
        return And(states + tps)

    def get_substitution(self):
        """Returns the substitution described by the mapping.

        The substitution matches each boolean and temporal symbol for
        the node n1 to the corresponding symbol for n2, for each pair
        (n1,n2) in the mapping.
        """
        subs = {}
        for n1, n2 in self.mapping:
            b1 = node_state_symbol(n1)
            b2 = get_primed_symbol(node_state_symbol(n2))
            t1 = node_time_point_symbol(n1)
            t2 = get_primed_symbol(node_time_point_symbol(n2))

            subs[b2] = b1
            subs[t2] = t1
        return subs

    def __str__(self):
        res = ["%s %s" % (k.name, v.name) for (k, v) in self.mapping]
        return "\n".join(res)

    def write_to_file(self, fname):
        with open(fname, "w") as fout:
            for k_v in self:
                fout.write("%s %s" % k_v)


def parse_scenario_file(fname):
    """Parses a scenario from the given file."""

    res = {}
    for line in open(fname, "r").readlines():
        split = line.split()
        if len(split) != 2:
            continue
        key, value = split
        if key[0] == "#":
            if key == "#mode":
                pass
            else:
                value = (value == "True")
        else:
            value = float(value)
        res[key] = value
    return res


def parse_mapping_file(fname):
    """ Mapping of variables converted to symbols. Does not include modes.
    Returns a list of Equals"""

    with open(fname, "r") as fin:
        res = Mapping()
        for line in fin.readlines():
            key, value = line.split()
            res.add(key, value)
        return res


def print_model(model):
    for (k1, v1) in model.items():
        print(k1, ":=", v1)
