from __future__ import print_function

# %HEADER

from builtins import range
import os
import sys
import subprocess
import tempfile
import time
from collections import OrderedDict

try:
    from xsap.utils.options import FT_ENGINE_BMC, FT_ENGINE_MSAT_BMC, \
        FT_ENGINE_PARAM, FT_ENGINE_BDD
except ImportError:
    from compass.options import FT_ENGINE_BMC, FT_ENGINE_MSAT_BMC, \
        FT_ENGINE_PARAM, FT_ENGINE_BDD

from xsap.features.common.tfpg_node_reachability import generate_pruned_assoc_struct
from xsap.features.common.tfpg_sys_extender import dump_extended_sys_as_smv
from xsap.features.tfpg_synthesis.tfpg_synthesizer import TfpgSynthesizer
from xsap.features.tfpg_synthesis.tle_list_generator import generate_tle_dic
from xsap.utils.misc import log
from xsap.utils.processes.xsap_process import XsapProcess


# INTERFACES

def perform_tfpg_synthesis(
        tfpg_assoc_struct,
        smv_model,
        out_tfpg_fname,
        bmc_length=10,
        engine=FT_ENGINE_BMC,
        disable_coi=False,
        xsap_dynamic_reorder=False,
        xsap_boolean_conversion=False,
        xsap_verbose_level=0,
        boolic3=False,
        outdir='.',
        print_debug_on_failure=False,
        xsap_process=None,
        enable_node_pruning=False):
    """Given an internal representation of associations, and an smv model,
    the tfpg synthesis is executed.

    @param tfpg_assoc_struct: the internal representation of associations
    @type  tfpg_assoc_struct: TfpgAssociations

    @param smv_model: the name of the smv model to be extended
    @type  smv_model: string

    @param bmc_length: bound for bmc analysis
    @type  bmc_length: int

    @param engine: the engine to be used for ft analysis
    @type  engine: string

    @param disable_coi: disables use of COI for TFPG simplification
    @type  disable_coi: boolean

    @param xsap_dynamic_reorder: enables bdd dynamic reordering
    @type  xsap_dynamic_reorder: boolean

    @param xsap_boolean_conversion: enables predicate normalization
                                    during boolean conversion
    @type  xsap_boolean_conversion: boolean

    @param xsap_verbose_level: sets the verbosity level for xsap
    @type  xsap_verbose_level: int

    @param boolic3: force boolean version of ic3
    @type  boolic3: boolean

    @param outdir: directory for all generated files
    @type  outdir: string

    @param xsap_process: the xsap process we may have created and
                         we want to use
    @type  xsap_process: utils.processes.xsap_process.xSAPProcess

    @param enable_node_pruning: if true, unreachable nodes will be ignored.
    @type  enable_node_pruning: Boolean
    """

    # identifying model properties

    need_msat = XsapProcess.has_infinite_domain_vars(smv_model)

    # 0. preprocessing

    if enable_node_pruning:

        print()
        print("pre-processing > pruning unreachable nodes")

        pruning_bound = None
        if engine not in (FT_ENGINE_BDD, FT_ENGINE_PARAM):
            # unbounded pruning only when IC3 or BDD is used; all other engines
            # are BMC engines (see options file).
            pruning_bound = bmc_length

        pruned_tfpg_assoc_struct = \
            generate_pruned_assoc_struct(smv_model, need_msat, tfpg_assoc_struct,
                                         pruning_bound, outdir=outdir)

        if len(pruned_tfpg_assoc_struct.get_var_names_of_failure_modes()) == 0:
            return (
                "error",
                "No failure mode is reachable%s." % ("" if pruning_bound is None else " (within bound)" ))

        unreach_nodes = [n for n in list(tfpg_assoc_struct.keys()) if n not in list(pruned_tfpg_assoc_struct.keys())]

        if len(unreach_nodes) > 0:
            print()
            print("  The following nodes are unreachable%s" % ('' if pruning_bound is None else ' within the analysis bound'), end=' ')
            print("and will be ignored during synthesis:")

            for node in unreach_nodes:
                print("  * %s" % node)

        else:
            print("  All nodes are reachable, no pruning performed.")

        tfpg_assoc_struct = pruned_tfpg_assoc_struct

    # 1. compute minimal cut-sets of discrepancies

    minimal_cutsets = OrderedDict()

    # extract model name
    model_fname = os.path.splitext(os.path.basename(smv_model))[0]

    # get tle list
    dic_tle = generate_tle_dic(tfpg_assoc_struct)

    # get list of failure modes
    fm_list = tfpg_assoc_struct.get_var_names_of_failure_modes()

    # time FTA step
    t0 = time.time()

    user_spec_proc = xsap_process is not None

    # generate extended SMV model

    extended_model_fname = model_fname + "_synth_extended" + '.smv'
    extended_model_fname = os.path.join(outdir, extended_model_fname)

    nodes_to_dump = fm_list + list(dic_tle.keys())
    log.xsap.user.debug("NODES_TO_DUMP: " + str(nodes_to_dump))

    dump_extended_sys_as_smv(
        tfpg_struct=None,
        associations=tfpg_assoc_struct,
        fn_smv=smv_model,
        output_fname=extended_model_fname,
        nodes_as_defines=True,
        filtered_nodes=nodes_to_dump)

    sys.stdout.write("\ntfpg synthesis > extended model dumped to '%s'\n" % extended_model_fname)
    sys.stdout.flush()

    #add reachable fm to minimal cut sets

    print()
    print("tfpg synthesis > Reachability Analysis of Failure Modes")
    res_fm, retval_fm = add_fm_to_minimal_cut_sets(
        need_msat, extended_model_fname, fm_list, minimal_cutsets, bmc_length, engine,
        xsap_dynamic_reorder, xsap_verbose_level, boolic3, outdir, xsap_process, enable_node_pruning)

    if res_fm == "error":
        return retval_fm

    minimal_cutsets = retval_fm

    # MCS computation with COI
    if not disable_coi:

        print()
        print("pre-processing > computing node dependencies")

        discr_deps = compute_node_dependencies(
            smv_model, need_msat, tfpg_assoc_struct, xsap_verbose_level, outdir,
            xsap_process)

        print()

        for i in range(len(list(dic_tle.keys()))):
            discrepancy = list(dic_tle.keys())[i]

            tle_list = [dic_tle[discrepancy]]
            filtered_nodes = list(discr_deps[discrepancy]) + [discrepancy]
            sa_fm_list_name = model_fname + "_" + \
                discrepancy + "_sa_fm_list.xml"

            sys.stdout.write("tfpg synthesis > minimal cut-set analysis for "
                   "discrepancy '%s' (%d/%d)\n" % (tle_list[0], i+1,
                                                 len(list(dic_tle.keys()))))
            sys.stdout.flush()

            res, retval = compute_minimal_cut_sets(
                need_msat, extended_model_fname, tle_list, fm_list,
                tfpg_assoc_struct, filtered_nodes,
                sa_fm_list_name, bmc_length, engine, xsap_dynamic_reorder,
                xsap_verbose_level, xsap_boolean_conversion, boolic3, outdir,
                print_debug_on_failure, xsap_process)

            log.xsap.user.debug("COMPUTE_MINIMAL_CUT_SETS " + str(retval))
            print()

            if res == "error":
                return retval

            assert discrepancy not in minimal_cutsets

            for n in retval:
                minimal_cutsets[n] = retval[n]

    # MCS computation without COI
    else:

        tle_list = [dic_tle[n] for n in dic_tle]
        filtered_nodes = fm_list + list(dic_tle.keys())
        sa_fm_list_name = model_fname + "_sa_fm_list.xml"

        print()
        print("tfpg synthesis > minimal cut-set analysis")
        res, retval = compute_minimal_cut_sets(
            need_msat, extended_model_fname, tle_list, fm_list,
            tfpg_assoc_struct, filtered_nodes, sa_fm_list_name, bmc_length,
            engine, xsap_dynamic_reorder, xsap_verbose_level,
            xsap_boolean_conversion, boolic3, outdir, print_debug_on_failure,
            xsap_process)
        log.xsap.user.debug("COMPUTE_MINIMAL_CUT_SETS " + str(retval))

        if res == "error":
            return retval

        for n in retval:
            minimal_cutsets[n] = retval[n]

    # all nodes must be reachable when pruning is used
    if enable_node_pruning:
        for n in minimal_cutsets:
            if len(minimal_cutsets[n]) == 0:
                # not even the empty set triggers the discrepancy, i.e. the
                # discrepancy is not reachable.
                raise Exception("  Error! Node '%s' has been identified as reachable " \
                      "during node pre-pruning, but was found not to be " \
                      "reachable during computation of minimal cut-sets.")

    log.xsap.tfpg_synth.debug("precedence constraints computed in" +
                              str(time.time()-t0) + "seconds")

    if user_spec_proc:
        xsap_process.kill_and_close_streams()

    # 2. create graph from minimal cut-sets

    t0 = time.time()

    discrepancies = get_discrepancies_for_synthesis(tfpg_assoc_struct)
    log.xsap.user.debug("DISCREPANCIES " + str(discrepancies))
    system_modes = tfpg_assoc_struct.get_var_names_of_tfpg_modes()
    log.xsap.user.debug("SYSTEM_MODES " + str(system_modes))
    tfpg_synthesizer = TfpgSynthesizer(
        system_modes, fm_list, discrepancies, minimal_cutsets)
    res = tfpg_synthesizer.synthesize_tfpg(out_tfpg_fname)

    log.xsap.tfpg_synth.debug("graph built in in" +
                              str(time.time()-t0) + "seconds")

    return res


# HELPERS

def dump_discrepancies_for_synthesis_to_xml(
        associations, filtered_nodes, output_fname):
    """ Dumps the discrepancies to an XML file. This is used in TFPG synthesis,
    specifically in the minimal cutset computation.

    associations:   the TFPG SLIM associations
    output fname:   file where to dump smv code
    """

    f = open(output_fname, 'w')

    f.write('<?xml version="1.0"?>\n')
    f.write('<compass>\n')
    f.write('<fmlist>\n')

    for k in associations.keys():
        (_, _, _type) = associations[k]
        if _type != "TfpgMode" and k in filtered_nodes:
            f.write('<fm name="tfpgsynth_%s" nominal_value="FALSE"/>\n' % k)

    f.write('</fmlist>\n')
    f.write('<obslist>\n')
    f.write('</obslist>\n')
    f.write('</compass>\n')

    f.close()

    return


def get_discrepancies_for_synthesis(tfpg_assoc_struct):
    """Returns the list of discrepancies in the format required by the
    tfpg synthesizer. Essentially, for each discrepancy, a
    tupe <discr_name, ["monitored"|"not_monitored"]>
    is created, depending on whether the discrepancy is monitored.

    @param tfpg_assoc_struct: the internal representation of the associations
    @type  tfpg_assoc_struct: TfpgAssociations
    """

    monitored_discr = tfpg_assoc_struct.get_var_names_of_monitored_discr()
    unmonitored_discr = tfpg_assoc_struct.get_var_names_of_unmonitored_discr()
    discrepancies = []

    for md in monitored_discr:
        discrepancies.append((md, "monitored"))
    for ud in unmonitored_discr:
        discrepancies.append((ud, "not_monitored"))

    return discrepancies


def compute_node_dependencies(
        smv_model,
        need_msat,
        tfpg_assoc_struct,
        xsap_verbose_level,
        outdir,
        xsap_process):

    # get discrepancies and node definitions for dependency computation

    ls_discrepancies = []
    dic_assoc_str = OrderedDict()

    user_spec_proc = xsap_process is not None

    for k in tfpg_assoc_struct.keys():
        (_, val, _type) = tfpg_assoc_struct[k]

        if _type == "TfpgMode":
            continue

        dic_assoc_str[k] = val

        if _type in ("MonitoredDiscr", "UnmonitoredDiscr"):
            ls_discrepancies.append(k)

    # instantiate process

    model_fname = os.path.splitext(os.path.basename(smv_model))[0]
    log_out, logfname = tempfile.mkstemp(suffix="_comp_coi.tfpgout",
                                         prefix=model_fname,
                                         dir=outdir, text=True)
    log.xsap.tfpg_synth.debug("Invoking xSAP, with redirect to " + logfname)

    if not user_spec_proc:
        xsap_process = XsapProcess(
            smv_fname=smv_model,
            need_msat=need_msat,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            dumpers=[os.fdopen(log_out, "w")],
            dynamic_reorder=False,
            verbose_level=xsap_verbose_level,
            boolean_conversion_uses_predicate_normalization=False)
        xsap_process.set_tmp_filename_prefix(outdir)
    else:
        xsap_process.reset()
        xsap_process.read_model(smv_model)
        # the log file can be closed as it is not needed
        os.close(log_out)

    xsap_process.set_tmp_filename_prefix(
        os.path.join(outdir, model_fname + "__temp_coi_tfpg__"))

    if need_msat:
        ok = xsap_process.send_init_commands(build_msat=True)
    else:
        ok = xsap_process.send_init_commands(build_bmc=True)

    if not ok:
        xsap_process.kill_and_close_streams()
        return None

    # compute dependencies of each discrepancy
    discr_deps = xsap_process.compute_discrepancy_dependencies(ls_discrepancies,
                                                            dic_assoc_str)

    if not user_spec_proc:
        xsap_process.kill_and_close_streams()

    return discr_deps

def compute_minimal_cut_sets(
        need_msat,
        extended_model_fname,
        tle_list,
        fm_list,
        tfpg_assoc_struct,
        filtered_nodes,
        sa_fm_list_name,
        bmc_length,
        engine,
        xsap_dynamic_reorder,
        xsap_verbose_level,
        xsap_boolean_conversion,
        boolic3,
        outdir,
        print_debug_on_failure,
        xsap_process):

    # create name for temporary file to store FTA results
    tmp_basename = os.path.splitext(os.path.basename(extended_model_fname))[0]
    tmp_out, tmp_filename = tempfile.mkstemp(suffix=".ftaout",
                                             prefix=tmp_basename,
                                             dir=outdir, text=True)
    os.close(tmp_out)

    # read log file for debugging (if something will fail)
    debug_lines = []
    # if print_debug_on_failure:
    #     with open(tmp_filename, 'r') as f:
    #         debug_lines = f.read().splitlines()

    # extend file names with output directory
    sa_fm_list_name = os.path.join(outdir, sa_fm_list_name)

    # create fm list
    dump_discrepancies_for_synthesis_to_xml(
        tfpg_assoc_struct, filtered_nodes, sa_fm_list_name)

    sys.stdout.write("  --> Failure modes list dumped to %s\n" % sa_fm_list_name)
    sys.stdout.flush()

    # instantiate process

    user_spec_proc = xsap_process is not None
    need_msat = need_msat or (engine in [FT_ENGINE_MSAT_BMC, FT_ENGINE_PARAM] and not boolic3)

    args = [
        "-sa_ft_layering",
        "-sa_compass",
        "-sa_compass_task",
        sa_fm_list_name] #args for xsap process

    if not user_spec_proc:
        xsap_process, _ = get_xsap_process(
            extended_model_fname,
            args,
            need_msat,
            xsap_dynamic_reorder,
            xsap_verbose_level,
            xsap_boolean_conversion,
            outdir)
    else:
        xsap_process.reset()
        xsap_process.send_cmd("set sa_ft_layering")
        xsap_process.send_cmd("set sa_compass")
        xsap_process.send_cmd(
            "set sa_compass_task_file '%s'" % sa_fm_list_name)
        xsap_process.read_model(model=extended_model_fname)

    minimal_cutsets = xsap_process.compute_minimal_cutsets(
        tle_list, tmp_filename, bmc_length, engine, xsap_boolean_conversion)

    if not user_spec_proc:
        xsap_process.kill_and_close_streams()

    if minimal_cutsets is None:
        # retrieve debug information
        if print_debug_on_failure:
            with open(tmp_filename, 'r') as f:
                debug_lines += f.read().splitlines()
            with open(extended_model_fname, 'r') as f:
                debug_lines += f.read().splitlines()
            print()
            print("DEBUGGING INFORMATION (START)")
            print()
            for line in debug_lines:
                print(line)
            print()
            print("DEBUGGING INFORMATION (END)")
            print()

        # something went wrong during minimal cutset computation
        return (
            "error",
            "error during minimal cutsets computation. "
            "See the log file for more information.")

    return ("ok", minimal_cutsets)


def get_xsap_process(
        smv_model,
        args,
        need_msat,
        xsap_dynamic_reorder,
        xsap_verbose_level,
        xsap_boolean_conversion,
        outdir):

    model_fname = os.path.splitext(os.path.basename(smv_model))[0]

    log_out, logfname = tempfile.mkstemp(suffix=".tfpgout",
                                         prefix=model_fname,
                                         dir=outdir, text=True)

    log.xsap.tfpg_synth.debug("Invoking xSAP, with redirect to " + logfname)

    if xsap_dynamic_reorder:
        args = args + ["-dynamic"]

    xsap_process = XsapProcess(
        smv_fname=smv_model,
        need_msat=need_msat,
        args=args,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        dumpers=[os.fdopen(log_out, "w")],
        dynamic_reorder=xsap_dynamic_reorder,
        verbose_level=xsap_verbose_level,
        boolean_conversion_uses_predicate_normalization=xsap_boolean_conversion)

    xsap_process.set_tmp_filename_prefix(os.path.join(outdir, model_fname + "__temp_tfpg__"))

    return (xsap_process, log_out)


def add_fm_to_minimal_cut_sets(
        need_msat,
        model_fname,
        fm_list,
        minimal_cutsets,
        bmc_length,
        engine,
        xsap_dynamic_reorder,
        xsap_verbose_level,
        boolic3,
        outdir,
        xsap_process,
        node_pruning):

    # instantiate process

    user_spec_proc = xsap_process is not None

    need_msat = need_msat or (engine in [FT_ENGINE_MSAT_BMC, FT_ENGINE_PARAM] and not boolic3)

    args = []

    if not user_spec_proc:
        xsap_process, _ = get_xsap_process(
            model_fname,
            args,
            need_msat,
            xsap_dynamic_reorder,
            xsap_verbose_level,
            False,
            outdir)
    else:
        xsap_process.reset()
        if xsap_dynamic_reorder:
            xsap_process.send_cmd("set dynamic_reorder")
        xsap_process.read_model(model=model_fname)

    try:
        minimal_cutsets = xsap_process.add_fm_to_minimal_cutsets(fm_list, minimal_cutsets,
                                                                 bmc_length, engine, node_pruning)
    except Exception as e:
        return (
            "error",
            "Error during reachability analysis of FM nodes. "
            "See the log file for more information.")

    if not user_spec_proc:
        xsap_process.kill_and_close_streams()

    if minimal_cutsets is None:
        # something went wrong during reachability analysis of FM
        return (
            "error",
            "error during reachability analysis of FM nodes. "
            "See the log file for more information.")

    return ('ok', minimal_cutsets)
