from __future__ import print_function
# %HEADER

from builtins import range
from builtins import object
import sys
import re
from collections import OrderedDict

from xsap.utils.misc import log

from mathsat import (
    msat_declare_function,
    msat_get_bool_type,
    MSAT_ERROR_DECL,
    msat_make_term,
    MSAT_ERROR_TERM,
    msat_make_and,
    msat_make_not,
    msat_term_id,
    msat_make_true,
    msat_make_false,
    msat_make_or,
    msat_create_env,
    msat_assert_formula,
    msat_solve_with_assumptions,
    MSAT_UNSAT,
    MSAT_SAT,
    msat_get_unsat_assumptions,
    msat_destroy_env)

# =========================== CLASS TfpgSynthesizer ======================


class TfpgSynthesizer (object):

    # ============================================== CONSTRUCTOR & HELPERS

    def __init__(
            self,
            arg_system_modes,
            arg_failure_modes,
            arg_discrepancies,
            arg_minimal_cutsets):

        # input parameters
        assert len(set(arg_failure_modes).intersection(
                set(arg_discrepancies))) == 0
        assert len(arg_failure_modes) > 0
        assert len(arg_discrepancies) > 0
        self.system_modes = arg_system_modes
        self.failure_modes = arg_failure_modes
        self.discrepancies = arg_discrepancies
        self.minimal_cutsets = arg_minimal_cutsets
        #self.generated_node_contexts = {}

        self.discrepancy_names = [d[0] for d in self.discrepancies]
        self.monitored_discrepancies = [d[0] for d in self.discrepancies if d[1] == "monitored"]

        # synthesis helpers
        self.gen_id = 0

        # data structure for instantiating graph
        self.causality_graph = OrderedDict()

    @staticmethod
    def is_generated_node_name(name):
        return re.match(r'intermediateNode_\d+', name) is not None

    @staticmethod
    def generate_random_id(reserved_names):

        trials = 0

        def randomword(length):
            import random, string
            return ''.join(random.choice(string.ascii_lowercase) for i in range(length))

        while trials < 1000:
            trials += 1
            random_id = "aux_" + randomword(6)
            if random_id not in reserved_names:
                return random_id

        raise Exception("Error: Failed to generate random TFPG node ID.")

    def generate_fresh_id(self):
        self.gen_id += 1
        return "intermediateNode_%d" % self.gen_id

    @staticmethod
    def appears_in_some_list(elem, lists):
        for l in lists:
            if elem in l:
                return True
        return False

    # MAIN SYNTHESIS ROUTINE

    def synthesize_tfpg(self, out_tfpg_fname, tfpg_name="", ignore_independent_discrepancies=False):
        """Perform the tfpg synthesis, dumping the result to
           the file out_tfpg_fname.

           Returns the string "ok" if the synthesis was performed properly,
           an error message otherwise.

           @param out_tfpg_fname: the name of the file where to dump
                                  the synthesized TFPG
           @type  out_tfpg_fname: string

           @param tfpg_name: the optional name of the synthesized TFPG
                             (if empty, the name will be "DependencyGraph")
           @type  tfpg_name: string

           @param ignore_independent_discrepancies: The check for independent discrepancies will
                                                    not be performed.
           @type ignore_independent_discrepancies: Boolean
        """

        print("tfpg synthesis > validating node names and cut-sets")

        check_nodenames = self.validate_node_names()
        if check_nodenames != "ok":
            print("The TFPG contains invalid names.")
            log.xsap.tfpg_synth.debug(
                "The TFPG contains invalid names. Returning..")
            log.xsap.tfpg_synth.debug(check_nodenames)
            return check_nodenames

        check_cutsets = self.validate_cutsets()
        if check_cutsets != "ok":
            print("Some errors in the cut-sets found.")
            log.xsap.tfpg_synth.debug(
                "Some errors in the cut-sets found. Returning..")
            log.xsap.tfpg_synth.debug(check_cutsets)
            return check_cutsets

        print("tfpg synthesis > processing unreachable discrepancies")
        self.process_unreachable_discrepancies()

        print("tfpg synthesis > processing failure mode nodes")
        self.process_failuremode_nodes()

        if ignore_independent_discrepancies:
            exist_independent_discrepancies = False
        else:
            print("tfpg synthesis > checking for independent discrepancies")
            exist_independent_discrepancies = \
                self.check_independent_discrepancies()

            if exist_independent_discrepancies:
                print ("  Only basic simplification will be performed due to "
                       "independent discrepancies.")

        print("tfpg synthesis > checking for correlated discrepancies")
        res_correlations = \
            self.exist_correlated_discrepancy_activations(self.minimal_cutsets)

        if res_correlations[0]:
            print ("  There are at least two discrepancies that have some "
                   "correlation (%s, %s)." % (res_correlations[1],
                                              res_correlations[2]))
            log.xsap.tfpg_synth.debug(
                "At least two correlated discrepancies "
                "found (%s, %s)." % (res_correlations[1],
                                     res_correlations[2]))
        else:
            log.xsap.tfpg_synth.debug("No correlated discrepancies found "
                                      "in the TFPG")

        print("tfpg synthesis > creating causality graph")
        log.xsap.tfpg_synth.debug("creating causality graph..")
        self.synthesize_causality_graph()
        log.xsap.user.debug("CAUSALITY GRAPH %s" % self.causality_graph)
        print("tfpg synthesis > calling graph simplification routines")

        log.xsap.tfpg_synth.debug("calling graph simplication routines")

        num_nodes, num_children = \
            self.simplify_get_num_nodes_children(self.causality_graph)
        log.xsap.tfpg_synth.debug(
                    "num. nodes (before simplification): %d" % num_nodes)
        log.xsap.tfpg_synth.debug(
                    "num. edges (before simplification): %d" % num_children)

        self.simplify_causality_graph(self.causality_graph,
                                      exist_independent_discrepancies)

        num_nodes, num_children = \
            self.simplify_get_num_nodes_children(self.causality_graph)
        log.xsap.tfpg_synth.debug(
                    "num. nodes (after simplification): %d" % num_nodes)
        log.xsap.tfpg_synth.debug(
                    "num. edges (after simplification): %d" % num_children)

        print("tfpg synthesis > writing result to file")
        f_out = open(out_tfpg_fname, 'w')
        # self.print_human_readable_causality_graph(f_out, tfpg_name)
        log.xsap.debug("FINAL CAUSALITY GRAPH: %s" % self.causality_graph)
        self.print_xml_causality_graph(f_out, tfpg_name)
        f_out.close()

        return "ok"

    # INPUT VALIDATION

    def validate_node_names(self):
        for fm in self.failure_modes:
            if self.is_generated_node_name(fm):
                return "illegal failure mode name (%s)" % fm
        for d in self.discrepancy_names:
            if self.is_generated_node_name(d):
                return "illegal discrepancy name (%s)" % d
        return "ok"

    def validate_cutsets(self):

        # check presence of minimal cutsets
        if self.minimal_cutsets is None:
            return "minimal cut-sets are not registered in synthesizer"
        if len(self.minimal_cutsets) == 0:
            return "list of minimal cut-sets is empty"

        # all keys of the minimal cutset dictionary must be either failure
        # modes or discrepancies
        for n in self.minimal_cutsets:
            if n not in self.discrepancy_names and n not in self.failure_modes:
                return "found minimal cut-sets of a node that is neither " \
                       "failure mode nor a discrepancy (%s)" % n

        # all discrepancies must have been processed by cut-set analysis
        for d in self.discrepancy_names:
            if d not in self.minimal_cutsets:
                return ("not all known discrepancies were processed by cutset "
                        "analysis (eg %s)" % d)

        # all failure modes must have been processed by cut-set analysis
        for fm in self.failure_modes:
            if fm not in self.minimal_cutsets:
                return "not all known failure modes were processed by " \
                       "cut-set analysis (eg %s)" % fm
            if len(self.minimal_cutsets[fm]) > 0 and \
                    len(self.minimal_cutsets[fm][0]) > 0:
                return "one failure mode has a non-empty minimal cut-set, " \
                       "this is an error (%s)" % fm

        # the list of minimal cut-sets must be initialized for all nodes
        for n in self.minimal_cutsets:
            if self.minimal_cutsets[n] is None:
                return "the minimal cut-sets of '%s' are not initialized" % n

        # every node that has the empty set as the single minimal cut-set
        # mustn't have any other minimal cut-set
        for d in self.minimal_cutsets:
            has_empty_mcs = False
            for mcs in self.minimal_cutsets[d]:
                if len(mcs) == 0:
                    has_empty_mcs = True
                    break
            if has_empty_mcs and len(self.minimal_cutsets[d]) != 1:
                return "node '%s' has empty set and at least one" \
                       "other cut-set as minimal cut-sets" % d

        return "ok"

    # ===================================== PRCESS UNREACHABLE/ISOLATED NODES

    def process_unreachable_discrepancies(self):
        """ Checks whether some discrepancies are found not to be
        reachable by FTA.
        If so, the user is warned and the discrepancy is inserted into the
        causality graph as an isolated OR node. """

        # for every discrepancy d
        for d in self.discrepancy_names:
            # check whether d has no cut-sets at all
            if len(self.minimal_cutsets[d]) == 0:
                print("  discrepancy '%s' is not reachable within " \
                       "analysis bound." % d)
                # self.insert_node_into_causality_graph(d, 'OR', [])
                log.xsap.tfpg_synth.debug(
                    "discrepancy '%s' is not reachable within analysis"
                    "bound. It will be inserted into the causality graph"
                    "as isolated OR node." % d)

    def exist_correlated_discrepancy_activations(self, minimal_cutsets):
        """ Checks whether some discrepancy in some context always fires
        simulateneously  with some other discrepancy.

        Algorithm: for each pair of nodes, checks whether they have a minimal
        cutset containing the respective other node and are otherwise equal.
        """

        for d in minimal_cutsets:

            # skip nodes with no minimal cutsets or just the empty one
            if len(minimal_cutsets[d]) == 0:
                continue
            if len(
                    minimal_cutsets[d]) == 1 and len(
                    minimal_cutsets[d][0]) == 0:
                continue

            for d_prime in minimal_cutsets:
                if d is d_prime:
                    continue

                # skip nodes with no minimal cutsets or just the empty one
                if len(minimal_cutsets[d_prime]) == 0:
                    continue
                if len(
                        minimal_cutsets[d_prime]) == 1 and len(
                        minimal_cutsets[d_prime][0]) == 0:
                    continue

                for mcs in minimal_cutsets[d]:
                    # skip cutsets not containing the other node
                    if d_prime not in mcs:
                        continue

                    # first cutset without the second discrepancy
                    tmp_mcs = [x for x in mcs if x != d_prime]

                    for mcs_prime in minimal_cutsets[d_prime]:
                        # skip cutsets not containing the other node
                        if d not in mcs_prime:
                            continue

                        # second cutset without the first discrepancy
                        tmp_mcs_prime = [x for x in mcs_prime if x != d]

                        if set(tmp_mcs) == set(tmp_mcs_prime):
                            return (True, d, d_prime)

        return (False, None, None)

    def check_independent_discrepancies(self):
        """ Checks whether discrepancies can occur independently of any FM
        node. If this is the case, a warning is printed.
        NB: These nodes are not inserted yet into the causality graph. """

        exist_independent_discrepancies = False

        # for every discrepancy d
        for d in self.discrepancy_names:

            # has no cut-sets: this is dealt with in
            # process_unreachable_discrepancies();
            if len(self.minimal_cutsets[d]) == 0:
                continue

            # has the empty set as the only cut-set
            if len(self.minimal_cutsets[d]) == 1 and \
                    len(self.minimal_cutsets[d][0]) == 0:
                print("  discrepancy '%s' can occur independently of any " \
                       "failure mode and any other discrepancy." % d)
                log.xsap.tfpg_synth.debug(
                    "discrepancy '%s' can occur independently of any "
                    "failure mode and any other discrepancy." % d)
                exist_independent_discrepancies = True

            else:
                for mcs in self.minimal_cutsets[d]:
                    assert len(mcs) > 0

                    has_fm_node = False
                    for node in mcs:
                        if node in self.failure_modes:
                            has_fm_node = True
                            break

                    if not has_fm_node:
                        print("  discrepancy '%s' can occur independently " \
                            "of any failure mode, but depends on other " \
                            "discrepancies." % d)
                        log.xsap.tfpg_synth.debug(
                            "discrepancy '%s' can occurr independently of "
                            "any failure mode, but depends on other "
                            "discrepancies." % d)
                        exist_independent_discrepancies = True
                        break

        return exist_independent_discrepancies

    def process_failuremode_nodes(self):
        """ Inserts all FM nodes into the causality graph. If an FM node is
        found to be isolated, an explanation is printed. """

        for fm in self.failure_modes:

            self.insert_node_into_causality_graph(fm, 'FM', [])
            log.xsap.tfpg_synth.debug(
                "failure mode '%s' added into the causality graph." %
                fm)

            # the FM node has no cut-sets, not even the empty one
            if len(self.minimal_cutsets[fm]) == 0:
                log.xsap.tfpg_synth.debug(
                    "failure mode '%s' is not reachable within analysis "
                    "bound." % fm)
                print ("  failure mode '%s' is not reachable within "
                       "analysis bound." % fm)

            else:
                # one minimal cut-set
                assert len(self.minimal_cutsets[fm]) == 1
                # the empty one
                assert len(self.minimal_cutsets[fm][0]) == 0

                fm_appears_in_some_mcs = False
                for d in self.discrepancy_names:
                    if self.appears_in_some_list(fm, self.minimal_cutsets[d]):
                        fm_appears_in_some_mcs = True
                        break

                if not fm_appears_in_some_mcs:
                    print("  failure mode '%s' has no effect on " \
                        "discrepancies within the analysis bound." % fm)
                    log.xsap.tfpg_synth.debug(
                        "failure mode '%s' has no effect on discrepancies "
                        "within the analysis bound." % fm)

    # =============================================== CAUSALITY GRAPH SYNTHESIS

    def synthesize_causality_graph(self):

        # for every discrepancy d
        for d in self.discrepancy_names:

            assert d not in self.causality_graph

            # instantiate every discrepancy as an OR node
            self.insert_node_into_causality_graph(d, 'OR', [])

            # unreachable nodes (i.e. no minimal cut-sets, not even
            # the empty one) don't need to be further processed
            if len(self.minimal_cutsets[d]) == 0:
                continue

            # for every mcs of d
            assert len(self.minimal_cutsets[d]) > 0
            for mcs in self.minimal_cutsets[d]:

                # don't instantiate AND node for spontaneous discrepancies
                if len(mcs) == 0:
                    assert len(self.minimal_cutsets[d]) == 1
                    break

                # instantiate a new virtual AND node v
                mcs_id = self.generate_fresh_id()
                self.insert_node_into_causality_graph(mcs_id, 'AND', [])

                # create edge (v,d)
                self.causality_graph[d]['children'].append(mcs_id)

                # for every v' in mcs
                for v_prime in mcs:

                    # create edge (v',v)
                    self.causality_graph[mcs_id]['children'].append(v_prime)

    def insert_node_into_causality_graph(self, node_name, node_type, children):
        assert node_name not in self.causality_graph
        assert children is not None
        self.causality_graph[node_name] = {'type': node_type,
                                           'children': children}

    def simplify_get_num_nodes_children(self, causality_graph):
        num_nodes = len(causality_graph)
        num_children = sum(
            [len(causality_graph[k]['children']) for k in causality_graph])
        return (num_nodes, num_children)

    def simplify_causality_graph(self, causality_graph,
                                 exist_independent_discrepancies):

        # 1. perform basic graph-based simplification routines

        num_nodes, num_children = \
            self.simplify_get_num_nodes_children(causality_graph)
        log.xsap.tfpg_synth.debug(
            "Size of the causality graph:\n  nodes: %d, "
            "children: %d" % (int(num_nodes), int(num_children)))

        while True:
            log.xsap.tfpg_synth.debug(
                "looking for transitive links to be simplified..")
            self.simplify_causality_graph_transitive_links(causality_graph)

            num_nodes_new, num_children_new = \
                self.simplify_get_num_nodes_children(causality_graph)
            log.xsap.tfpg_synth.debug(
                "Size of the causality graph:\n  nodes: %d, "
                "children: %d" % (int(num_nodes_new), int(num_children_new)))

            if num_nodes_new == num_nodes and num_children_new == num_children:
                log.xsap.tfpg_synth.debug("simplification methods converged")
                break

            num_nodes = num_nodes_new
            num_children = num_children_new

        # 2. perform SAT-based simplification

        if not exist_independent_discrepancies:
            log.xsap.tfpg_synth.debug("perform SAT-based simplification..")
            self.simplify_causality_graph_sat_based(causality_graph)
        else:
            log.xsap.tfpg_synth.debug(
                "skipping SAT-based simplification because of "
                "spontaneous discrepancies")

        # 3. perform simple clean-up simplifications

        num_nodes, num_children = \
            self.simplify_get_num_nodes_children(causality_graph)
        log.xsap.tfpg_synth.debug("Size of the causality graph:\n  nodes: %d, "
                                  "children: %d" % (int(num_nodes),
                                                    int(num_children)))

        while True:
            log.xsap.tfpg_synth.debug("removing duplicate AND nodes..")
            self.simplify_causality_graph_duplicate_and_nodes(causality_graph)

            log.xsap.tfpg_synth.debug(
                "looking for transitive links to be simplified..")
            self.simplify_causality_graph_transitive_links(causality_graph)

            log.xsap.tfpg_synth.debug(
                "looking for OR nodes to be converted into AND nodes..")
            self.simplify_causality_graph_or_to_and(causality_graph)

            log.xsap.tfpg_synth.debug(
                "looking for single generated nodes to be simplified..")
            self.simplify_causality_graph_single_generated_nodes(
                causality_graph)

            log.xsap.tfpg_synth.debug("duplicate edges..")
            self.simplify_causality_graph_duplicate_edges(causality_graph)

            log.xsap.tfpg_synth.debug(
                "looking for AND nodes to be converted to OR nodes..")
            self.simplify_causality_graph_and_to_or(causality_graph)

            num_nodes_new, num_children_new = \
                self.simplify_get_num_nodes_children(causality_graph)
            log.xsap.tfpg_synth.debug(
                "Size of the causality graph:\n  nodes: %d, "
                "children: %d" % (int(num_nodes_new), int(num_children_new)))

            if num_nodes_new == num_nodes and num_children_new == num_children:
                log.xsap.tfpg_synth.debug("simplification methods converged")
                break

            num_nodes = num_nodes_new
            num_children = num_children_new
        

    def simp_sat_get_param_name(self, node_source, node_target):
        return "param_edge_%s_to_%s" % (node_source, node_target)

    def simp_sat_get_node_names_from_param_name(self, param_name):
        m = re.match(r'^param_edge_(.+)_to_(.+)$', param_name)
        assert m is not None
        return (m.group(1), m.group(2))

    def simp_sat_get_formula(self, menv, causality_graph):

        phi = None
        term2assump = OrderedDict()
        assump2term = OrderedDict()

        symb_done = set()
        var2term = OrderedDict()

        def declare_var(v):
            assert v not in symb_done
            symb_done.add(v)
            decl = msat_declare_function(menv, v, msat_get_bool_type(menv))
            assert not MSAT_ERROR_DECL(decl)
            t_var = msat_make_term(menv, decl, [])
            assert not MSAT_ERROR_TERM(t_var)
            var2term[v] = t_var

        # all user-defined discrepancies are in graph and are OR discrepancies
        for v in self.discrepancies:
            assert v[0] in list(causality_graph.keys())
            assert causality_graph[v[0]]['type'] == 'OR'

            # ignore discrepancies with no children (unreachable discrepancies)
            if len(causality_graph[v[0]]['children']) == 0:
                # but check that they are not children themselves of some other
                # node
                for d in causality_graph:
                    assert v[0] not in causality_graph[d]['children']
                continue

            # all OR nodes have some children, which are all AND nodes
            for v_prime in causality_graph[v[0]]['children']:
                assert causality_graph[v_prime]['type'] == 'AND'

        # all discrepancies of graph that are not in discrepancy list are AND
        # or FM nodes

        discrepancy_names = [v[0] for v in self.discrepancies]

        for v in list(causality_graph.keys()):
            if v in discrepancy_names:
                continue
            assert causality_graph[v]['type'] in ('AND', 'FM')
            if causality_graph[v]['type'] == 'AND':
                assert len(causality_graph[v]['children']) > 0
                for v_prime in causality_graph[v]['children']:
                    assert causality_graph[v_prime]['type'] in ('OR', 'FM')

        # create node symbols
        for v in self.failure_modes:
            declare_var(v)
        for v in self.discrepancies:
            # ignore discrepancies with no children (unreachable discrepancies)
            if len(causality_graph[v[0]]['children']) == 0:
                continue
            declare_var(v[0])

        # create edge parameters
        for v in self.causality_graph:
            if causality_graph[v]['type'] != 'AND':
                continue

            for v_prime in causality_graph[v]['children']:
                param_name = self.simp_sat_get_param_name(v_prime, v)
                declare_var(param_name)
                term2assump[msat_term_id(var2term[param_name])] = param_name
                assump2term[param_name] = var2term[param_name]

        # build formula
        phi_g = self.simp_sat_build_phi(menv, causality_graph, False, var2term)
        phi_g_prime = self.simp_sat_build_phi(
            menv, causality_graph, True, var2term)
        phi = msat_make_and(menv, msat_make_not(menv, phi_g), phi_g_prime)

        return (phi, assump2term, term2assump)

    def simp_sat_build_phi(self, menv, causality_graph, with_params, var2term):

        res = msat_make_true(menv)

        for d in self.discrepancies:

            # ignore discrepancies with no children (unreachable discrepancies)
            if len(causality_graph[d[0]]['children']) == 0:
                continue

            t_d = var2term[d[0]]
            big_or = msat_make_false(menv)
            for v in causality_graph[d[0]]['children']:
                big_and = msat_make_true(menv)
                for v_prime in causality_graph[v]['children']:
                    t_v_prime = var2term[v_prime]
                    expr = None
                    if with_params:
                        param = self.simp_sat_get_param_name(v_prime, v)
                        t_param = var2term[param]
                        expr = msat_make_or(menv,
                                            msat_make_not(menv, t_param),
                                            t_v_prime)
                    else:
                        expr = t_v_prime
                    big_and = msat_make_and(menv, big_and, expr)
                big_or = msat_make_or(menv, big_or, big_and)

            res = msat_make_and(menv, res, msat_make_or(menv,
                                                        msat_make_not(menv,
                                                                      t_d),
                                                        big_or))

        return res

    def simp_sat_drop_edge(self, causality_graph, edge):

        (node_source, node_target) = \
            self.simp_sat_get_node_names_from_param_name(edge)

        assert node_target in causality_graph
        assert causality_graph[node_target]['type'] == 'AND'
        assert node_source in causality_graph[node_target]['children']

        causality_graph[node_target]['children'] = \
            [v for v in causality_graph[node_target]['children']
             if v != node_source]

    def simp_sat_drop_nonmandatory(self, cur_assump, mandatory_assump):

        assert set(mandatory_assump) <= set(cur_assump)

        if set(mandatory_assump) == set(cur_assump):
            return [x for x in mandatory_assump]

        res = [x for x in cur_assump]
        pick_assump = None
        for assump in cur_assump:
            if assump not in mandatory_assump:
                pick_assump = assump
                break
        assert pick_assump is not None
        res.remove(pick_assump)
        return res

    def simplify_causality_graph_sat_based(self, causality_graph):

        log.xsap.user.debug("SIMPLIFY CAUSALITY GRAPH SAT BASED")
        menv = msat_create_env({"model_generation": "true"})

        # generate and assert formula
        (phi, assump2term, term2assump) = \
            self.simp_sat_get_formula(menv, causality_graph)

        #log.xsap.user.debug("PHI "+ str(phi))
        #log.xsap.user.debug("ASSUMP2TERM "+ str(assump2term))
        #log.xsap.user.debug("TERM2ASSUMP "+ str(term2assump))

        assert not MSAT_ERROR_TERM(phi)
        msat_assert_formula(menv, phi)

        # set previous/current UNSAT core to all assumptions
        all_assump = list(assump2term.keys())
        cur_assump = all_assump
        prev_assump = all_assump
        mandatory_assump = []

        # flag to check SAT call is unsat at first execution of loop
        first_run = True
        # while some assumptions still need to be tested
        while set(mandatory_assump) != set(prev_assump):

            assert set(mandatory_assump) < set(prev_assump)
            assert set(mandatory_assump) <= set(cur_assump)

            if not first_run:
                assert set(cur_assump) < set(prev_assump)

            cur_assump_terms = []
            for a in cur_assump:
                cur_assump_terms.append(assump2term[a])
            log.xsap.user.debug("CUR_ASSUMP_TERMS %s" % str(cur_assump_terms))

            res = msat_solve_with_assumptions(menv, cur_assump_terms)

            if first_run:
                assert res == MSAT_UNSAT
                first_run = False

            if res == MSAT_SAT:

                # assert that exactly one assumption is missing in current
                # assumptions wrt prev unsat core
                assert set(cur_assump) < set(prev_assump)
                #assump_diff = prev_assump - cur_assump
                assump_diff = [x for x in prev_assump if x not in cur_assump]
                assert len(assump_diff) == 1

                # assert that it is not marked as mandatory yet
                new_mandatory_assump = assump_diff.pop()

                # mark it as mandatory
                assert new_mandatory_assump not in mandatory_assump
                mandatory_assump.append(new_mandatory_assump)

                # drop another assumption not marked as mandatory
                cur_assump = \
                    self.simp_sat_drop_nonmandatory(prev_assump,
                                                    mandatory_assump)

            else:
                # get UNSAT assumptions
                assert res == MSAT_UNSAT

                log.xsap.debug("FIRST EXECUTION ")
                core_tmp = list(msat_get_unsat_assumptions(menv))
                core = []
                for x in core_tmp:
                    if x not in core:
                        core.append(x)

                core_assump = [term2assump[msat_term_id(t)] for t in core]

                assert set(core_assump) <= set(cur_assump)

                # backup current assumptions
                prev_assump = cur_assump

                # set next candidate to check
                if set(core_assump) == set(cur_assump):
                    log.xsap.debug("SIMP SAT DROP NONMANDATORY")
                    cur_assump = \
                        self.simp_sat_drop_nonmandatory(prev_assump,
                                                        mandatory_assump)
                else:
                    cur_assump = core_assump

        # drop all edges from graph that are not in mandatory_assump
        #redundant_assump = all_assump - mandatory_assump
        redundant_assump = [x for x in all_assump if x not in mandatory_assump]
        
        for edge in redundant_assump:
            self.simp_sat_drop_edge(causality_graph, edge)

        msat_destroy_env(menv)

    # BASIC SIMPLIFICATION RULES

    def simplify_causality_graph_duplicate_and_nodes(self, causality_graph):
        # remove nodes that have identical incoming and outgoing edges
        # (works through side-effect on causality_graph)

        for d in causality_graph:
            if len(causality_graph[d]['children']) < 2:
                continue
            if causality_graph[d]['type'] != 'OR':
                continue

            for child_one in causality_graph[d]['children']:
                if not self.is_generated_node_name(child_one):
                    continue

                for child_two in causality_graph[d]['children']:
                    if not self.is_generated_node_name(child_two):
                        continue
                    if child_one is child_two:
                        continue

                    assert causality_graph[child_one][
                        'type'] == causality_graph[child_two]['type']
                    assert causality_graph[child_one]['type'] == 'AND'
                    assert len(causality_graph[child_one]['children']) > 0
                    assert len(causality_graph[child_two]['children']) > 0

                    # same children?
                    if set(causality_graph[child_one]['children']) != \
                       set(causality_graph[child_two]['children']):
                        continue

                    # drop child_two of them from children of d
                    causality_graph[d]['children'] = \
                        [v for v in causality_graph[d]['children']
                         if v is not child_two]

                    # drop child_two from causality graph
                    del causality_graph[child_two]

                    return

    def simplify_causality_graph_transitive_links(self, causality_graph):
        # remove transitive links where possible (works through side-effect
        # on causality_graph)

        for d in causality_graph:

            if causality_graph[d]['type'] != 'AND':
                continue
            assert len(causality_graph[d]['children']) > 0

            for child_one in causality_graph[d]['children']:
                for child_two in causality_graph[d]['children']:
                    if child_one is child_two:
                        continue

                    if len(causality_graph[child_two]['children']) == 0:
                        continue

                    if (causality_graph[child_two]['type'] == 'AND' or
                        len(causality_graph[child_two]['children']) == 1) and \
                            child_one in \
                            causality_graph[child_two]['children']:

                        # child_one is guaranteed to occur by child_two; it's
                        # safe to remove edge from child_one to d

                        log.xsap.tfpg_synth.debug(
                            "SIMP TRANS1 > remove edge '%s -> "
                            "%s'" % (child_one, d))

                        causality_graph[d]['children'] = \
                            [v for v in causality_graph[d]['children']
                             if v != child_one]

                        return

    def simplify_causality_graph_single_generated_nodes(self, causality_graph):
        # remove generated cutset AND nodes with a single input (works through
        # side-effect on causality_graph)

        keys = list(causality_graph.keys())
        while len(keys) > 0:
            d = keys[0]

            if self.is_generated_node_name(d):
                assert self.causality_graph[d]['type'] == 'AND'
                assert len(causality_graph[d]['children']) > 0
                if len(causality_graph[d]['children']) == 1:

                    # 'd' can be removed from the graph
                    log.xsap.tfpg_synth.debug(
                        "SIMP SINGLE > remove node '%s'" % (d))

                    # identify single child node
                    single_child = causality_graph[d]['children'][0]

                    # for all other nodes that have the generated node as a
                    # child
                    for d_prime in causality_graph:
                        if d_prime is d:
                            continue
                        assert causality_graph[d_prime]['children'] is not None

                        # replace the generated node with the single child node
                        # in their cutsets
                        if d in causality_graph[d_prime]['children']:
                            causality_graph[d_prime]['children'].remove(d)
                            causality_graph[d_prime][
                                'children'].append(single_child)

                    del causality_graph[d]

            keys = keys[1:]

    def simplify_causality_graph_or_to_and(self, causality_graph):
        # simplify an OR node with a single cutset by changing it to an AND
        # node, linking the incoming edges of the generated cutset AND node
        # to the new AND node, and deleting the generated cutset AND node.
        # (works through side-effect on causality_graph)

        # for each discrepancy:
        keys = list(causality_graph.keys())
        while len(keys) > 0:
            d = keys[0]

            # if not generated and has only one child which is a generated
            # node:
            if d in self.discrepancy_names and \
               len(causality_graph[d]['children']) == 1 and \
               self.is_generated_node_name(causality_graph[d]['children'][0]):

                log.xsap.tfpg_synth.debug(
                    "SIMP OR2AND > redeclare '%s' as AND node" % (d))

                # get generated node id
                child = causality_graph[d]['children'][0]

                # duplicate 'child' with name of 'd'
                causality_graph[d] = causality_graph[child]
                assert causality_graph[d]['type'] == 'AND'

                # remove generated node
                del causality_graph[child]

            keys = keys[1:]

    def simplify_causality_graph_and_to_or(self, causality_graph):
        # an AND discrepancy that has a single input is redeclared as an OR
        for d in causality_graph:
            if causality_graph[d]['type'] == 'AND' and len(
                    causality_graph[d]['children']) == 1:
                log.xsap.tfpg_synth.debug(
                    "SIMP AND2OR > redeclare '%s' as OR node" % (d))
                causality_graph[d]['type'] = 'OR'

    def simplify_causality_graph_duplicate_edges(self, causality_graph):
        for d in causality_graph:
            tmp = []
            for x in causality_graph[d]['children']:
                if x not in tmp: 
                    tmp.append(x)
            causality_graph[d]['children'] = tmp

    # OUTPUT

    def print_human_readable_causality_graph(self, f, tfpg_name):

        # tfpg_name: it is the arbitrary name of the tfpg: str

        assert self.causality_graph is not None

        # TFPG name

        if tfpg_name is None or tfpg_name == "":
            f.write("NAME DependencyGraph\n")
        else:
            f.write("NAME " + tfpg_name + "\n")

        # nodes

        for d in self.causality_graph:
            d_type = self.causality_graph[d]['type']
            f.write("%s %s\n" % (d_type, d))

        # modes

        all_modes = ','.join(self.system_modes)
        f.write("MODES %s\n" % all_modes)

        # edges

        edge_counter = 1

        for target_node in self.causality_graph:
            for source_node in self.causality_graph[target_node]['children']:
                e_name = "EDGE%d" % edge_counter
                edge_counter += 1
                f.write(
                    "EDGE %s %s %s 0.0 -1.0 (%s)\n" % (e_name, source_node,
                                                       target_node, all_modes))

    def print_xml_causality_graph(self, f, tfpg_name):

        # tfpg_name: it is the arbitrary name of the tfpg: str

        assert self.causality_graph is not None

        f.write("<?xml version='1.0' encoding='UTF-8'?>\n")

        if tfpg_name is None or tfpg_name == "":
            f.write('<tfpg name="DependencyGraph">\n')
        else:
            f.write('<tfpg name="' + tfpg_name + '">\n')

        f.write("  <nodesList>\n")

        for d in self.causality_graph:
            f.write(
                '    <node name="%s" isMonitored="%s">\n' %
                (d, "true" if d in self.monitored_discrepancies else "false"))
            f.write(
                "      <type>%s</type>\n" %
                self.causality_graph[d]['type'])
            f.write("    </node>\n")

        f.write("  </nodesList>\n")
        f.write("  <modesList>\n")
        for mode in self.system_modes:
            f.write("    <mode>%s</mode>\n" % mode)
        f.write("  </modesList>\n")

        f.write("  <edgesList>\n")

        edge_counter = 0
        for target_node in self.causality_graph:
            for source_node in self.causality_graph[target_node]['children']:
                edge_counter += 1

                f.write('    <edge name="EDGE%d">\n' % edge_counter)
                f.write("      <srcNode>%s</srcNode>\n" % source_node)
                f.write("      <tMin>0.0</tMin>\n")
                f.write("      <tMax>-1.0</tMax>\n")
                f.write("      <modesList>\n")
                for mode in self.system_modes:
                    f.write("      <mode>%s</mode>\n" % mode)
                f.write("      </modesList>\n")
                f.write("      <destNode>%s</destNode>\n" % target_node)
                f.write("    </edge>\n")

        f.write("  </edgesList>\n")
        f.write("</tfpg>\n")


# =============================== CLASS CutSet ===========================


class CutSet(object):

    def __init__(self, top_level_event, cutset_events):
        assert isinstance(cutset_events, list)
        self.top_level_event = top_level_event
        self.cutset_events = [v for v in cutset_events if v != top_level_event]

    def get_tle(self):
        return self.top_level_event

    def get_events(self):
        return self.cutset_events

    @staticmethod
    def parse_cutset_list(filename):
        """ Static method for parsing list of cutsets from an SA log file. """

        f = open(filename, 'r')

        cutset_list = None
        current_cutset = None

        top_level_event = None

        for line in f:

            match = re.search(
                r'-- minimal cut sets for specification (\S+)', line)
            if match:
                assert top_level_event is None
                top_level_event = match.group(1)
                continue

            # skip everything until the minimal cut sets
            if top_level_event is None:
                continue

            match = re.search(r'\[no cut set found\]', line)
            if match:
                assert top_level_event is not None
                continue

            match = re.search(r'\d+\) -------------------------', line)
            if match:
                assert top_level_event is not None
                # create obj for prev cutset
                if current_cutset is not None:
                    if cutset_list is None:
                        cutset_list = []
                    cutset_list.append(CutSet(top_level_event, current_cutset))
                # set-up new cutset
                current_cutset = []
                continue

            match = re.search(r'^\s*(_hist_var_\d+_)?(\S+)_TRUE = TRUE', line)
            if match:
                assert top_level_event is not None
                # add to current cutset
                current_cutset.append(match.group(2))
                continue

            match = re.search(r'\[cut set with no variables\]', line)
            if match:
                assert top_level_event is not None
                if cutset_list is None:
                    cutset_list = []
                for cs in cutset_list:
                    assert cs.top_level_event != top_level_event
                cutset_list.append(CutSet(top_level_event, []))
                continue

        f.close()

        # create obj for last cutset
        if current_cutset is not None:
            if cutset_list is None:
                cutset_list = []
            cutset_list.append(CutSet(top_level_event, current_cutset))

        return cutset_list


# MAIN

"""
create main function that takes as input a file with all the minimal cutsets,
the discrepancies, and the failuremodes, and creates an output file with the
resulting TFPG.
"""

if __name__ == "__main__":

    # ok
    tfpg_01 = ('01',
               {'F1': [],
                'F2': [],
                'A': [['F1']],
                'B': [['F1', 'A', 'F2']]})

    # ok
    tfpg_02 = ('02',
               {'F1': [],
                'F2': [],
                'A': [['F1'], ['F2']]})

    # ok
    tfpg_03 = ('03',
               {'F1': [],
                'F2': [],
                'A': [['F1', 'F2']]})

    # ok
    tfpg_04 = ('04',
               {'F1': [],
                'A': [['F1']],
                'B': [['A', 'F1']]})

    # ??
    tfpg_05 = ('05',
               {'F1': [],
                'F2': [],
                'A': [['F1'], ['F2', 'B', 'C']],
                'B': [['F2'], ['F1', 'A']],
                'C': [['F2', 'B'], ['F1', 'A', 'B']]})

    # ??
    tfpg_06 = ('06',
               {'F1': [],
                'F2': [],
                'A': [['F1'], ['F2', 'B']],
                'B': [['F2'], ['F1', 'A']]})

    # circular dependency (output warning)
    tfpg_07 = ('07',
               {'F1': [],
                'alpha': [['F1']],
                'beta': [['F1']],
                'delta': [['F1']],
                'A': [['alpha', 'B'], ['delta', 'B']],
                'B': [['beta'], ['alpha', 'A'], ['delta', 'A']]})

    ls_all_tfpg = [tfpg_01, tfpg_02, tfpg_03, tfpg_04, tfpg_05, tfpg_06]

    my_tfpg = tfpg_05

    system_modes = ['std']
    failure_modes = [x for x in my_tfpg[1] if x[0] == 'F']
    discrepancies = [x for x in my_tfpg[1] if x[0] != 'F']
    minimal_cutsets = my_tfpg[1]

    tfpg_synthesizer = TfpgSynthesizer(
        system_modes,
        failure_modes,
        discrepancies,
        minimal_cutsets)

    print("validating node names and cutsets ...", end=' ')
    res_nodenames = tfpg_synthesizer.validate_node_names()
    res_cutsets = tfpg_synthesizer.validate_cutsets()

    if res_nodenames == "ok" and res_cutsets == "ok":
        print("calling synthesis ...", end=' ')
        tfpg_synthesizer.synthesize_causality_graph()
        print("done")

        print("calling simplification ...", end=' ')
        tfpg_synthesizer.simplify_causality_graph(
            tfpg_synthesizer.causality_graph, False)
        # tfpg_synthesizer.simplify_causality_graph(tfpg_synthesizer.causality_graph,
        # False)
        print("done")

        print()
        tfpg_synthesizer.print_human_readable_causality_graph(sys.stdout, "")

    else:
        print()
        print("res_nodenames:", res_nodenames)
        print("res_cutsets:", res_cutsets)
