
import os
import re

from xsap.data_structure.tfpg.tfpg import Tfpg
from xsap.data_structure.tfpg.tfpg_associations import TfpgAssociations
from xsap.data_structure.tfpg.txt.tfpg import dump_tfpg_to_txt_file
from xsap.features.common.ProofObligationMTL import ProofObligationMTL
from xsap.features.common.tfpg_symb_converter import compute_smv_em_params_name
from xsap.features.common.tfpg_symb_converter import compute_smv_etmax_bool_params_name
from xsap.features.common.tfpg_symb_converter import compute_smv_etmax_real_params_name
from xsap.features.common.tfpg_symb_converter import compute_smv_etmin_params_name
from xsap.features.tfpg_tightening.tightening_manual_time import tighten_tfpg_manualtime
from xsap.features.tfpg_tightening.tightening_native_time import tighten_tfpg_nativetime


def perform_edge_tightening(fn_smv, fn_assoc, fn_tfpg, options):
    """ This is the top-level tightening function, that takes the user-defined inputs,
    calls the tightening function, and dumps the new TFPG.

    @param fn_smv: original SMV model (in native or manual time)
    @type  fn_smv: str

    @param fn_assoc: association file
    @type  fn_assoc: str

    @param fn_tfpg: the TFPG to be tightened (in XML or TXT format)
    @type  fn_tfpg: str

    @param options: various tightening options
    @type  options: dict

    @return: A tuple is returned with two elements: if everything went well, the first element is
             the file name pointing to the tight TFPG, and the second element is None; if an error
             occurred, the first element will be None and the second one will be an error message
             describing the problem.
    @rtype:  tuple
    """

    # parse assoc

    assoc_struct = TfpgAssociations.parse_assoc(fn_assoc)
    if assoc_struct is None:
        return None, "Given association file contains errors."

    # parse tfpg

    tfpg_struct = Tfpg.parse_tfpg(fn_tfpg)
    if tfpg_struct is None:
        return None, "Given TFPG contains errors."

    # create proof obligation generator
    po_mtl = ProofObligationMTL(tfpg_struct, assoc_struct, parametric=True, tags=options["tags"])

    # compute parameters
    tmax_bound = options["tmax_bound"] if "tmax_bound" in options else None
    dic_params, err_msg = generate_parameters_dictionary(tfpg_struct, tmax_bound)

    if err_msg is not None:
        return None, err_msg

    # call tightener (with local tightening, call it one time for each local TFPG)
    if is_native_time(fn_smv):
        res_params, err_msg = \
            tighten_tfpg_nativetime(fn_smv, po_mtl, options, dic_params, tfpg_struct)
    else:
        res_params, err_msg = \
            tighten_tfpg_manualtime(fn_smv, po_mtl, options, dic_params, tfpg_struct)

    if err_msg is not None:
        return None, err_msg

    # dump solution
    fn_tfpg_tight, err_msg = \
        dump_solution(tfpg_struct, res_params, fn_tfpg, options["outdir"], options["tags"])

    if err_msg is not None:
        return None, err_msg

    return fn_tfpg_tight, None


def is_native_time(fn_smv):
    """ Determines whether the given SMV file is in nuXmv native time or not. """

    with open(fn_smv, 'r') as f_in:
        for line in f_in:

            # drop comments (if any)
            line = line[:line.find('--')]

            # see if time domain tag is present
            m = re.search(r'@TIME_DOMAIN (none|discrete|continuous)', line)
            if m is not None and m.group(1) in ("discrete", "continuous"):
                return True

    return False


def generate_parameters_dictionary(tfpg_struct, tmax_bound):
    """ Generates a dictionary with all edge parameters of the given TFPG that need
    to be considered for tightening.

    @param tfpg_struct: the TFPG to consider
    @type  tfpg_struct: Tfpg

    @param tmax_bound: the tmax bound (if any) to validate it against tmin/tmax values in TFPG
    @type  tmax_bound: int or None

    @return tuple with the first element being a dictionary with parameters and respective
            information, and the second one an error message (if some error occurred)
    @rtype  tuple
    """

    def gen_param(ptype, monotonicity, tag, low_bound, up_bound):
        return {"type": ptype,
                "monotonicity": monotonicity,
                "tag": tag,
                "lower_bound": low_bound,
                "upper_bound": up_bound}

    res_dict = {}

    for e in tfpg_struct.get_edges_list():

        if tmax_bound is not None:
            if e.get_tMin() > tmax_bound or e.get_tMax() > tmax_bound:
                return None, "Edge %s has incompatible time bounds with chosen tmax bound." % \
                       e.get_name()

        if len(e.get_modes()) > 1:
            for m in e.get_modes():
                res_dict[compute_smv_em_params_name(e, m)] = \
                    gen_param("boolean", "upward", "modes", "false", "true")

        val_tmax = e.get_tMax()
        upper_real_bound = "none" if val_tmax == -1 else val_tmax
        upper_bool_bound = "true" if val_tmax == -1 else "false"

        res_dict[compute_smv_etmin_params_name(e)] = \
            gen_param("real", "downward", "tmin", e.get_tMin(), upper_real_bound)
        res_dict[compute_smv_etmax_real_params_name(e)] = \
            gen_param("real", "upward", "tmax", e.get_tMin(), upper_real_bound)
        res_dict[compute_smv_etmax_bool_params_name(e)] = \
            gen_param("boolean", "upward", "tmax", "false", upper_bool_bound)

    return res_dict, None


def dump_solution(tfpg_struct, param_values, fn_original_tfpg, outdir, tags):
    """ Updates the TFPG edge parameters with the given new values and dumps the result to a new
    TFPG file in the outdir directory.

    @param tfpg_struct: original TFPG structure
    @type  tfpg_struct: Tfpg

    @param param_values: dictionary where the keys are edge parameters (see dump_parameters()),
                         and the values the new values.
    @type  param_values: dict

    @param fn_original_tfpg: the filename of the original TFPG; used to compute the filename and
                             output format.
    @type  fn_original_tfpg: str

    @param outdir: the directory where the result should be dumped to
    @type  outdir: str

    @param tags: tags on which the TFPG was tightened
    @type  tags: tuple
    """

    # update TFPG name
    tfpg_struct.set_name(tfpg_struct.get_name() + "_tightened")

    # update edge constraints
    for e in tfpg_struct.get_edges_list():

        if "modes" in tags and len(e.get_modes()) > 1:

            new_modes = []
            for m in e.get_modes():
                # print "res mode", m
                p_name_em = compute_smv_em_params_name(e, m)
                assert p_name_em in param_values
                if param_values[p_name_em].lower() == "true":
                    new_modes.append(m)
            e.set_modes(new_modes)

        if "tmin" in tags:

            p_name_tmin = compute_smv_etmin_params_name(e)
            assert p_name_tmin in param_values

            # improved tmin needs to be the same or larger
            assert e.get_tMin() <= param_values[p_name_tmin]
            # but still smaller than the old tmax (if finite)
            if e.get_tMax() >= 0:
                assert param_values[p_name_tmin] <= e.get_tMax()

            e.set_tMin(param_values[p_name_tmin])

        if "tmax" in tags:

            p_name_tmax_bool = compute_smv_etmax_bool_params_name(e)
            assert p_name_tmax_bool in param_values

            p_name_tmax_real = compute_smv_etmax_real_params_name(e)
            assert p_name_tmax_real in param_values

            if param_values[p_name_tmax_bool].lower() == "true":
                # if the improved value is infinity, it must be infinity also before optimization,
                # and the tmax value doesn't need to be updated.
                assert e.get_tMax() == -1
            else:
                assert param_values[p_name_tmax_bool].lower() == "false"
                # new tmax must be larger than the old or updated tmin
                assert e.get_tMin() <= param_values[p_name_tmax_real]
                e.set_tMax(param_values[p_name_tmax_real])

    # dump file

    tfpg_basename = os.path.basename(fn_original_tfpg)
    tfpg_file_base, tfpg_file_ext = os.path.splitext(tfpg_basename)

    fn_tfpg_tight = tfpg_file_base + "_tight" + tfpg_file_ext
    fn_tfpg_tight = os.path.join(outdir, fn_tfpg_tight)

    if tfpg_file_ext in [".txml", ".xml"]:
        tfpg_struct.dump_to_xml(fn_tfpg_tight)
    elif tfpg_file_ext == ".tfpg":
        dump_tfpg_to_txt_file(tfpg_struct, fn_tfpg_tight)
    else:
        return None, "unknown TFPG format (%s)" % tfpg_file_ext

    return fn_tfpg_tight, None
