
from builtins import object
from mathsat import *

from xsap.utils.misc import log

import sys, time


class Msatic3Checker(object):

    def __init__(self, menv, vmt_string, prover_mode, bmc_len, ic3_len, ic3_mode):

        assert prover_mode in ('standard', 'reuse-invariant', 'reuse-prover')
        self.prover_mode = prover_mode

        assert ic3_mode in ('conc', 'ia')
        self.ic3_mode = ic3_mode

        self.bmc_len = bmc_len
        self.ic3_len = ic3_len

        self.msat_env = menv
        self.vmt_string = vmt_string
        self.prop = None

        self.inductive_invariant = msat_make_true(self.msat_env)

        # extract property

        self.mic3 = msat_create_ic3(self.msat_env)
        msat_ic3_from_smtlib2(self.mic3, self.vmt_string)

        if self.ic3_len is not None:
            assert 0 == msat_ic3_setopt(self.mic3, "ic3core.max_frames", str(self.ic3_len))

        if self.ic3_mode == 'conc':
            msat_ic3_init(self.mic3, MSAT_IC3_CONCRETE)
        else:
            assert 0 == msat_ic3_setopt(self.mic3, "ic3ia.abstract_bool_vars", "false")
            assert 0 == msat_ic3_setopt(self.mic3, "ic3ia.add_initial_reset", "false")
            assert 0 == msat_ic3_setopt(self.mic3, "ic3ia.make_simple_property", "false")
            msat_ic3_init(self.mic3, MSAT_IC3_IA)

        assert msat_ic3_num_properties(self.mic3) == 1
        self.prop = msat_ic3_get_property(self.mic3, 0)
        assert not MSAT_ERROR_TERM(self.prop)
        assert MSAT_IC3_PROP_INVAR == msat_ic3_get_property_kind(self.mic3, 0)

        self.fixed_constraints = []

        if self.prover_mode != 'reuse-prover':
            msat_destroy_ic3(self.mic3)

    ############################################################################

    def add_fixed_constraint(self, term):
        self.fixed_constraints.append(term)

    ############################################################################

    def get_property(self):
        return self.prop

    ############################################################################

    def get_inductive_invariant(self):
        return self.inductive_invariant

    ############################################################################

    def do_check_with_bmc(self, candidate, proof_obl):

        prop = self.compute_candidate_valuation(candidate, [])
        prop = msat_make_or(self.msat_env,
                            msat_make_not(self.msat_env, prop),
                            proof_obl)

        mic3 = msat_create_ic3(self.msat_env)

        msat_ic3_from_smtlib2(mic3, self.vmt_string)

        # add property to solver
        idx = msat_ic3_add_invar_property(mic3, prop)
        assert idx != -1

        # set msatic3 options
        assert 0 == msat_ic3_setopt(mic3, "bmc.max_k", str(self.bmc_len))

        # initialize prover
        assert 0 == msat_ic3_init(mic3, MSAT_IC3_BMC)

        # add inductive invariant to initial states and transition relation
        if self.prover_mode == 'reuse-invariant':
            res = msat_ic3_add_init_constraint(mic3, self.inductive_invariant)
            assert res == 0
            res = msat_ic3_add_trans_constraint(mic3, self.inductive_invariant)
            assert res == 0

        # add additional constraints to prover
        for term in self.fixed_constraints:
            res = msat_ic3_add_init_constraint(mic3, term)
            assert res == 0
            res = msat_ic3_add_trans_constraint(mic3, term)
            assert res == 0

        # main call to prover

        t0 = time.time()
        res = msat_ic3_prove(mic3, idx)
        t1 = time.time()
        msat_destroy_ic3(mic3)
        log.xsap.tfpg_tight.debug("  msat_ic3_prove (bmc) duration: %f" % (t1-t0))

        # done
        if res == 0:
            return 'verified'
        elif res == 1:
            return 'falsified'
        else:
            assert res is MSAT_IC3_ERR_MAX_BOUND
            return 'unknown'

    ############################################################################

    def do_check_with_invariant(self, candidate, proof_obl):

        checkenv = msat_create_env({}, self.msat_env)

        # assert candidate valuation as hard constraint
        candidate_term = self.compute_candidate_valuation(candidate, [])
        msat_assert_formula(checkenv, candidate_term)

        # assert negated property (ie the bad states)
        neg_prop = msat_make_not(self.msat_env, proof_obl)
        msat_assert_formula(checkenv, neg_prop)

        # assert inductive invariant
        msat_assert_formula(checkenv, self.inductive_invariant)

        # solve
        res = msat_solve(checkenv)

        # unsat -> prop holds
        prop_holds = res == MSAT_UNSAT

        msat_destroy_env(checkenv)
        return prop_holds

    ############################################################################

    def do_check(self, candidate, proof_obl, do_bmc_step):

        # check with inductive invariant
        prop_holds = self.do_check_with_invariant(candidate, proof_obl)
        if prop_holds:
            log.xsap.tfpg_tight.debug("proved by inductive invariant")
            return True

        # first try with BMC

        if do_bmc_step:
            log.xsap.tfpg_tight.debug("calling bmc")

            res = self.do_check_with_bmc(candidate, proof_obl)

            if res == 'verified':
                return True
            elif res == 'falsified':
                return False
            else:
                # unknown, try now with IC3
                pass

        # then with IC3

        log.xsap.tfpg_tight.debug("computing valuation")
        prop = self.compute_candidate_valuation(candidate, [])
        prop = msat_make_or(self.msat_env,
                            msat_make_not(self.msat_env, prop),
                            proof_obl)

        log.xsap.tfpg_tight.debug("calling msatic3")

        if self.prover_mode == 'reuse-prover':
            return self.call_msatic3(self.mic3, prop, False, False)

        else:
            mic3 = msat_create_ic3(self.msat_env)
            msat_ic3_from_smtlib2(mic3, self.vmt_string)
            prop_holds = self.call_msatic3(mic3, prop, True,
                self.prover_mode == 'reuse-invariant')

            msat_destroy_ic3(mic3)
            return prop_holds

    ############################################################################

    def call_msatic3(self, mic3, prop, call_init, add_invar_to_trans):

        # add property to solver
        idx = msat_ic3_add_invar_property(mic3, prop)
        assert idx != -1

        # set msatic3 options and initialize prover

        if call_init:
            if self.ic3_mode == 'conc':
                if self.ic3_len is not None:
                    assert 0 == msat_ic3_setopt(mic3, "ic3core.max_frames", str(self.ic3_len))
                log.xsap.tfpg_tight.debug("calling msat_ic3_init (MSAT_IC3_CONCRETE)")
                msat_ic3_init(mic3, MSAT_IC3_CONCRETE)
            else:
                assert 0 == msat_ic3_setopt(mic3, "ic3ia.abstract_bool_vars", "false")
                assert 0 == msat_ic3_setopt(mic3, "ic3ia.add_initial_reset", "false")
                assert 0 == msat_ic3_setopt(mic3, "ic3ia.make_simple_property", "false")
                log.xsap.tfpg_tight.debug("calling msat_ic3_init (MSAT_IC3_IA)")
                msat_ic3_init(mic3, MSAT_IC3_IA)

        # add inductive invariant to initial states and transition relation
        if add_invar_to_trans:

            # TODO: performance seems better when not setting init (just trans)
            #       more benchmarks needed on this point.
            #res = msat_ic3_add_init_constraint(mic3, self.inductive_invariant)
            #assert res == 0

            res = msat_ic3_add_trans_constraint(mic3, self.inductive_invariant)
            assert res == 0

        # add additional constraints to prover

        for term in self.fixed_constraints:
            assert 0 == msat_ic3_add_init_constraint(mic3, term)
            assert 0 == msat_ic3_add_trans_constraint(mic3, term)

        # main call to prover

        log.xsap.tfpg_tight.debug("calling msat_ic3_prove()")
        t0 = time.time()
        res = msat_ic3_prove(mic3, idx)
        t1 = time.time()

        log.xsap.tfpg_tight.debug("msat_ic3_prove duration: %f" % (t1-t0))
        log.xsap.tfpg_tight.debug("result: %d" % res)

        # msat_ic3_prove returns 0 if the property holds, 1 if it is falsifed,
        # and an error code otherwise.

        # error handling
        if res not in (0, 1):
            return None

        prop_holds = res == 0

        # strengthen current inductive invariant if property holds
        if prop_holds:
            witness = msat_ic3_witness(mic3)
            assert witness is not None

            inv = self.compute_inductive_invariant(witness)

            # combine current and new invariant
            self.inductive_invariant = msat_make_and(self.msat_env,
                self.inductive_invariant, inv)

        # done
        return prop_holds

    ############################################################################

    def compute_inductive_invariant(self, witness):

        inv = msat_make_true(self.msat_env)
        clause = None

        for el in witness:
            # clauses within the same section, with sections being separated by
            # error terms, need to be put in disjunction. on the top level,
            # those disjunctions then need to be put in conjunction.

            if MSAT_ERROR_TERM(el):
                sys.stdout.flush()
                if clause is not None:
                    inv = msat_make_and(self.msat_env, inv, clause)
                    clause = None

            else:
                if clause is None:
                    clause = el
                else:
                    clause = msat_make_or(self.msat_env, clause, el)

        return inv

    ############################################################################

    def compute_candidate_valuation(self, candidate, params_to_be_skipped):

        prop = msat_make_true(self.msat_env)

        # compute the term representing the candidate valuation
        # of the parameters

        for p_name in candidate:
            if p_name in params_to_be_skipped:
                continue

            p_name_term = msat_from_string(self.msat_env, p_name)

            if MSAT_ERROR_TERM(p_name_term):
                log.xsap.tfpg_tight.debug("MSAT_ERROR_TERM(p_name_term) for p_name='%s'" % str(p_name))
            assert not MSAT_ERROR_TERM(p_name_term)

            p_value = candidate[p_name]

            # p_value is None when the parameter is fully unconstrained
            if p_value is None: continue

            p_value_term = None
            if isinstance(p_value, float):
                p_value_term = msat_make_number(self.msat_env, str(p_value))
            else:
                p_value_term = msat_from_string(self.msat_env, p_value)

            if MSAT_ERROR_TERM(p_value_term):
                log.xsap.tfpg_tight.debug("MSAT_ERROR_TERM(p_value_term) for p_value='%s'" % str(p_value))
            assert not MSAT_ERROR_TERM(p_value_term)

            p_term = None

            if msat_term_is_boolean_constant(self.msat_env, p_name_term):
                assert msat_term_is_true(self.msat_env, p_value_term) or \
                       msat_term_is_false(self.msat_env, p_value_term)
                p_term = msat_make_iff(self.msat_env, p_name_term, p_value_term)
            else:
                p_term = msat_make_equal(self.msat_env, p_name_term, p_value_term)

            prop = msat_make_and(self.msat_env, prop, p_term)

        return prop

