from __future__ import print_function

from builtins import object
import sys
import re

from mathsat import (
    msat_create_env,
    msat_destroy_env,
    msat_solve,
    msat_solve_with_assumptions,
    msat_get_unsat_assumptions,
    msat_push_backtrack_point,
    msat_pop_backtrack_point,
    msat_term_is_boolean_constant,
    msat_assert_formula,
    msat_make_and,
    msat_make_or,
    msat_make_true,
    msat_make_false,
    msat_make_not,
    msat_from_string,
    msat_make_iff,
    msat_make_equal,
    msat_make_number,
    msat_make_leq,
    MSAT_SAT,
    MSAT_UNKNOWN,
    MSAT_UNSAT,
    MSAT_ERROR_TERM
)

from xsap.features.tfpg_tightening.tightening_engine_bisection.ParameterManager \
    import ParameterManager
from xsap.features.tfpg_tightening.tightening_engine_bisection.Msatic3Checker \
    import Msatic3Checker
from xsap.features.tfpg_tightening.tightening_engine_bisection.SearchBoundManager \
    import SearchBoundManager
import xsap.features.tfpg_tightening.tightening_engine_bisection.utils as utils
from xsap.utils.misc import log


class ParamSynthesizer(object):

    def __init__(self, fn_vmt, dic_params, prover_mode, bmc_len, ic3_len, ic3_mode, skip_bmc,
                 tmax_bound, step_size):

        # parse and set model
        with open(fn_vmt, "r") as file_model:
            self.vmt_string = file_model.read()

        # create and initialize the global mathsat environment
        self.msat_env = msat_create_env()
        utils.init_msat_env(self.msat_env, self.vmt_string)

        # set up parameter manager
        self.params_mgr = ParameterManager(dic_params, tmax_bound, step_size)

        # set up model checker
        self.mchecker = Msatic3Checker(self.msat_env, self.vmt_string, prover_mode, bmc_len,
                                       ic3_len, ic3_mode)
        self.skip_bmc = skip_bmc

    ###############################################################################################
    # main optimization function
    ###############################################################################################

    def tfpg_find_one_optimal(self, tags, untimed, silent=False):

        for t in tags:
            assert t in ['tmin', 'tmax', 'modes']

        candidate = self.params_mgr.get_lattice_top()
        prop = self.mchecker.get_property()

        # setting invariant parameters as fixed constraints

        modes_params_shortlist = self.params_mgr.get_filtered("modes")
        tmin_params_shortlist = self.params_mgr.get_filtered("tmin")
        tmax_params_shortlist = self.params_mgr.get_filtered("tmax")

        # some sanity checks

        for p_name in modes_params_shortlist:
            assert p_name in candidate
        for p_name in tmin_params_shortlist:
            assert p_name in candidate
        for p_name in tmax_params_shortlist:
            if 'bool' in p_name:
                assert p_name in candidate
            else:
                assert 'real' in p_name
                if p_name not in candidate:
                    p_obj = self.params_mgr.get_param(p_name)

                    # top val must be none
                    assert p_obj.get_top_val() is None

                    # corresponding bool must be true and have singular domain
                    m = re.match(r'param_tmax_real_(.+)__AT0', p_name)
                    assert m is not None

                    p_name_bool = "param_tmax_bool_%s__AT0" % m.group(1)
                    assert p_name_bool in candidate

                    p_obj_bool = self.params_mgr.get_param(p_name_bool)
                    assert p_obj_bool.is_boolean()
                    assert p_obj_bool.has_singular_range()
                    assert p_obj_bool.get_top_val() == "true"

        if 'modes' not in tags:
            log.xsap.tfpg_tight.debug("setting all mode parameters to true")
            for p in modes_params_shortlist:
                candidate[p] = "true"
                term = self.get_assignment_term(p, "true")
                self.mchecker.add_fixed_constraint(term)

        # modify candidate in case of untimed tightening

        if untimed:
            for p_name in tmin_params_shortlist:
                # set tmin to 0
                candidate[p_name] = 0.0
            for p_name in tmax_params_shortlist:
                # set tmax to infinity
                if 'bool' in p_name:
                    candidate[p_name] = "true"

        # perform initial completeness check

        log.xsap.tfpg_tight.debug("initial completeness check")
        if not silent:
            print()
            print("  > initial completeness check")

        prop_holds = self.mchecker.do_check(candidate, prop, False)

        # check whether an error occurred
        if prop_holds is None:
            return {'result': 'error', 'message': 'internal solver error'}

        if not prop_holds:
            log.xsap.tfpg_tight.debug("given TFPG is not complete")
            if not silent:
                print("  > given TFPG is not complete")
            return {'result': 'error', 'message': 'given TFPG is not complete'}

        log.xsap.tfpg_tight.debug("given TFPG is complete")
        if not silent:
            print("  > given TFPG is complete")
            print()

        ###########################################################################################
        # optimize modes
        ###########################################################################################

        if 'modes' in tags:

            log.xsap.tfpg_tight.debug("tightening of modes")
            if not silent:
                print("  > tightening of modes")

            candidate = \
                self.optimize_valuation_linear_search(candidate, prop, modes_params_shortlist,
                                                      silent=silent)
            if not silent:
                print()

        ###########################################################################################
        # optimize tmin values
        ###########################################################################################

        if 'tmin' in tags:

            log.xsap.tfpg_tight.debug("tightening of tmin values")
            if not silent:
                print()
                print("  > tightening of tmin values")

            # sort parameter names to guarantee reproducability
            tmin_params_shortlist = sorted(tmin_params_shortlist)

            candidate =\
                self.optimize_valuation_binary_search(candidate, prop, tmin_params_shortlist,
                                                      try_minimal_improvement_first=True,
                                                      silent=silent)

        ###########################################################################################
        # optimize tmax values
        ###########################################################################################

        if 'tmax' in tags:

            tmax_bool = []
            tmax_real = []

            for p in tmax_params_shortlist:
                if "bool" in p:
                    tmax_bool.append(p)
                else:
                    assert "real" in p
                    if p not in candidate:
                        log.xsap.tfpg_tight.debug("skipping parameter '%s'" % p)
                    else:
                        tmax_real.append(p)

            # sort parameter names to guarantee reproducability
            tmax_bool = sorted(tmax_bool)
            tmax_real = sorted(tmax_real)

            if len(tmax_bool) > 0:

                log.xsap.tfpg_tight.debug("tightening of tmax values (tmax=infinity)")
                if not silent:
                    print()
                    print()
                    print("  > tightening of tmax values (tmax=infinity)")

                candidate = self.optimize_valuation_linear_search(candidate, prop, tmax_bool,
                                                                  silent=silent)

            if len(tmax_real) > 0:

                def get_corresponding_tmin(par_tmax, ls_tmin_params):
                    # eg, param_tmax_real_EDGE1__AT0
                    m_par_name = re.match(r'param_tmax_real_(.+)__AT0', par_tmax)
                    assert m_par_name is not None
                    # eg, param_tmin_EDGE1__AT0
                    par_tmin = "param_tmin_%s__AT0" % m_par_name.group(1)
                    assert par_tmin in ls_tmin_params
                    return par_tmin

                # initialize tmax lower bounds with tmin assignments
                for p_tmax in tmax_real:
                    p_tmin = get_corresponding_tmin(p_tmax, tmin_params_shortlist)
                    p_obj = self.params_mgr.get_param(p_tmax)
                    p_obj.set_bot_val(candidate[p_tmin])

                log.xsap.tfpg_tight.debug("tightening of tmax values (tmax=finite)")
                if not silent:
                    print()
                    print()
                    print("  > tightening of tmax values (tmax=finite)")

                candidate = \
                    self.optimize_valuation_binary_search(candidate, prop, tmax_real,
                                                          try_maximal_improvement_first=True,
                                                          silent=silent)
                if not silent:
                    print()

        return {'result': 'ok', 'candidate': candidate}

    ############################################################################

    def optimize_valuation_binary_search(self, candidate, prop, params_shortlist,
                                         try_minimal_improvement_first=False,
                                         try_maximal_improvement_first=False,
                                         silent=False):

        # these flags cannot be true both at the same time
        assert not (try_maximal_improvement_first and try_minimal_improvement_first)

        log.xsap.tfpg_tight.debug("starting optimize_valuation_binary_search()")

        cur_candidate = candidate
        cur_best = None

        iteration_cnt = 1

        # initialize search bounds for all parameters in shortlist
        sbmgr = SearchBoundManager(self.params_mgr, params_shortlist)

        var_names_sorted = sorted(candidate.keys())
        params_sorted = sorted(params_shortlist)

        # to provide feedback on progress
        len_prev_shortlist = len(params_shortlist)

        # iterate while some potential improvement exists
        while cur_candidate is not None and len(params_shortlist) > 0:

            if silent:
                pass
            elif iteration_cnt == 1:
                sys.stdout.write("    %d parameters remaining ." % len(params_shortlist))
            elif len(params_shortlist) < len_prev_shortlist:
                sys.stdout.write("\n    %d parameters remaining ." % len(params_shortlist))
            else:
                sys.stdout.write(".")
            len_prev_shortlist = len(params_shortlist)

            log.xsap.tfpg_tight.debug("iteration %d" % iteration_cnt)
            iteration_cnt += 1

            log.xsap.tfpg_tight.debug("cur_candidate:")

            for p_name in var_names_sorted:
                if cur_best is None or \
                                p_name not in cur_best or \
                                cur_candidate[p_name] != cur_best[p_name]:
                    if p_name in params_sorted:
                        log.xsap.tfpg_tight.debug("  + %s = %s (%s)" %
                                                  (p_name, cur_candidate[p_name],
                                                   sbmgr.get_range_string(p_name)))
                    else:
                        log.xsap.tfpg_tight.debug("  + %s = %s" % (p_name, cur_candidate[p_name]))

            # check if property holds under current candidate
            prop_holds = self.mchecker.do_check(cur_candidate, prop, not self.skip_bmc)

            # quit if original candidate is not feasible
            if not prop_holds and cur_candidate is candidate:
                log.xsap.tfpg_tight.debug("original candidate is not feasible")
                break

            # optimize candidate using inductive invariant, update search ranges
            # where possible;
            if prop_holds:
                cur_best = self.do_reduce(cur_candidate, params_shortlist, prop)

                reduce_count = 0
                for p_name in params_shortlist:

                    # update search range for all parameters
                    sbmgr.update_w_good_val(p_name, cur_best[p_name])

                    # print info on reduced parameters
                    if cur_best[p_name] != cur_candidate[p_name]:
                        reduce_count += 1
                        log.xsap.tfpg_tight.debug("  red. %s: %s -> %s" %
                                                  (p_name, cur_candidate[p_name],
                                                   cur_best[p_name]))

                log.xsap.tfpg_tight.debug("reduced %d parameters" % reduce_count)

            # identify bad assignment and update respective search range
            else:
                bad_param = None
                for p_name in cur_candidate:
                    if p_name not in cur_best:
                        continue
                    if cur_best[p_name] != cur_candidate[p_name]:
                        assert bad_param is None
                        bad_param = p_name

                sbmgr.update_w_bad_val(bad_param, cur_candidate[bad_param])

            # remove parameters from shortlist that have a singular search range

            new_params_shortlist = []

            for p_name in params_shortlist:
                if sbmgr.range_is_singular(p_name):
                    p_value = sbmgr.get_search_val(p_name)
                    assert str(cur_best[p_name]) == str(p_value)
                else:
                    new_params_shortlist.append(p_name)

            params_shortlist = new_params_shortlist

            # pick one parameter and improve it

            if len(params_shortlist) > 0:

                cur_candidate = cur_best.copy()

                p_name = params_shortlist[0]
                p_obj = self.params_mgr.get_param(p_name)

                p_value = cur_candidate[p_name]

                if try_maximal_improvement_first and p_value == candidate[p_name]:
                    # p_value is the original value, let's try to improve it to the
                    # maximal value first; this helps eg when optimizing tmax, and
                    # the tight value is the one of the corresponding tmin parameter.
                    # if this succeeds, no bisection iterations have to be performed.

                    bot_val = p_obj.get_bot_val()

                    if bot_val is not None and sbmgr.val_in_range(p_name, bot_val):
                        p_value_improved = bot_val
                    else:
                        p_value_improved = sbmgr.get_search_val(p_name)

                elif try_minimal_improvement_first and p_value == candidate[p_name]:
                    # p_value is the original value, let's try a minimal improvement
                    # first to see whether it is the tight value. if so, the
                    # actual bisection iterations can be avoided.
                    p_value_improved = p_obj.get_minimal_improvement(p_value)
                    assert sbmgr.val_in_range(p_name, p_value_improved)
                else:
                    p_value_improved = sbmgr.get_search_val(p_name)

                if p_value_improved == p_value:
                    p_value_improved = p_obj.get_minimal_improvement(p_value)
                    assert sbmgr.val_in_range(p_name, p_value_improved)

                cur_candidate[p_name] = p_value_improved

        log.xsap.tfpg_tight.debug("optimize_valuation_binary_search() done")

        return cur_best

    ############################################################################

    def optimize_valuation_linear_search(self, candidate, prop, params_shortlist, silent=False):

        cur_candidate = candidate
        cur_best = None

        iteration_cnt = 1

        # to provide feedback on progress
        len_prev_shortlist = len(params_shortlist)

        # iterate while some potential improvement exists
        while cur_candidate is not None:

            if silent:
                pass
            elif iteration_cnt == 1:
                sys.stdout.write("    %d parameters remaining ." % len(params_shortlist))
            elif len(params_shortlist) < len_prev_shortlist:
                sys.stdout.write("\n    %d parameters remaining ." % len(params_shortlist))
            else:
                sys.stdout.write(".")
            len_prev_shortlist = len(params_shortlist)

            log.xsap.tfpg_tight.debug("iteration %d" % iteration_cnt)
            iteration_cnt += 1

            log.xsap.tfpg_tight.debug("cur_candidate:")

            for p_name in cur_candidate:
                if cur_best is None or \
                                p_name not in cur_best or \
                                cur_candidate[p_name] != cur_best[p_name]:
                    log.xsap.tfpg_tight.debug("  %s = %s" % (p_name, cur_candidate[p_name]))

            # check if property holds under current candidate
            prop_holds = self.mchecker.do_check(cur_candidate, prop, False)

            # quit if original candidate is not feasible
            if not prop_holds and cur_candidate is candidate:
                break

            if prop_holds:
                # optimize candidate using approximation of reachable states
                cur_best = self.do_reduce(cur_candidate, params_shortlist, prop)

                reduce_count = 0
                for p_name in cur_best:
                    if cur_best[p_name] != cur_candidate[p_name]:
                        reduce_count += 1
                        log.xsap.tfpg_tight.debug("  red. %s: %s -> %s" %
                                                  (p_name, cur_candidate[p_name],
                                                   cur_best[p_name]))

                log.xsap.tfpg_tight.debug("reduced %d parameters" % reduce_count)

            else:
                # drop parameter associated with illegal minimal improvement from shortlist
                params_shortlist = self.get_updated_shortlist(params_shortlist, cur_candidate,
                                                              cur_best)

                log.xsap.tfpg_tight.debug("new shortlist:" + str(params_shortlist))

            # get minimal improvement of current best candidate
            cur_candidate = self.get_minimal_candidate_improvement(cur_best, params_shortlist)

        return cur_best

    ############################################################################

    def get_updated_shortlist(self, params_shortlist, cur_candidate, cur_best):

        # it cannot be minimally improved thus the current value is the
        # best, thus the parameter can be ignored in further
        # optimization rounds.

        # identify bad parameter improvement
        modified_params = []
        for p_name in cur_candidate:
            if cur_candidate[p_name] != cur_best[p_name]:
                modified_params.append(p_name)

        # exactly one parameter of the shortlist must have a minimally
        # improved value wrt the current best valid candidate
        assert len(modified_params) == 1
        assert modified_params[0] in params_shortlist

        p_name = modified_params[0]
        p_obj = self.params_mgr.get_param(p_name)

        p_best_val = cur_best[p_name]
        p_bad_value = cur_candidate[p_name]
        p_improved_val = p_obj.get_minimal_improvement(p_best_val)

        assert p_bad_value == p_improved_val

        # return shortlist minus the affected parameter
        return [p for p in params_shortlist if p != p_name]

    ############################################################################

    def get_minimal_candidate_improvement(self, candidate, params_shortlist):

        # always pick first in alphabetical order for which an improvement exists
        # in order to guarantee deterministic performance
        params_shortlist.sort()

        p_name = None
        p_value = None

        for cname in params_shortlist:
            p_obj = self.params_mgr.get_param(cname)
            cval = p_obj.get_minimal_improvement(candidate[cname])
            if cval is not None:
                p_name = cname
                p_value = cval
                break

        # return if nothing can be improved
        if p_name is None:
            return None

        # create new candidate
        new_candidate = candidate.copy()
        new_candidate[p_name] = p_value

        return new_candidate

    ############################################################################

    def get_assignment_term(self, p_name, p_value):

        p_obj = self.params_mgr.get_param(p_name)
        p_name_term = self.mk_from_string(p_name)

        if p_obj.is_boolean():
            p_value_term = self.mk_from_string(p_value)
        else:
            p_value_term = self.mk_num(p_value)

        return self.mk_eq(p_name_term, p_value_term)

    ############################################################################

    def has_bad_val(self, p_name, candidate, invar_prop, reach, reach_is_region):

        p_obj = self.params_mgr.get_param(p_name)
        p_name_term = self.mk_from_string(p_name)

        # create environment

        cenv = msat_create_env({"model_generation": "true"}, self.msat_env)

        # assert reachable states and property violation

        if reach_is_region:
            self.f_assert(cenv, self.mk_not(reach))
        else:
            self.f_assert(cenv, reach)
            self.f_assert(cenv, self.mk_not(invar_prop))

        # fix all parameters except p_name to the value given in candidate

        for p_name_2 in candidate:
            if p_name_2 == p_name:
                continue

            # parameters where no value exists such that the property
            # can be violated are assigned the top value
            if candidate[p_name_2] is None:
                p2_obj = self.params_mgr.get_param(p_name_2)
                assert p2_obj.has_top_val()
                self.f_assert(cenv, self.get_assignment_term(p_name_2, p2_obj.get_top_val()))

            else:
                self.f_assert(cenv, self.get_assignment_term(p_name_2, candidate[p_name_2]))

        # compute domain constraint of parameter and assert it

        domain_constraint = self.mk_true()

        if p_obj.is_real():
            p_range = p_obj.get_range()

            if p_range['min'] is not None:
                domain_constraint = self.mk_geq(p_name_term, self.mk_num(p_range['min']))

            if p_range['max'] is not None:
                domain_constraint = self.mk_and(
                    domain_constraint,
                    self.mk_leq(p_name_term, self.mk_num(p_range['max'])))

        self.f_assert(cenv, domain_constraint)

        # try to see if the problem is SAT (ie some assignment to the
        # parameter exists s.t. the property is violated)

        res = msat_solve(cenv)

        # done

        has_bad_value = res == MSAT_SAT

        msat_destroy_env(cenv)
        return has_bad_value

    ############################################################################

    def do_reduce(self, candidate, params_shortlist, invar_prop):

        # get type of parameters and check that they're all of the same type
        params_are_bool = None
        for p_name in params_shortlist:
            p_obj = self.params_mgr.get_param(p_name)
            if p_obj.is_boolean():
                assert params_are_bool is not False
                params_are_bool = True
            else:
                assert params_are_bool is not True
                params_are_bool = False

        # remove parameters from shortlist that already have the best
        # value possible
        updated_shortlist = []
        for p_name in params_shortlist:

            # skip parameters not specified in the candidate
            # (this can be the case for tmax=infinity)
            if p_name not in candidate:
                continue

            p_obj = self.params_mgr.get_param(p_name)
            if not p_obj.is_bot_val(candidate[p_name]):
                updated_shortlist.append(p_name)

        if len(updated_shortlist) == 0:
            return candidate

        params_shortlist = updated_shortlist

        # initialize environment
        optenv = msat_create_env({"model_generation": "true"}, self.msat_env)

        # add inductive invariant
        inductive_invariant = self.mchecker.get_inductive_invariant()
        self.f_assert(optenv, inductive_invariant)

        # add negated property
        negated_prop = self.mk_not(invar_prop)
        self.f_assert(optenv, negated_prop)

        # parameters not in shortlist should be treated as constants, i.e. not
        # as assumptions.

        for p_name in candidate:
            if p_name in params_shortlist or candidate[p_name] is None:
                continue

            p_fixed_constraint = self.get_assignment_term(p_name, candidate[p_name])
            self.f_assert(optenv, p_fixed_constraint)

        # compute smallest valuation s.t. property violation is UNSAT in invar
        # call reduce function according to parameter type

        if params_are_bool:
            log.xsap.tfpg_tight.debug("calling boolean reduce")
            reduced_candidate = self.do_reduce_bool(optenv, candidate, params_shortlist)

        else:
            log.xsap.tfpg_tight.debug("calling real reduce")
            reduced_candidate = self.do_reduce_real(optenv, candidate, params_shortlist, invar_prop)

        msat_destroy_env(optenv)
        return reduced_candidate

    def do_reduce_bool(self, optenv, candidate, params_shortlist):

        # an assumption is a less-than-optimal value necessary to guarantee
        # property satisfaction; this function tries to drop assumptions not
        # necessary to guarantee property satisfaction assuming the states
        # represented by the inductive invariant are reachable.

        assumptions = []
        term2param = {}

        # compute list of assumptions
        for p_name in params_shortlist:
            p_obj = self.params_mgr.get_param(p_name)
            assert p_obj.is_boolean()

            p_value = candidate[p_name]

            # skip parameter if it already has assigned the best value
            if p_obj.is_bot_val(p_value):
                continue

            p_name_term = self.mk_from_string(p_name)
            p_value_term = self.mk_from_string(p_value)
            assumption_term = self.mk_eq(p_name_term, p_value_term)

            assumptions.append(assumption_term)
            term2param[assumption_term] = p_name

        # solve with assumptions associated with shortlisted parameters
        res = msat_solve_with_assumptions(optenv, assumptions)
        assert res == MSAT_UNSAT

        # retrieve unsat assumptions and set them as new assumptions
        new_assumptions = msat_get_unsat_assumptions(optenv)

        assert set(new_assumptions) <= set(assumptions)
        assumptions = new_assumptions

        # repeat process with new assumptions until unsat core
        # consists of all assumptions
        while True:
            res = msat_solve_with_assumptions(optenv, assumptions)
            assert res == MSAT_UNSAT

            new_assumptions = msat_get_unsat_assumptions(optenv)
            assert set(new_assumptions) <= set(assumptions)

            if len(new_assumptions) < len(assumptions):
                assumptions = new_assumptions
            else:
                break

        # optimize parameters that are not part of the core assumptions

        reduced_candidate = candidate.copy()
        core_params = [term2param[t] for t in assumptions]

        for p_name in params_shortlist:
            if p_name not in core_params:
                p_obj = self.params_mgr.get_param(p_name)
                opt_p_value = p_obj.get_minimal_improvement(candidate[p_name])
                assert opt_p_value is not None
                reduced_candidate[p_name] = opt_p_value

        return reduced_candidate

    def do_reduce_real(self, optenv, candidate, params_shortlist, invar_prop):

        candidate = candidate.copy()
        reach = self.mchecker.get_inductive_invariant()

        # try to reduce every parameter in the shortlist
        for p_name in params_shortlist:

            p_obj = self.params_mgr.get_param(p_name)
            assert p_obj.is_real()

            p_name_term = self.mk_from_string(p_name)

            # corner case: no illegal assignment for the parameter exists
            if not self.has_bad_val(p_name, candidate, invar_prop, reach, False):
                if p_obj.has_bot_val():
                    candidate[p_name] = p_obj.get_bot_val()
                else:
                    candidate[p_name] = None
                log.xsap.tfpg_tight.debug("no bad val exists (setting %s to %s)" %
                                          (p_name, candidate[p_name]))
                continue

            # push backtrack point
            msat_push_backtrack_point(optenv)

            # assert equality for fixed parameters

            for p_name_2 in params_shortlist:
                if p_name_2 is p_name:
                    continue
                p_name_2_term = self.mk_from_string(p_name_2)

                p_value_2 = candidate[p_name_2]
                if p_value_2 is None:
                    continue

                p_value_2_term = self.mk_num(p_value_2)
                self.f_assert(optenv, self.mk_eq(p_name_2_term, p_value_2_term))

            while True:

                # improve by one delta
                p_value = p_obj.get_minimal_improvement(candidate[p_name])
                if p_value is None:
                    break

                # push backtrack point
                msat_push_backtrack_point(optenv)

                # assert name=value in optenv
                self.f_assert(optenv, self.mk_eq(p_name_term, self.mk_num(p_value)))

                # solve
                res = msat_solve(optenv)
                assert res != MSAT_UNKNOWN

                # pop backtrack point
                msat_pop_backtrack_point(optenv)

                # if unsat, i.e. improvement doesn't cause property violation:
                if MSAT_UNSAT == res:
                    candidate[p_name] = p_value

                # if sat, break
                else:
                    break

            # pop backtrack point
            msat_pop_backtrack_point(optenv)

        return candidate

    ############################################################################

    def mk_term(self, fn, *args):
        res = fn(self.msat_env, *args)
        assert not MSAT_ERROR_TERM(res)
        return res

    def t_is_bool(self, t):
        return msat_term_is_boolean_constant(self.msat_env, t)

    def f_assert(self, menv, f):
        assert not MSAT_ERROR_TERM(f)
        msat_assert_formula(menv, f)

    def mk_from_string(self, string):
        return self.mk_term(msat_from_string, string)

    def mk_false(self):
        return msat_make_false(self.msat_env)

    def mk_true(self):
        return msat_make_true(self.msat_env)

    def mk_or(self, t1, t2):
        return self.mk_term(msat_make_or, t1, t2)

    def mk_and(self, t1, t2):
        return self.mk_term(msat_make_and, t1, t2)

    def mk_eq(self, t1, t2):
        if self.t_is_bool(t1) or self.t_is_bool(t2):
            return self.mk_term(msat_make_iff, t1, t2)
        else:
            return self.mk_term(msat_make_equal, t1, t2)

    def mk_neq(self, t1, t2):
        return self.mk_not(self.mk_eq(t1, t2))

    def mk_not(self, t):
        return self.mk_term(msat_make_not, t)

    def mk_num(self, val):
        return self.mk_term(msat_make_number, str(val))

    def mk_leq(self, t1, t2):
        return self.mk_term(msat_make_leq, t1, t2)

    def mk_geq(self, t1, t2):
        return self.mk_not(self.mk_lt(t1, t2))

    def mk_lt(self, t1, t2):
        return self.mk_and(self.mk_neq(t1, t2), self.mk_leq(t1, t2))

    def mk_gt(self, t1, t2):
        return self.mk_not(self.mk_leq(t1, t2))

    def mk_iff(self, t1, t2):
        return self.mk_term(msat_make_iff, t1, t2)
