from __future__ import division
from __future__ import print_function

from builtins import object
from past.utils import old_div
import math

import xsap.features.tfpg_tightening.tightening_engine_bisection.utils as utils


class Parameter(object):

    def __init__(self, param_string):

        # param_string format (all fields must be present):
        #   name / domain / monotonicity_direction / val_min /
        #   val_max / step_delta / tag

        param_tokens = param_string.split("/")
        param_tokens = [pt.strip() for pt in param_tokens]

        assert len(param_tokens) == 7
        for pt in param_tokens:
            assert len(pt) > 0

        self.name = param_tokens[0]

        self.domain = param_tokens[1]
        assert self.domain in ("boolean", "real")

        self.monotonicity_direction = param_tokens[2]
        assert self.monotonicity_direction in ("upward", "downward")

        self.val_min = param_tokens[3].lower()
        if self.is_boolean():
            assert self.val_min == "false"
        elif self.val_min == "none":
            self.val_min = None
        else:
            assert utils.is_float(self.val_min)
            self.val_min = float(self.val_min)

        self.val_max = param_tokens[4].lower()
        if self.is_boolean():
            assert self.val_max in ("false", "true")
        elif self.val_max == "none":
            self.val_max = None
        else:
            assert utils.is_float(self.val_max)
            self.val_max = float(self.val_max)

        self.step_delta = param_tokens[5].lower()
        if self.is_boolean():
            assert self.step_delta == "none"
            self.step_delta = None
        else:
            assert utils.is_float(self.step_delta)
            self.step_delta = float(self.step_delta)
            # we assume this value is strictly positive, if given;
            assert self.step_delta > 0

        self.tag = param_tokens[6]

        # top value: the worst value for the parameter
        # bot value: the best value for the parameter

        if self.monotonicity_direction == "upward":
            self.top_val = self.val_max
            self.bot_val = self.val_min
        else:
            self.top_val = self.val_min
            self.bot_val = self.val_max

    def pprint(self):
        print("parameter '%s'" %  self.name)
        print("  domain:", self.domain)
        print("  monotonicity_direction:", self.monotonicity_direction)
        print("  val_min:", self.val_min)
        print("  val_max:", self.val_max)
        print("  step_delta:", self.step_delta)
        print("  tag:", self.tag)
        print("  top_val:", self.top_val)
        print("  bot_val:", self.bot_val)

    def is_boolean(self):
        return self.domain == "boolean"

    def is_real(self):
        return self.domain == "real"

    def is_upward_monotonic(self):
        return self.monotonicity_direction == "upward"

    def is_downward_monotonic(self):
        return self.monotonicity_direction == "downward"

    def get_range(self):
        return {'min':self.val_min, 'max':self.val_max}

    def has_singular_range(self):
        if self.val_min is None or self.val_max is None:
            return False
        else:
            return self.val_min == self.val_max

    def has_top_val(self):
        return self.top_val is not None

    def is_top_val(self, val):
        return self.normalize_value(val) == self.top_val

    def get_top_val(self):
        return self.top_val

    def set_bot_val(self, new_top_val):
        normalized_val = self.normalize_value(new_top_val)
        assert normalized_val is not None
        self.bot_val = normalized_val

        if self.monotonicity_direction == "upward":
            self.val_min = normalized_val
        else:
            self.val_max = normalized_val

    def has_bot_val(self):
        return self.bot_val is not None

    def is_bot_val(self, val):
        return self.normalize_value(val) == self.bot_val

    def get_bot_val(self):
        return self.bot_val

    def get_step_delta(self):
        return self.step_delta

    def normalize_value(self, val):

        # self is boolean
        if self.is_boolean():
            if not isinstance(val, str):
                return None

            # convert to lower-case and drop mathsat string delimiters if present
            val = val.lower().strip('`')

            if val in ("false", "true"):
                return val
            else:
                return None

        # self is real
        elif isinstance(val, str):
            if utils.is_float(val):
                return float(val)
            elif utils.is_fraction(val):
                return utils.fraction_string_to_float(val)
            else:
                return None
        elif isinstance(val, float) or isinstance(val, int):
            return float(val)
        else:
            return None

    def get_minimal_improvement(self, val):

        # cannot improve a non-existing value
        if val is None:
            return None

        val = self.normalize_value(val)

        if val == self.bot_val:
            # val is already the best value possible
            return None
        else:
            # compute better value
            if self.is_boolean():
                return self.bot_val
            else:
                # the value should be a multiple of the step delta
                assert str(round(old_div(val, self.step_delta), 4)).endswith(".0") 
                 
                if self.is_upward_monotonic():
                    # lower is better
                    return val - self.step_delta
                else:
                    # higher is better
                    return val + self.step_delta
