
import os
import re
import subprocess

from xsap.features.tfpg_tightening.tightening_engine_bisection.tightening_engine \
    import do_tighening_bisection
from xsap.utils.misc import log
from xsap.utils.processes.xsap_process import XsapProcess


def perform_tightening_invar(fn_smv_ext, invar_prop, options, dic_params, tfpg_struct):
    """ Performs tightening with a bisection approach.

    @param fn_smv_ext: system model (without parameters)
    @type  fn_smv_ext: str

    @param invar_prop: parametric invariant property expressing completeness, to be evaluated
                       on the given SMV model
    @type  invar_prop: str

    @param options: tightening options (eg engine options, tags, ...)
    @type  options: dict

    @param dic_params: parameter dictionary generated by generate_parameters_dictionary()
    @type  dic_params: dict

    @param tfpg_struct: the TFPG to be tightened
    @type  tfpg_struct: Tfpg

    @return: returns a tuple where the first element is a dictionary with the resulting
             parameter values if everything went well and None otherwise, and the second
             element is an error message in case errors occurred.
    @rtype:  tuple
    """

    # check all needed options are present
    assert "tags" in options
    assert "t_delta_var" in options
    assert "tmax_bound" in options
    assert "engine" in options
    assert "skip_bmc" in options
    assert "outdir" in options

    # dump VMT

    fn_vmt = compute_vmt_fname(fn_smv_ext, options["outdir"])
    res = dump_smv_to_vmt(fn_smv_ext, fn_vmt, str(invar_prop))

    if res is None:
        return None, "dumping of VMT model failed"

    # get tmax bound

    if options["tmax_bound"] is None:
        tmax_bound = get_largest_constant(tfpg_struct)
    else:
        tmax_bound = options["tmax_bound"]

    # get search precision

    step_size = get_step_size(tfpg_struct)

    # call bisection tightening function
    return do_tighening_bisection(fn_vmt, options, dic_params, tmax_bound, step_size)


def compute_vmt_fname(fn_smv, outdir):
    full_filename = os.path.basename(fn_smv)
    f_name, f_ext = os.path.splitext(full_filename)
    assert f_ext == '.smv'
    return os.path.join(outdir, f_name + '.vmt')


def dump_smv_to_vmt(fn_smv, fn_vmt, inv_prop):

    # create xsap instance
    need_msat = True  # msat is always needed due to the real-valued counters
    xsap_proc = XsapProcess(smv_fname=fn_smv, ord_fname=None, need_msat=need_msat,
                            args=[], stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                            dumpers=[], dynamic_reorder=False, verbose_level=0)

    # dump vmt file
    log.xsap.tfpg_tight.debug("Dumping parameterized SMV to VMT.")

    # parse model
    success = xsap_proc.send_init_commands(build_msat=True)
    if not success:
        return None

    xsap_proc.send_cmd('write_vmt_model -i \"%s\" -o %s' % (inv_prop, fn_vmt))
    xsap_proc.wait_for_command_finish(end_message="END OF VMT DUMPING")

    xsap_proc.send_cmd('quit -x')

    # close xsap instance
    xsap_proc.kill_and_close_streams()

    # return vmt file name
    return fn_vmt


def get_largest_constant(tfpg_struct):
    max_constant = -1
    for e in tfpg_struct.get_edges_list():
        max_constant = max(max_constant, float(e.get_tMin()))
        max_constant = max(max_constant, float(e.get_tMax()))
    return max_constant


def get_step_size(tfpg_struct):

    precision = 0

    for e in tfpg_struct.get_edges_list():
        for val in [e.get_tMin(), e.get_tMax()]:
            s_val = str(val)
            assert s_val.find('.') > 0

            # strip trailing zeroes in case there are any
            s_val = re.sub(r'0+$', '', s_val)

            # keep only decimal part
            s_val_dec = s_val[s_val.find('.') + 1:]

            # update precision if necessary
            precision = max(precision, len(s_val_dec))

    assert precision >= 0

    if precision == 0:
        return 1.0
    else:
        return float('0.' + (precision-1) * '0' + '1')
