from __future__ import print_function
# %HEADER

from builtins import range
from collections import OrderedDict
import os.path
import re
import sys
import string
import random
import tempfile
import subprocess

try:
    import xsap.utils.options as options
    from xsap.utils.options import MC_ENGINE_BMC, MC_ENGINE_MSAT_BMC, \
        MC_ENGINE_IC3, MC_ENGINE_BDD
    from xsap.utils.options import FT_ENGINE_BMC, FT_ENGINE_BDD, \
        FT_ENGINE_MSAT_BMC, FT_ENGINE_SBMC, FT_ENGINE_PARAM, \
        FT_ENGINE_BMC_INC_BDD, FT_ENGINE_SBMC_INC_BDD
except ImportError:
    from compass import options # in case we are in COMPASS project
    from compass.options import MC_ENGINE_BMC, MC_ENGINE_MSAT_BMC, MC_ENGINE_IC3, MC_ENGINE_BDD
    from compass.options import FT_ENGINE_BMC, FT_ENGINE_BDD, \
        FT_ENGINE_MSAT_BMC, FT_ENGINE_SBMC, FT_ENGINE_PARAM, \
        FT_ENGINE_BMC_INC_BDD, FT_ENGINE_SBMC_INC_BDD
from xsap.utils.processes.nuxmv_process import nuXmvProcess
from xsap.utils.processes.process import get_exec_name, TextLineGrepper
import xsap.utils.misc.log as log
from xsap.utils.misc.exceptions import ProcessInternalException, TraceException
from xsap.features.tfpg_synthesis.tfpg_synthesizer import CutSet


class XsapProcess (nuXmvProcess):
    """Process handler for running Xsap interactively"""

    # prefix for generated files
    _fname_prefix = "__xsap_prefix__"

    def __exit__(self, exc_type, exc_val, exc_tb):
        """This method is called when exiting from the with statement.
        This overrides the __exit__ method of the KillablePipe base class"""
        self.kill_and_close_streams()

    def __init__(self, smv_fname, fail_obs_alm_fname=None,
                 ord_fname=None, need_msat=False,
                 args=[], stdout=None, stderr=None, dumpers=[],
                 dynamic_reorder=True, verbose_level=0,
                 boolean_conversion_uses_predicate_normalization=False):

        if fail_obs_alm_fname is not None:
            args = list(args) + ["-sa_compass", "-sa_ft_layering",
                                 "-sa_compass_task", fail_obs_alm_fname]

        # store handles of dumpers
        self.dumpers = dumpers

        # first tries the fully operative version of xsap, then if
        # failing falls back to the standard xsap
        self.__degraded = False
        try:
            path = get_exec_name(options.XSAP)
        except IOError:
            path = get_exec_name(
                options.DEGRADED_MODE)
            self.__degraded = True
            pass

        super(XsapProcess, self).__init__(path, smv_fname, ord_fname,
                                          need_msat, args,
                                          stdout=stdout, stderr=stderr,
                                          dumpers=dumpers,
                                          dynamic_reorder=dynamic_reorder,
                                          verbose_level=verbose_level,
                                          boolean_conversion_uses_predicate_normalization=boolean_conversion_uses_predicate_normalization)

        self._fail_obs_alm_fname = fail_obs_alm_fname

    @staticmethod
    def has_infinite_domain_vars(smv_file):
        """Checks if SMV model contains variables of infinite domain."""

        has_inf_domain_vars = True

        # create xsap process
        xsap_process = XsapProcess(
            smv_fname=smv_file,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT)

        xsap_process.send_cmd('set on_failure_script_quits')

        # read and flatten model

        xsap_process.send_cmd('read_model')
        xsap_process.send_cmd('flatten_hierarchy')

        # print and grep variable information

        xsap_process.send_cmd('show_vars')
        xsap_process.send_cmd('echo DONE_SHOW_VARS')

        grep = TextLineGrepper(r".*", "DONE_SHOW_VARS")
        (m, txt) = grep.match(xsap_process)
        assert m is not None

        m = re.search(r'\s*\d+\s*:\s*(real|integer)', txt)
        if m is None:
            has_inf_domain_vars = False

        # before returning, close process
        xsap_process.kill_and_close_streams()

        return has_inf_domain_vars

    def kill_and_close_streams(self):
        """Terminate parent and close all pipes and files associated
        with this process."""

        super(XsapProcess, self).kill()

        for f in self.dumpers:
            f.close()

        self.stdout.close()

    def wait_for_command_finish(self,
                                end_message="END OF XSAP COMMAND EXECUTION ~",
                                tool="xSAP"):
        # this messages is used to detect the end of FSAP command
        # execution.  it must be unique (not output by normal Xsap
        # run), without spaces at the ends and simple enough for python's re.
        return super(XsapProcess, self).wait_for_command_finish(end_message, tool)

    def _check_handler_errors(self, txt, tool="xSAP"):
        super(XsapProcess, self)._check_handler_errors(txt, tool)

    def set_tmp_filename_prefix(self, prefix):
        """Sets the prefix of temporary generated files to the provided one."""
        if os.path.isdir(prefix):
            self._fname_prefix = prefix + "/"
        else:
            self._fname_prefix = prefix

    def is_degraded_mode(self):
        """Returns True if the process supports full operations"""
        return self.__degraded

    # ======================================================================
    # TFPG Analyses
    # ======================================================================

    # ======================================================================
    # == Behavioral Validation

    def check_tfpg_completeness(self, proof_obligations, bmc_k,
                                engine=MC_ENGINE_MSAT_BMC,
                                boolean_conversion=False):
        """ This function performs the completeness check of behavioral
        validation for TFPGs. It tries to find a system trace for which no
        valid corresponding TFPG trace exists, ie which causes the TFPG to
        violate some of its constraints when observing that trace.

        The function returns a tuple containing the analysis result and a
        completeness counterexample if one has been found. A completeness
        counterexample consists of a system trace and a corresponding TFPG
        trace that violates some TFPG constraints.

        The tfpg passed as a parameter is assumed to be the same used for
        creating the extended SMV file.

        @param bmc_k: maximum depth for bounded model checking
        @type bmc_k: integer """

        # result can be: {error, complete, incomplete, complete_bounded}
        # counterexample can be: {None, pair of traces}
        # description is a string, containing a high-level description of the
        # incompleteness cause (eg. "missed activation of some_tfpg_node")
        res = {'result': None, 'counterexample': None, 'description': None}

        # we start by assuming completeness; if a counterexample is found or
        # model checking can prove absence of counterexamples only up to the
        # given bound, this value is changed.
        res['result'] = 'complete'

        # check sanity of parameters
        assert isinstance(bmc_k, int)
        assert bmc_k >= 0

        ok = self.send_init_commands(build_msat=True)
        if not ok:
            return {'result': 'error', 'description': 'Couldn\'t build model.'}

        # model check the proof obligations

        for proof_obl_idx in range(len(proof_obligations)):

            prop = proof_obligations[proof_obl_idx]['prop']
            assert proof_obligations[proof_obl_idx]['proptype'] == 'invarspec'

            print("  > checking proof obligation %d/%d ..." % \
                (proof_obl_idx + 1, len(proof_obligations)), end=' ')
            sys.stdout.flush()

            prop_check_res = self.verify_prop(prop, 'invarspec', engine, bmc_k)

            if prop_check_res['result'] == 'error':
                res['result'] = 'error'
                print("error")
                res['description'] = 'Error in verifying proof obligation %d/%d ...' % \
                  (proof_obl_idx+1, len(proof_obligations))
                break

            elif prop_check_res['result'] == 'true':
                print("satisfied")

            elif prop_check_res['result'] == 'false':
                res['result'] = 'incomplete'
                if prop_check_res['trace'] == "No traces available":
                    res['counterexample'] = None
                else:
                    res['counterexample'] = prop_check_res['trace']
                res['description'] = proof_obligations[
                    proof_obl_idx]['description']
                print("not satisfied (%s)" % res['description'])
                break

            else:
                assert prop_check_res['result'] == 'unknown'
                print("satisfied (within bound)")
                res['result'] = 'complete_bounded'

        return res

    # ======================================================================
    # == Effectiveness validation

    def check_tfpg_edge_tightness(self, proof_obligations, bmc_length,
                                  engine, boolean_conversion, tags):

        # check sanity of parameters
        assert isinstance(bmc_length, int)
        assert bmc_length >= 0
        for t in tags:
            assert t in ("tmin", "tmax", "tmax-inf", "modes")

        ok = self.send_init_commands(build_msat=True)
        if not ok:
            return {'result': 'error', 'description': 'Couldn\'t build model.'}

        # flag to indicate result for tmax=infinity cases
        tmaxinf_tight = "tight"

        for proof_obl_idx in range(len(proof_obligations)):

            proof_obl = proof_obligations[proof_obl_idx]

            prop = proof_obl['prop']
            proptype = proof_obl['proptype']

            descr = proof_obl['description']

            print("  > checking proof obligation %d/%d ..." % \
                (proof_obl_idx + 1, len(proof_obligations)), end=' ')
            sys.stdout.flush()

            if descr.startswith("can tighten tmin of edge") and "tmin" not in tags:
                print("    .. skipping tmin proof obligation")
                continue

            if descr.startswith("can tighten tmax of edge") and "tmax" not in tags:
                print("    .. skipping tmax proof obligation")
                continue

            # check prop using invar_engine
            prop_check_res = self.verify_prop(
                prop, proptype, engine, bmc_length)

            if prop_check_res['result'] == 'error':
                print("error")
                return {'result': 'error', 'description': None}

            # if has trace, continue
            if prop_check_res['result'] == 'false':
                print("satisfied")
                continue

            # property is true (up to bound) but this is a tmax-infinity proof
            # obligation: unknown
            elif proof_obl['tmax-infinity']:
                assert prop_check_res['result'] in ('true', 'unknown')
                print("unknown (%s?)" % descr)
                tmaxinf_tight = "unknown"
                continue

            # property is true: not tight
            elif prop_check_res['result'] == 'true':
                assert not proof_obl['tmax-infinity']
                print("not satisfied (%s)" % descr)
                return {'result': 'not_tight', 'description': descr}

            # no trace within bound: not tight (within bound)
            else:
                assert prop_check_res['result'] == 'unknown'
                assert not proof_obl['tmax-infinity']
                print("not satisfied within bound (%s)" % descr)
                return {'result': 'not_tight_bounded', 'description': descr}

        return {'result': 'tight', 'description': None, 'tmaxinf_tight': tmaxinf_tight}

    def check_tfpg_diagnosability(self, obs_fname, diagcond_fname, smv_prop,
                                  use_bmc=False, bmc_k=10):
        """ Checks whether a bound exists within which any behavior caused by
        failures from target_fm_set can be distinguished from any other
        behavior not including any of these failures."""

        # result can be: {error, diagnosable, not_diagnosable,
        # diagnosability_unknown}
        result = {'result': None, 'critical_pair': None}

        # check sanity of parameters

        assert isinstance(bmc_k, int)
        assert bmc_k >= 0

        # cmd: set on_failure_script_quits

        self.send_cmd('set on_failure_script_quits')

        # cmd: build twin plant
        self.build_twin_plant(obs_fname, None,
                              use_history_vars=False, async_composition=True)

        # cmd: load_diag_conditions
        self.load_diag_conditions(diagcond_fname, -1.0)

        # cmd: go
        if use_bmc:
            self.build_bmc_fsm()
        else:
            self.build_bdd_fsm()

        if not self.is_alive():
            # For Windows: it does not kills automatically also the
            # threads created by this process
            self.kill_and_close_streams()
            # If something goes wrong, try to print the line error.
            self.search_and_print_errors(self.get_stderr_text())
            return None

        # cmd: check_property
        if use_bmc:
            (res, trace_fname) = self.model_check_bmc_ltl(smv_prop, bmc_k,
                                                          use_inc=False)
        else:
            (res, trace_fname) = self.model_check_bdd_ltl(smv_prop)

        # retrieve results
        if res == 'error':
            result['result'] = 'error'
            result['critical_pair'] = None

        elif res == 'true':
            result['result'] = 'diagnosable'
            result['critical_pair'] = None

        elif res == 'false':
            result['result'] = 'not_diagnosable'
            orig_trace_fname = trace_fname

            if orig_trace_fname is None:
                result['result'] = 'error'
                return result

            result['critical_pair'] = orig_trace_fname

        else:
            assert res == 'unknown'
            result['result'] = 'diagnosability_unknown'
            result['critical_pair'] = None

        return result

    # ======================================================================
    # == Synthesis

    def compute_minimal_cutsets(self, tle_list, tmp_filename, bmc_k,
                                engine=FT_ENGINE_BMC,
                                boolean_conversion=False):
        """ Returns the minimal cutsets list for a given smv model. If some error
        occurs, None is returned. ""

        @param tle_list: the list of top level events
        @type  tle_list: list of strings

        @param bmc_k: bound for bmc analysis
        @type  bmc_k: int
        """

        build_bdd = engine in [FT_ENGINE_BDD,
                               FT_ENGINE_BMC_INC_BDD,
                               FT_ENGINE_SBMC_INC_BDD]

        build_bmc = engine in [FT_ENGINE_BMC,
                               FT_ENGINE_SBMC,
                               FT_ENGINE_BMC_INC_BDD,
                               FT_ENGINE_SBMC_INC_BDD] or \
                    (engine == FT_ENGINE_PARAM and not self._need_msat)

        build_msat = engine == FT_ENGINE_MSAT_BMC or \
                     (engine == FT_ENGINE_PARAM and self._need_msat)

        ok = self.send_init_commands(
            build_bdd=build_bdd,
            build_bmc=build_bmc,
            build_msat=build_msat)
        if not ok:
            return None

        from collections import OrderedDict
        minimal_cutsets = OrderedDict()

        # for each tle
        for tle_idx in range(len(tle_list)):

            tle = tle_list[tle_idx]

            m = re.match(r"^tfpgsynth_(\S+)$", tle)
            assert m is not None
            discr_name = m.group(1)

            if len(tle_list) > 1:
                print(
                    "  discrepancy %d/%d (%s)" %
                    (tle_idx + 1, len(tle_list), discr_name))

            minimal_cutsets[discr_name] = []

            # wipe logfile
            open(tmp_filename, 'w').close()

            log.xsap.tfpg_synth.debug(
                "Computing minimal cutsets for discrepancy '%s' "
                "(discrepancy %d/%d)." % (discr_name, tle_idx + 1,
                                          len(tle_list)))

            # compute all mcs (read results from stdout)

            if engine == FT_ENGINE_BDD:
                self.compute_fault_tree(tle, out_fname=tmp_filename,
                                        disable_ft_generation=True)

            elif engine == FT_ENGINE_PARAM:
                self.compute_fault_tree(tle,
                                        use_bdd=False,
                                        use_param=True,
                                        out_fname=tmp_filename,
                                        force_smt=self._need_msat,
                                        disable_ft_generation=True)

            elif engine == FT_ENGINE_BMC:
                assert not self.needs_msat()
                self.compute_fault_tree(tle,
                                        bmc_length=bmc_k,
                                        use_bdd=False,
                                        out_fname=tmp_filename,
                                        disable_ft_generation=True)

            elif engine == FT_ENGINE_SBMC:
                assert not self.needs_msat()
                self.compute_fault_tree(tle,
                                        bmc_length=bmc_k,
                                        use_bdd=False,
                                        sat_sbmc=True,
                                        out_fname=tmp_filename,
                                        disable_ft_generation=True)

            elif engine == FT_ENGINE_BMC_INC_BDD:
                self.compute_fault_tree(tle,
                                        bmc_length=bmc_k,
                                        use_bdd=False,
                                        use_bdd_bmc=True,
                                        out_fname=tmp_filename,
                                        disable_ft_generation=True)

            elif engine == FT_ENGINE_SBMC_INC_BDD:
                self.compute_fault_tree(tle,
                                        bmc_length=bmc_k,
                                        use_bdd=False,
                                        use_bdd_bmc=True,
                                        sat_sbmc=True,
                                        out_fname=tmp_filename,
                                        disable_ft_generation=True)

            else:
                assert engine == FT_ENGINE_MSAT_BMC
                assert self.needs_msat()
                self.compute_fault_tree(tle,
                                        bmc_length=bmc_k,
                                        use_bdd=False,
                                        out_fname=tmp_filename,
                                        disable_ft_generation=True)

            # parse mcs using cutset class
            log.xsap.tfpg_synth.debug("parsing cutset file ... ")
            cur_cutsets = CutSet.parse_cutset_list(tmp_filename)

            if cur_cutsets is None:
                log.xsap.tfpg_synth.debug(
                    "'%s' is not reachable within given bound" %
                    discr_name)
                print ("    warning: '%s' is not reachable within "
                       "given bound." % discr_name)
            else:
                assert len(cur_cutsets) > 0
                log.xsap.tfpg_synth.debug(
                    "found %d cutsets for '%s'" %
                    (len(cur_cutsets), discr_name))
                for cs in cur_cutsets:
                    # tle = cs.get_tle().replace('tfpgsynth_','')
                    # assert tle == discr_name
                    events = [e.replace('tfpgsynth_', '')
                              for e in cs.get_events()]
                    log.xsap.tfpg_synth.debug(" cutset events: %s" % events)
                    minimal_cutsets[discr_name].append(events)

            log.xsap.tfpg_synth.debug("parsing cutset file ... done")

        return minimal_cutsets

    def add_fm_to_minimal_cutsets(self, failure_modes, minimal_cutsets,
                                  bmc_k, engine=FT_ENGINE_BMC, node_pruning=False):
        """ Helper function for TFPG synthesis. It adds all reachable FM nodes to
        the minimal cutset dictionary, with the key being the FM node name and the
        value the empty set. """

        # add the empty cutset for all FMs that occur in discrepancy cutsets
        for fm in failure_modes:

            # fm node has already been inserted in MCS dictionary
            if fm in minimal_cutsets:
                continue

        # for all failure modes not in minimal cutsets

        fm_not_in_minimal_cutsets = [
            fm for fm in failure_modes if fm not in minimal_cutsets]


        for fm_idx, fm in enumerate(fm_not_in_minimal_cutsets):

            if not node_pruning:
             print("  failure mode reachability %d/%d (%s)" %
                   (fm_idx + 1, len(fm_not_in_minimal_cutsets), fm))

             # check whether they are reachable at all
             log.xsap.tfpg_synth.debug(
                 "Verifying reachability of failure mode '%s' (failure mode "
                 "reachability %d/%d)." % (fm, fm_idx + 1,
                                           len(fm_not_in_minimal_cutsets)))

             tle = "tfpgsynth_" + fm
             inv_prop = "!(%s)" % tle

             if engine in [FT_ENGINE_BDD, FT_ENGINE_BMC_INC_BDD, FT_ENGINE_SBMC_INC_BDD]:
                 res = self.model_check_bdd_ltl("G %s" % inv_prop, generate_trace=False)
             elif engine in FT_ENGINE_PARAM:
                 res = self.model_check_ic3_invar(inv_prop, bmc_k, generate_trace=False)
             elif engine in [FT_ENGINE_BMC, FT_ENGINE_SBMC]:
                 res = self.model_check_bmc_invar(inv_prop, bmc_k, generate_trace=False)
             else:
                 assert engine == FT_ENGINE_MSAT_BMC
                 res = self.model_check_msat_bmc_invar(inv_prop, bmc_k, use_inc=True, generate_trace=False)

             if res[0] == "true":
                 log.xsap.tfpg_synth.debug("The FM node '%s' is not reachable." % fm)
                 print("    The FM node '%s' is not reachable." % fm)
                 minimal_cutsets[fm] = []

             elif res[0] == "unknown":
                 log.xsap.tfpg_synth.debug("The FM node '%s' is not reachable "
                                           "within the given bound." % fm)
                 print("    The FM node '%s' is not reachable within the given bound." % fm)
                 minimal_cutsets[fm] = []

             else:
                 assert res[0] == "false"
                 log.xsap.tfpg_synth.debug("The FM node '%s' is reachable within "
                                           "the given bound." % fm)
                 print("    The FM node '%s' is reachable within the given "
                       "bound." % fm)
                 minimal_cutsets[fm] = [[]]
            else:
                minimal_cutsets[fm] = [[]]

        return minimal_cutsets

    def compute_discrepancy_dependencies(
            self, discrepancies, node_associations):
        """Computes for every discrepancy d the TFPG nodes whose COI covers d.

        A discrepancy d is in the COI of another node v iff the set of defining
        variables of d intersect with the COI of the expression associated
        with v.

        @param discrepancies: list of discrepancies to be analyzed
        @type  discrepancies: list of strings

        @param node_associations: dictionary with keys being all TFPG nodes and
                                  values being the respective associated
                                  expressions
        @type  node_associations: dictionary of strings
        """

        discrepancy_deps = OrderedDict()

        # retrieve variables associated with each node

        vars_of_nodes = OrderedDict()
        for node in node_associations:
            node_assoc_str = node_associations[node]
            vars_of_nodes[node] = self.compute_var_deps(
                node_assoc_str, depth_zero=True)

        # compute for each discrepancy the nodes it functionally depends on

        for d in discrepancies:

            log.xsap.tfpg_synth.debug("computing dependencies of '%s'" % d)
            sys.stdout.write("  computing dependencies of '%s'\n" % d)

            discrepancy_deps[d] = []

            d_assoc_str = node_associations[d]
            deps_of_d = self.compute_var_deps(d_assoc_str, depth_zero=False)

            # pick all nodes v which are defined by at least one
            # variable in the coi of the target discrepancy

            for node in vars_of_nodes:
                if node == d:
                    continue

                if len(set(vars_of_nodes[node]).intersection(set(deps_of_d))) > 0:
                    discrepancy_deps[d].append(node)
                    # enable only for debugging, can result in very verbose
                    # output when many nodes are given in input.
                    #log.xsap.tfpg_synth.debug("  > %s" % node)
                    #sys.stdout.write("    > %s\n" % node)

        return discrepancy_deps

    def compute_var_deps(self, expression, depth_zero):
        """Computes the COI for the given expression.

        @param expression: the expression for which the COI should be computed
        @type  expression: string

        @param depth_zero: whether COI at depth 0 (depth_zero=True) or
                           unbounded (depth_zero=False) should be computed
        @type  depth_zero: boolean
        """

        if depth_zero:
            log.xsap.tfpg_behav.debug(
                "sending 'show_dependencies -k 0 -e \"%s\"' to xSAP" %
                expression)
            self.show_dependencies(0, expression)
        else:
            log.xsap.tfpg_behav.debug(
                "sending 'show_dependencies -e \"%s\"' to xSAP" %
                expression)
            self.show_dependencies(None, expression)

        # read result: {var1, var2, varN}

        grep = TextLineGrepper(r"{(.*)}")
        (m, txt) = grep.match(self)

        assert m is not None
        res_split = m.groups()[0].split(", ")

        # put each variable into coi set

        res_set = []
        for var in res_split:
            res_set.append(var)

        return res_set

    def assert_discrepancy_unreachability(self, discr_expr, bmc_len):
        """Checks whether the given discrepancy is reachable, by performing
        a check_invar command.

        If the discrepancy is indeed unreachable within the bound, returns True,
        otherwise False.

        @param discr: the name of the variable to be checked
        @type  discr: string

        @param bmc_len: the bound for bmc analysis
        @type  bmc_len: int
        """

        reach_res = self.analyze_node_reachability(node_expr=discr_expr,
                                                   analysis_bound=bmc_len)

        if reach_res == 'reachable':
            return False
        else:
            return True

    def identify_unreachable_nodes(self, ls_nodes, analysis_bound=None):
        """ Identifies what TFPG nodes among the given ones are not reachable.

        @param ls_nodes: the TFPG nodes whose reachability shall be checked
        @type  ls_nodes: list of two-keyed dictionaries (name, expression)
        """

        unreachable_nodes = []

        num_nodes = len(ls_nodes)

        for i in range(num_nodes):
            node = ls_nodes[i]

            sys.stdout.write("  [%d/%d] %s .. " % (i+1, num_nodes, node['name']))
            sys.stdout.flush()

            reach_res = self.analyze_node_reachability(node['expression'], analysis_bound)
            if reach_res == 'not_reachable':
                unreachable_nodes.append(node)
                sys.stdout.write("not reachable\n")
            elif reach_res == 'not_reachable_within_bound':
                assert analysis_bound is not None
                unreachable_nodes.append(node)
                sys.stdout.write("not reachable within bound\n")
            else:
                sys.stdout.write("reachable\n")

        return unreachable_nodes

    def analyze_node_reachability(self, node_expr, analysis_bound=None):
        """ Analyzes the reachability of a given TFPG node expression (either
        failure mode or discrepancy). This is done by negating the property
        associated to the node, given as a parameter to this function, and
        checking whether this invariant property universally holds, possibly
        within a given bound.

        The following cases are then possible:
          1. node is reachable (witness is found): returns 'reachable'.
          2. node is not reachable (bounded): returns 'not_reachable_within_bound'.
          3. node is not reachable (unbounded): returns 'not_reachable'.

        @param node_expr: the name of the variable to be checked
        @type  node_expr: string

        @param analysis_bound: the bound for bmc analysis (None for unbounded)
        @type  analysis_bound: int / None
        """

        # generate invariant property

        invar_prop = "!(%s)" % node_expr

        # build model in model-checker instance

        self.send_init_commands(
            build_bdd=False,
            build_bmc=(not self.needs_msat()),
            build_msat=self.needs_msat())

        # run the relevant invariant checking procedure

        check_res = None
        if analysis_bound is None:
            check_res = self.verify_prop(
                invar_prop, 'invarspec', MC_ENGINE_IC3, 0, gen_trace=False)
        elif self.needs_msat():
            check_res = self.verify_prop(
                invar_prop, 'invarspec', MC_ENGINE_MSAT_BMC, analysis_bound, gen_trace=False)
        else:
            check_res = self.verify_prop(
                invar_prop, 'invarspec', MC_ENGINE_BMC, analysis_bound, gen_trace=False)

        if check_res['result'] == 'error':
            raise Exception("Error encountered during reachability analysis of "
                            "TFPG node.")

        assert check_res['result'] in ('true', 'false', 'unknown')

        # process and return result

        if check_res['result'] == 'true':
            # all reachable states satisfy the negation of the node definition
            # property; this means that the node is not reachable.
            return "not_reachable"

        elif check_res['result'] == 'false':
            # a state can be reached (see witness) where the node definition
            # property is true, i.e. its negation is violated; this means that
            # the node is reachable.
            return "reachable"

        else:
            # the negation of the node property is not violated within the
            # analysis bound, which means that the node is not reachable within
            # the given bound.
            return "not_reachable_within_bound"

    def flatten_and_dump(self, out_smv, in_smv=None):
        """Asks xSAP to parse the SMV file, flatten it, and dump the result
        to the specified output file.

        @param out_smv: destination for dump
        @type  out_smv: string

        @param in_smv: input SMV file (optional)
        @type  in_smv: string
        """

        self.send_cmd('set on_failure_script_quits')

        if in_smv is not None:
            self.reset()
            self.read_model(model=in_smv)
        else:
            self.read_model()

        self.send_cmd('flatten_hierarchy')
        self.write_flat_model(out_smv)

        return

    # ============= MODEL CHECKING PRIMITIVES

    def send_init_commands(self, build_bdd=False, build_bmc=False,
        build_msat=False):

        # maybe some timeouts occurred or the user stopped the analysis
        if not self.is_alive():
            # For Windows: it does not kills automatically also the
            # threads created by this process
            self.kill()
            # If something goes wrong, try to print the line error.
            self.search_and_print_errors(self.get_stderr_text())
            return False

        # at least one needs to be set
        assert build_bdd or build_bmc or build_msat, self.kill_and_close_streams()
        # if msat is set, the others shouldn't be set
        assert (not build_msat) or (not build_bdd and not build_bmc), self.kill_and_close_streams()

        if self._need_msat and (build_bdd or build_bmc):
            print(" >> error: incompatible engine selected")
            return False

        self.send_cmd('set on_failure_script_quits')

        if build_bdd:
            self.build_bdd_fsm()
            if not self.is_alive():
                print(" >> error: building bdd-fsm failed")

        if build_bmc:
            self.build_bmc_fsm()
            if not self.is_alive():
                print(" >> error: building bmc-fsm failed")

        if build_msat:
            self.build_msat_fsm()
            if not self.is_alive():
                print(" >> error: building msat-fsm failed")

        if not self.is_alive():
            # For Windows: it does not kills automatically also the
            # threads created by this process
            self.kill_and_close_streams()
            # If something goes wrong, try to print the line error.
            self.search_and_print_errors(self.get_stderr_text())
            return False

        return True

    def verify_prop(self, prop, proptype, engine, bound=10, gen_trace=True):

        # res
        #   'result': false/true/unknown/error
        #   'trace': path of counterexample

        # check sanity of parameters

        assert isinstance(bound, int)
        assert bound >= 0

        # assert that process is running

        assert self.is_alive()

        # send model checking command to xsap instance

        if engine == MC_ENGINE_MSAT_BMC and proptype == 'ltlspec':
            log.xsap.tfpg_behav.debug(
                "Sending 'msat_check_ltlspec_sbmc_inc' command to xSAP.")
            (res, trace_fname) = self.model_check_msat_bmc_ltl(prop, bound, generate_trace=gen_trace)

        elif engine == MC_ENGINE_BMC and proptype == 'ltlspec':
            log.xsap.tfpg_behav.debug(
                "Sending 'check_ltlspec_sbmc_inc' command to xSAP.")
            assert options.BMC_USE_SBMC
            (res, trace_fname) = self.model_check_bmc_ltl(prop, bound, generate_trace=gen_trace)

        elif engine == MC_ENGINE_MSAT_BMC and proptype == 'invarspec':
            log.xsap.tfpg_behav.debug(
                "Sending 'msat_check_invar_bmc' command to xSAP.")
            (res, trace_fname) = self.model_check_msat_bmc_invar(prop, bound,
                                                                 use_inc=True,
                                                                 generate_trace=gen_trace)

        elif engine == MC_ENGINE_BMC and proptype == 'invarspec':
            log.xsap.tfpg_behav.debug(
                "Sending 'check_invar_bmc_inc' command to xSAP.")
            (res, trace_fname) = self.model_check_bmc_invar(prop, bound, generate_trace=gen_trace)

        else:
            assert engine == MC_ENGINE_IC3 and proptype == 'invarspec'
            log.xsap.tfpg_behav.debug(
                "Sending 'check_invar_ic3' command to xSAP.")
            (res, trace_fname) = self.model_check_ic3_invar(prop, bound, generate_trace=gen_trace)

        # retrieve results

        # result_patterns = [
        #     "is false",
        #     "is true",
        #     "is unknown",
        #     "no counterexample or proof found for invariant .* up to %d" %
        #     bound,
        #     "no counterexample found for invariant .* up to %d" %
        #     bound,
        #     "no counterexample found with bound %d" %
        #     bound,
        #     "no proof or counterexample found with bound %d" %
        #     bound,
        #     "terminating with bound %d" %
        #     bound]

        # grep = TextLineGrepper(r".*\s+(%s)" % '|'.join(result_patterns))
        # (m, txt) = grep.match(self)

        return {'result': res, 'trace': trace_fname}


    # ======================================================================
    # FAULT TREES
    # ======================================================================

    def compute_dyn_fault_tree(self, property,
                               bmc_length=options.BMC_LENGTH,
                               out_fname=None):
        """Compute a Fault Tree for a given propositional property,
        outputs the result into a file (one for events and one for gates)
        and returns the names of the generated files (events, gates, xml).
        Raises Exception if a problem in execution is detected.

        @param property: a propositional SMV expression.
        @type property: string
        @param bmc_length: number of transitions used in the hybrid case and
                           BMC
        @type bmc_length: int
        """
        return self.compute_fault_tree(property,
                                       dynamic=True,
                                       bmc_length=bmc_length,
                                       out_fname=out_fname)

    def compute_fault_tree(self, property, dynamic=False,
                           bmc_length=options.BMC_LENGTH,
                           use_bdd=True,
                           use_bdd_bmc=False,
                           sat_sbmc=False,
                           sbmc_completeness=False,
                           compute_probabilities=False,
                           use_param=False,
                           use_param_bmc_ic3=False,
                           fails_nbr=None,
                           out_fname=None,
                           force_smt=False,
                           symbolic=False,
                           quick=False,
                           intermediate=False,
                           gen_trace=False,
                           gen_random_prefix=True,
                           disable_ft_generation=False):
        """Compute a Fault Tree for a given propositional property,
        outputs the result into a file (one for events and one for gates)
        and returns the names of the generated files (events, gates, xml).
        Raises Exception if a problem in execution is detected.

        @param property: a propositional SMV expression.
        @type  property: string

        @param dynamic: a flag to generate a DFT
        @type  dynamic: boolean

        @param bmc_length: number of transitions used in the hybrid case
                           and BMC
        @type  bmc_length: int

        @param use_bdd: a flag to generate FT using BDDs (default)
        @type  use_bdd: boolean

        @param sat_sbmc: a flag to select SBMC SAT algorithms
        @type  sat_sbmc: boolean

        @param sbmc_completeness: a flag to enable completeness check
        @type  sbmc_completeness: boolean

        @param out_fname: the name of the file where to store results
        @type  out_fname: string

        @param force_smt: forces the use of smt for infinite models
                          (to be used with param)
        @type  force_smt: boolean

        @param symbolic: flag to enable symbolic probabilities
        @type  symbolic: boolean

        @param quick: flag to speed up the computation
        @type  quick: boolean

        @param intermediate: flag to generate intermediate fault trees
        @type intermediate: boolean

        @param gen_trace: flag to print counterex traces
        @type  gen_trace: boolean

        @param gen_random_prefix: flag to add a random prefix to the generated files
        @type  gen_random_prefix: boolean

        @param disable_ft_generation: flag to disable generation of FT files
        @type  disable_ft_generation: boolean
        """
        # note for developers: here are a list of possible commands to xSAP:
        # compute_fault_tree -t -x prefix_
        #                    "root.sc_Ctrl.sc_error_bit_hardware = 1"
        # compute_fault_tree_bmc -t -x prefix_ -k 12
        #                    "root.sc_Ctrl.sc_error_bit_hardware = 1"
        # compute_fault_tree_bmc_inc -t -x prefix_ -k 12
        #                    "root.sc_Ctrl.sc_error_bit_hardware = 1"
        # compute_fault_tree_sbmc_inc -t -x prefix_ -k 12
        #                    "root.sc_Ctrl.sc_error_bit_hardware = 1"

        # produce random name

        if gen_random_prefix:
            digits = string.ascii_letters + string.digits + "_"
            rnd_name = random.choice(string.ascii_letters) + \
            ''.join((random.choice(digits) for x in range(5)))
        else:
            rnd_name = ""
        rnd_prefix = self._fname_prefix + rnd_name

        if self.needs_msat():
            self.build_msat_fsm()
            if dynamic:
                log.xsap.user.warning(
                    "xSAPProcess: dynamic hybrid FT are not supported, "
                    "falling back to static")

            if use_param:
                self.compute_fault_tree_param(compute_probabilities,
                                              fails_nbr,
                                              property,
                                              rnd_prefix,
                                              out_fname,
                                              symbolic,
                                              quick,
                                              intermediate,
                                              gen_trace,
                                              force_smt,
                                              disable_ft_generation)
            elif use_param_bmc_ic3:
                self.compute_fault_tree_param_bmc_ic3(bmc_length,
                                                      compute_probabilities,
                                                      fails_nbr,
                                                      property,
                                                      rnd_prefix,
                                                      out_fname,
                                                      symbolic,
                                                      quick,
                                                      intermediate,
                                                      gen_trace,
                                                      force_smt,
                                                      disable_ft_generation)

            else:
                self.compute_fault_tree_msat_bmc(bmc_length,
                                                 rnd_prefix,
                                                 property,
                                                 compute_probabilities,
                                                 out_fname,
                                                 symbolic,
                                                 quick,
                                                 gen_trace,
                                                 disable_ft_generation)

        elif use_bdd: #  BDD FT computation (default)
            self.build_bdd_fsm()
            self.compute_fault_tree_bdd(dynamic,
                                        rnd_prefix,
                                        property,
                                        compute_probabilities,
                                        out_fname,
                                        symbolic,
                                        quick,
                                        gen_trace,
                                        disable_ft_generation)

        elif use_bdd_bmc:
            self.build_bdd_fsm()
            self.build_bmc_fsm()

            if sat_sbmc:
                self.compute_fault_tree_sbmc_inc_bdd(bmc_length,
                                                     rnd_prefix,
                                                     property,
                                                     compute_probabilities,
                                                     out_fname,
                                                     symbolic,
                                                     quick,
                                                     gen_trace,
                                                     disable_ft_generation)
            else:
                self.compute_fault_tree_bmc_inc_bdd(bmc_length,
                                                    rnd_prefix,
                                                    property,
                                                    compute_probabilities,
                                                    out_fname,
                                                    symbolic,
                                                    quick,
                                                    gen_trace,
                                                    disable_ft_generation)
        else:
            self.build_bmc_fsm()

            if use_param:
                self.build_msat_fsm()
                self.compute_fault_tree_param(compute_probabilities,
                                              fails_nbr,
                                              property,
                                              rnd_prefix,
                                              out_fname,
                                              symbolic,
                                              quick,
                                              intermediate,
                                              gen_trace,
                                              force_smt,
                                              disable_ft_generation)

            elif use_param_bmc_ic3:
                self.build_msat_fsm()
                self.compute_fault_tree_param_bmc_ic3(bmc_length,
                                                      compute_probabilities,
                                                      fails_nbr,
                                                      property,
                                                      rnd_prefix,
                                                      out_fname,
                                                      symbolic,
                                                      quick,
                                                      intermediate,
                                                      gen_trace,
                                                      force_smt,
                                                      disable_ft_generation)

            else:
                sbmc_closure = sat_sbmc and sbmc_completeness

                if sat_sbmc:
                    self.compute_fault_tree_sbmc_inc(bmc_length, sbmc_closure,
                                                     rnd_prefix, property,
                                                     compute_probabilities,
                                                     out_fname,
                                                     symbolic,
                                                     quick,
                                                     gen_trace,
                                                     disable_ft_generation)
                else:
                    # Note that in this case sbmc_closure is always False
                    # because in order to be true it requires sat_sbmc to be True
                    # which would make the flow go in the previous if
                    self.compute_fault_tree_bmc_inc(bmc_length,
                                                    rnd_prefix, property,
                                                    compute_probabilities,
                                                    out_fname,
                                                    symbolic,
                                                    quick,
                                                    gen_trace,
                                                    disable_ft_generation)

        if not disable_ft_generation:
            return (rnd_prefix + "events.txt",
                    rnd_prefix + "gates.txt",
                    rnd_prefix + "ft.xml")
        return (None, None, None)

    def compute_fault_tree_msat_bmc(self, bmc_len, prefix, prop,
                                    compute_probabilities,
                                    out_fname, symbolic, quick,
                                    gen_trace, disable_ft_generation):

        cmd = 'compute_fault_tree_msat_bmc -k %d -x "%s"' % (bmc_len, prefix)

        if not disable_ft_generation:
            cmd += ' -t'
        if compute_probabilities:
            cmd += ' -p'
        if out_fname:
            cmd += ' -o "%s"' % out_fname
        if symbolic:
            cmd += ' -S'
        if quick:
            cmd += ' -Q'
        if gen_trace:
            cmd += ' -e'
        cmd += ' "%s"' % prop

        self.execute_command_and_check(cmd, 'END OF COMPUTE_FAULT_TREE_MSAT_BMC')
        return

    def compute_fault_tree_bdd(self, dynamic, prefix, prop,
                               compute_probabilities,
                               out_fname, symbolic, quick,
                               gen_trace, disable_ft_generation):

        cmd = "compute_fault_tree -x %s" % prefix

        if not disable_ft_generation:
            cmd += ' -t'
        if dynamic:
            cmd += " -d"
        if compute_probabilities:
            cmd += " -p"
        if out_fname:
            cmd += ' -o "%s"' % out_fname
        if symbolic:
            cmd += ' -S'
        if quick:
            cmd += ' -Q'
        if gen_trace:
            cmd += ' -e'
        cmd += ' "%s"' % prop

        self.execute_command_and_check(cmd, 'END OF COMPUTE_FAULT_TREE_BDD')
        return

    def compute_fault_tree_sbmc_inc(self, bmc_len, sbmc_closure, prefix, prop,
                                    compute_probabilities, out_fname, symbolic, quick,
                                    gen_trace, disable_ft_generation):

        cmd = 'compute_fault_tree_sbmc_inc -k %d -x "%s"' % (bmc_len, prefix)

        if not disable_ft_generation:
            cmd += ' -t'
        if sbmc_closure:
            cmd += ' -c'
        if compute_probabilities:
            cmd += ' -p'
        if out_fname:
            cmd += ' -o "%s"' % out_fname
        if symbolic:
            cmd += ' -S'
        if quick:
            cmd += ' -Q'
        if gen_trace:
            cmd += ' -e'
        cmd += ' "%s"' % prop

        self.execute_command_and_check(cmd, 'END OF COMPUTE_FAULT_TREE_SBMC_INC')
        return

    def compute_fault_tree_bmc_inc(self, bmc_len, prefix, prop,
                                   compute_probabilities, out_fname, symbolic, quick,
                                   gen_trace, disable_ft_generation):

        cmd = 'compute_fault_tree_bmc_inc -k %d -x "%s"' % (bmc_len, prefix)

        if not disable_ft_generation:
            cmd += ' -t'
        if compute_probabilities:
            cmd += ' -p'
        if out_fname:
            cmd += ' -o "%s"' % out_fname
        if symbolic:
            cmd += ' -S'
        if quick:
            cmd += ' -Q'
        if gen_trace:
            cmd += ' -e'
        cmd += ' "%s"' % prop

        self.execute_command_and_check(cmd, 'END OF COMPUTE_FAULT_TREE_BMC_INC')
        return

    def compute_fault_tree_sbmc_inc_bdd(self, bmc_len, prefix, prop,
                                        compute_probabilities, out_fname,
                                        symbolic, quick, gen_trace, disable_ft_generation):

        cmd = 'compute_fault_tree_sbmc_inc_bdd -k %d -x "%s"' % (bmc_len, prefix)

        if not disable_ft_generation:
            cmd += ' -t'
        if compute_probabilities:
            cmd += ' -p'
        if out_fname:
            cmd += ' -o "%s"' % out_fname
        if symbolic:
            cmd += ' -S'
        if quick:
            cmd += ' -Q'
        if gen_trace:
            cmd += ' -e'
        cmd += ' "%s"' % prop

        self.execute_command_and_check(cmd, 'END OF COMPUTE_FAULT_TREE_SBMC_INC_BDD')
        return

    def compute_fault_tree_bmc_inc_bdd(self, bmc_len, prefix, prop,
                                       compute_probabilities, out_fname,
                                       symbolic, quick, gen_trace, disable_ft_generation):

        cmd = 'compute_fault_tree_bmc_inc_bdd -k %d -x "%s"' % (bmc_len, prefix)

        if not disable_ft_generation:
            cmd += ' -t'
        if compute_probabilities:
            cmd += ' -p'
        if out_fname:
            cmd += ' -o "%s"' % out_fname
        if symbolic:
            cmd += ' -S'
        if quick:
            cmd += ' -Q'
        if gen_trace:
            cmd += ' -e'
        cmd += ' "%s"' % prop

        self.execute_command_and_check(cmd, 'END OF COMPUTE_FAULT_TREE_BMC_INC_BDD')
        return

    def compute_fault_tree_param(self, compute_probabilities, nr_fails, prop,
                                 prefix, out_fname, symbolic, quick,
                                 intermediate, gen_trace, force_smt=False,
                                 disable_ft_generation=False):

        cmd = 'compute_fault_tree_param -x "%s"' % prefix

        if not disable_ft_generation:
            cmd += ' -t'
        if compute_probabilities:
            cmd += ' -p'
        if nr_fails:
            cmd += ' -N %d' % nr_fails
        if out_fname:
            cmd += ' -o %s' % out_fname
        if force_smt:
            cmd += ' -i'
        if symbolic:
            cmd += ' -S'
        if quick:
            cmd += ' -Q'
        if intermediate:
            cmd += ' -I'
        if gen_trace:
            cmd += ' -e'
        cmd += ' "%s"' % prop

        self.execute_command_and_check(cmd, 'END OF COMPUTE_FAULT_TREE_PARAM')
        return

    def compute_fault_tree_param_bmc_ic3(self, bmc_length, compute_probabilities, nr_fails, prop,
                                         prefix, out_fname, symbolic, quick,
                                         intermediate, gen_trace, force_smt=False,
                                         disable_ft_generation=False):

        cmd = 'compute_fault_tree_param -x "%s" -k %d' % (prefix, bmc_length)

        if not disable_ft_generation:
            cmd += ' -t'
        if compute_probabilities:
            cmd += ' -p'
        if nr_fails:
            cmd += ' -N %d' % nr_fails
        if out_fname:
            cmd += ' -o %s' % out_fname
        if force_smt:
            cmd += ' -i'
        if symbolic:
            cmd += ' -S'
        if quick:
            cmd += ' -Q'
        if intermediate:
            cmd += ' -I'
        if gen_trace:
            cmd += ' -e'
        cmd += ' "%s"' % prop

        self.execute_command_and_check(cmd, 'END OF COMPUTE_FAULT_TREE_PARAM_BMC_IC3')
        return

    def compute_fault_tree_param_klive(self, compute_probabilities, prop,
                                       out_fname, symbolic, quick,
                                       intermediate, gen_trace,
                                       disable_ft_generation):

        cmd = 'compute_fault_tree_param_klive -x "%s"' % self._fname_prefix

        if not disable_ft_generation:
            cmd += ' -t'
        if compute_probabilities:
            cmd += ' -p'
        if out_fname:
            cmd += ' -o "%s"' % out_fname
        if symbolic:
            cmd += ' -S'
        if quick:
            cmd += ' -Q'
        if intermediate:
            cmd += ' -I'
        if gen_trace:
            cmd += ' -e'
        cmd += ' "%s"' % prop

        self.execute_command_and_check(cmd, 'END OF COMPUTE_FAULT_TREE_PARAM_KLIVE')
        return

    def execute_command_and_check(self, cmd, end_msg):
        self.send_cmd(cmd)
        self.send_cmd(f'echo {end_msg}')

        grep = TextLineGrepper(end_msg, ["No property named .+"])
        (m, txt) = grep.match(self)

        if m is None or grep.is_match_unwanted(m):
            raise ValueError(txt)

    # ======================================================================
    # MODE TRANSITION CUT SETS
    # ======================================================================

    def compute_mode_transition_cut_sets(self, use_vars, non_monotone,
                                         use_strict, single_source, use_smt,
                                         layering, xml_out, visual_out, paging,
                                         exprs, prefix):

        if self.needs_msat():
            self.build_msat_fsm()
        else:
            self.build_bmc_fsm()

        status = "Error"
        delim = "" if prefix == "" else "/"
        index = "X" if paging else ""
        xml_file = '%s%soutput.xml' % (prefix, delim)
        dot_file = '%s%soutput%s.dot' % (prefix, delim, index)
        tex_file = '%s%soutput.tex' % (prefix, delim)
        expr_str = ' %s' % " ".join([ '"' + x + '"' for x in exprs])
        # [-h] [-v] [-M] [-s] [-e] [-i] [-L] [-t] [-o format] [-g] [-x]
        cmd = 'compute_mode_transition_cut_sets '

        if use_vars:
            cmd += ' -v'
        if non_monotone:
            cmd += ' -M'
        if use_strict:
            cmd += ' -s'
        if single_source:
            cmd += ' -e'
        if use_smt:
            cmd += ' -i'
        if layering:
            cmd += ' -L'
        if xml_out:
            cmd += ' -t'
        if visual_out:
            cmd += ' -o "%s"' % visual_out
        if paging:
            cmd += ' -g'
        cmd += ' -x %s' % prefix
        cmd += expr_str

        self.send_cmd(cmd)

        self.send_cmd("echo END OF COMPUTE_MODE_TRANSITION_CUT_SETS")

        grep = TextLineGrepper("END OF COMPUTE_MODE_TRANSITION_CUT_SETS")
        (m, txt) = grep.match(self)

        if m is None or m.groups() is None:
            status = "Error"

        if "Mode transition cut sets successfully computed." in txt:
            status = "OK"

        if "Internal error in ComputeModeTransitionCutSets" in txt:
            status = "Error"

        return (status, xml_file,
                dot_file if visual_out == "dot" else None,
                tex_out if visual_out == "tex" else None)

    # ======================================================================
    # FMEA TABLES
    # ======================================================================

    def compute_dyn_fmea_table(self, property_list, fault_cardinality,
                               bmc_length=options.BMC_LENGTH,
                               _compact=options.XSAP_COMPACT_FMEA_TABLE):
        """Compute a FMEA (Failure Mode and Effects Analysis) Table
        for a given set of propositional property,
        outputs the result into a file
        and returns the names of the generated files.
        Raises Exception if a problem in execution is detected.

        @param porperty_list: a list of properties
        @type porperty_list: list of strings

        @param fault_cardinality: a natural number specifying cardinality
            of a set of fault configurations. A positive number.
        @type fault_cardinality: int

        @param bmc_length: number of transitions used in the hybrid case
                           and BMC
        @type bmc_length: int
        """
        return self.compute_fmea_table(property_list, fault_cardinality,
                                       dynamic=True, bmc_length=bmc_length,
                                       compact=_compact)

    def compute_fmea_table(self, property_list, fault_cardinality,
                           dynamic=False, bmc_length=options.BMC_LENGTH,
                           compact=options.XSAP_COMPACT_FMEA_TABLE,
                           engine=options.FMEA_ENGINE_BDD):
        """Compute a FMEA (Failure Mode and Effects Analysis) Table
        for a given set of propositional property,
        outputs the result into a file
        and returns the names of the generated files.
        Raises Exception if a problem in execution is detected.

        @param property_list: a list of properties
        @type porperty_list: list of strings

        @param fault_cardinality: a natural number specifying cardinality
            of a set of fault configurations. A positive number.
        @type fault_cardinality: int

        @param dynamic: a flag to generate a FMEA
        @type dynamic: boolean

        @param bmc_length: number of transitions used in the hybrid case
                           and BMC
        @type bmc_length: int

        @param compact: whether compact fmea table (without
        redundances) has to be generated. This works only if dynamic
        is False. If dynaic is True, compact is ignored.
        @type compact: boolean
        """
        # note for developers: a typical xSAP command to compute FMEA table
        # compute_fmea_table -t -x prefix_ -N 2 \
        #         "root.sc_Ctrl.sc_error_bit_hardware = 1" \
        #         "root.sc_Ctrl.sc_error_bit_hardware != 1" \
        #         "root.sc_Ctrl.mode = mode_completely_failed"

        if compact and not dynamic:
            compact_opt = "-c"
        else:
            if compact and dynamic:
                log.xsap.user.warning(
                    "xSAPProcess: compact & dynamic FMEA are not supported, "
                    "falling back to dynamic only")
            compact_opt = ""
            pass

        if self.needs_msat() or engine == options.FMEA_ENGINE_MSAT:
            if dynamic:
                log.xsap.user.warning(
                    "xSAPProcess: dynamic hybrid FMEA are not supported, "
                    "falling back to static")
            self.build_msat_fsm()
            self.compute_fmea_table_msat_bmc(bmc_length, self._fname_prefix,
                                             fault_cardinality, compact_opt,
                                             '"%s"' % '" "'.join(property_list))

        elif engine == options.FMEA_ENGINE_BDD:
            self.build_bdd_fsm()
            self.compute_fmea_table_bdd(dynamic, self._fname_prefix,
                                        fault_cardinality, compact_opt,
                                             '"%s"' % '" "'.join(property_list))
        else:
            assert engine == options.FMEA_ENGINE_BMC
            self.build_bmc_fsm()
            self.compute_fmea_table_bmc(bmc_length, self._fname_prefix,
                                        fault_cardinality, compact_opt,
                                             '"%s"' % '" "'.join(property_list))
            pass

        return self._fname_prefix + "fmea_table.txt"

    def compute_fmea_table_bdd(self, dynamic, prefix, cardinality, compact_opt,
                               props_list):

        cmd = 'compute_fmea_table -t %s -x "%s" -N %d %s %s' % \
            ("-d" if dynamic else "",
             prefix, cardinality,
             compact_opt, props_list)
        self.send_cmd(cmd)
        self.wait_for_command_finish(
            end_message="END OF COMPUTE_FMEA_TABLE")
        return

    def compute_fmea_table_msat_bmc(self, bmc_len, prefix, cardinality,
                                    compact_opt, props_list):

        cmd = 'compute_fmea_table_msat_bmc -k %d -t -x "%s" -N %d %s %s' % \
            (bmc_len, prefix, cardinality, compact_opt, props_list)
        self.send_cmd(cmd)
        self.wait_for_command_finish(
            end_message="END OF COMPUTE_FMEA_TABLE_MSAT_BMC")
        return

    def compute_fmea_table_bmc(self, bmc_len, prefix, cardinality,
                               compact_opt, props_list):

        cmd = 'compute_fmea_table_bmc_inc -k %d -t -x "%s" -N %d %s %s' % \
              (bmc_len, prefix, cardinality, compact_opt, props_list)
        self.send_cmd(cmd)
        self.wait_for_command_finish(
            end_message="END OF COMPUTE_FMEA_TABLE_BMC_INC")
        return

    # ======================================================================
    # FDIR
    # ======================================================================

    def compute_fault_detection(self, property,
                                bmc_length=options.BMC_LENGTH,
                                delay_counter_var=None, detection_delay=None,
                                out_fname=None):
        """Perform Fault Detection Analysis and
        returns a set of alarms for a given propositional property (a fault).
        Outputs the set of alarm into file
        and returns the name of the generated file.
        If empty set is returned then the fault is not detectable.
        Raises Exception if a problem in execution is detected.
        @param property: a fault, i.e. a propositional properties
        @type property: string
        @param bmc_length: number of transitions used in the hybrid case
                           and BMC
        @type bmc_length: int
        @param delay_counter_var: a variable that counts the time/steps after
               the occurrence of some property of interest (incompatible with
               parameter 'property'; requires parameter 'detection_delay')
        @type delay_counter_var: string
        @param detection_delay: a bound within which the alarm has to be
               raised for being considered as a valid fault detection means
               (requires parameter 'delay_counter_var')
        @type detection_delay: double
        """
        # note for developers: a typical xSAP command to run Fault Detection
        # compute_fault_detection -t -x prefix_ \
        #         "root.sc_Ctrl.sc_error_bit_hardware = 1"

        assert (detection_delay is None and delay_counter_var is None) or \
            (detection_delay is not None and delay_counter_var is not None)
        assert (detection_delay is None or detection_delay == -1 or
                detection_delay >= 0)

        if self.needs_msat():
            self.build_msat_fsm()
            self.compute_fault_detection_msat_bmc(bmc_length,
                                                  self._fname_prefix,
                                                  detection_delay,
                                                  delay_counter_var,
                                                  out_fname,
                                                  property)
        else:
            assert delay_counter_var is None
            self.build_bdd_fsm()
            self.compute_fault_detection_bdd(self._fname_prefix,
                                             out_fname, property)
            pass

        return self._fname_prefix + "detection_means.txt"

    def compute_fault_detection_bdd(self, prefix, out_fname, prop):

        cmd = 'compute_fault_detection -t -x "%s" ' % prefix
        if out_fname is not None:
            cmd += '-o "%s" ' % out_fname
        cmd += '"%s"' % prop
        self.send_cmd(cmd)
        return

    def compute_fault_detection_msat_bmc(self, bmc_len, prefix,
                                         detection_delay,
                                         delay_counter_var,
                                         out_fname, prop):
        if detection_delay is None:
            assert delay_counter_var is None
        else:
            assert delay_counter_var is not None

        cmd = 'compute_fault_detection_msat_bmc -k %d -t -x "%s" %s %s %s' %\
            (bmc_len, prefix,
             ("-d %f" % detection_delay) if delay_counter_var else "",
             ("-c %s" % delay_counter_var) if delay_counter_var else "",
             ('"%s"' % prop) if delay_counter_var is None else "")
        self.send_cmd(cmd)
        return

    def compute_fault_isolation(self, number_of_alarms,
                                bmc_length=options.BMC_LENGTH):
        """Perform Fault Isolation Analysis and
        generates a set of Fault Trees, one for each alarm.
        Returns the set of generated file names (set of pairs: events file
        and gates files).
        Raises Exception if a problem in execution is detected.
        @param number_of_alarms: the number of alarms (int)
        @param bmc_length: number of transitions used in the hybrid case
                           and BMC
        @type bmc_length: int
        """
        # note for developers: a typical xSAP command to run Fault Isolation
        #compute_fault_isolation -t -x prefix_
        if self.needs_msat():
            self.build_msat_fsm()
            self.compute_fault_isolation_msat_bmc(bmc_length,
                                                  self._fname_prefix)
        else:
            self.build_bdd_fsm()
            self.compute_fault_isolation_bdd(self._fname_prefix)
            pass

        # create a list of generated files.
        res = []
        for num in range(1,number_of_alarms+1) :
            res.append(("%s_%d_events.txt" % (self._fname_prefix, num), \
                        "%s_%d_gates.txt" % (self._fname_prefix, num)))
            pass
        return res

    def compute_fault_isolation_msat_bmc(self, bmc_len, prefix):

        cmd = 'compute_fault_isolation_msat_bmc -k %d -t -x "%s"' % \
            (bmc_len, prefix)
        self.send_cmd(cmd)
        return

    def compute_fault_isolation_bdd(self, prefix):

        cmd = 'compute_fault_isolation -t -x "%s"' % prefix
        self.send_cmd(cmd)
        return

    def print_sensor_configurations(self, xsap_output_txt, f_out, no_print_stdout):
        """ Extracts sensor configurations from xSAP output"""

        lines = xsap_output_txt.split('\n')

        header = "\nMinimal Sets of Observables\n"
        f_out.write(header + '\n')
        if not no_print_stdout:
            print(header)

        reading_sc = False
        for line in lines:
            line = line.strip()

            assert "No activation makes system diagnosable!" not in line
            assert "No solution exists!" not in line

            if re.search(r'\d+\) -------------------------', line):
                assert not reading_sc
                reading_sc = True
                f_out.write("%s\n" % line)
                if not no_print_stdout:
                    print(line)
                continue

            if not reading_sc: continue

            match = re.search(r'\s*cost: (\d+)',line)
            if match:
                line_out = "   %s\n" % line
                f_out.write(line_out + '\n')
                if not no_print_stdout:
                    print(line_out)
                reading_sc = False
                continue

            match = re.search(r'\s*> (\S+)', line)
            assert match
            line_out = "   %s" % line
            f_out.write(line_out + '\n')
            if not no_print_stdout:
                print(line_out)

        assert not reading_sc

    def synthesize_FD(self, asl_file, obs_file, target_smv, combined,
                      composition_semantics, no_dag):
        # exit immediately on errors
        self.send_cmd('set on_failure_script_quits')
        if composition_semantics == "a":
            self.send_cmd('set synth_asynchronous_composition_semantics 1')

        # build model
        log.xsap.user.debug('Building model.')
        self.build_bdd_fsm()

        # call synthesis
        log.xsap.user.debug('Calling synth_FD.')
        # ... synth_FD -o <> -a <>
        self.send_cmd("synth_FD -o %s -f %s" % (obs_file, asl_file))

        # handle result
        self.send_cmd("echo SYNTH DONE")
        grep = TextLineGrepper("SYNTH DONE")
        (m, txt) = grep.match(self)

        if m is None or m.groups() is None:
            print(txt)
            return False

        if "0 states." in txt:
            print(txt)
            return False

        synth_dump_fdir_cmd = "synth_dump_fdir -o %s" % target_smv
        if combined:
            synth_dump_fdir_cmd += " -c"
        if no_dag:
            synth_dump_fdir_cmd += " -v"
        self.send_cmd(synth_dump_fdir_cmd)

        grep = TextLineGrepper("Synth_DumpFDIR completed.")
        (m, txt) = grep.match(self)
        if m is None or m.groups() is None:
            print(txt)
            return False
        return True


    def optimize_observables_ASL(self, diag_cond, alarm_pattern, delay_bound,
                                 ltl_context, obs_file, asl_file, engine,
                                 no_print_stdout):

        # exit immediately on errors
        self.send_cmd('set on_failure_script_quits')

        # build twin plant with new SMV file
        log.xsap.user.debug('Building twin plant.')
        self.build_twin_plant(obs_file, None, use_history_vars=True,
                              async_composition=False, asl_spec=True)

        # build model
        log.xsap.user.debug('Building model.')
        if self.needs_msat():
            self.build_msat_fsm()
        else:
            self.build_bmc_fsm()

        # call synthesis
        log.xsap.user.debug('Calling synthesis procedure for diagnosability.')
        if asl_file is None:
            res_txt = self.diag_optimize_observables_asl(alarm_pattern,
                                                         None if "finite" == alarm_pattern else delay_bound,
                                                         diag_cond,
                                                        ltl_context,
                                                         engine)
        else:
            res_txt = self.diag_optimize_observables_asl_from_spec(asl_file, engine)

        # handle result
        self.send_cmd("echo DIAGSYNTH DONE")
        grep = TextLineGrepper("DIAGSYNTH DONE")
        (m, txt) = grep.match(self)

        if asl_file is not None and res_txt is not None:
                txt = res_txt + txt

        if m is None or m.groups() is None:
            return ("Error", None)

        if "No solution exists!" in txt:
            return ("NotDiag", None)

        if "Diagnosable under any set of observables!" in txt:
            return ("AllDiag", None)

        basename = os.path.splitext(os.path.basename(self._smv_fname))[0]
        res_filename = "%s_minimal.txt" % basename

        if os.path.exists(res_filename):
            (res_filehandle, res_filename) = tempfile.mkstemp(suffix='.txt', prefix=basename, dir='.')
            f_out = os.fdopen(res_filehandle, "w")
        else:
            f_out = open(res_filename, "w")

        self.print_sensor_configurations(txt, f_out, no_print_stdout)
        f_out.close()

        return ("OK", res_filename)

    def check_diagnosability_ASL(self, diag_cond, alarm_pattern, delay_bound,
                                 ltl_context, obs_file, engine, bmc_len, asl_file):

        # sanity checks
        assert not self.needs_msat() or engine in (MC_ENGINE_MSAT_BMC, MC_ENGINE_IC3)

        # exit immediately on errors
        self.send_cmd('set on_failure_script_quits')

        # build twin plant with new SMV file
        log.xsap.user.debug('Building twin plant.')
        self.build_twin_plant(obs_file, None, use_history_vars=True,
                              async_composition=False, asl_spec=False)

        # generate LTL property
        log.xsap.user.debug('Generating LTL proof obligation for diagnosabiltiy.')
        if asl_file is None:
            self.diag_load_asl_spec(alarm_pattern,
                                    None if "finite" == alarm_pattern else delay_bound,
                                    diag_cond,
                                    ltl_context)
        else:
            alarm_pattern, txt = self.diag_load_asl_spec_from_file(asl_file)
            
        # List of patterns, where absence of critical pairs does not imply diagnosability
        incomplete_patterns = ["bounded", "BoundDel", "finite", "FiniteDel"]

        # build model
        log.xsap.user.debug('Building model.')
        if self.needs_msat():
            self.build_msat_fsm()
        elif MC_ENGINE_BDD == engine:
            self.build_bdd_fsm()
        else:
            self.build_bmc_fsm()

        # Call verification for different indexes
        diag_results = []
        for i in range(self.get_properties_count()):
            (res, trace_A_fn, trace_B_fn) = self._check_diagnosability_property(i, engine, bmc_len)
            # The absence of critical pair does not prove diagnosability
            if (res == "True") and (alarm_pattern in incomplete_patterns):
              # Cannot decide considering only critical pairs
              res = "Unknown-inc"
            diag_results.append((res, trace_A_fn, trace_B_fn))
            if res == "Error":
                return diag_results
        return diag_results

    def _check_diagnosability_property(self, prop_idx, engine, bmc_len):
        # call verification
        log.xsap.user.debug('Calling verification for diagnosability proof obligation %d.' % prop_idx)
        if MC_ENGINE_IC3 == engine:
            (res, trace_fname) = self.model_check_klive_ltl(None, None, force_smt_use=self.needs_msat(), prop_idx=prop_idx)

        elif self.needs_msat():
            assert MC_ENGINE_MSAT_BMC == engine
            (res, trace_fname) = self.model_check_msat_bmc_ltl(None, bmc_len, prop_idx=prop_idx)

        elif MC_ENGINE_BDD == engine:
            (res, trace_fname) = self.model_check_bdd_ltl(None, prop_idx=prop_idx)

        else:
            assert MC_ENGINE_BMC == engine
            assert not options.BMC_USE_SBMC
            (res, trace_fname) = self.model_check_bmc_ltl(None, bmc_len, prop_idx=prop_idx)

        # handle result

        if res == "true":
            return ("True", None, None)

        elif res == "false":
            (trace_A_fn, trace_B_fn) = self.split_tracefile(trace_fname)
            # delete twin-plant trace, keep only pair of traces
            os.remove(trace_fname)
            return ("False", trace_A_fn, trace_B_fn)

        elif res == 'unknown':
            return ("Unknown", None, None)

        else:
            assert res == 'error'
            return ("Error", None, None)

    def split_tracefile(self, twin_trace_fn):
        """This method takes a trace as generated by xSAP in
        check_diagnosability(), and splits them into two traces,
        hiding internal implementation details of xSAP.
        The filenames of the two new trace files are returned.
        @param trace_filename: the filename of the original trace
        @type trace_filename: string
        """

        trace_base_name = os.path.splitext(os.path.basename(twin_trace_fn))[0]
        trace_L_fn = os.path.join(os.path.dirname(twin_trace_fn),
                                  trace_base_name + '_A.xml')
        trace_R_fn = os.path.join(os.path.dirname(twin_trace_fn),
                                  trace_base_name + '_B.xml')

        out_L = open(trace_L_fn, 'w')
        out_R = open(trace_R_fn, 'w')

        trace_in = open(twin_trace_fn, 'r')
        for line in trace_in:
            if 'twin_L.' in line:
                out_L.write('%s' % line.replace('twin_L.',''))
            elif 'twin_R.' in line:
                out_R.write('%s' % line.replace('twin_R.',''))
            elif '<value variable' in line:
                continue
            else:
                out_L.write('%s' % line)
                out_R.write('%s' % line)

        out_L.close()
        out_R.close()

        return (trace_L_fn, trace_R_fn)

    def check_diagnosability(self, property, context, observables,
                             path_restrictions, delta_t_var_name=None):
        """Performs Diagnosability Check and returns ("True",null,null) if
        the two given propositional properties are guaranteed to be
        distinguishable given a set of observable variables and a set
        of (zero or more) path restrictions (specified in chronological order),
        where the latter restrict the state space to be considered.
        In case the two properties are not distinguishable, two traces
        are generated, where under the same observable behavior one ends up
        in property_A and the other in property_B. The return value is
        ("False", trace_A_fn, trace_B_fn), with trace_A/B
        being the names of the xml trace files.
        In the case of hybrid models with an infinite domain, not BDD but BMC
        is used; if no counterexample is found, the return value is
        ("Unknown", None, None), and in case it is, the answer's again
        ("False", trace_A_fn, trace_B_fn).
        @param property: a propositional smv formula
        @type property: string
        @param context: a propositional smv formula
        @type context: string
        @param observables: all flattened smv names of the observable variables
        @type observables: list of strings
        @param path_restrictions: propositional properties which have to apply
                                  to states of the FSM in the given order
                                  before diagnosability is considered (order
                                  in list is important; length of list
                                  can be zero, as this is an optional
                                  parameter)
        @type path_restrictions: list of strings
        """

        # create temporary observables file
        obs_filename = self._fname_prefix + "diag_obs.tmp"
        log.xsap.user.debug(
            'Creating helper file for observables (%s).' % obs_filename )
        obs_file = open(obs_filename, "w")
        for obs in observables:
            obs_file.write(obs + "\n")
        obs_file.close()

        # build the twin plant of the model
        self.build_twin_plant(obs_filename, delta_t_var_name)

        # create temporary file for diagnosability conditions (xSAP format)
        diagcond_filename = self._fname_prefix + "diag_cond.tmp"
        log.xsap.user.debug(
            'Creating helper file for diagnosability '
            'condition (%s).' % diagcond_filename )
        cond_file = open(diagcond_filename, "w")
        cond_file.write(
            "CONDITION_L (" + context + " & (" + property + "))\n")
        cond_file.write(
            "CONDITION_R (" + context + " & !(" + property + "))\n")
        cond_file.write("CONTEXT_R !(" + property + ")\n")
        for res in path_restrictions:
            cond_file.write("RESTRICTION (" + res + ")\n")
        cond_file.close()

        # load the diagnosability condition into the property manager
        # nb: all properties within the property manager are removed
        # first (with '-p')
        self.load_diag_conditions(diagcond_filename, options.DIAG_DELAY)

        # with hybrid models, bmc has to be used
        if self.needs_msat():
            self.build_msat_fsm()
            if options.DIAG_DELAY == -1.0:
                options.BMC_USE_SBMC = False
                (res, trace_fname) = self.model_check_msat_bmc_ltl(
                    None, options.BMC_LENGTH)
            elif options.DIAG_DELAY >= 0.0:
                (res, trace_fname) = self.model_check_msat_bmc_invar(
                    None, options.BMC_LENGTH, use_inc=True)

        # discrete models are dealt with using bdd
        else:
            self.build_bdd_fsm()
            (res, trace_fname) = self.model_check_bdd_ltl(None)

        # return results
        if res == "error":
            return ('Error', None, None)

        elif res == 'true':
            return ('True', None, None)

        elif res == 'unknown':
            return ('Unknown', None, None)

        else:
            assert res == 'false'
            (trace_A_fn, trace_B_fn) = self.split_tracefile(trace_fname)
            return ('False', trace_A_fn, trace_B_fn)


    def build_twin_plant(self, obs_filename, delta_t_var_name,
                         use_history_vars=True,
                         async_composition=True,
                         asl_spec=False,
                         c1_c2_spec=False):

        cmd  = 'build_twin_plant -o "%s"' % obs_filename
        if delta_t_var_name:
            cmd += ' -t "%s"' % delta_t_var_name
        if async_composition:
            cmd += ' -a'
        if use_history_vars:
            cmd += ' -p'
        if asl_spec:
            cmd += ' -A'
        if c1_c2_spec:
            cmd += ' -f'

        self.send_cmd(cmd)
        return

    def load_diag_conditions(self, diagcond_fname, diag_delay):

        if diag_delay == -1.0: # NEEDS LTL CHECKING
            cmd = 'load_diag_conditions -p -c %s -d -1' % (diagcond_fname)
        elif diag_delay >= 0.0: # NEEDS INVAR CHECKING
            cmd = 'load_diag_conditions -p -c "%s" -d %f -r' % \
                (diagcond_fname, diag_delay)
        else:
            ## we should never reach this code!!
            raise ProcessInternalException(
                self.name,
                "Delay must be >= 0 or 'FINITE'")
        self.send_cmd(cmd)
        return


    def diag_optimize_observables_asl(self, alarm_pattern, delay,
                                      diag_cond, context, engine):

        cmd = 'diag_optimize_observables_asl'
        if alarm_pattern:
            cmd += ' -a "%s"' % alarm_pattern
        if delay is not None:
            cmd += ' -d %d' % delay
        if diag_cond:
            cmd += ' -c "%s"' % diag_cond
        if context:
            cmd += ' -x "%s"' % context
        if engine:
            cmd += ' -e %s' % engine

        self.send_cmd(cmd)
        return

    def diag_optimize_observables_asl_from_spec(self, asl_file, engine):
        cmd = 'diag_optimize_observables_asl -f %s' % asl_file
        if engine:
            cmd += ' -e %s' % engine

        self.send_cmd(cmd)

        grep = TextLineGrepper(r"cost: \d+")
        m, txt = grep.match(self)
        if m is None:
            log.xsap.user.error("Process failed on command:\n" + cmd)
            self._log_returned_text(txt)
            return None

        return txt

    def diag_load_asl_spec(self, alarm_pattern, delay, diag_cond, context):

        cmd = 'diag_load_asl_spec -p'
        if alarm_pattern:
            cmd += ' -a "%s"' % alarm_pattern
        if delay:
            cmd += ' -d %d' % delay
        if diag_cond:
            cmd += ' -c "%s"' % diag_cond
        if context:
            cmd += ' -x "%s"' % context
        self.send_cmd(cmd)
        return

    def diag_load_asl_spec_from_file(self, asl_file, reset_props=True):
        cmd = 'diag_load_asl_spec '
        if reset_props:
            cmd += '-p '
        cmd += '-f %s' % asl_file
        self.send_cmd(cmd)

        grep = TextLineGrepper(
            "Generated (BoundDel|ExactDel|FiniteDel) proof obligation.")
        m, txt = grep.match(self)  # wait for result!
        if m is None:
            log.xsap.user.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return None

        # Return also matched alarm pattern
        return m.groups()[0], txt

    pass  # end of class


    # ===
    # TODO: the following commands are offered by xSAP but there is
    #       no related interface
    # ===
    # check_diagnosability
    # compute_fault_tree_bmc
    # compute_fault_tree_bmc_bdd
    # compute_fmea_table_bmc_inc
    # compute_optimal_observables
    # compute_pareto_optimal_frontier
    # compute_prime_implicants
    # compute_prima_implicants_param
    # fame_FR
    # fame_dump_change
    # fame_dump_result_xml
    # fame_dump_slim
    # fame_dump_stats
    # fame_parse_xml
    # fame_test_fdirconf
    # fame_test_sys2tfpg
    # fe_extend_module
    # fe_load_doc
    # generate_allocations
    # simplify_twin_plant
    # synth_FD
    # synth_FR
    # synth_annotate_belief_explorer
    # synth_build_belief_explorer
    # synth_build_plan
    # synth_build_recoverer
    # synth_compute_obs_filter
    # synth_dump_change
    # synth_dump_diagnoser
    # synth_dump_fdir
    # synth_dump_plan
    # synth_dump_result_xml
    # synth_edit_alarms_spec
    # synth_edit_controllables
    # synth_edit_observables
    # synth_edit_plans_spec
    # synth_extend_with_history_vars
