#!/usr/bin/env python3
# common utilities for scripts

import platform
import logging
import sys
import os.path
import glob

import _paths


def get_exec_fullname(name="xSAP"):
    """Given generic name (e.g. "xSAP") provides the
    full name of the executable, considering platform
    and bits"""

    # AMa2RC: The following two return statements are wrong: What if
    # on windows, or different SO? Just return "xsap"??
    if platform.system() == "Linux":
        if os.path.isfile(name):
            return name

        if os.path.isfile(os.path.join(_paths.BIN_DIR, name)):
            return os.path.join(_paths.BIN_DIR, name)

    bits = int(platform.architecture()[0][0:-3])
    try:
        system, exeext = {'Linux': ('linux', '',),
                          'Windows': ('win', '.exe'),
                          'Darwin': ('mac', '')}[platform.system()]

    except KeyError:
        logging.error("System was not recognized: %s", platform.system())
        sys.exit(1)

    # now searches if it is available
    path = os.path.join(_paths.BIN_DIR, "%s-%s%d%s" %
                        (name, system, bits, exeext))
    if os.path.isfile(path):
        return path

    # if 64 bit not found, try with 32:
    if bits == 64:
        path = os.path.join(_paths.BIN_DIR, "%s-%s%d" % (name, system, 32))
        if os.path.isfile(path):
            logging.warning("Using 32 bit version for executable %s, "
                            "even if bits are %d", name, bits)
            return path

    # no lucky: raise an exception
    raise IOError("Not found a valid executable for %s, "
                  "platform is %s, bits are %d" %
                  (name, system, bits))


def get_suitable_smv_file(smv_fname, outdir):
    """Searches for a good smv filename (only one canditate must match
    to succeeed) if non specified"""

    if smv_fname and os.path.isfile(smv_fname):
        return smv_fname

    if smv_fname:
        logging.error("Specified smv file was not found: '%s'" % smv_fname)
        return None

    files = glob.glob(os.path.join(outdir, "extended_*.smv"))

    if len(files) == 0:
        logging.error("No suitable smv file was specified or found")
        return None

    if len(files) > 1:
        logging.error("Multiple suitable smv files were found")
        logging.error(", ".join(files))
        logging.error("Please specify one only, with option --smv-file")
        return None

    return files.pop()


def get_suitable_fms_file(fms_fname, smv_fname, outdir):
    """Searches for a good fms filename (only one canditate must match
    to succeeed) if non specified"""
    if fms_fname and os.path.isfile(fms_fname):
        return fms_fname

    if fms_fname:
        logging.error("Specified fms file was not found: '%s'" % fms_fname)
        return None

    fms_fname = os.path.join(outdir, get_fms_fname(smv_fname))
    if not os.path.isfile(fms_fname):
        logging.error("No suitable fms file was specified or found")
        logging.error("I have searched for '%s' with no success" % fms_fname)
        return None

    return fms_fname


def get_fms_fname(expanded_smv_fname):
    full_base, ext = os.path.splitext(expanded_smv_fname)
    base = os.path.basename(full_base)
    prefix = "extended_"
    if base.startswith(prefix):
        base = base[len(prefix):]

    return "fms_" + base + ".xml"
