#!/usr/bin/env python3
# Use this script to check all requirements to run scripts and tools

import sys
import _paths


# ----------------------------------------------------------------------
# Requirements predicates
# ----------------------------------------------------------------------
def check_python_version():
    """Python version must be 3.8.x"""
    return (3, 8) <= sys.version_info


def check_platform():
    """Platform not supported"""
    import platform
    is_64bits = sys.maxsize > 2**32
    system = platform.system().lower()
    return (system == "linux" or
            (system == "windows" and is_64bits))


req_packages = ("gi", "lxml", "future", "antlr4", "networkx", "PySide2", "PIL", "cairo", "shiboken2")
# ----------------------------------------------------------------------


def check_import(module_name):
    def __importer():
        try:
            __import__(module_name)
        except ImportError:
            return False
        return True

    __importer.__name__ = "check_import_module__" + module_name
    __importer.__doc__ = "Module '%s' could not be imported" % module_name
    return __importer


def main():
    reqs = (check_python_version,
            check_platform,
            ) + tuple(check_import(mod)
                      for mod in req_packages)

    checked_reqs = [(req(), req) for req in reqs]
    failed = 0
    print("-" * 70)
    for rid, req in enumerate(checked_reqs):
        if not req[0]:
            print("ERROR-{rid} {name}: {message}".format(
                rid="%02d" % (rid+1),
                name=req[1].__name__,
                message=req[1].__doc__))
            failed += 1

        else:
            print("PASSED-{rid} {name}".format(
                rid="%02d" % (rid+1),
                name=req[1].__name__))

    print("-" * 70)
    if failed > 0:
        print("FAILED! There are %d critical requirements unsatisfied" %
              failed)
    else:
        print("SUCCESS! All requirements are satisfied")
    print("-" * 70)


if "__main__" == __name__:
    _paths.setup_path()
    main()
