#!/usr/bin/env python3

from __future__ import print_function
import sys
import os.path
from argparse import ArgumentParser, RawTextHelpFormatter

import _paths
_paths.setup_path()

from xsap.data_structure.tfpg.tfpg_errors import UnsupportedInfinitySemantics
from xsap.data_structure.tfpg.tfpg import OPEN_INFINITY
from xsap.data_structure.tfpg.tfpg import Tfpg
from xsap.features.tfpg_smt.utils import parse_mapping_file, print_model
from xsap.features.tfpg_smt.reasoner import TFPGReasoner


def main():
    parser = ArgumentParser(formatter_class=RawTextHelpFormatter,
                            description="""
TFPG Refinement (SMT):  Checks if a inputTFPG is a refinement of the original TFPG.
""")

    parser.add_argument('inputTFPG', metavar='inputTFPG',
                        type=str,
                        help='The TFPG model to use (TFPG 1)')

    parser.add_argument('--mapping', '-m', metavar='mapping',
                        type=str,
                        help='The refinement mapping')

    parser.add_argument('--refine', '-r', metavar='original',
                        type=str,
                        help='TFPG 2')

    parser.add_argument('--open-infinity', action='store_true',
                        help='Force the usage of Open Infinity semantics')

    args = parser.parse_args()


    # Command line handling
    tfpg = Tfpg.parse_tfpg(args.inputTFPG)
    if tfpg is None:
        sys.exit(1)

    if args.open_infinity:
        tfpg.set_infinity_semantics(OPEN_INFINITY)

    try:
        reasoner = TFPGReasoner(tfpg, include_health=False)
    except UnsupportedInfinitySemantics:
        print("The current version of the algorithm only supports Open Infinity Semantics.")
        print("To force the implicit conversion, use the option --open-infinity")
        sys.exit(1)


    if args.refine:
        original_tfpg = Tfpg.parse_tfpg(args.refine)
        if original_tfpg is None:
            sys.exit(1)

        mapping_f = args.mapping

        if original_tfpg is None:
            print("The given tfpg is invalid")
            sys.exit(1)

        if mapping_f is None:
            print("You must specify the mapping by using the -m option")
            print("See help for further details.")
            sys.exit(1)

        mapping = parse_mapping_file(mapping_f)
        if mapping is None:
            print("The given mapping is invalid")
            sys.exit(1)

        print("Checking the refinement...")
        res, model = reasoner.check_refinement(original_tfpg, mapping)

        if res:
            print("The given TFPG is a refinement of original_tfpg")
        else:
            print("The given TFPG is NOT a refinement of original_tfpg")
            print("A trace that has no refinement is:")
            print_model(model)

    else:
        print("Invalid command line parameters.")
        print("See help for further details.")
        sys.exit(1)

if __name__ == "__main__":
    main()
