#!/usr/bin/env python3
# takes: fms and smv,
# and produces xml, dot or tex output file

import sys
import os.path
import subprocess

import _paths
import _utils


# WARNING! ordering is important here!
FT_ENGINES = "bdd bmc bddbmc msat ic3 bmc_ic3".split()
(FT_ENGINE_BDD,
 FT_ENGINE_BMC,
 FT_ENGINE_BDD_BMC,
 FT_ENGINE_MSAT,
 FT_ENGINE_IC3,
 FT_ENGINE_BMC_IC3 ) = FT_ENGINES



def handle_options():
    """returns (options, input_file):
    values (as a map) and the name of the input file"""

    import argparse
    parser = argparse.ArgumentParser(description="Computes MTCS for given"
                                     " modes, smv file and fms file")

    parser.add_argument("--smv-file",
                        metavar="SMV-FILE",
                        help="The input extended smv file name")

    parser.add_argument("--fms-file",
                        metavar="FMS-FILE",
                        help="The input fault mode xml file name")

    parser.add_argument('--verbose', '-v',
                        action='store_true',
                        default=False,
                        help='Sets verbosity to high level')

    parser.add_argument('--use-vars', '-V',
                        action='store_true',
                        default=False,
                        help='Treats expressions as state mode variables')

    parser.add_argument('--paging', '-g',
                        action='store_true',
                        default=False,
                        help='In visual output, print each transitions separately')

    parser.add_argument("--visual-out", "-o",
                        metavar="FORMAT",
                        help="Visual output format, either tex or dot")

    parser.add_argument("--layering", "-L",
                        action='store_false',
                        default=True,
                        help="Turn off layering")

    parser.add_argument("--single-source", "-e",
                        action='store_true',
                        default=False,
                        help="Compute transitions only from the first mode")

    parser.add_argument('--expressions',
                        nargs=argparse.REMAINDER,
                        required=True,
                        help='List of expressions defining mode variables or modes')

    parser.add_argument("-d", "--outdir",
                        default="out",
                        metavar="PATH",
                        help="Output directory, where all generated file should be "
                        "put into (default:%(default)s)")

    parser.add_argument("--boolean-conversion-uses-predicate-normalization", "-b",
                        dest="boolean_conversion",
                        action="store_true",
                        default=False,
                        help="Enables predicate normalization during boolean conversion")

    args = parser.parse_args()

    for fn in (args.smv_file, args.fms_file):
        if fn and not os.path.isfile(fn):
            parser.error("File not found: " + fn)

    if os.path.exists(args.outdir) and not os.path.isdir(args.outdir):
        parser.error("Not a directory: " + args.outdir)

    return args


if "__main__" == __name__:

    _paths.setup_path()

    import xsap.utils.misc.log
    import xsap.utils.options
    from xsap.utils.processes.xsap_process import XsapProcess

    opts = handle_options()

    if not os.path.exists(opts.outdir):
        os.makedirs(opts.outdir)

    if opts.verbose:
        xsap.utils.options.LOGGING_XSAP_LEVEL = "debug"
        xsap.utils.options.LOGGING_XSAP_CONSOLE_LEVEL = "info"

    xsap.utils.misc.log.init()

    # automatically searches for inputs when not explicitly provided
    smv_fname = _utils.get_suitable_smv_file(opts.smv_file, opts.outdir)
    if smv_fname is None:
        sys.exit(1)

    fms_fname = _utils.get_suitable_fms_file(opts.fms_file,
                                             smv_fname,
                                             opts.outdir)
    if fms_fname is None:
        sys.exit(1)

    stdouterr_fname = os.path.join(opts.outdir, "xsap_compute_mtcs.out")
    if os.path.exists(stdouterr_fname):
        os.remove(stdouterr_fname)

    stdouterr_file = open(stdouterr_fname, "w")

    # instantiate the process
    xsap.utils.misc.log.xsap.user.info("Invoking xsap to compute mode transition cut sets")

    need_msat = XsapProcess.has_infinite_domain_vars(smv_fname)
    with  XsapProcess(smv_fname, fms_fname,
                      ord_fname=None, need_msat=need_msat,
                      args=["-v", "%d" % (5 if opts.verbose else 0)],
                      stdout=subprocess.PIPE,
                      stderr=subprocess.STDOUT,
                      dumpers=[stdouterr_file],
                      boolean_conversion_uses_predicate_normalization=opts.boolean_conversion) as xsap_process:

        # set the prefix
        full_base, _ext = os.path.splitext(smv_fname)
        base = os.path.basename(full_base)
        xsap_process.set_tmp_filename_prefix(opts.outdir + "/" + base)

        stdouterr_file.write("\n" + "-" * 70)
        stdouterr_file.write("\nInvoking: %s %s\n" % (xsap_process.get_path(),
                                                  " ".join(xsap_process.get_args())))
        stdouterr_file.write("\n" + "-" * 70)
        stdouterr_file.write("\n")

        status, xml_file, dot_file, tex_file = xsap_process.compute_mode_transition_cut_sets(
            use_vars=opts.use_vars,
            non_monotone=False,
            use_strict=False,
            single_source=opts.single_source,
            use_smt=need_msat,
            layering=opts.layering,
            xml_out=True,
            visual_out=opts.visual_out,
            paging=opts.paging,
            exprs=opts.expressions,
            prefix=opts.outdir)

    stdouterr_file.close()

    if status == "OK":
        xsap.utils.misc.log.xsap.user.info("xsap produced mode transition cut sets in: ")
        if xml_file:
          xsap.utils.misc.log.xsap.user.info("  xml file: '%s'" % xml_file)
        if dot_file:
          xsap.utils.misc.log.xsap.user.info("  dot file%s: '%s'" % ("s" if opts.paging else "", dot_file))
        if tex_file:
          xsap.utils.misc.log.xsap.user.info("  tex file: '%s'" % tex_file)
    else:
        xsap.utils.misc.log.xsap.user.info("xsap process failed, see xsap_compute_mtcs.out file for more info")
