#!/usr/bin/env python3
from __future__ import print_function

import sys
import os
import re
import subprocess

import _paths


def handle_options():
    """Returns options values as map."""

    import argparse
    parser = argparse.ArgumentParser(description="")

    parser.add_argument("--engine", "-E",
                        dest="engine",
                        choices=["bmc", "bmc_ic3"],
                        default="bmc",
                        help="Use given engine (default: %(default)s)")

    parser.add_argument("--bmc-length", "-k",
                        dest="bmc_len",
                        action="store",
                        type=int,
                        default=10,
                        help="Maximum path length for BMC")

    parser.add_argument("--smv-file", "-m",
                        dest="smv",
                        action="store",
                        default=None,
                        help="SMV file")

    parser.add_argument("--diagnosis-condition", "-c",
                        dest="diag_cond",
                        action="store",
                        default=None,
                        help="Diagnosis condition")

    parser.add_argument("--alarm-pattern", "-a",
                        dest="alarm_pattern",
                        choices=['exact', 'bounded', 'finite'],
                        default=None,
                        help="Alarm pattern (default: %(default)s)")

    parser.add_argument("--delay-bound", "-d",
                        dest="delay_bound",
                        action="store",
                        type=int,
                        default=None,
                        help="Alarm pattern delay bound")

    parser.add_argument("--context-expression", "-x",
                        dest="ltl_context",
                        action="store",
                        default=None,
                        help="LTL context expression")

    parser.add_argument("--observables-file", "-o",
                        dest="obs_file",
                        action="store",
                        default=None,
                        help="File specifying observable variables")

    parser.add_argument("--asl-file", "-f",
                        dest="asl_file",
                        action="store",
                        default=None,
                        help="File containing the alarms specification")

    parser.add_argument("--verbosity-level", "-v",
                        dest="verbose_level",
                        action="store",
                        type=int,
                        default=0,
                        help="Sets the output verbosity level")

    parser.add_argument("--no-stdout-print",
                        dest="no_print_stdout",
                        action="store_true",
                        default=False,
                        help="Result is not printed to stdout.")

    args = parser.parse_args()

    # check arguments

    if args.smv is None or not os.path.exists(args.smv):
        parser.error("The SMV model was not specified.")

    if args.bmc_len < 0:
        parser.error("Illegal BMC length.")

    if args.obs_file is None or not os.path.exists(args.obs_file):
        parser.error("The observables file was not specified.")

    if args.asl_file is None:
        if args.diag_cond is None:
            parser.error("The diagnosis condition was not specified.")

        if args.alarm_pattern is None:
            parser.error("The alarm pattern was not chosen.")

        if args.alarm_pattern in ['exact', 'bounded'] and args.delay_bound is None:
            parser.error("For the chosen alarm pattern a delay bound needs to be specified.")

        if args.alarm_pattern in ['finite'] and args.delay_bound is not None:
            parser.error("A delay bound cannot be specified for finite-delay.")

        if args.delay_bound is not None and int(args.delay_bound) < 0:
            parser.error("Illegal delay bound.")
    else:
        if not os.path.exists(args.asl_file):
            parser.error("The ASL file does not exist")
        if args.diag_cond is not None:
            parser.error("Cannot specify both a diagnosis condition and an ASL file")

    return args


if __name__ == "__main__":

    _paths.setup_path()

    # setup logging
    import xsap.utils.misc.log
    xsap.utils.misc.log.init()
    args = handle_options()

    from xsap.utils.options import MC_ENGINE_MSAT_BMC
    from xsap.utils.processes.xsap_process import XsapProcess

    need_msat = XsapProcess.has_infinite_domain_vars(args.smv) or MC_ENGINE_MSAT_BMC == args.engine

    xsap_process = XsapProcess(args.smv, fail_obs_alm_fname=None,
                               ord_fname=None, need_msat=need_msat,
                               args=["-v", str(args.verbose_level)],
                               stdout=subprocess.PIPE, stderr=subprocess.STDOUT, dumpers=[])

    xsap_process.send_cmd("set bmc_length %d" % args.bmc_len)

    delay_bound = -1 if args.delay_bound is None else int(args.delay_bound)

    (res, res_filename) = xsap_process.optimize_observables_ASL(args.diag_cond,
        args.alarm_pattern, delay_bound, args.ltl_context, args.obs_file,
        args.asl_file, args.engine, args.no_print_stdout)

    xsap_process.kill_and_close_streams()

    if "Error" == res:
        print("--------------------------------------------------")
        print("Result: An error has occurred during the analysis.")
        print("--------------------------------------------------")

    elif "AllDiag" == res:
        print("---------------------------------------------------")
        print("Result: The specification is vacuously diagnosable.")
        print("---------------------------------------------------")

    elif "NotDiag" == res:
        print("------------------------------------------------------------------")
        print("Result: The specification is not diagnosable with all observables.")
        print("------------------------------------------------------------------")

    elif "OK" == res:
        num_dashes = max(len("Result: Minimal sets of observables have been computed"),
                         len(res_filename))
        print("-" * num_dashes)
        print("Result: Minimal sets of observables have been computed")
        print("        and stored in file '%s'." % res_filename)
        print("-" * num_dashes)

    else:
        assert False, "BUG: unexpected result of diagnosability verification."
