#!/usr/bin/env python3
from __future__ import print_function

## Input
## - Associations file -> in SMV
## - SMV model
## - (optional) BMC bound


import sys
import os
import _paths


def handle_options():
    """Returns options values as map."""

    from xsap.utils.options import FT_ENGINES, FT_ENGINE_BMC

    import argparse
    parser = argparse.ArgumentParser(description="")

    parser.add_argument("--associations-file", "-a",
                        dest="assoc",
                        action="store",
                        default=None,
                        help="TFPG SMV associations file")

    parser.add_argument("--smv-file", "-m",
                        dest="smv",
                        action="store",
                        default=None,
                        help="SMV File")

    parser.add_argument("--output-tfpg", "-o",
                        dest="out_tfpg_fname",
                        action="store",
                        default=None,
                        help="Name of the synthesized TFPG")

    parser.add_argument("--engine", "-E",
                        dest="engine",
                        choices=FT_ENGINES,
                        default=FT_ENGINE_BMC,
                        help="Use given engine (default: %(default)s)")

    parser.add_argument("--force-boolean-ic3", "-B",
                        dest="boolic3",
                        action="store_true",
                        default=False,
                        help="Force use of Boolean version of IC3 engine (for finite-state models)")

    parser.add_argument("--bmc-length", "-k",
                        dest="bmc_len",
                        action="store",
                        type=int,
                        default=10,
                        help="Maximum path length for BMC")

    parser.add_argument("--dynamic", "-D",
                        dest="dynamic_reorder",
                        action="store_true",
                        default=False,
                        help="Enables dynamic reordering of variables")

    parser.add_argument("--boolean-conversion-uses-predicate-normalization", "-b",
                        dest="boolean_conversion",
                        action="store_true",
                        default=False,
                        help="Enables predicate normalization during boolean conversion")

    parser.add_argument("--disable-coi-reduction", "-c",
                        dest="no_coi_reduction",
                        action="store_true",
                        default=False,
                        help="Disables TFPG graph simplification based on cone-of-influence")

    parser.add_argument("--prune-unreachable-nodes",
                        dest="prune_unreachable_nodes",
                        action="store_true",
                        default=False,
                        help="Enables a preprocessing step that removes "
                             "unreachable nodes before synthesis is started.")

    parser.add_argument("--verbosity-level", "-v",
                        dest="verbose_level",
                        action="store",
                        type=int,
                        default=0,
                        help="Sets the output verbosity level")

    parser.add_argument("--outdir", "-O",
                        dest="outdir",
                        default="out",
                        metavar="PATH",
                        help="Output directory, where all generated file should be "
                        "put into (default: %(default)s)")

    args = parser.parse_args()

    if args.assoc is None or not os.path.exists(args.assoc):
        parser.error("The TFPG SMV associations file was not specified.")

    if args.smv is None or not os.path.exists(args.smv):
        parser.error("The SMV model was not specified.")

    if args.boolic3 and not args.engine == xsap.utils.options.FT_ENGINE_PARAM:
        parser.error("Option '--force-boolean-ic3' requires IC3 engine.")

    if os.path.exists(args.outdir) and not os.path.isdir(args.outdir):
        parser.error("Not a directory: " + args.outdir)

    if args.bmc_len < 0:
        parser.error("BMC length cannot be negative.")

    return args



if __name__ == "__main__":

    _paths.setup_path()
    # setup logging
    import xsap.utils.misc.log
    xsap.utils.misc.log.init()
    args = handle_options()

    assoc_fname = args.assoc                # The name of the tfpg SMV associations file
    smv_fname   = args.smv                   # The name of the SMV model
    bmc_length  = args.bmc_len               # Maximum path length for BMC
    engine      = args.engine
    boolic3     = args.boolic3
    prune_unreachable_nodes = args.prune_unreachable_nodes

    # create directory for generated files
    outdir = args.outdir
    if not os.path.exists(outdir):
        os.makedirs(outdir)

    if args.out_tfpg_fname is not None:     # The name of the output tfpg
        out_tfpg_fname = args.out_tfpg_fname
    else:
        out_tfpg_fname = os.path.join(outdir, os.path.splitext(os.path.basename(smv_fname))[0] + "_synth_tfpg.txml")

    disable_coi = args.no_coi_reduction

    # xsap options
    xsap_dynamic_reorder    = args.dynamic_reorder
    xsap_boolean_conversion = args.boolean_conversion
    xsap_verbose_level      = args.verbose_level

    assert assoc_fname is not None
    assert smv_fname is not None

    from xsap.data_structure.tfpg.tfpg_associations import TfpgAssociations
    from xsap.data_structure.tfpg.tfpg_errors import UnreachableDiscrepancy
    from xsap.data_structure.tfpg.xml.tfpg import parse_tfpg_xml_file_to_internal_struct
    from xsap.features.tfpg_synthesis.synthesis import perform_tfpg_synthesis

    # parse associations file

    associations = TfpgAssociations.parse_assoc(assoc_fname)
    if associations is None:
        print("Error! Invalid TFPG associations file")
        sys.exit(1)

    num_failure_modes = len(associations.get_var_names_of_failure_modes())
    num_discrepancies = len(associations.get_var_names_of_unmonitored_discr()) + len(associations.get_var_names_of_monitored_discr())
    num_modes = len(associations.get_var_names_of_tfpg_modes())

    if num_discrepancies is 0 or num_failure_modes is 0 or num_modes is 0:
        print("Error! Associations file must contain at least a failure mode, a discrepancy and a system mode")
        sys.exit(1)

    # perform tfpg synthesis

    res = perform_tfpg_synthesis(associations, smv_fname, out_tfpg_fname, bmc_length, engine, disable_coi,
                                 xsap_dynamic_reorder, xsap_boolean_conversion,
                                 xsap_verbose_level, boolic3, outdir,
                                 enable_node_pruning=prune_unreachable_nodes)

    if res == "ok":
        print("  tfpg file:", out_tfpg_fname)
        tfpg, errors = parse_tfpg_xml_file_to_internal_struct(out_tfpg_fname)
        if len(tfpg.get_errors_list()) > 0:
            print("The synthesized TFPG contains some errors!")
            for err in tfpg.get_errors_list():
                if isinstance(err, UnreachableDiscrepancy):
                    print(" ",err)
                else:
                    print(" ",err)
                    raise err
        if len(errors) > 0:
            for err in errors:
                raise err
    else:
        # Print the error message
        print("tfpg synthesis > %s" %res)
