#!/usr/bin/env python3
from __future__ import print_function

## Input:
## - Tfpg file (in XML)
## - Associations file -> in SMV
## - SMV model
## - (optional) BMC bound


import os
import sys

import _paths


def handle_options():
    """Returns options values as map."""

    import argparse
    parser = argparse.ArgumentParser(description="")

    parser.add_argument("--tfpg-file", "-t",
                        dest="tfpg",
                        action="store",
                        default=None,
                        help="The TFPG file (can be in XML or human readable format)")

    parser.add_argument("--associations-file", "-a",
                        dest="assoc",
                        action="store",
                        default=None,
                        help="TFPG SMV associations file")

    parser.add_argument("--smv-file", "-m",
                        dest="smv",
                        action="store",
                        default=None,
                        help="SMV File")

    parser.add_argument("--tags", "-T",
                        dest="tags",
                        action="store",
                        default="tmin;tmax;modes",
                        help="Parameters to be tightened (tmin/tmax/modes; semi-colon separated)")

    parser.add_argument("--delta-t-var", "-d",
                        dest="delta_t_var",
                        action="store",
                        default=None,
                        help="delta t variable name")

    parser.add_argument("--engine", "-E",
                        dest="engine",
                        choices=['conc', 'ia'],
                        default='conc',
                        help="IC3 mode (default: %(default)s)")

    parser.add_argument("--tmax-bound",
                        dest="tmax_bound",
                        action="store",
                        type=int,
                        default=None,
                        help="Upper bound for tmax (for tightening of tmax=infinity).")

    parser.add_argument("--bmc-length", "-k",
                        dest="bmc_len",
                        action="store",
                        type=int,
                        default=10,
                        help="Maximum path length for BMC")

    parser.add_argument("--skip-bmc",
                        dest="skip_bmc",
                        action="store_true",
                        default=False,
                        help="Use only IC3 instead of mixed BMC-IC3")

    parser.add_argument("--ic3-length", "-K",
                        dest="ic3_len",
                        action="store",
                        type=int,
                        default=None,
                        help="Maximum number of frames for IC3")

    parser.add_argument("--outdir", "-O",
                        dest="outdir",
                        default="out",
                        metavar="PATH",
                        help="Output directory, where all generated file should be "
                        "put into (default: %(default)s)")

    args = parser.parse_args()

    if args.tfpg is None or not os.path.exists(args.tfpg):
        parser.error("The TFPG file was not specified.")

    if args.assoc is None or not os.path.exists(args.assoc):
        parser.error("The TFPG SMV associations file was not specified.")

    if args.smv is None or not os.path.exists(args.smv):
        parser.error("The SMV model was not specified.")

    if os.path.exists(args.outdir) and not os.path.isdir(args.outdir):
        parser.error("Not a directory: " + args.outdir)

    chosen_tags = args.tags.split(';')

    for tag in chosen_tags:
        if tag not in ('tmin', 'tmax', 'modes'):
            parser.error("Invalid tag string (%s)." % args.tags)

    if len(chosen_tags) is 0:
        parser.error("At least one tag must be chosen.")

    if args.bmc_len < 1:
        parser.error("'bmc_len' must be a positive integer.")

    assert args.engine in ('conc', 'ia')

    return args


if __name__ == "__main__":

    # paths and logging set up

    _paths.setup_path()

    import xsap.utils.misc.log
    xsap.utils.misc.log.init()

    # parse arguments

    cmd_args = handle_options()

    tfpg_fname = cmd_args.tfpg              # The name of the tfpg model (XML format)
    assoc_fname = cmd_args.assoc            # The name of the tfpg SMV associations file
    smv_fname = cmd_args.smv                # The name of the SMV model

    tmax_bound = cmd_args.tmax_bound        # Upper search bound for tmax values
    delta_t_var = cmd_args.delta_t_var      # delta t variable name
    chosen_tags = cmd_args.tags.split(';')  # parameters to be tightened

    bmc_len = cmd_args.bmc_len              # Maximum path length for BMC
    skip_bmc = cmd_args.skip_bmc            # Flag to skip BMC in engine
    ic3_len = cmd_args.ic3_len              # Maximum path length for IC3
    engine = cmd_args.engine                # IC3 mode (conc / ia)

    outdir = cmd_args.outdir
    if not os.path.exists(cmd_args.outdir):
        os.makedirs(cmd_args.outdir)

    assert tfpg_fname is not None
    assert assoc_fname is not None
    assert smv_fname is not None

    # perform tightening

    print()
    print("TFPG Tightening")
    print()

    tags = []

    if delta_t_var is None:

        if 'modes' not in chosen_tags:
            print('--------------------------------------------------------------')
            print('Message: Nothing to be tightened (time delta variable not set,')
            print('         selected tags do not contain "modes" tag).')
            print('--------------------------------------------------------------')
            sys.exit(0)

        for tag in chosen_tags:
            if tag != 'modes':
                print("  > skipping tightening of parameter '%s'" % tag)

        tags = ['modes']

    else:
        tags = chosen_tags

    options = {"tags": tags, "bmc_len": bmc_len, "ic3_len": ic3_len, "t_delta_var": delta_t_var,
               "tmax_bound": tmax_bound, "engine": engine, "skip_bmc": skip_bmc, "outdir": outdir}

    from xsap.features.tfpg_tightening.tightening import perform_edge_tightening
    fn_tfpg_tight, err_msg = perform_edge_tightening(smv_fname, assoc_fname, tfpg_fname, options)

    # present results

    print()

    if err_msg is None:

        gen_res_string = 'Result: The TFPG has been tightened.'
        res_string = "Output file: " + fn_tfpg_tight

        print('-' * max(len(gen_res_string), len(res_string)))
        print(gen_res_string)
        print(res_string)
        print('-' * max(len(gen_res_string), len(res_string)))

    else:

        gen_res_string = "Result: An error has occurred during the analysis."

        print('-' * max(len(gen_res_string), len(err_msg)+len('Message: ')))
        print(gen_res_string)
        print('Message:', err_msg)
        print('-' * max(len(gen_res_string), len(err_msg)+len('Message: ')))

    print()
