#!/usr/bin/env python3
from __future__ import print_function

## Input:
## - Tfpg file (in XML)
## - Associations file -> in SMV
## - SMV model
## - (optional) BMC bound


import sys
import os
import _paths


def handle_options():
    """Returns options values as map."""

    from xsap.utils.options import MC_ENGINE_MSAT_BMC, MC_ENGINE_IC3

    import argparse
    parser = argparse.ArgumentParser(description="")

    parser.add_argument("--tfpg-file", "-t",
                        dest="tfpg",
                        action="store",
                        default=None,
                        help="The TFPG file (can be in XML or human readable format)")

    parser.add_argument("--associations-file", "-a",
                        dest="assoc",
                        action="store",
                        default=None,
                        help="TFPG SMV associations file")

    parser.add_argument("--smv-file", "-m",
                        dest="smv",
                        action="store",
                        default=None,
                        help="SMV File")

    parser.add_argument("--boolean-conversion-uses-predicate-normalization", "-b",
                        dest="boolean_conversion",
                        action="store_true",
                        default=False,
                        help="Enables predicate normalization during boolean conversion")

    parser.add_argument("--dynamic-reordering", "-D",
                        dest="dynamic_reorder",
                        action="store_true",
                        default=False,
                        help="Enables dynamic reordering of BDD variables")

    parser.add_argument("--delta-t-var", "-d",
                        dest="delta_t_var",
                        action="store",
                        default=None,
                        help="delta t variable name")

    parser.add_argument("--bmc-length", "-k",
                        dest="bmc_len",
                        action="store",
                        type=int,
                        default=10,
                        help="Maximum path length for BMC")

    parser.add_argument("--engine", "-E",
                        dest="engine",
                        choices=[MC_ENGINE_MSAT_BMC, MC_ENGINE_IC3],
                        default=MC_ENGINE_MSAT_BMC,
                        help="Use given engine (default: %(default)s)")

    parser.add_argument("--property-to-validate", "-p",
                        dest="property_to_validate",
                        choices=['completeness', 'edge_tightness'],
                        default='completeness',
                        help="Validate given property (default: %(default)s)")

    parser.add_argument("--monolithic-check", "-M",
                        dest="monolithic_check",
                        action="store_true",
                        default=False,
                        help="Check completeness proof obligations using a single monolithic check")

    parser.add_argument("--generate-only", "-g",
                        dest="generate_only",
                        action="store_true",
                        default=False,
                        help="Only generate files for model checking")

    parser.add_argument("--parametric-dump", "-P",
                        dest="parametric_dump",
                        action="store_true",
                        default=False,
                        help="Dump parameterized completeness property")

    parser.add_argument("--outdir", "-O",
                        dest="outdir",
                        default="out",
                        metavar="PATH",
                        help="Output directory, where all generated file should be "
                        "put into (default: %(default)s)")

    args = parser.parse_args()

    if args.tfpg is None or not os.path.exists(args.tfpg):
        parser.error("The TFPG file was not specified.")
        pass

    if args.assoc is None or not os.path.exists(args.assoc):
        parser.error("The TFPG SMV associations file was not specified.")
        pass

    if args.smv is None or not os.path.exists(args.smv):
        parser.error("The SMV model was not specified.")
        pass

    if args.parametric_dump and args.property_to_validate != 'completeness':
        parser.error("Parametric dumping only possible for completeness check.")
        pass

    if args.parametric_dump and not args.generate_only:
        parser.error("Parametric dumping only possible in generate-only mode.")
        pass

    if os.path.exists(args.outdir) and not os.path.isdir(args.outdir):
        parser.error("Not a directory: " + args.outdir)

    return args


if __name__ == "__main__":

    # set up paths
    _paths.setup_path()

    # set up logging
    import xsap.utils.misc.log
    xsap.utils.misc.log.init()

    # parse arguments
    args = handle_options()

    tfpg_fname = args.tfpg                       # The name of the tfpg model (XML format)
    assoc_fname = args.assoc                     # The name of the tfpg SMV associations file
    smv_fname = args.smv                         # The name of the SMV model
    bmc_length = args.bmc_len                    # Maximum path length for BMC
    engine = args.engine                         # Model checking engine
    delta_t_var = args.delta_t_var               # delta t variable name
    boolean_conversion = args.boolean_conversion # predicate normalization
    dynamic_reorder    = args.dynamic_reorder    # BDD dynamic reordering
    property_to_validate = args.property_to_validate # property to validate (completeness or edge_tightness)
    monolithic_check = args.monolithic_check     # completeness po's as single monolithic check
    generate_only = args.generate_only           # don't solve mc problems
    parametric_dump = args.parametric_dump       # dump parameterized model and proof obligations

    assert tfpg_fname is not None
    assert assoc_fname is not None
    assert smv_fname is not None

    # create directory for generated files
    outdir = args.outdir
    if not os.path.exists(outdir):
        os.makedirs(outdir)

    # parse the TFPG into internal struct
    tfpg_file_ext = os.path.splitext(tfpg_fname)[1]
    tfpg = None

    if tfpg_file_ext in [".txml", ".xml"]:
        from xsap.data_structure.tfpg.xml.tfpg import parse_tfpg_xml_file_to_internal_struct
        tfpg, errors = parse_tfpg_xml_file_to_internal_struct(tfpg_fname)
    elif tfpg_file_ext == ".tfpg":
        from xsap.data_structure.tfpg.txt.tfpg import parse_tfpg_txt_file_to_internal_struct
        tfpg = parse_tfpg_txt_file_to_internal_struct(tfpg_fname)
    else:
        print("Error! Unknown format for the TFPG file: %s." %tfpg_file_ext)
        sys.exit(1)

    if len(tfpg.get_errors_list()) > 0:
        for err in tfpg.get_errors_list():
            raise err
    if len(errors) > 0:
        for err in errors:
            raise err

    # parse associations file
    assoc_file_ext = os.path.splitext(assoc_fname)[1]
    associations = None

    if assoc_file_ext in [".axml", ".xml"]:
        from xsap.data_structure.tfpg.xml.tfpg_associations import parse_assoc_xml_file_to_internal_struct
        associations = parse_assoc_xml_file_to_internal_struct(assoc_fname)
    elif assoc_file_ext == ".tfpga":
        from xsap.data_structure.tfpg.txt.tfpg_associations import parse_assoc_txt_file_to_internal_struct
        associations = parse_assoc_txt_file_to_internal_struct(assoc_fname)
    else:
        print("Error! Unknown format for the TFPG associations file: %s." %assoc_file_ext)
        sys.exit(1)

    if (len(associations.get_var_names_of_unmonitored_discr()) == 0 and len(associations.get_var_names_of_monitored_discr()) == 0) or \
       len(associations.get_var_names_of_failure_modes()) == 0 or \
       len(associations.get_var_names_of_tfpg_modes()) == 0:
        print("Error! Associations file must contain at least a failure mode, a discrepancy and a system mode")
        sys.exit(1)

    # perform behavioral validation
    from xsap.features.tfpg_behavioral_validation.behavioral_validation import perform_behavioral_validation
    res = perform_behavioral_validation(property_to_validate, tfpg, associations, smv_fname, bmc_length, delta_t_var, engine, boolean_conversion, dynamic_reorder, monolithic_check, generate_only, parametric_dump, outdir)

    # present results

    print()

    if res['result'] == "error":
        print("--------------------------------------------------")
        print("Result: An error has occurred during the analysis.")
        print("--------------------------------------------------")

    elif res['result'] == "complete":
        print("-------------------------------------------------------")
        print("Result: The TFPG is complete with respect to the model.")
        print("-------------------------------------------------------")

    elif res['result'] == "complete_bounded":
        print("-------------------------------------------------------------------------------")
        print("Result: The TFPG is complete with respect to the model (within analysis bound).")
        print("-------------------------------------------------------------------------------")

    elif res['result'] == "incomplete":
        assert res['counterexample'] is not None
        print("-----------------------------------------------------------------------------------------------")
        print("Result: The TFPG is incomplete with respect to the model. A pair of traces have been generated.")
        for trace in res['counterexample']:
            print("%s:" % trace, end=' ')
            print(res['counterexample'][trace])
        print("-----------------------------------------------------------------------------------------------")

    elif res['result'] == "tight":
        print("------------------------------------------------------------------")

        if res['tmaxinf_tight'] is None or res['tmaxinf_tight'] == "tight":
            print("Result: The edges of the TFPG are tight with respect to the model.")

        else:
            assert res['tmaxinf_tight'] == "unknown"
            print("Result: The edges of the TFPG are tight with respect to the model,")
            print("        with the exception of some tmax=inf assignments for which")
            print("        tightness couldn't be proven or disproven.")

        print("------------------------------------------------------------------")

    elif res['result'] == "not_tight":
        print("----------------------------------------------------------------------")
        print("Result: The edges of the TFPG are not tight with respect to the model.")
        print("----------------------------------------------------------------------")

    elif res['result'] == "not_tight_bounded":
        print("----------------------------------------------------------------------------------------------")
        print("Result: The edges of the TFPG are not tight with respect to the model (within analysis bound).")
        print("----------------------------------------------------------------------------------------------")

    else:
        message = "Result: %s." % res['result']
        print(len(message) * '-')
        print(message)
        print(len(message) * '-')

    print()
