from __future__ import print_function

# %HEADER

from builtins import range
from builtins import object
from decimal import *  # MG: Be explicit about this import

import os

from xsap.data_structure.tfpg.tfpg_errors import (
    FMNotCorrectError,
    TimeNotCorrectError,
    NodeNotInEdge,
    FMNodeIsNotDestNode,
    NodeLoopError,
    NotExistingModes,
    NotModesInTheEdge,
    InvalidModeName,
    EmptyNodeName)

#from xsap.utils.misc import log

# MG: The use of the decimal package could, in principle, allow the
# representation of the infinity. However, we currently do not exploit.
# Therefore, we define a constant here.
INFINITY = -1.0

# Constants for the semantics of infinity tmax:
# CLOSED : The propagation might not occur in the edges in which tmax=INFINITY
# OPEN: The propagation will occur in the edges in which tmax=INFINITY
CLOSED_INFINITY, OPEN_INFINITY = list(range(2))


class Tfpg(object):
    """Represents a TFPG"""

    def __init__(self):
        """Constructor Method"""
        self.nodes_list = []
        self.edges_list = []
        self.modes_list = []
        self.name = ""
        self.infinity_semantics = CLOSED_INFINITY
        self.errors_list = []
        return

    @staticmethod
    def parse_tfpg(fn_tfpg, ignore_errors=False):
        """ This is a parsing function for TFPG files accepting both XML
        and TXT formats. Error checking is performed.

        @param   fn_assoc: name of file containing the TFPG
        @type    fn_assoc: string

        @param   ignore_errors: parsing proceeds also on the presence of errors
        @type    ignore_errors: bool

        @return: The parsed Tfpg object, or None in case of errors.
        @rtype:  Tfpg / None
        """

        tfpg_file_ext = os.path.splitext(fn_tfpg)[1]
        tfpg_struct = None

        if tfpg_file_ext in [".txml", ".xml"]:
            from xsap.data_structure.tfpg.xml.tfpg import (
                TfpgDocument,
                parse_tfpg_xml_file_to_internal_struct)

            tfpg_doc, validation_errors = TfpgDocument().parse(fn_tfpg)
            if len(validation_errors) > 0:
                err_msg = str(validation_errors[0])
                print("Error! Validation of the TFPG failed (%s)." % err_msg)
                if not ignore_errors:
                    return None

            syntax_errors = TfpgDocument().get_syntax_errors_in_tfpg_document(tfpg_doc)
            if len(syntax_errors) > 0:
                err_msg = str(syntax_errors[0])
                print("Error! The TFPG contains syntax errors (%s)." % err_msg)
                if not ignore_errors:
                    return None

            tfpg_struct, errors = parse_tfpg_xml_file_to_internal_struct(fn_tfpg)
            if not ignore_errors:
                assert len(errors) == 0

        elif tfpg_file_ext == ".tfpg":
            from xsap.data_structure.tfpg.txt.tfpg import (
                parse_tfpg_txt_file_to_internal_struct)

            try:
                tfpg_struct, errors = parse_tfpg_txt_file_to_internal_struct(fn_tfpg)
                if not ignore_errors:
                    assert len(errors) == 0
            except Exception as parsing_exception:
                print("Error! TFPG parsing failed (%s)." % str(parsing_exception))
                #log.xsap.tfpg_other.debug("Exception during TFPG parsing: " \
                #                          "'%s'" % str(parsing_exception))
                return None

        else:
            # unknown file format
            print("Error! Unknown file extension (%s)." % tfpg_file_ext)
            return None

        assert tfpg_struct is not None

        if len(tfpg_struct.get_errors_list()) > 0 and not ignore_errors:
            return None

        return tfpg_struct

    def drop_isolated_discrepancies(self):
        """ Removes all discrepancies that are not involved in any edges. """

        new_nodes_list = []

        for d in self.get_or_nodes() + self.get_and_nodes():
            for e in self.get_edges_list():
                if e.get_srcNode() == d or e.get_destNode() == d:
                    new_nodes_list.append(d)
                    break

        self.nodes_list = new_nodes_list + self.get_fm_nodes()

    def has_non_maximally_permissive_edges(self):
        """ Returns true if at least one edge has tmin>0, or tmax!=inf, or modes!=all,
        and false otherwise. """

        for e in self.edges_list:
            if e.get_tMin() != 0:
                #print "tmin not 0 for the following edge:"
                #print e
                return True
            if e.get_tMax() != INFINITY:
                #print "tmax not infinity for the following edge:"
                #print e
                return True
            if set(e.get_modes()) != set(self.modes_list):
                #print "modes not equivalent with all modes for following edge:"
                #print e
                return True

        return False

    def get_nodes_list(self):
        """Gets the list of the nodes of the TFPG.
            @return   self.nodes_list:  The list of the nodes of the TFPG.
            @rtype    self.nodes_list:  list
        """
        return self.nodes_list

    def get_fm_nodes(self):
        """Gets the list of the failure mode nodes of the TFPG.
            @return   nodes_list:  The list of the failure mode nodes of
                                   the TFPG.
            @rtype    nodes_list:  list
        """
        return [node for node in self.nodes_list
                if node.get_type() == NodeType.FM]

    def get_or_nodes(self):
        """Gets the list of the OR discrepancy nodes of the TFPG.
            @return   nodes_list:  The list of the OR discrepancy nodes of
                                   the TFPG.
            @rtype    nodes_list:  list
        """
        return [node for node in self.nodes_list
                if node.get_type() == NodeType.OR]

    def get_and_nodes(self):
        """Gets the list of the AND discrepancy nodes of the TFPG.
            @return   nodes_list:  The list of the AND discrepancy nodes of
                                   the TFPG.
            @rtype    nodes_list:  list
        """
        return [node for node in self.nodes_list
                if node.get_type() == NodeType.AND]

    def get_monitored_discrepancies(self):
        """Returns the list of monitored discrepancies, i.e. the list of
        monitored AND or OR nodes.
        @return   nodes_list:  The list of monitored discrepancies
                                   of the TFPG.
        @rtype    nodes_list:  list
        """
        return [node for node in self.nodes_list if
                (node.get_type() == NodeType.AND or
                 node.get_type() == NodeType.OR) and
                node.is_monitored() is True]

    def get_unmonitored_discrepancies(self):
        """Returns the list of unmonitored discrepancies, i.e. the list of
        unmonitored AND or OR nodes.
            @return   nodes_list:  The list of unmonitored discrepancies of
                                   the TFPG.
            @rtype    nodes_list:  list
        """
        return [node for node in self.nodes_list if
                (node.get_type() == NodeType.AND or
                 node.get_type() == NodeType.OR) and
                node.is_monitored() is False]

    def get_edges_list(self):
        """Gets the list of the edges of the TFPG.
            @return   self.edges_list:  The list of the edges of the TFPG.
            @rtype    self.edges_list:  list
        """
        return self.edges_list

    def get_timings_constants(self, exclude_infinity):
        """Gets the list of timings constants of the edges of the TFPG.
        @param   exclude_infinity:  If the INFINITY time is excluded or not.
        @type    exclude_infinity:  boolean
        @return   timings:  The list of timings constants of the edges of
                            the TFPG
        @rtype    timings:  list of float
        """

        timings = []

        for e in self.edges_list:

            getcontext().prec = 20
            tmin = Decimal(str(e.get_tMin()))
            tmax = Decimal(str(e.get_tMax()))

            timings.append(tmin)

            if tmax >= 0 or not exclude_infinity:
                timings.append(tmax)

        return timings

    def get_name(self):
        """Gets the name of the TFPG.
            @return   self.name:  The name of the TFPG.
            @rtype    self.name:  string
        """
        return self.name

    def set_name(self, name):
        """Sets the name of the TFPG.
            @param   name:  The name of the TFPG.
            @type    name:  string
        """
        self.name = name
        return

    def has_node_by_name(self, name):
        """Checks whether there is a node with the same name in the Tfpg.
            @param   name:  The name of the Node.
            @type    name:  string

            @return   Boolean:  If there is a node with the same name in
                                the Tfpg.
            @rtype    Boolean:  boolean
        """
        for node in self.nodes_list:
            if node.get_name() == name:
                return True

        return False

    def get_node_by_name(self, name):
        """If exists, it returns the node having the passed parameter as name
           Notice that this function is useful while converting a TfpgDocument
           into a tfpg internal struct, as in the case of a TfpgDocument, the
           source node of an edge is represented as a string (the name of the
           node)
            @param  name: the name of the Node.
            @type   name: string

            @return node: The Node or None.
            @rtype  node: Node
        """
        for node in self.nodes_list:
            if node.get_name() == name:
                return node

        return None

    def has_edge_by_name(self, name):
        """Check whether the is an edge with the same name in the Tfpg.
            @param  name: the name of the Edge.
            @type   name: string

            @return Boolean: True in case there is an edge with the same in
                             the Tfpg
            @rtype  Boolean: boolean
        """
        for edge in self.edges_list:
            if edge.get_name() == name:
                return True

        return False

    def get_edge_by_name(self, name):
        """Returns the edge with the given name, if it exists.
            @param  name: the name of the Edge.
            @type   name: string

            @return edge: The Edge or None.
            @rtype  edge: Edge
        """
        for edge in self.edges_list:
            if edge.get_name() == name:
                return edge

        return None

    def get_edge(self, srcNode, destNode):
        """Returns the edge going from src to dst.
            @param  srcNode:  The source node.
            @type   name: Node
            @param  srcNode:  The destination node.
            @type   name: Node

            @return edge: The Edge or None.
            @rtype  edge: Edge
         """
        for edge in self.edges_list:
            if edge.get_srcNode() == srcNode and \
                    edge.get_destNode() == destNode:
                return edge
        return None

    def get_predecessors(self, node):
        """ Returns the list of predecessors of the given node.
        @param node: The node for which we return the predecessors.
        @type node: Node

        @rtype: list of Node
        """

        srcs = []
        for edge in self.get_edges_list():
            if edge.destNode == node:
                srcs.append(edge.srcNode)

        return srcs

    def add_fm_node(self, name):
        """Adds a Failure Mode Node.
            If errors are found, they are added to the global errors_list
            and None is returned
            @param   name:  The name of the Node.
            @type    name:  string

            @return   node:  The Failure Mode Node.
            @rtype    node:  Node
        """
        # it does not make sense to let the user specify also the isMonitored
        # parameter, since for Failure Modes it must always be False

        try:
            node = Node(NodeType.FM, name, False)
            self.nodes_list.append(node)
            return node

        except Exception as e:
            self.errors_list.append(e)
            return None

    def add_or_node(self, name, isMonitored):
        """Adds an OR Discrepancy Node.
            @param   name:  The name of the Node.
            @type    name:  string
            @param   isMonitored:  If the Node is monitored.
            @type    isMonitored:  bool

            @return   node:  The OR Discrepancy Node.
            @rtype    node:  Node
        """
        try:
            node = Node(NodeType.OR, name, isMonitored)
            if node.get_node_errors_list() == []:
                self.nodes_list.append(node)
                return node
            else:
                # MG: This code does not do what the programmer expected.
                for err in node.get_node_errors_list():
                    raise err

        except Exception as e:
            self.errors_list.append(e)
            return None

    def add_and_node(self, name, isMonitored):
        """Adds an AND Discrepancy Node.
            @param   name:  The name of the Node.
            @type    name:  string
            @param   isMonitored:  If the Node is monitored.
            @type    isMonitored:  bool

            @return   node:  The AND Discrepancy Node.
            @rtype    node:  Node
        """
        try:
            node = Node(NodeType.AND, name, isMonitored)
            if node.get_node_errors_list() == []:
                self.nodes_list.append(node)
                return node
            else:
                # MG: Same as above. This code is incorrect.
                for err in node.get_node_errors_list():
                    raise err

        except Exception as e:
            self.errors_list.append(e)
            return None

    def check_edge(self, edge):
        """Checks whether an edge has been properly built.
            In particular, it performs the following checks:
            - the srcNode must exist
            - the destNode must exist
            - the modes must exist
            - the destNode must not be a FM node

            @param   edge: the edge to be checked
            @type    edge: class Edge

            @return  errors_list: the (possibly empty) list of errors
            @rtype   errors_list: list of Exceptions
        """

        errors_list = []

        srcNode = edge.get_srcNode()
        destNode = edge.get_destNode()
        modes = edge.get_modes()
        name = edge.get_name()

        try:
            if not self.has_node_by_name(srcNode.get_name()):
                raise NodeNotInEdge(srcNode.get_name(), name)

            elif not self.has_node_by_name(destNode.get_name()):
                raise NodeNotInEdge(destNode.get_name(), name)

            elif self.has_modes_list(modes):
                raise NotModesInTheEdge(name)

            elif destNode.get_type() == NodeType.FM:
                raise FMNodeIsNotDestNode(destNode.name, name)

            elif self.get_unknown_modes(modes) != []:
                raise NotExistingModes(self.get_unknown_modes(modes))

        except Exception as e:
            errors_list.append(e)

        return errors_list

    def add_edge(self, srcNode, tMin, tMax, modes, destNode, name):
        """Adds an Edge.
            @param   srcNode:  The source node of the Edge.
            @type    srcNode:  class Node
            @param   tMin:  The initial time of the Edge.
            @type    tMin:  float
            @param   tMax:  The final time of the Edge.
            @type    tMax:  float
            @param   modes:  The list of modes of the Edge.
            @type    modes:  list
            @param   destNode:  The destination node of the Edge.
            @type    destNode:  class Node
            @param   name:  The name of the Edge.
            @type    name:  string

            @return   node:  The Edge.
            @rtype    node:  Edge
            @return   error_list:  The list of errors.
            @rtype    error_list:  list
        """

        # the init method of the Edge class implicitely checks that the Edge
        # values are valid. Additionally, other checks must be performed with
        # the check_edge() function

        new_edge = Edge(srcNode, tMin, tMax, modes, destNode, name)
        for err in new_edge.get_edge_errors_list():
            self.errors_list.append(err)

        error_list = self.check_edge(new_edge)

        for error in error_list:
            self.errors_list.append(error)

        if self.errors_list == []:
            self.edges_list.append(new_edge)
            return new_edge

        else:
            return None

    def get_modes_list(self):
        """Gets the list of the active modes of the TFPG.
            @return   self.modes_list:  The list of the active modes of
                                        the TFPG.
            @rtype    self.modes_list:  list of strings
        """
        return self.modes_list

    def set_modes_list(self, modes_list):
        """Sets the list of the modes of the TFPG.

           A check is performed within this function; there cannot be modes
           having the same name of a node/edge

            @param   modes_list:  The list of the modes of the TFPG.
            @type    modes_list:  list of strings
        """

        try:
            for md in modes_list:
                if self.has_node_by_name(str(md)):
                    raise InvalidModeName(str(md), isNode=True)
                elif self.has_edge_by_name(str(md)):
                    raise InvalidModeName(str(md), isNode=False)

            self.modes_list = modes_list

        except Exception as e:
            self.errors_list.append(e)

        return

    def has_mode(self, mode):
        """Checks whether a mode is contained in the modes list.
            @param   mode:  The mode of the modes_list.
            @type    mode:  string
        """
        for md in self.get_modes_list():
            if str(md) == str(mode):
                return True
        return False

    def has_modes_list(self, modes):
        """Checks whether a modes is not empty.
            @param   modes:  The modes_list.
            @type    modes:  list of string
        """
        for md in modes:
            if md is None:
                return True
        return False

    def get_unknown_modes(self, modes):
        """Returns the list of modes which are not contained in the structural tfpg.
        For every mode contained in 'modes', check whether it is contained in
        the list of modes belonging to the structural tfpg; the list of
        not-contained modes is returned. In case all the modes are contained in
        the strucutural tfpg, an empty list is returned.
        @param modes: the list of modes to be checked
        @type  modes: list of string
        """
        unknown_modes = []

        for md in modes:
            if self.has_mode(md) is False:
                unknown_modes.append(md)

        return unknown_modes

    def get_errors_list(self):
        """Gets the list of the errors of the TFPG.
            @return   self.errors_list:  The list of the errors of the TFPG.
            @rtype    self.errors_list:  list
        """
        return self.errors_list

    def remove_node(self, node_name):
        """ Removes node with given node_name from nodes_list. Also drops any edge involving the
        node. """

        node = self.get_node_by_name(node_name)

        if node is None:
            return

        # remove all edges involving the node
        new_edges = []
        for e in self.edges_list:
            if e.get_srcNode() is not node and e.get_destNode() is not node:
                new_edges.append(e)
        self.edges_list = new_edges

        # remove the node itself from the node list
        self.nodes_list.remove(node)

        # validate result: no node with given name must exist
        node = self.get_node_by_name(node_name)
        assert node is None

        # validate result: no edge must connect to a node with given name
        # or a node that has been removed.
        for e in self.edges_list:

            n_src = e.get_srcNode()
            assert n_src in self.nodes_list
            assert n_src.get_name() != node_name

            n_dest = e.get_destNode()
            assert n_dest in self.nodes_list
            assert n_dest.get_name() != node_name

    def is_node_reachable_from_fm_node(self, nd, avoid_nodes):
        """Check whether there exists a path starting from a failure
        mode reaching the given node"""

        # base case for recursion
        if nd.get_type() == NodeType.FM:
            return True

        # check if this node was already visited to
        # avoid infinite loops
        if nd in avoid_nodes:
            return False

        # recursive call on all source nodes of edges
        # leading to the current node
        incoming_edges = self.get_edges_from_dest_node(nd.get_name())

        # to avoid infinite loops
        avoid_nodes.append(nd)

        for e in incoming_edges:

            if self.is_node_reachable_from_fm_node(e.get_srcNode(),
                                                   avoid_nodes):
                return True

            # not reachable from fm nodes, no point in trying again;
            avoid_nodes.append(e.get_srcNode())
            # MG: This code is useless. srcNode has already been added during
            #     its visit

        return False

    def compute_reachable_from_set(self, node_names):
        """ Computes the set of TFPG nodes from which the nodes with the
        given names is reachable from, following the TFPG edges. Nb: the
        result will also contain all of the given node names. """

        # assert that all node names are in the TFPG (they can be both
        # FM or D nodes)

        assert isinstance(node_names, list)

        for name in node_names:
            if self.get_node_by_name(name) is None:
                print("Error: Node '%s' not found in TFPG." % name)
                return None

        # initialize result to empty set

        reachable_from_set = set()

        # recursive helper function

        def compute_reachable_from_set_aux(cur_node):
            # cur_node is a Node instance

            # if cur_node is already in set
            if cur_node.get_name() in reachable_from_set:
                # just return
                pass

            # if cur_node is an FM
            elif cur_node.get_type() == NodeType.FM:
                # add to set and return
                reachable_from_set.add(cur_node.get_name())

            else:
                # we must be dealing with an AND or OR node
                assert cur_node.get_type() == NodeType.AND or \
                       cur_node.get_type() == NodeType.OR

                # add node name to result set
                reachable_from_set.add(cur_node.get_name())

                # recurse on all source nodes of incoming edges
                for pred_node in self.get_predecessors(cur_node):
                    compute_reachable_from_set_aux(pred_node)

        # perform the computation for all nodes in the input list

        for name in node_names:
            compute_reachable_from_set_aux(self.get_node_by_name(name))

        return reachable_from_set

    def get_edges_from_dest_node(self, node_name):
        """Returns a list of the edges having the specified node as
        destination"""
        edges = []
        for eg in self.get_edges_list():
            if eg.get_destNode().get_name() == node_name:
                edges.append(eg)
        return edges

    def clear(self):
        """Clears the graph, deleting all its nodes and all its edges"""
        del self.nodes_list[0:len(self.nodes_list)]
        del self.edges_list[0:len(self.edges_list)]
        self.name = ""
        return

    def set_infinity_semantics(self, new_semantics):
        """Sets the semantics for the INFINITY in tmax.

        @param new_semantics: The semantics to set
        @type  new_semantics: One of OPEN_INFINITY or CLOSED_INFINITY constant
        """

        if new_semantics not in [OPEN_INFINITY, CLOSED_INFINITY]:
            raise ValueError(
                "Invalid Infinity Semantics type: %s" % new_semantics)
        self.infinity_semantics = new_semantics

    def has_closed_infinity_semantics(self):
        """Returns whether the TFPG uses a closed infinity semantics."""
        return self.infinity_semantics == CLOSED_INFINITY

    def has_open_infinity_semantics(self):
        """Returns whether the TFPG uses an open infinity semantics."""
        return self.infinity_semantics == OPEN_INFINITY

    def dump_to_concrete_syntax_file(self, out_filename):
        """Saves the current tfpg in concrete syntax format.
            @param out_filename: The name of the file in which the tfpg will
                                 be dumped.
            @type  out_filename: string
        """
        from xsap.data_structure.tfpg.txt.tfpg import dump_tfpg_to_txt_file
        return dump_tfpg_to_txt_file(self, out_filename)

    def dump_to_xml(self, out_filename):
        """Saves the current tfpg in xml format.
            @param out_filename: the name of the file in which the tfpg will
                                 be dumped.
            @type out_filename: string
        """
        from xsap.data_structure.tfpg.xml.tfpg \
            import tfpg_internal_struct_to_tfpg_doc
        with open(out_filename, "w") as f:
            f.write(str(tfpg_internal_struct_to_tfpg_doc(self)))
        return

    def eq_except_edge_names(self, other):
        return self._eq_aux(other, ignore_edge_names=True)

    def _eq_aux(self, other, ignore_edge_names):
        """ Function to compare two TFPGs. Two TFPGs are considered to be equal
        if all their attributes are equal. """

        #log.xsap.tfpg_other.debug("Comparing TFPGs")

        # compare basics

        if self is None or other is None:
            #log.xsap.tfpg_other.debug("self(%s), other(%s)" % (str(self), str(other)))
            return False

        if self.get_name() != other.get_name():
            #log.xsap.tfpg_other.debug("different names: '%s' != '%s'" % (self.get_name(), other.get_name()))
            return False

        if self.infinity_semantics != other.infinity_semantics:
            #log.xsap.tfpg_other.debug("different infinity semantics: '%s' != '%s'" % (self.infinity_semantics, other.infinity_semantics))
            return False

        # compare modes

        if set(self.get_modes_list()) != set(other.get_modes_list()):
            #log.xsap.tfpg_other.debug("different sets of modes:")
            #log.xsap.tfpg_other.debug("  self:  %s" % str(self.get_modes_list()))
            #log.xsap.tfpg_other.debug("  other: %s" % str(other.get_modes_list()))
            return False

        # compare nodes

        self_nodes = self.get_nodes_list()
        other_nodes = other.get_nodes_list()

        if len(self_nodes) != len(other_nodes):
            #log.xsap.tfpg_other.debug("different number of nodes (%d vs %d)" % (len(self_nodes), len(other_nodes)))
            return False

        self_nodes.sort(key=lambda n: n.get_name())
        other_nodes.sort(key=lambda n: n.get_name())

        for i in range(len(self_nodes)):
            if self_nodes[i] != other_nodes[i]:
                #log.xsap.tfpg_other.debug("node comparison failed")
                #log.xsap.tfpg_other.debug(self_nodes[i])
                #log.xsap.tfpg_other.debug(other_nodes[i])
                return False

        # compare edges

        self_edges = self.get_edges_list()
        other_edges = other.get_edges_list()

        if len(self_edges) != len(other_edges):
            #log.xsap.tfpg_other.debug("different number of edges (%d vs %d)" % (len(self_edges),
                                                                                #len(other_edges)))
            return False

        self_edges.sort(key=lambda e: e.get_name())
        other_edges.sort(key=lambda e: e.get_name())

        if ignore_edge_names:

            matched_edges_in_other = set()

            # for every edge in self ...
            for e_self in self_edges:

                # ... there must be a matching edge in other
                found_match = False
                for e_other in other_edges:

                    # but don't consider the same edge in other twice
                    if str(e_other) in matched_edges_in_other:
                        continue

                    # found a match
                    if e_self.eq_except_edge_names(e_other):
                        found_match = True
                        matched_edges_in_other.add(str(e_other))
                        break

                # didn't find a matching edge
                if not found_match:
                    return False

            # at this point, all edges in other must have been matched
            assert matched_edges_in_other == set([str(e) for e in other_edges]), "Error: Please " \
                                                                                 "report to the " \
                                                                                 "xSAP team."

        else:
            for i in range(len(self_edges)):
                if self_edges[i] != other_edges[i]:
                    #log.xsap.tfpg_other.debug("edge comparison failed")
                    #log.xsap.tfpg_other.debug(self_edges[i])
                    #log.xsap.tfpg_other.debug(other_edges[i])
                    return False

        return True

    def __ne__(self, other):
        """ Negation of __eq__ """
        return not self.__eq__(other)

    def __eq__(self, other):
        """ TFPG comparison function. Everything must be equivalent, including edge names. """
        return self._eq_aux(other=other, ignore_edge_names=False)

    def is_isomorphic(self, other, check_attributes):
        import networkx as nx

        def _convert_tfpg_to_graph(tfpg):
            g = nx.Graph(name=tfpg.get_name())
            node_map = dict()
            for idx, node in enumerate(tfpg.nodes_list):
                node_map[node.get_name()] = idx+1
                g.add_node(idx+1, name=node.get_name(), type=node.get_type(), is_monitored=node.is_monitored())
            #print("NODE MAP: " + str(node_map))
            for edge in tfpg.edges_list:
                g.add_edge(node_map[edge.get_srcNode().get_name()],
                           node_map[edge.get_destNode().get_name()],
                           tMin=edge.get_tMin(),
                           tMax=edge.get_tMax(),
                           modes=sorted(edge.get_modes()))
            return g

        tfpg1 = _convert_tfpg_to_graph(self)
        tfpg2 = _convert_tfpg_to_graph(other)

        def em(e1, e2):
            if e1['tMin'] != e2['tMin']:
                return False
            if e1['tMax'] != e2['tMax']:
                return False
            if e1['modes'] != e2['modes']:
                return False
            return True

        def nm(n1, n2):
            if n1['type'] != n2['type']:
                return False
            if n1['is_monitored'] != n2['is_monitored']:
                return False
            return True

        if check_attributes:
            return nx.is_isomorphic(tfpg1, tfpg2, edge_match=em, node_match=nm) 
        else:
            return nx.is_isomorphic(tfpg1, tfpg2)


    # end of Tfpg class


# ------------------------------------------------------------------------------

class Edge(object):
    """Represents a TFPG Edge"""

    def __init__(self, srcNode, tMin, tMax, modes, destNode, name):
        """Constructor Method

        This function creates an edge with the values specified as parameters.
        Notice that these values are set with the internal set_<attribute>
        functions. Internally, these functions make some comparisons,
        e.g. the set_tMin() function checks that tMin is lower than tMax etc.
        In order to do that, tMax must containe a value, i.e. being different
        from None (otherwise python will complain about that).
        This is the reason why, before setting the values, the attributes are
        initialized to "standard values".

        """

        # Initialization of the attributes
        self.name = ""
        self.srcNode = Node(None, "", False)
        self.destNode = Node(None, "", False)
        self.tMin = 0.0
        self.tMax = INFINITY
        self.modes = []
        self.edge_errors_list = []  # LB: added this list with all
        #                                 errors of Edges

        # The constructor parameters are assigned to the internal attributes
        self.set_name(name)
        self.set_srcNode(srcNode)
        self.set_tMin(tMin)
        self.set_tMax(tMax)
        self.set_modes(modes)
        self.set_destNode(destNode)

        return

    def __str__(self):
        return "Edge '%s': %s --[%.2f, %.2f](%s)--> %s" % (self.name, self.srcNode.get_name(),
                                                           self.tMin, self.tMax, ', '.join(self.modes),
                                                           self.destNode.get_name())

    def get_srcNode(self):
        """Gets the source node of the Edge.
            @return   self.srcNode:  The source node of the Edge.
            @rtype    self.srcNode:  class Node
        """
        return self.srcNode

    def set_srcNode(self, srcNode):
        """Sets the source node of the Edge.
            @param   srcNode:  The source node of the Edge.
            @type    srcNode: class Node
        """

        try:
            if srcNode.name == self.get_destNode().name:
                raise NodeLoopError(srcNode.name)
            else:
                self.srcNode = srcNode

        except Exception as e:
            self.edge_errors_list.append(e)

        return

    def get_tMin(self):
        """Gets the initial time of the Edge.
            @return   self.tMin:  The initial time of the Edge.
            @rtype    self.tMin:  float
        """
        return self.tMin

    def set_tMin(self, tMin):
        """Sets the initial time of the Edge.
            @param   tMin:  The initial time of the Edge.
            @type    tMin:  float
        """
        try:
            if float(tMin) < 0 or ((float(self.get_tMax()) < float(tMin)) and
                                   (float(self.get_tMax()) != INFINITY)):
                raise TimeNotCorrectError(self.get_name(),
                                          tMin,
                                          self.get_tMax())
            else:
                self.tMin = tMin

        except Exception as e:
            self.edge_errors_list.append(e)
        return

    def get_tMax(self):
        """Gets the final time of the Edge.
            @return   self.tMax:  The final time of the Edge.
            @rtype    self.tMax:  float
        """
        return self.tMax

    def set_tMax(self, tMax):
        """Sets the final time of the Edge.
            @param   tMax:  The final time of the Edge.
            @type    tMax:  float
        """
        try:
            if (float(tMax) < float(self.get_tMin())) and \
                    float(tMax) != INFINITY:
                raise TimeNotCorrectError(self.get_name(),
                                          self.get_tMin(),
                                          tMax)
            else:
                self.tMax = tMax

        except Exception as e:
            self.edge_errors_list.append(e)
        return

    def get_modes(self):
        """Gets the list of modes of the Edge.
            @return   self.modes:  The list of modes of the Edge.
            @rtype    self.modes:  list
        """
        return self.modes

    def set_modes(self, modes):
        """Sets the list of the modes of the Edge.
            @param   modes: the list of the modes of the Edge
            @type    modes: list[string]
        """
        self.modes = modes
        return

    def get_destNode(self):
        """Gets the destination node of the Edge.
            @return   self.destNode:  The destination node of the Edge.
            @rtype    self.destNode:  class Node
        """
        return self.destNode

    def set_destNode(self, destNode):
        """Sets the destination node of the Edge.
            @param   destNode:  The destination node of the Edge.
            @type    destNode:  class Node
        """

        try:
            if self.get_srcNode().name == destNode.name:
                raise NodeLoopError(destNode.name)
            else:
                self.destNode = destNode

        except Exception as e:
            self.edge_errors_list.append(e)
        return

    def get_name(self):
        """Gets the name of the Edge.
            @return   self.name:  The name of the Edge.
            @rtype    self.name:  string
        """
        return self.name

    def set_name(self, name):
        """Sets the name of the Edge.
            @param   name:  The name of the Edge.
            @type    name:  string
        """
        self.name = name
        return

    def _eq_aux(self, other, ignore_edge_names):
        """Builtin function to compare two edges. Notice that two edges are
        considered to be equal if all their attributes are equal. """

        if not ignore_edge_names and self.get_name() != other.get_name():
            return False
        if self.get_srcNode() != other.get_srcNode():
            return False
        if self.get_destNode() != other.get_destNode():
            return False
        if float(self.get_tMin()) != float(other.get_tMin()):
            return False
        if float(self.get_tMax()) != float(other.get_tMax()):
            return False

        self_modes = sorted(self.get_modes())
        other_modes = other.get_modes()
        other_modes.sort()

        if self_modes != other_modes:
            return False

        return True

    def eq_except_edge_names(self, other):
        return self._eq_aux(other, ignore_edge_names=True)

    def __ne__(self, other):
        """ Negation of __eq__ """
        return not self.__eq__(other)

    def __eq__(self, other):
        """ TFPG Edge comparison function. Everything must be equivalent,
        including edge names. """
        return self._eq_aux(other, ignore_edge_names=False)

    def get_edge_errors_list(self):
        """Gets the errors of the Edge.
            @return   self.edge_errors_list:  The errors of the Edge.
            @rtype    self.edge_errors_list:  list
        """
        return self.edge_errors_list

    # end of Edge class


# ------------------------------------------------------------------------------

class Node(object):
    """Represents a TFPG Node"""

    def __init__(self, _type, name, isMonitored):
        """Constructor Method.

           This function creates an edge with the values specified as
           parameters.
           Notice that these values are set with the internal set_<attribute>
           functions. Internally, these functions make some comparisons,
           e.g. the set_type() function checks that whether the attribute
           isMonitored is False etc.
           In order to do that, isMonitored must containe a value, i.e.
           being different from None (otherwise python will complain about
           that).
           This is the reason why, before setting the values, the attributes
           are initialized to "standard values".
        """

        # Initialization of the attributes
        self._type = None
        self.name = ""
        self.isMonitored = False
        # LB: added this list with all errors of Nodes
        self.node_errors_list = []

        # The constructor parameters are assigned to the internal attributes
        self.set_type(_type)
        self.set_name(name)
        self.set_isMonitored(isMonitored)

        return

    def get_type(self):
        """Gets the type of the Node (FM, AND or OR).
            @return   self.type:  The type of the Node.
            @rtype    self.type:  class NodeType
        """
        return self.type

    def set_type(self, _type):
        """Sets the type of the Node.
            @param   type:  The type of the Node.
            @type    type:  class Node
        """
        try:
            if _type == NodeType.FM and self.is_monitored():
                raise FMNotCorrectError(self.get_name())
            else:
                self.type = _type

        except Exception as e:
            self.node_errors_list.append(e)
        return

    def get_name(self):
        """Gets the name of the Node.
            @return   self.name:  The name of the Node.
            @rtype    self.name:  string
        """
        return self.name

    def set_name(self, name):
        """Sets the name of the Node.
            @param   name:  The name of the Node.
            @type    name:  string
        """
        try:
            if name is None or name == "" or len(name.replace(" ", "")) == 0:
                raise EmptyNodeName()
            else:
                self.name = name

        except Exception as e:
            self.node_errors_list.append(e)
        return

    def is_monitored(self):
        """Says whether the Node is monitored or not.
            @return   self.isMonitored:  If the Node is monitored.
            @rtype    self.isMonitored:  boolean
        """
        return self.isMonitored

    def set_isMonitored(self, isMonitored):
        """Sets whether the Node is monitored or not.
            @param   isMonitored:  If the Node is monitored.
            @type    isMonitored:  boolean
        """
        try:
            if isMonitored and self.get_type() == NodeType.FM:
                raise FMNotCorrectError(self.get_name())
            else:
                self.isMonitored = isMonitored

        except Exception as e:
            self.node_errors_list.append(e)
        return

    def __ne__(self, other):
        """Negation of __eq__"""
        return not self.__eq__(other)

    def __eq__(self, other):
        """Builting function to compare two nodes. Notice that two nodes are considered
            to be equal only if all thei attributes are equal
        """

        if self.get_name() != other.get_name():
            return False
        if self.get_type() != other.get_type():
            return False
        if self.is_monitored() != other.is_monitored():
            return False

        return True

    def get_node_errors_list(self):
        """Gets the errors of the Node.
            @return   self.node_errors_list:  The errors of the Node.
            @rtype    self.node_errors_list:  list
        """
        return self.node_errors_list

    def __str__(self):
        return "%s %s" % (self.get_type(), self.get_name())

    # end of Node class


# ------------------------------------------------------------------------------

class NodeType(object):
    """Represents a TFPG Node type"""

    FM = "FM"
    AND = "AND"
    OR = "OR"

    # end of NodeType class
